@@ -82,6 +82,16 @@ F2x_to_Flx(GEN x)
   return z;
 }
 
+GEN
+F2x_to_F2xX(GEN z, long sv)
+{
+  long i, d = F2x_degree(z);
+  GEN x = cgetg(d+3,t_POL);
+  for (i=0; i<=d; i++)
+    gel(x,i+2) = F2x_coeff(z,i) ? pol1_F2x(sv): pol0_F2x(sv);
+  x[1] = evalsigne(d+1!=0)| z[1]; return x;
+}
+
 GEN
 Z_to_F2x(GEN x, long v)
 {
@@ -189,7 +199,7 @@ Rg_to_F2xq(GEN x, GEN T)
       return F2xq_div(a,b, T);
   }
   pari_err_TYPE("Rg_to_F2xq",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 ulong
@@ -600,6 +610,12 @@ F2x_shift(GEN y, long d)
   return d<0 ? F2x_shiftneg(y,-d): F2x_shiftpos(y,d);
 }
 
+GEN
+F2x_get_red(GEN T)
+{
+  return T;
+}
+
 /* separate from F2x_divrem for maximal speed. */
 GEN
 F2x_rem(GEN x, GEN y)
@@ -923,6 +939,18 @@ F2xq_powu(GEN x, ulong n, GEN pol)
   return gerepileupto(av, y);
 }
 
+GEN
+F2xq_pow_init(GEN x, GEN n, long k,  GEN T)
+{
+  return gen_pow_init(x, n, k, (void*)T, &_F2xq_sqr, &_F2xq_mul);
+}
+
+GEN
+F2xq_pow_table(GEN R, GEN n, GEN T)
+{
+  return gen_pow_table(R, n, (void*)T, &_F2xq_one, &_F2xq_mul);
+}
+
 /* generates the list of powers of x of degree 0,1,2,...,l*/
 GEN
 F2xq_powers(GEN x, long l, GEN T)
@@ -1067,7 +1095,7 @@ F2x_is_smooth(GEN g, long r)
 static GEN
 F2x_factorel(GEN h)
 {
-  GEN F = F2x_factcantor(h, 0);
+  GEN F = F2x_factor(h);
   GEN F1 = gel(F, 1), F2 = gel(F, 2);
   long i, l1 = lg(F1)-1;
   GEN p2 = cgetg(l1+1, t_VECSMALL);
@@ -1236,7 +1264,7 @@ F2xq_log_Coppersmith_rec(GEN W, long r2, GEN a, long r, long n, GEN T, GEN m)
       if (!R) break;
       Ao = Fp_add(Ao, mulis(R, E[i]), m);
     }
-    if (i==l) return subis(Ao,AV);
+    if (i==l) return subiu(Ao,AV);
   }
 }
 
@@ -1261,48 +1289,75 @@ rel_Coppersmith(GEN u, GEN v, long h, GEN R, long r, long n, long d)
   return famatsmall_reduce(M);
 }
 
+GEN
+F2xq_log_Coppersmith_worker(GEN u, long i, GEN V, GEN R)
+{
+  long r = V[1], h = V[2], n = V[3], d = V[4];
+  pari_sp ltop = avma;
+  GEN v = mkF2(0,u[1]);
+  GEN L = cgetg(2*i+1, t_VEC);
+  pari_sp av = avma;
+  long j;
+  long nbtest=0, rel = 1;
+  for(j=1; j<=i; j++)
+  {
+    v[2] = j;
+    avma = av;
+    if (F2x_degree(F2x_gcd(u,v))==0)
+    {
+      GEN z = rel_Coppersmith(u, v, h, R, r, n, d);
+      nbtest++;
+      if (z) { gel(L, rel++) = z; av = avma; }
+      if (i==j) continue;
+      z = rel_Coppersmith(v, u, h, R, r, n, d);
+      nbtest++;
+      if (z) { gel(L, rel++) = z; av = avma; }
+    }
+  }
+  setlg(L,rel);
+  return gerepilecopy(ltop, mkvec2(stoi(nbtest), L));
+}
+
 static GEN
 F2xq_log_Coppersmith(long nbrel, long r, long n, GEN T)
 {
   long dT = F2x_degree(T);
   long h = dT>>n, d = dT-(h<<n);
   GEN R = F2x_add(F2x_shift(pol1_F2x(T[1]), dT), T);
-  GEN u = mkF2(0,T[1]), v = mkF2(0,T[1]);
+  GEN u = mkF2(0,T[1]);
   long rel = 1, nbtest = 0;
   GEN M = cgetg(nbrel+1, t_VEC);
-  pari_sp av = avma;
-  long i,j;
+  long i = 0;
+  GEN worker = snm_closure(is_entry("_F2xq_log_Coppersmith_worker"),
+               mkvec2(mkvecsmall4(r, h, n, d), R));
+  struct pari_mt pt;
+  long running, pending = 0, stop=0;
+  mt_queue_start(&pt, worker);
   if (DEBUGLEVEL) err_printf("Coppersmith (R = %ld): ",F2x_degree(R));
-  for (i=1; ; i++)
+  while ((running = !stop) || pending)
   {
+    GEN L, done;
+    long j, l;
     u[2] = i;
-    for(j=1; j<=i; j++)
+    mt_queue_submit(&pt, 0, running ? mkvec2(u, stoi(i)): NULL);
+    done = mt_queue_get(&pt, NULL, &pending);
+    if (!done) continue;
+    L = gel(done, 2); nbtest += itos(gel(done,1));
+    l = lg(L);
+    if (l > 1)
     {
-      v[2] = j;
-      avma = av;
-      if (F2x_degree(F2x_gcd(u,v))==0)
+      for (j=1; j<l; j++)
       {
-        GEN z = rel_Coppersmith(u, v, h, R, r, n, d);
-        nbtest++;
-        if (z)
-        {
-          gel(M,rel++) = gerepilecopy(av, z); av = avma;
-          if (DEBUGLEVEL && (rel&511UL)==0)
-            err_printf("%ld%%[%ld] ",rel*100/nbrel,i);
-        }
         if (rel>nbrel) break;
-        if (i==j) continue;
-        z = rel_Coppersmith(v, u, h, R, r, n, d);
-        nbtest++;
-        if (!z) continue;
-        gel(M,rel++) = gerepilecopy(av, z); av = avma;
+        gel(M,rel++) = gel(L,j);
         if (DEBUGLEVEL && (rel&511UL)==0)
           err_printf("%ld%%[%ld] ",rel*100/nbrel,i);
-        if (rel>nbrel) break;
       }
     }
-    if (rel>nbrel) break;
+    if (rel>nbrel) stop=1;
+    i++;
   }
+  mt_queue_end(&pt);
   if (DEBUGLEVEL) err_printf(": %ld tests\n", nbtest);
   return M;
 }
@@ -1317,27 +1372,30 @@ smallirred_F2x(ulong n, long sv)
 }
 
 static GEN
-check_kernel(long N, GEN M, GEN T, GEN m)
+check_kernel(long N, GEN M, long nbi, GEN T, GEN m)
 {
   pari_sp av = avma;
   GEN K = FpMs_leftkernel_elt(M, N, m);
-  long i, f=0;
+  long i, f=0, tbs;
   long l = lg(K), lm = lgefint(m);
-  GEN idx = diviiexact(subis(int2n(F2x_degree(T)),1),m);
+  GEN idx = diviiexact(int2um1(F2x_degree(T)),m);
   GEN g = F2xq_pow(polx_F2x(T[1]), idx, T);
+  GEN tab;
   pari_timer ti;
   if (DEBUGLEVEL) timer_start(&ti);
   K = FpC_Fp_mul(K, Fp_inv(gel(K,2), m), m);
+  tbs = maxss(1, expu(nbi/expi(m)));
+  tab = F2xq_pow_init(g, int2n(F2x_degree(T)), tbs, T);
   for(i=1; i<l; i++)
   {
     GEN k = gel(K,i);
-    if (signe(k)==0 || !F2x_equal(F2xq_pow(g, k, T),
+    if (signe(k)==0 || !F2x_equal(F2xq_pow_table(tab, k, T),
                                   F2xq_pow(mkF2(i,T[1]), idx, T)))
       gel(K,i) = cgetineg(lm);
     else
       f++;
   }
-  if (DEBUGLEVEL) timer_printf(&ti,"found %ld logs", f);
+  if (DEBUGLEVEL) timer_printf(&ti,"found %ld/%ld logs", f, nbi);
   return gerepileupto(av, K);
 }
 
@@ -1365,7 +1423,7 @@ F2xq_log_index(GEN a0, GEN b0, GEN m, GEN T0)
   M = F2xq_log_Coppersmith(nbrel,r,d,T);
   if(DEBUGLEVEL)
     timer_printf(&ti,"relations");
-  W = check_kernel(N, M, T, m);
+  W = check_kernel(N, M, nbi, T, m);
   timer_start(&ti);
   Ao = F2xq_log_Coppersmith_rec(W, r2, a, r, d2, T, m);
   if (DEBUGLEVEL) timer_printf(&ti,"smooth element");
@@ -1444,7 +1502,7 @@ F2xq_sqrtn(GEN a, GEN n, GEN T, GEN *zeta)
       *zeta=pol1_F2x(T[1]);
     return pol0_F2x(T[1]);
   }
-  return gen_Shanks_sqrtn(a,n,addis(powuu(2,F2x_degree(T)),-1),zeta,(void*)T,&F2xq_star);
+  return gen_Shanks_sqrtn(a,n,subiu(powuu(2,F2x_degree(T)),1),zeta,(void*)T,&F2xq_star);
 }
 
 GEN
@@ -1458,7 +1516,7 @@ gener_F2xq(GEN T, GEN *po)
     if (po) *po = mkvec2(gen_1, trivial_fact());
     return pol1_F2x(vT);
   }
-  q = subis(powuu(2,f), 1);
+  q = int2um1(f);
   o = factor_pn_1(gen_2,f);
   L2 = leafcopy( gel(o, 1) );
   for (i = j = 1; i < lg(L2); i++)
@@ -1480,7 +1538,7 @@ gener_F2xq(GEN T, GEN *po)
   }
   if (!po) g = gerepilecopy(av0, g);
   else {
-    *po = mkvec2(subis(int2n(f), 1), o);
+    *po = mkvec2(int2um1(f), o);
     gerepileall(av0, 2, &g, po);
   }
   return g;
@@ -1665,33 +1723,15 @@ Flv_to_F2v(GEN x)
 
 GEN
 ZM_to_F2m(GEN x)
-{
-  long j, l = lg(x);
-  GEN y = cgetg(l,t_MAT);
-  if (l == 1) return y;
-  for (j=1; j<l; j++) gel(y,j) = ZV_to_F2v(gel(x,j));
-  return y;
-}
+{ pari_APPLY_same(ZV_to_F2v(gel(x,i))) }
 
 GEN
 RgM_to_F2m(GEN x)
-{
-  long j, l = lg(x);
-  GEN y = cgetg(l,t_MAT);
-  if (l == 1) return y;
-  for (j=1; j<l; j++) gel(y,j) = RgV_to_F2v(gel(x,j));
-  return y;
-}
+{ pari_APPLY_same(RgV_to_F2v(gel(x,i))) }
 
 GEN
 Flm_to_F2m(GEN x)
-{
-  long j, l = lg(x);
-  GEN y = cgetg(l,t_MAT);
-  if (l == 1) return y;
-  for (j=1; j<l; j++) gel(y,j) = Flv_to_F2v(gel(x,j));
-  return y;
-}
+{ pari_APPLY_same(Flv_to_F2v(gel(x,i))) }
 
 GEN
 const_F2v(long m)
@@ -1733,39 +1773,19 @@ F2v_add_inplace(GEN x, GEN y)
 
 GEN
 FlxC_to_F2xC(GEN x)
-{
-  long i, l=lg(x);
-  GEN z = cgetg(l,t_COL);
-  for (i=1; i<l ; i++) gel(z,i) = Flx_to_F2x(gel(x,i));
-  return z;
-}
+{ pari_APPLY_type(t_COL, Flx_to_F2x(gel(x,i))) }
 
 GEN
 F2xC_to_FlxC(GEN x)
-{
-  long i, l=lg(x);
-  GEN z = cgetg(l,t_COL);
-  for (i=1; i<l ; i++) gel(z,i) = F2x_to_Flx(gel(x,i));
-  return z;
-}
+{ pari_APPLY_type(t_COL, F2x_to_Flx(gel(x,i))) }
 
 GEN
-F2xC_to_ZXC(GEN v)
-{
-  long j, N = lg(v);
-  GEN y = cgetg(N, t_COL);
-  for (j=1; j<N; j++) gel(y,j) = F2x_to_ZX(gel(v,j));
-  return y;
-}
+F2xC_to_ZXC(GEN x)
+{ pari_APPLY_type(t_COL, F2x_to_ZX(gel(x,i))) }
 
 GEN
-F2xV_to_F2m(GEN v, long n)
-{
-  long j, N = lg(v);
-  GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = F2x_to_F2v(gel(v,j), n);
-  return y;
-}
+F2xV_to_F2m(GEN x, long n)
+{ pari_APPLY_type(t_MAT, F2x_to_F2v(gel(x,i), n)) }
 
 /***********************************************************************/
 /**                                                                   **/
@@ -1849,6 +1869,18 @@ F2xX_deriv(GEN z)
   return F2xX_renormalize(x,l);
 }
 
+static GEN
+F2xX_addspec(GEN x, GEN y, long lx, long ly)
+{
+  long i,lz;
+  GEN z;
+  if (ly>lx) swapspec(x,y, lx,ly);
+  lz = lx+2; z = cgetg(lz, t_POL);
+  for (i=0; i<ly; i++) gel(z,i+2) = F2x_add(gel(x,i), gel(y,i));
+  for (   ; i<lx; i++) gel(z,i+2) = F2x_copy(gel(x,i));
+  return F2xX_renormalize(z, lz);
+}
+
 GEN
 F2xX_add(GEN x, GEN y)
 {
@@ -1863,6 +1895,20 @@ F2xX_add(GEN x, GEN y)
   return F2xX_renormalize(z, lz);
 }
 
+GEN
+F2xX_F2x_add(GEN x, GEN y)
+{
+  long i, lz = lg(y);
+  GEN z;
+  if (signe(y) == 0) return scalarpol(x, varn(y));
+  z = cgetg(lz,t_POL); z[1] = y[1];
+  gel(z,2) = F2x_add(gel(y,2), x);
+  if (lz == 3) z = F2xX_renormalize(z,lz);
+  else
+    for(i=3;i<lz;i++) gel(z,i) = F2x_copy(gel(y,i));
+  return z;
+}
+
 GEN
 F2xX_F2x_mul(GEN P, GEN U)
 {
@@ -1894,13 +1940,32 @@ F2xY_F2xq_evalx(GEN P, GEN x, GEN T)
   return gerepileupto(av, F2xY_F2xqV_evalx(P, xp, T));
 }
 
+static GEN
+F2xX_to_Kronecker_spec(GEN P, long n, long d)
+{
+  long i, k, l, N = 2*d + 1;
+  long dP = n+1;
+  GEN x;
+  if (n == 0) return pol0_Flx(P[1]&VARNBITS);
+  l = nbits2nlong(N*dP+d+1);
+  x = zero_zv(l+1);
+  for (k=i=0; i<n; i++, k+=N)
+  {
+    GEN c = gel(P,i);
+    F2x_addshiftip(x, c, k);
+  }
+  x[1] = P[1]&VARNBITS; return F2x_renormalize(x, l+2);
+}
+
 GEN
 F2xX_to_Kronecker(GEN P, long d)
 {
-  long i, k, N = 2*d + 1;
+  long i, k, l, N = 2*d + 1;
   long dP = degpol(P);
-  long l = nbits2nlong(N*dP+d+1);
-  GEN x = zero_zv(l+1);
+  GEN x;
+  if (dP < 0) return pol0_Flx(P[1]&VARNBITS);
+  l = nbits2nlong(N*dP+d+1);
+  x = zero_zv(l+1);
   for (k=i=0; i<=dP; i++, k+=N)
   {
     GEN c = gel(P,i+2);
@@ -1963,24 +2028,12 @@ Kronecker_to_F2xqX(GEN z, GEN T)
 /***********************************************************************/
 
 GEN
-FlxXC_to_F2xXC(GEN B)
-{
-  long i, l = lg(B);
-  GEN z = cgetg(l, t_COL);
-  for (i=1; i<l; i++)
-    gel(z,i) = FlxX_to_F2xX(gel(B,i));
-  return z;
-}
+FlxXC_to_F2xXC(GEN x)
+{ pari_APPLY_type(t_COL, FlxX_to_F2xX(gel(x,i))); }
 
 GEN
-F2xXC_to_ZXXC(GEN B)
-{
-  long i, l = lg(B);
-  GEN z = cgetg(l, t_COL);
-  for (i=1; i<l; i++)
-    gel(z,i) = F2xX_to_ZXX(gel(B,i));
-  return z;
-}
+F2xXC_to_ZXXC(GEN x)
+{ pari_APPLY_type(t_COL, F2xX_to_ZXX(gel(x,i))); }
 
 /***********************************************************************/
 /**                                                                   **/
@@ -1988,6 +2041,13 @@ F2xXC_to_ZXXC(GEN B)
 /**                                                                   **/
 /***********************************************************************/
 
+static GEN
+get_F2xqX_red(GEN T, GEN *B)
+{
+  if (typ(T)!=t_VEC) { *B=NULL; return T; }
+  *B = gel(T,1); return gel(T,2);
+}
+
 GEN
 random_F2xqX(long d1, long v, GEN T)
 {
@@ -2050,6 +2110,18 @@ F2xqX_mul(GEN x, GEN y, GEN T)
   return gerepileupto(ltop,z);
 }
 
+static GEN
+F2xqX_mulspec(GEN x, GEN y, GEN T, long nx, long ny)
+{
+  pari_sp ltop=avma;
+  GEN z,kx,ky;
+  kx= F2xX_to_Kronecker_spec(x, nx, F2x_degree(T));
+  ky= F2xX_to_Kronecker_spec(y, ny, F2x_degree(T));
+  z = F2x_mul(ky, kx);
+  z = Kronecker_to_F2xqX(z,T);
+  return gerepileupto(ltop,z);
+}
+
 GEN
 F2xqX_sqr(GEN x, GEN T)
 {
@@ -2072,8 +2144,8 @@ F2xqX_sqr(GEN x, GEN T)
   return FlxX_renormalize(z, l);
 }
 
-GEN
-F2xqX_divrem(GEN x, GEN y, GEN T, GEN *pr)
+static GEN
+F2xqX_divrem_basecase(GEN x, GEN y, GEN T, GEN *pr)
 {
   long vx, dx, dy, dz, i, j, sx, lr;
   pari_sp av0, av, tetpil;
@@ -2157,13 +2229,259 @@ F2xqX_divrem(GEN x, GEN y, GEN T, GEN *pr)
   *pr = rem; return z-2;
 }
 
+static GEN
+F2xX_recipspec(GEN x, long l, long n, long vs)
+{
+  long i;
+  GEN z = cgetg(n+2,t_POL);
+  z[1] = 0; z += 2;
+  for(i=0; i<l; i++)
+    gel(z,n-i-1) = F2x_copy(gel(x,i));
+  for(   ; i<n; i++)
+    gel(z,n-i-1) = pol0_F2x(vs);
+  return F2xX_renormalize(z-2,n+2);
+}
+
+static GEN
+F2xqX_invBarrett_basecase(GEN T, GEN Q)
+{
+  long i, l=lg(T)-1, lr = l-1, k;
+  long sv=Q[1];
+  GEN r=cgetg(lr,t_POL); r[1]=T[1];
+  gel(r,2) = pol1_F2x(sv);
+  for (i=3;i<lr;i++)
+  {
+    pari_sp ltop=avma;
+    GEN u = gel(T,l-i+2);
+    for (k=3;k<i;k++)
+      u = F2x_add(u, F2xq_mul(gel(T,l-i+k),gel(r,k),Q));
+    gel(r,i) = gerepileupto(ltop, u);
+  }
+  r = F2xX_renormalize(r,lr);
+  return r;
+}
+
+/* Return new lgpol */
+static long
+F2xX_lgrenormalizespec(GEN x, long lx)
+{
+  long i;
+  for (i = lx-1; i>=0; i--)
+    if (lgpol(gel(x,i))) break;
+  return i+1;
+}
+
+static GEN
+F2xqX_invBarrett_Newton(GEN S, GEN T)
+{
+  pari_sp av = avma;
+  long nold, lx, lz, lq, l = degpol(S), i, lQ;
+  GEN q, y, z, x = cgetg(l+2, t_POL) + 2;
+  long dT = F2x_degree(T);
+  ulong mask = quadratic_prec_mask(l-2); /* assume l > 2 */
+  for (i=0;i<l;i++) gel(x,i) = pol0_F2x(T[1]);
+  q = F2xX_recipspec(S+2,l+1,l+1,dT);
+  lQ = lgpol(q); q+=2;
+  /* We work on _spec_ FlxX's, all the l[xzq] below are lgpol's */
+
+  /* initialize */
+  gel(x,0) = F2xq_inv(gel(q,0),T);
+  if (lQ>1 && F2x_degree(gel(q,1)) >= dT)
+    gel(q,1) = F2x_rem(gel(q,1), T);
+  if (lQ>1 && lgpol(gel(q,1)))
+  {
+    GEN u = gel(q, 1);
+    if (!F2x_equal1(gel(x,0))) u = F2xq_mul(u, F2xq_sqr(gel(x,0), T), T);
+    else u = F2x_copy(u);
+    gel(x,1) = u; lx = 2;
+  }
+  else
+    lx = 1;
+  nold = 1;
+  for (; mask > 1; )
+  { /* set x -= x(x*q - 1) + O(t^(nnew + 1)), knowing x*q = 1 + O(t^(nold+1)) */
+    long i, lnew, nnew = nold << 1;
+
+    if (mask & 1) nnew--;
+    mask >>= 1;
+
+    lnew = nnew + 1;
+    lq = F2xX_lgrenormalizespec(q, minss(lQ,lnew));
+    z = F2xqX_mulspec(x, q, T, lx, lq); /* FIXME: high product */
+    lz = lgpol(z); if (lz > lnew) lz = lnew;
+    z += 2;
+    /* subtract 1 [=>first nold words are 0]: renormalize so that z(0) != 0 */
+    for (i = nold; i < lz; i++) if (lgpol(gel(z,i))) break;
+    nold = nnew;
+    if (i >= lz) continue; /* z-1 = 0(t^(nnew + 1)) */
+
+    /* z + i represents (x*q - 1) / t^i */
+    lz = F2xX_lgrenormalizespec (z+i, lz-i);
+    z = F2xqX_mulspec(x, z+i, T, lx, lz); /* FIXME: low product */
+    lz = lgpol(z); z += 2;
+    if (lz > lnew-i) lz = F2xX_lgrenormalizespec(z, lnew-i);
+
+    lx = lz+ i;
+    y  = x + i; /* x -= z * t^i, in place */
+    for (i = 0; i < lz; i++) gel(y,i) = gel(z,i);
+  }
+  x -= 2; setlg(x, lx + 2); x[1] = S[1];
+  return gerepilecopy(av, x);
+}
+
+/* x/polrecip(P)+O(x^n) */
+GEN
+F2xqX_invBarrett(GEN T, GEN Q)
+{
+  pari_sp ltop=avma;
+  long l=lg(T), v = varn(T);
+  GEN r;
+  GEN c = gel(T,l-1);
+  if (l<5) return pol_0(v);
+  if (l<=F2xqX_INVBARRETT_LIMIT)
+  {
+    if (!F2x_equal1(c))
+    {
+      GEN ci = F2xq_inv(c,Q);
+      T = F2xqX_F2xq_mul(T, ci, Q);
+      r = F2xqX_invBarrett_basecase(T,Q);
+      r = F2xqX_F2xq_mul(r,ci,Q);
+    } else
+      r = F2xqX_invBarrett_basecase(T,Q);
+  } else
+    r = F2xqX_invBarrett_Newton(T,Q);
+  return gerepileupto(ltop, r);
+}
+
+GEN
+F2xqX_get_red(GEN S, GEN T)
+{
+  if (typ(S)==t_POL && lg(S)>F2xqX_BARRETT_LIMIT)
+    retmkvec2(F2xqX_invBarrett(S, T), S);
+  return S;
+}
+
+/* Compute x mod S where 2 <= degpol(S) <= l+1 <= 2*(degpol(S)-1)
+ *  * and mg is the Barrett inverse of S. */
+static GEN
+F2xqX_divrem_Barrettspec(GEN x, long l, GEN mg, GEN S, GEN T, GEN *pr)
+{
+  GEN q, r;
+  long lt = degpol(S); /*We discard the leading term*/
+  long ld, lm, lT, lmg;
+  ld = l-lt;
+  lm = minss(ld, lgpol(mg));
+  lT  = F2xX_lgrenormalizespec(S+2,lt);
+  lmg = F2xX_lgrenormalizespec(mg+2,lm);
+  q = F2xX_recipspec(x+lt,ld,ld,0);               /* q = rec(x)     lq<=ld*/
+  q = F2xqX_mulspec(q+2,mg+2,T,lgpol(q),lmg);   /* q = rec(x) * mg lq<=ld+lm*/
+  q = F2xX_recipspec(q+2,minss(ld,lgpol(q)),ld,0);/* q = rec (rec(x) * mg) lq<=ld*/
+  if (!pr) return q;
+  r = F2xqX_mulspec(q+2,S+2,T,lgpol(q),lT);     /* r = q*pol        lr<=ld+lt*/
+  r = F2xX_addspec(x,r+2,lt,minss(lt,lgpol(r)));/* r = x - r   lr<=lt */
+  if (pr == ONLY_REM) return r;
+  *pr = r; return q;
+}
+
+static GEN
+F2xqX_divrem_Barrett_noGC(GEN x, GEN mg, GEN S, GEN T, GEN *pr)
+{
+  long l = lgpol(x), lt = degpol(S), lm = 2*lt-1;
+  GEN q = NULL, r;
+  long i;
+  if (l <= lt)
+  {
+    if (pr == ONLY_REM) return RgX_copy(x);
+    if (pr == ONLY_DIVIDES) return signe(x)? NULL: pol_0(varn(x));
+    if (pr) *pr =  RgX_copy(x);
+    return pol_0(varn(x));
+  }
+  if (lt <= 1)
+    return F2xqX_divrem_basecase(x,S,T,pr);
+  if (pr != ONLY_REM && l>lm)
+  {
+    long vT = get_F2x_var(T);
+    q = cgetg(l-lt+2, t_POL);
+    for (i=0;i<l-lt;i++) gel(q+2,i) = pol0_F2x(vT);
+  }
+  r = l>lm ? shallowcopy(x): x;
+  while (l>lm)
+  {
+    GEN zr, zq = F2xqX_divrem_Barrettspec(r+2+l-lm,lm,mg,S,T,&zr);
+    long lz = lgpol(zr);
+    if (pr != ONLY_REM)
+    {
+      long lq = lgpol(zq);
+      for(i=0; i<lq; i++) gel(q+2+l-lm,i) = gel(zq,2+i);
+    }
+    for(i=0; i<lz; i++) gel(r+2+l-lm,i) = gel(zr,2+i);
+    l = l-lm+lz;
+  }
+  if (pr != ONLY_REM)
+  {
+    if (l > lt)
+    {
+      GEN zq = F2xqX_divrem_Barrettspec(r+2,l,mg,S,T,&r);
+      if (!q) q = zq;
+      else
+      {
+        long lq = lgpol(zq);
+        for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
+      }
+    }
+    else
+    { setlg(r, l+2); r = RgX_copy(r); }
+  }
+  else
+  {
+    if (l > lt)
+      (void) F2xqX_divrem_Barrettspec(r+2,l,mg,S,T,&r);
+    else
+    { setlg(r, l+2); r = RgX_copy(r); }
+    r[1] = x[1]; return F2xX_renormalize(r, lg(r));
+  }
+  if (pr) { r[1] = x[1]; r = F2xX_renormalize(r, lg(r)); }
+  q[1] = x[1]; q = F2xX_renormalize(q, lg(q));
+  if (pr == ONLY_DIVIDES) return signe(r)? NULL: q;
+  if (pr) *pr = r;
+  return q;
+}
+
+GEN
+F2xqX_divrem(GEN x, GEN S, GEN T, GEN *pr)
+{
+  GEN B, y = get_F2xqX_red(S, &B);
+  long dy = degpol(y), dx = degpol(x), d = dx-dy;
+  if (pr==ONLY_REM) return F2xqX_rem(x, y, T);
+  if (!B && d+3 < F2xqX_DIVREM_BARRETT_LIMIT)
+    return F2xqX_divrem_basecase(x,y,T,pr);
+  else
+  {
+    pari_sp av=avma;
+    GEN mg = B? B: F2xqX_invBarrett(y, T);
+    GEN q = F2xqX_divrem_Barrett_noGC(x,mg,y,T,pr);
+    if (!q) {avma=av; return NULL;}
+    if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q);
+    gerepileall(av,2,&q,pr);
+    return q;
+  }
+}
+
 GEN
 F2xqX_rem(GEN x, GEN S, GEN T)
 {
-  GEN y = S;
+  GEN B, y = get_F2xqX_red(S, &B);
   long dy = degpol(y), dx = degpol(x), d = dx-dy;
   if (d < 0) return F2xqX_red(x, T);
-  return F2xqX_divrem(x,y, T, ONLY_REM);
+  if (!B && d+3 < F2xqX_REM_BARRETT_LIMIT)
+    return F2xqX_divrem_basecase(x,y, T, ONLY_REM);
+  else
+  {
+    pari_sp av=avma;
+    GEN mg = B? B: F2xqX_invBarrett(y, T);
+    GEN r = F2xqX_divrem_Barrett_noGC(x, mg, y, T, ONLY_REM);
+    return gerepileupto(av, r);
+  }
 }
 
 GEN
@@ -2183,6 +2501,44 @@ F2xqX_gcd(GEN a, GEN b, GEN T)
   avma = av; return a;
 }
 
+GEN
+F2xqX_extgcd(GEN a, GEN b, GEN T,  GEN *ptu, GEN *ptv)
+{
+  pari_sp av=avma;
+  GEN u,v,d,d1,v1;
+  long vx = varn(a);
+  d = a; d1 = b;
+  v = pol_0(vx); v1 = pol1_F2xX(vx, get_F2x_var(T));
+  while (signe(d1))
+  {
+    GEN r, q = F2xqX_divrem(d, d1, T, &r);
+    v = F2xX_add(v,F2xqX_mul(q,v1,T));
+    u=v; v=v1; v1=u;
+    u=r; d=d1; d1=u;
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_extgcd (d = %ld)",degpol(d));
+      gerepileall(av,5, &d,&d1,&u,&v,&v1);
+    }
+  }
+  if (ptu) *ptu = F2xqX_div(F2xX_add(d,F2xqX_mul(b,v, T)), a, T);
+  *ptv = v; return d;
+}
+
+static GEN
+_F2xqX_mul(void *data,GEN a,GEN b)
+{ return F2xqX_mul(a,b, (GEN) data); }
+
+static GEN
+_F2xqX_sqr(void *data,GEN a)
+{ return F2xqX_sqr(a, (GEN) data); }
+
+GEN
+F2xqX_powu(GEN x, ulong n, GEN T)
+{
+  return gen_powu(x, n, (void*)T, &_F2xqX_sqr, &_F2xqX_mul);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                             F2xqXQ                                **/
@@ -2199,6 +2555,25 @@ F2xqXQ_sqr(GEN x, GEN S, GEN T) {
   return F2xqX_rem(F2xqX_sqr(x,T),S,T);
 }
 
+GEN
+F2xqXQ_invsafe(GEN x, GEN S, GEN T)
+{
+  GEN V, z = F2xqX_extgcd(get_F2xqX_mod(S), x, T, NULL, &V);
+  if (degpol(z)) return NULL;
+  z = F2xq_invsafe(gel(z,2),T);
+  if (!z) return NULL;
+  return F2xqX_F2xq_mul(V, z, T);
+}
+
+GEN
+F2xqXQ_inv(GEN x, GEN S, GEN T)
+{
+  pari_sp av = avma;
+  GEN U = F2xqXQ_invsafe(x, S, T);
+  if (!U) pari_err_INV("F2xqXQ_inv",x);
+  return gerepileupto(av, U);
+}
+
 struct _F2xqXQ {
   GEN T, S;
 };
@@ -2231,12 +2606,12 @@ _F2xqXQ_sqr(void *data, GEN x) {
 static GEN
 _F2xqXQ_zero(void *data) {
   struct _F2xqXQ *d = (struct _F2xqXQ*) data;
-  return pol_0(varn(d->S));
+  return pol_0(get_F2xqX_var(d->S));
 }
 static GEN
 _F2xqXQ_one(void *data) {
   struct _F2xqXQ *d = (struct _F2xqXQ*) data;
-  return pol1_F2xX(varn(d->S),d->T[1]);
+  return pol1_F2xX(get_F2xqX_var(d->S),get_F2x_var(d->T));
 }
 
 static struct bb_algebra F2xqXQ_algebra = { _F2xqXQ_red,
@@ -2247,10 +2622,10 @@ F2xqXQ_pow(GEN x, GEN n, GEN S, GEN T)
 {
   struct _F2xqXQ D;
   long s = signe(n);
-  if (!s) return pol1_F2xX(varn(S), T[1]);
-  if (s < 0) pari_err_IMPL("F2xqXQ_inv");
+  if (!s) return pol1_F2xX(get_F2xqX_var(S), get_F2x_var(T));
+  if (s < 0) x = F2xqXQ_inv(x,S,T);
   if (is_pm1(n)) return s < 0 ? x : gcopy(x);
-  if (degpol(x)>=degpol(S)) x = F2xqX_rem(x,S,T);
+  if (degpol(x) >= get_F2xqX_degree(S)) x = F2xqX_rem(x,S,T);
   D.S = S; D.T = T;
   return gen_pow(x, n, (void*)&D, &_F2xqXQ_sqr, &_F2xqXQ_mul);
 }
@@ -2259,7 +2634,7 @@ GEN
 F2xqXQ_powers(GEN x, long l, GEN S, GEN T)
 {
   struct _F2xqXQ D;
-  int use_sqr = 2*degpol(x) >= degpol(S);
+  int use_sqr = 2*degpol(x) >= get_F2xqX_degree(S);
   D.S = S; D.T = T;
   return gen_powers(x, l, use_sqr, (void*)&D, &_F2xqXQ_sqr, &_F2xqXQ_mul,&_F2xqXQ_one);
 }
@@ -2277,7 +2652,7 @@ GEN
 F2xqX_F2xqXQ_eval(GEN Q, GEN x, GEN S, GEN T)
 {
   struct _F2xqXQ D;
-  int use_sqr = 2*degpol(x) >= degpol(S);
+  int use_sqr = 2*degpol(x) >= get_F2xqX_degree(S);
   D.S = S; D.T = T;
   return gen_bkeval(Q, degpol(Q), x, use_sqr, (void*)&D, &F2xqXQ_algebra,
                                                     _F2xqXQ_cmul);
@@ -2313,7 +2688,7 @@ F2xqXQ_autpow_mul(void * E, GEN x, GEN y)
 }
 
 GEN
-F2xqXQV_autpow(GEN aut, long n, GEN S, GEN T)
+F2xqXQ_autpow(GEN aut, long n, GEN S, GEN T)
 {
   struct _F2xqXQ D;
   D.S = S; D.T = T;
@@ -2345,9 +2720,13 @@ F2xqXQ_auttrace_sqr(void *E, GEN x)
 { return F2xqXQ_auttrace_mul(E, x, x); }
 
 GEN
-F2xqXQV_auttrace(GEN aut, long n, GEN S, GEN T)
+F2xqXQ_auttrace(GEN aut, long n, GEN S, GEN T)
 {
   struct _F2xqXQ D;
   D.S = S; D.T = T;
   return gen_powu(aut,n,&D,F2xqXQ_auttrace_sqr,F2xqXQ_auttrace_mul);
 }
+
+GEN
+F2xqXQV_red(GEN x, GEN S, GEN T)
+{ pari_APPLY_type(t_VEC, F2xqX_rem(gel(x,i), S, T)) }
