@@ -23,15 +23,6 @@ get_FpX_red(GEN T, GEN *B)
   *B = gel(T,1); return gel(T,2);
 }
 
-GEN
-get_FpX_mod(GEN T) { return typ(T)==t_VEC? gel(T,2): T; }
-
-long
-get_FpX_var(GEN T) { return typ(T)==t_VEC? varn(gel(T,2)): varn(T); }
-
-long
-get_FpX_degree(GEN T) { return typ(T)==t_VEC? degpol(gel(T,2)): degpol(T); }
-
 /***********************************************************************/
 /**                                                                   **/
 /**                              FpX                                  **/
@@ -54,7 +45,8 @@ to_Flx(GEN *P, GEN *Q, GEN p)
 {
   ulong pp = uel(p,2);
   *P = ZX_to_Flx(*P, pp);
-  *Q = ZX_to_Flx(*Q, pp); return pp;
+  if(Q) *Q = ZX_to_Flx(*Q, pp);
+  return pp;
 }
 
 static ulong
@@ -85,27 +77,18 @@ FpX_red(GEN z, GEN p)
   for (i=2; i<l; i++) gel(x,i) = modii(gel(z,i),p);
   x[1] = z[1]; return FpX_renormalize(x,l);
 }
-GEN
-FpXV_red(GEN z, GEN p)
-{
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_VEC);
-  for (i=1; i<l; i++) gel(x,i) = FpX_red(gel(z,i), p);
-  return x;
-}
 
 GEN
-FpXT_red(GEN z, GEN p)
+FpXV_red(GEN x, GEN p)
+{ pari_APPLY_type(t_VEC, FpX_red(gel(x,i), p)) }
+
+GEN
+FpXT_red(GEN x, GEN p)
 {
-  if (typ(z) == t_POL)
-    return FpX_red(z, p);
+  if (typ(x) == t_POL)
+    return FpX_red(x, p);
   else
-  {
-    long i,l = lg(z);
-    GEN x = cgetg(l, t_VEC);
-    for (i=1; i<l; i++) gel(x,i) = FpXT_red(gel(z,i), p);
-    return x;
-  }
+    pari_APPLY_type(t_VEC, FpXT_red(gel(x,i), p))
 }
 
 GEN
@@ -124,6 +107,14 @@ FpX_center(GEN T, GEN p, GEN pov2)
   for(i=2; i<l; i++) gel(P,i) = Fp_center(gel(T,i), p, pov2);
   P[1] = T[1]; return P;
 }
+GEN
+FpX_center_i(GEN T, GEN p, GEN pov2)
+{
+  long i, l = lg(T);
+  GEN P = cgetg(l,t_POL);
+  for(i=2; i<l; i++) gel(P,i) = Fp_center_i(gel(T,i), p, pov2);
+  P[1] = T[1]; return P;
+}
 
 GEN
 FpX_add(GEN x,GEN y,GEN p)
@@ -276,14 +267,44 @@ Fp_FpX_sub(GEN x, GEN y, GEN p)
 }
 
 GEN
-FpX_mul(GEN x,GEN y,GEN p) { return FpX_red(ZX_mul(x, y), p); }
+FpX_convol(GEN x, GEN y, GEN p)
+{
+  long lx = lg(x), ly = lg(y), i;
+  GEN z;
+  if (lx < ly) swapspec(x,y, lx,ly);
+  z = cgetg(lx,t_POL); z[1] = x[1];
+  for (i=2; i<ly; i++) gel(z,i) = Fp_mul(gel(x,i),gel(y,i), p);
+  for (   ; i<lx; i++) gel(z,i) = modii(gel(x,i), p);
+  z = ZX_renormalize(z, lx);
+  if (!lgpol(z)) { avma = (pari_sp)(z + lx); return pol_0(varn(x)); }
+  return z;
+}
+
+GEN
+FpX_mul(GEN x,GEN y,GEN p)
+{
+  if (lgefint(p) == 3)
+  {
+    ulong pp = to_Flx(&x, &y, p);
+    return Flx_to_ZX(Flx_mul(x, y, pp));
+  }
+  return FpX_red(ZX_mul(x, y), p);
+}
 
 GEN
 FpX_mulspec(GEN a, GEN b, GEN p, long na, long nb)
 { return FpX_red(ZX_mulspec(a, b, na, nb), p); }
 
 GEN
-FpX_sqr(GEN x,GEN p) { return FpX_red(ZX_sqr(x), p); }
+FpX_sqr(GEN x,GEN p)
+{
+  if (lgefint(p) == 3)
+  {
+    ulong pp = to_Flx(&x, NULL, p);
+    return Flx_to_ZX(Flx_sqr(x, pp));
+  }
+  return FpX_red(ZX_sqr(x), p);
+}
 
 GEN
 FpX_mulu(GEN y, ulong x,GEN p)
@@ -503,7 +524,7 @@ FpX_digits(GEN x, GEN T, GEN p)
 }
 
 GEN
-FpX_fromdigits(GEN x, GEN T, GEN p)
+FpXV_FpX_fromdigits(GEN x, GEN T, GEN p)
 {
   pari_sp av = avma;
   struct _FpXQ D;
@@ -707,25 +728,28 @@ FpX_gcd(GEN x, GEN y, GEN p)
   return gerepileupto(av, FpX_gcd_basecase(x,y,p));
 }
 
-/*Return 1 if gcd can be computed
- * else return a factor of p*/
-
+/* Return NULL if gcd can be computed else return a factor of p */
 GEN
 FpX_gcd_check(GEN x, GEN y, GEN p)
 {
+  pari_sp av = avma;
   GEN a,b,c;
-  pari_sp av=avma;
 
   a = FpX_red(x, p);
   b = FpX_red(y, p);
   while (signe(b))
   {
-    GEN lead = leading_coeff(b);
-    GEN g = gcdii(lead,p);
+    GEN g = gcdii(p, leading_coeff(b));
     if (!equali1(g)) return gerepileuptoint(av,g);
-    c = FpX_rem(a,b,p); a=b; b=c;
+    c = FpX_rem(a,b,p); a = b; b = c;
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM>1)
+        pari_warn(warnmem,"FpX_gcd_check (d = %ld)",degpol(b));
+      gerepileall(av,2,&a,&b);
+    }
   }
-  avma = av; return gen_1;
+  avma = av; return NULL;
 }
 
 static GEN
@@ -824,6 +848,72 @@ FpX_rescale(GEN P, GEN h, GEN p)
 GEN
 FpX_deriv(GEN x, GEN p) { return FpX_red(ZX_deriv(x), p); }
 
+GEN
+FpX_integ(GEN x, GEN p)
+{
+  long i, lx = lg(x);
+  GEN y;
+  if (lx == 2) return ZX_copy(x);
+  y = cgetg(lx+1, t_POL); y[1] = x[1];
+  gel(y,2) = gen_0;
+  for (i=3; i<=lx; i++)
+    gel(y,i) = Fp_div(gel(x,i-1), utoipos(i-2), p);
+  return ZX_renormalize(y, lx+1);;
+}
+
+INLINE GEN
+FpXn_recip(GEN P, long n)
+{ return RgXn_recip_shallow(P, n); }
+
+GEN
+FpX_Newton(GEN P, long n, GEN p)
+{
+  pari_sp av = avma;
+  GEN dP = FpX_deriv(P, p);
+  GEN Q = FpXn_recip(FpX_div(RgX_shift(dP,n), P, p), n);
+  return gerepilecopy(av, Q);
+}
+
+GEN
+FpX_fromNewton(GEN P, GEN p)
+{
+  pari_sp av = avma;
+  long n = itos(modii(constant_coeff(P), p))+1;
+  GEN z = FpX_neg(FpX_integ(RgX_shift(P,-1),p),p);
+  GEN Q = FpXn_recip(FpXn_exp(z, n, p), n);
+  return gerepilecopy(av, Q);
+}
+
+GEN
+FpX_invLaplace(GEN x, GEN p)
+{
+  pari_sp av = avma;
+  long i, e = 0, l = lg(x);
+  GEN t = gen_1, y = cgetg(l,t_POL);
+  y[1] = x[1];
+  for (i=2; i<l; i++)
+  {
+    gel(y,i) = Fp_div(gel(x,i), t, p);
+    e++; t = Fp_mulu(t,e,p);
+  }
+  return gerepilecopy(av, y);
+}
+
+GEN
+FpX_Laplace(GEN x, GEN p)
+{
+  pari_sp av = avma;
+  long i, e = 0, l = lg(x);
+  GEN t = gen_1, y = cgetg(l,t_POL);
+  y[1] = x[1];
+  for (i=2; i<l; i++)
+  {
+    gel(y,i) = Fp_mul(gel(x,i), t, p);
+    e++; t = Fp_mulu(t,e,p);
+  }
+  return gerepilecopy(av, y);
+}
+
 int
 FpX_is_squarefree(GEN f, GEN p)
 {
@@ -917,6 +1007,15 @@ FpX_resultant(GEN a, GEN b, GEN p)
   GEN c,lb, res = gen_1;
 
   if (!signe(a) || !signe(b)) return gen_0;
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flx(&a, &b, p);
+    long r = Flx_resultant(a, b, pp);
+    avma = av;
+    return utoi(r);
+  }
+
   da = degpol(a);
   db = degpol(b);
   if (db > da)
@@ -931,7 +1030,7 @@ FpX_resultant(GEN a, GEN b, GEN p)
     lb = gel(b,db+2);
     c = FpX_rem(a,b, p);
     a = b; b = c; dc = degpol(c);
-    if (dc < 0) { avma = av; return NULL; }
+    if (dc < 0) { avma = av; return gen_0; }
 
     if (both_odd(da,db)) res = subii(p, res);
     if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, da - dc, p), p);
@@ -954,7 +1053,7 @@ FpX_disc(GEN P, GEN p)
   pari_sp av = avma;
   GEN L, dP = FpX_deriv(P,p), D = FpX_resultant(P, dP, p);
   long dd;
-  if (!D || !signe(D)) return gen_0;
+  if (!signe(D)) return gen_0;
   dd = degpol(P) - 2 - degpol(dP); /* >= -1; > -1 iff p | deg(P) */
   L = leading_coeff(P);
   if (dd && !equali1(L))
@@ -1282,17 +1381,6 @@ FpX_rem(GEN x, GEN T, GEN p)
   }
 }
 
-static GEN
-deg2pol_shallow(GEN x2, GEN x1, GEN x0, long v)
-{
-  GEN x = cgetg(5,t_POL);
-  x[1] = evalsigne(1) | evalvarn(v);
-  gel(x,2) = x0;
-  gel(x,3) = x1;
-  gel(x,4) = x2;
-  return normalizepol_lg(x,5);
-}
-
 static GEN
 FpV_producttree(GEN xa, GEN s, GEN p, long vs)
 {
@@ -1859,9 +1947,8 @@ FpXQ_pow_Frobenius(GEN x, GEN n, GEN aut, GEN T, GEN p)
 {
   pari_sp av=avma;
   long d = get_FpX_degree(T);
-  GEN an = absi(n), z, q;
-  if (cmpii(an,p)<0 || cmpis(an,d)<=0)
-    return FpXQ_pow(x, n, T, p);
+  GEN an = absi_shallow(n), z, q;
+  if (cmpii(an,p)<0 || cmpis(an,d)<=0) return FpXQ_pow(x, n, T, p);
   q = powiu(p, d);
   if (dvdii(q, n))
   {
@@ -1876,7 +1963,7 @@ FpXQ_pow_Frobenius(GEN x, GEN n, GEN aut, GEN T, GEN p)
     long m, r, k = is_pm1(bb) ? 1 : bounded_order(p,bb,d);
     if (!k || d-vb<k) return FpXQ_pow(x,n, T, p);
     m = (d-vb)/k; r = (d-vb)%k;
-    u = diviiexact(subis(powiu(p,k),1),bb);
+    u = diviiexact(subiu(powiu(p,k),1),bb);
     v = diviiexact(subii(powiu(p,r+vb),a),b);
     autk = k==1 ? aut: FpXQ_autpow(aut,k,T,p);
     if (r)
@@ -1942,9 +2029,9 @@ Fp_FpXQ_log(GEN a, GEN g, GEN o, GEN T, GEN p)
   if (equali1(a)) return gen_0;
   /* p > 2 */
 
-  ordp = subis(p, 1); /* even */
+  ordp = subiu(p, 1); /* even */
   ord  = get_arith_Z(o);
-  if (!ord) ord = T? subis(powiu(p, get_FpX_degree(T)), 1): ordp;
+  if (!ord) ord = T? subiu(powiu(p, get_FpX_degree(T)), 1): ordp;
   if (equalii(a, ordp)) /* -1 */
     return gerepileuptoint(av, shifti(ord,-1));
   ordp = gcdii(ordp,ord);
@@ -2101,7 +2188,7 @@ FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zeta)
   {
     void *E;
     const struct bb_group *S = get_FpXQ_star(&E,T,p);
-    GEN o = addis(powiu(p,get_FpX_degree(T)),-1);
+    GEN o = subiu(powiu(p,get_FpX_degree(T)),1);
     z = gen_Shanks_sqrtn(a,n,o,zeta,E,S);
     if (!z) return NULL;
     if (!zeta) return gerepileupto(av, z);
@@ -2156,12 +2243,6 @@ FpXQ_charpoly(GEN x, GEN T, GEN p)
 /* cf Shoup 'Efficient Computation of Minimal Polynomials */
 /*          in Algebraic Extensions of Finite Fields'     */
 
-static GEN
-FpXn_mul(GEN a, GEN b, long n, GEN p)
-{
-  return FpX_red(RgXn_red_shallow(ZX_mul(a, b), n), p);
-}
-
 /* Let v a linear form, return the linear form z->v(tau*z)
    that is, v*(M_tau) */
 
@@ -2341,7 +2422,7 @@ gener_FpXQ(GEN T, GEN p, GEN *po)
   Lq = leafcopy( gel(o, 1) );
   for (i = j = 1; i < lg(Lq); i++)
   {
-    if (remii(p_1, gel(Lq,i)) == gen_0) continue;
+    if (dvdii(p_1, gel(Lq,i))) continue;
     gel(Lq,j++) = diviiexact(N, gel(Lq,i));
   }
   setlg(Lq, j);
@@ -2354,25 +2435,6 @@ gener_FpXQ(GEN T, GEN p, GEN *po)
   return g;
 }
 
-#if 0 /* generic version: slower */
-GEN
-gener_FpXQ2(GEN T, GEN p, GEN *po)
-{
-  pari_sp av = avma;
-  void *E;
-  long f = get_FpX_degree(T);
-  GEN g, o = factor_pn_1(p,f);
-  const struct bb_group *S = get_FpXQ_star(&E,T,p);
-  g = gen_gener(o,E,S);
-  if (!po) g = gerepilecopy(av, g);
-  else {
-    *po = mkvec2(powiu(p,f), o);
-    gerepileall(av, 2, &g, po);
-  }
-  return g;
-}
-#endif
-
 GEN
 gener_FpXQ_local(GEN T, GEN p, GEN L)
 {
@@ -2400,3 +2462,110 @@ gener_FpXQ_local(GEN T, GEN p, GEN L)
   setlg(Lq, iq);
   return gener_FpXQ_i(T, p, p_1, Lp, Lq);
 }
+
+/***********************************************************************/
+/**                                                                   **/
+/**                              FpXn                                 **/
+/**                                                                   **/
+/***********************************************************************/
+
+INLINE GEN
+FpXn_red(GEN a, long n)
+{ return RgXn_red_shallow(a, n); }
+
+GEN
+FpXn_mul(GEN a, GEN b, long n, GEN p)
+{
+  return FpX_red(ZXn_mul(a, b, n), p);
+}
+
+GEN
+FpXn_sqr(GEN a, long n, GEN p)
+{
+  return FpX_red(ZXn_sqr(a, n), p);
+}
+
+GEN
+FpXn_exp(GEN h, long e, GEN p)
+{
+  pari_sp av = avma, av2;
+  long v = varn(h), n=1;
+  GEN f = pol_1(v), g = pol_1(v);
+  ulong mask = quadratic_prec_mask(e);
+  av2 = avma;
+  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
+    pari_err_DOMAIN("FpXn_exp","valuation", "<", gen_1, h);
+  for (;mask>1;)
+  {
+    GEN q, w;
+    long n2 = n;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    g = FpX_sub(FpX_mulu(g,2,p), FpXn_mul(f, FpXn_sqr(g, n2, p), n2, p), p);
+    q = FpX_deriv(FpXn_red(h,n2), p);
+    w = FpX_add(q, FpXn_mul(g, FpX_sub(FpX_deriv(f, p), FpXn_mul(f,q,n-1, p), p),n-1, p), p);
+    f = FpX_add(f, FpXn_mul(f, FpX_sub(FpXn_red(h, n), FpX_integ(w,p), p), n, p), p);
+    if (gc_needed(av2,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXn_exp, e = %ld", n);
+      gerepileall(av2, 2, &f, &g);
+    }
+  }
+  return gerepileupto(av, f);
+}
+
+/* (f*g) \/ x^n */
+static GEN
+FpX_mulhigh_i(GEN f, GEN g, long n, GEN p)
+{
+  return RgX_shift_shallow(FpX_mul(f,g, p),-n);
+}
+
+static GEN
+FpXn_mulhigh(GEN f, GEN g, long n2, long n, GEN p)
+{
+  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
+  return FpX_add(FpX_mulhigh_i(fl, g, n2, p), FpXn_mul(fh, g, n - n2, p), p);
+}
+
+GEN
+FpXn_inv(GEN f, long e, GEN p)
+{
+  pari_sp av = avma, av2;
+  ulong mask;
+  GEN W, a;
+  long v = varn(f), n = 1;
+
+  if (!signe(f)) pari_err_INV("FpXn_inv",f);
+  a = Fp_inv(gel(f,2), p);
+  if (e == 1) return scalarpol(a, v);
+  else if (e == 2)
+  {
+    GEN b;
+    if (degpol(f) <= 0) return scalarpol(a, v);
+    b = Fp_neg(gel(f,3),p);
+    if (signe(b)==0) return scalarpol(a, v);
+    if (!is_pm1(a)) b = Fp_mul(b, Fp_sqr(a, p), p);
+    W = deg1pol_shallow(b, a, v);
+    return gerepilecopy(av, W);
+  }
+  W = scalarpol_shallow(Fp_inv(gel(f,2), p),v);
+  mask = quadratic_prec_mask(e);
+  av2 = avma;
+  for (;mask>1;)
+  {
+    GEN u, fr;
+    long n2 = n;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    fr = FpXn_red(f, n);
+    u = FpXn_mul(W, FpXn_mulhigh(fr, W, n2, n, p), n-n2, p);
+    W = FpX_sub(W, RgX_shift_shallow(u, n2), p);
+    if (gc_needed(av2,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"FpXn_inv, e = %ld", n);
+      W = gerepileupto(av2, W);
+    }
+  }
+  return gerepileupto(av, W);
+}
