@@ -72,6 +72,11 @@ gen_bkeval_powers(GEN P, long d, GEN V, void *E, const struct bb_algebra *ff,
   if (d < 0) return ff->zero(E);
   if (d < l) return gerepileupto(av, gen_RgXQ_eval_powers(P,V,0,d,E,ff,cmul));
   if (l<2) pari_err_DOMAIN("gen_RgX_bkeval_powers", "#powers", "<",gen_2,V);
+  if (DEBUGLEVEL>=8)
+  {
+    long cnt = 1 + (d - l) / (l-1);
+    err_printf("RgX_RgXQV_eval(%ld/%ld): %ld RgXQ_mul\n", d, l-1, cnt);
+  }
   d -= l;
   z = gen_RgXQ_eval_powers(P,V,d+1,l-1,E,ff,cmul);
   while (d >= l-1)
@@ -84,11 +89,6 @@ gen_bkeval_powers(GEN P, long d, GEN V, void *E, const struct bb_algebra *ff,
   }
   u = gen_RgXQ_eval_powers(P,V,0,d,E,ff,cmul);
   z = ff->add(E,u, ff->mul(E,z,gel(V,d+2)));
-  if (DEBUGLEVEL>=8)
-  {
-    long cnt = 1 + (d - l) / (l-1);
-    err_printf("RgX_RgXQV_eval: %ld RgXQ_mul [%ld]\n", cnt, l-1);
-  }
   return gerepileupto(av, ff->red(E,z));
 }
 
@@ -159,6 +159,24 @@ get_Rg_algebra(void)
   return &Rg_algebra;
 }
 
+static struct bb_ring Rg_ring = {  _gen_add, _gen_mul, _gen_sqr };
+
+static GEN
+_RgX_divrem(void *E, GEN x, GEN y, GEN *r)
+{
+  (void) E;
+  return RgX_divrem(x, y, r);
+}
+
+GEN
+RgX_digits(GEN x, GEN T)
+{
+  pari_sp av = avma;
+  long d = degpol(T), n = (lgpol(x)+d-1)/d;
+  GEN z = gen_digits(x,T,n,NULL, &Rg_ring, _RgX_divrem);
+  return gerepileupto(av, z);
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                         RgX                                     */
@@ -179,10 +197,10 @@ RgX_equal(GEN x, GEN y)
 /* Returns 1 in the base ring over which x is defined */
 /* HACK: this also works for t_SER */
 GEN
-RgX_get_1(GEN x)
+Rg_get_1(GEN x)
 {
   GEN p, T;
-  long i, lx, tx = RgX_type(x, &p, &T, &lx);
+  long i, lx, tx = Rg_type(x, &p, &T, &lx);
   if (RgX_type_is_composite(tx))
     RgX_type_decode(tx, &i /*junk*/, &tx);
   switch(tx)
@@ -196,10 +214,10 @@ RgX_get_1(GEN x)
 /* Returns 0 in the base ring over which x is defined */
 /* HACK: this also works for t_SER */
 GEN
-RgX_get_0(GEN x)
+Rg_get_0(GEN x)
 {
   GEN p, T;
-  long i, lx, tx = RgX_type(x, &p, &T, &lx);
+  long i, lx, tx = Rg_type(x, &p, &T, &lx);
   if (RgX_type_is_composite(tx))
     RgX_type_decode(tx, &i /*junk*/, &tx);
   switch(tx)
@@ -242,9 +260,9 @@ RgX_unscale(GEN P, GEN h)
   }
   return Q;
 }
-/* P a ZX, h a t_INT. Return P(h * x), not memory clean; optimize for h = -1 */
+/* P a ZX, Return P(h * x), not memory clean; optimize for h = -1 */
 GEN
-ZX_unscale(GEN P, GEN h)
+ZX_z_unscale(GEN P, long h)
 {
   long i, l = lg(P);
   GEN Q = cgetg(l, t_POL);
@@ -252,8 +270,8 @@ ZX_unscale(GEN P, GEN h)
   if (l == 2) return Q;
   gel(Q,2) = gel(P,2);
   if (l == 3) return Q;
-  if (equalim1(h))
-    for (i=3; i<l; i++)
+  if (h == -1)
+    for (i = 3; i < l; i++)
     {
       gel(Q,i) = negi(gel(P,i));
       if (++i == l) break;
@@ -261,16 +279,38 @@ ZX_unscale(GEN P, GEN h)
     }
   else
   {
-    GEN hi = h;
-    gel(Q,3) = mulii(gel(P,3), hi);
-    for (i=4; i<l; i++)
+    GEN hi;
+    gel(Q,3) = mulis(gel(P,3), h);
+    hi = sqrs(h);
+    for (i = 4; i < l; i++)
     {
-      hi = mulii(hi,h);
       gel(Q,i) = mulii(gel(P,i), hi);
+      if (i != l-1) hi = mulis(hi,h);
     }
   }
   return Q;
 }
+/* P a ZX, h a t_INT. Return P(h * x), not memory clean; optimize for h = -1 */
+GEN
+ZX_unscale(GEN P, GEN h)
+{
+  long i, l;
+  GEN Q, hi;
+  i = itos_or_0(h); if (i) return ZX_z_unscale(P, i);
+  l = lg(P); Q = cgetg(l, t_POL);
+  Q[1] = P[1];
+  if (l == 2) return Q;
+  gel(Q,2) = gel(P,2);
+  if (l == 3) return Q;
+  hi = h;
+  gel(Q,3) = mulii(gel(P,3), hi);
+  for (i = 4; i < l; i++)
+  {
+    hi = mulii(hi,h);
+    gel(Q,i) = mulii(gel(P,i), hi);
+  }
+  return Q;
+}
 /* P a ZX. Return P(x << n), not memory clean */
 GEN
 ZX_unscale2n(GEN P, long n)
@@ -516,7 +556,7 @@ QXQ_to_mod_copy(GEN x, GEN T)
       if (d == 0) return gcopy(gel(x,2));
       return mkpolmod(RgX_copy(x), T);
     default: pari_err_TYPE("QXQ_to_mod",x);
-             return NULL;/* not reached */
+             return NULL;/* LCOV_EXCL_LINE */
   }
 }
 /* pure shallow version */
@@ -534,7 +574,7 @@ QXQ_to_mod(GEN x, GEN T)
       if (d == 0) return gel(x,2);
       return mkpolmod(x, T);
     default: pari_err_TYPE("QXQ_to_mod",x);
-             return NULL;/* not reached */
+             return NULL;/* LCOV_EXCL_LINE */
   }
 }
 /* T a ZX, z lifted from (Q[Y]/(T(Y)))[X], apply QXQ_to_mod_copy to all coeffs.
@@ -575,6 +615,25 @@ QXQV_to_mod(GEN V, GEN T)
   return z;
 }
 
+/* Apply QXQ_to_mod_copy to all entries. Memory-clean ! */
+GEN
+QXQC_to_mod_shallow(GEN V, GEN T)
+{
+  long i, l = lg(V);
+  GEN z = cgetg(l, t_COL);
+  for (i=1;i<l; i++) gel(z,i) = QXQ_to_mod(gel(V,i), T);
+  return z;
+}
+
+GEN
+QXQM_to_mod_shallow(GEN V, GEN T)
+{
+  long i, l = lg(V);
+  GEN z = cgetg(l, t_MAT);
+  for (i=1; i<l; i++) gel(z,i) = QXQC_to_mod_shallow(gel(V,i), T);
+  return z;
+}
+
 GEN
 RgX_renormalize_lg(GEN x, long lx)
 {
@@ -638,31 +697,17 @@ Rg_to_RgC(GEN x, long N)
 /* vector of polynomials (in v) whose coeffs are given by the columns of x */
 GEN
 RgM_to_RgXV(GEN x, long v)
-{
-  long j, lx = lg(x);
-  GEN y = cgetg(lx, t_VEC);
-  for (j=1; j<lx; j++) gel(y,j) = RgV_to_RgX(gel(x,j), v);
-  return y;
-}
+{ pari_APPLY_type(t_VEC, RgV_to_RgX(gel(x,i), v)) }
 
 /* matrix whose entries are given by the coeffs of the polynomials in
  * vector v (considered as degree n-1 polynomials) */
 GEN
-RgV_to_RgM(GEN v, long n)
-{
-  long j, N = lg(v);
-  GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = Rg_to_RgC(gel(v,j), n);
-  return y;
-}
+RgV_to_RgM(GEN x, long n)
+{ pari_APPLY_type(t_MAT, Rg_to_RgC(gel(x,i), n)) }
+
 GEN
-RgXV_to_RgM(GEN v, long n)
-{
-  long j, N = lg(v);
-  GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = RgX_to_RgC(gel(v,j), n);
-  return y;
-}
+RgXV_to_RgM(GEN x, long n)
+{ pari_APPLY_type(t_MAT, RgX_to_RgC(gel(x,i), n)) }
 
 /* polynomial (in v) of polynomials (in w) whose coeffs are given by the columns of x */
 GEN
@@ -738,23 +783,17 @@ RgXY_degreex(GEN b)
 GEN
 RgXn_red_shallow(GEN a, long n)
 {
-  long i, L, l = lg(a);
+  long i, L = n+2, l = lg(a);
   GEN  b;
-  if (l == 2 || !n) return pol_0(varn(a));
-  L = n+2; if (L > l) L = l;
+  if (L >= l) return a; /* deg(x) < n */
   b = cgetg(L, t_POL); b[1] = a[1];
   for (i=2; i<L; i++) gel(b,i) = gel(a,i);
   return normalizepol_lg(b,L);
 }
 
 GEN
-RgXnV_red_shallow(GEN P, long n)
-{
-  long i, l = lg(P);
-  GEN Q = cgetg(l, t_VEC);
-  for (i=1; i<l; i++) gel(Q,i) = RgXn_red_shallow(gel(P,i), n);
-  return Q;
-}
+RgXnV_red_shallow(GEN x, long n)
+{ pari_APPLY_type(t_VEC, RgXn_red_shallow(gel(x,i), n)) }
 
 /* return (x * X^n). Shallow */
 GEN
@@ -867,21 +906,21 @@ RgX_valrem_inexact(GEN x, GEN *Z)
 }
 
 GEN
-RgXQC_red(GEN P, GEN T)
-{
-  long i, l = lg(P);
-  GEN Q = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(Q,i) = grem(gel(P,i), T);
-  return Q;
-}
+RgXQC_red(GEN x, GEN T)
+{ pari_APPLY_type(t_COL, grem(gel(x,i), T)) }
 
 GEN
-RgXQV_red(GEN P, GEN T)
+RgXQV_red(GEN x, GEN T)
+{ pari_APPLY_type(t_VEC, grem(gel(x,i), T)) }
+
+GEN
+RgXQM_red(GEN x, GEN T)
+{ pari_APPLY_same(RgXQC_red(gel(x,i), T)) }
+
+GEN
+RgXQM_mul(GEN P, GEN Q, GEN T)
 {
-  long i, l = lg(P);
-  GEN Q = cgetg(l, t_VEC);
-  for (i=1; i<l; i++) gel(Q,i) = grem(gel(P,i), T);
-  return Q;
+  return RgXQM_red(RgM_mul(P, Q), T);
 }
 
 GEN
@@ -910,7 +949,8 @@ GEN
 RgX_recipspec_shallow(GEN x, long l, long n)
 {
   long i;
-  GEN z=cgetg(n+2,t_POL)+2;
+  GEN z = cgetg(n+2,t_POL);
+  z[1] = 0; z += 2;
   for(i=0; i<l; i++)
     gel(z,n-i-1) = gel(x,i);
   for(   ; i<n; i++)
@@ -918,6 +958,14 @@ RgX_recipspec_shallow(GEN x, long l, long n)
   return normalizepol_lg(z-2,n+2);
 }
 
+GEN
+RgXn_recip_shallow(GEN P, long n)
+{
+  GEN Q = RgX_recipspec_shallow(P+2, lgpol(P), n);
+  setvarn(Q, varn(P));
+  return Q;
+}
+
 /* return coefficients s.t x = x_0 X^n + ... + x_n */
 GEN
 RgX_recip(GEN x)
@@ -1068,31 +1116,49 @@ RgX_debug_spec(GEN x, long nx)
 #endif
 
 /* generic multiplication */
-
-static GEN
-addpol(GEN x, GEN y, long lx, long ly)
+GEN
+RgX_addspec_shallow(GEN x, GEN y, long nx, long ny)
 {
-  long i,lz;
-  GEN z;
-
-  if (ly>lx) swapspec(x,y, lx,ly);
-  lz = lx+2; z = cgetg(lz,t_POL) + 2;
-  for (i=0; i<ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i));
-  for (   ; i<lx; i++) gel(z,i) = gel(x,i);
-  z -= 2; z[1]=0; return normalizepol_lg(z, lz);
+  GEN z, t;
+  long i;
+  if (nx == ny) {
+    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    return normalizepol_lg(z, nx+2);
+  }
+  if (ny < nx) {
+    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < ny; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    for (   ; i < nx; i++) gel(t,i) = gel(x,i);
+    return normalizepol_lg(z, nx+2);
+  } else {
+    z = cgetg(ny+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    for (   ; i < ny; i++) gel(t,i) = gel(y,i);
+    return normalizepol_lg(z, ny+2);
+  }
 }
-
-static GEN
-addpolcopy(GEN x, GEN y, long lx, long ly)
+GEN
+RgX_addspec(GEN x, GEN y, long nx, long ny)
 {
-  long i,lz;
-  GEN z;
-
-  if (ly>lx) swapspec(x,y, lx,ly);
-  lz = lx+2; z = cgetg(lz,t_POL) + 2;
-  for (i=0; i<ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i));
-  for (   ; i<lx; i++) gel(z,i) = gcopy(gel(x,i));
-  z -= 2; z[1]=0; return normalizepol_lg(z, lz);
+  GEN z, t;
+  long i;
+  if (nx == ny) {
+    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    return normalizepol_lg(z, nx+2);
+  }
+  if (ny < nx) {
+    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < ny; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    for (   ; i < nx; i++) gel(t,i) = gcopy(gel(x,i));
+    return normalizepol_lg(z, nx+2);
+  } else {
+    z = cgetg(ny+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
+    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
+    for (   ; i < ny; i++) gel(t,i) = gcopy(gel(y,i));
+    return normalizepol_lg(z, ny+2);
+  }
 }
 
 /* Return the vector of coefficients of x, where we replace rational 0s by NULL
@@ -1150,16 +1216,16 @@ RgX_mulspec_basecase(GEN x, GEN y, long nx, long ny, long v)
 
 /* return (x * X^d) + y. Assume d > 0 */
 GEN
-addmulXn(GEN x, GEN y, long d)
+RgX_addmulXn_shallow(GEN x0, GEN y0, long d)
 {
-  GEN xd, yd, zd;
+  GEN x, y, xd, yd, zd;
   long a, lz, nx, ny;
 
-  if (!signe(x)) return y;
-  ny = lgpol(y);
-  nx = lgpol(x);
+  if (!signe(x0)) return y0;
+  ny = lgpol(y0);
+  nx = lgpol(x0);
   zd = (GEN)avma;
-  x += 2; y += 2; a = ny-d;
+  x = x0 + 2; y = y0 + 2; a = ny-d;
   if (a <= 0)
   {
     lz = nx+d+2;
@@ -1171,35 +1237,25 @@ addmulXn(GEN x, GEN y, long d)
   else
   {
     xd = new_chunk(d); yd = y+d;
-    x = addpol(x,yd, nx,a);
+    x = RgX_addspec_shallow(x,yd, nx,a);
     lz = (a>nx)? ny+2: lg(x)+d;
     x += 2; while (xd > x) *--zd = *--xd;
   }
   while (yd > y) *--zd = *--yd;
-  *--zd = evalsigne(1);
+  *--zd = x0[1];
   *--zd = evaltyp(t_POL) | evallg(lz); return zd;
 }
-
 GEN
-addshiftpol(GEN x, GEN y, long d)
+RgX_addmulXn(GEN x0, GEN y0, long d)
 {
-  long v = varn(x);
-  x = addmulXn(x,y,d);
-  setvarn(x,v); return x;
-}
-
-/* as above, producing a clean malloc */
-static GEN
-addmulXncopy(GEN x, GEN y, long d)
-{
-  GEN xd, yd, zd;
+  GEN x, y, xd, yd, zd;
   long a, lz, nx, ny;
 
-  if (!signe(x)) return RgX_copy(y);
-  nx = lgpol(x);
-  ny = lgpol(y);
+  if (!signe(x0)) return RgX_copy(y0);
+  nx = lgpol(x0);
+  ny = lgpol(y0);
   zd = (GEN)avma;
-  x += 2; y += 2; a = ny-d;
+  x = x0 + 2; y = y0 + 2; a = ny-d;
   if (a <= 0)
   {
     lz = nx+d+2;
@@ -1211,12 +1267,12 @@ addmulXncopy(GEN x, GEN y, long d)
   else
   {
     xd = new_chunk(d); yd = y+d;
-    x = addpolcopy(x,yd, nx,a);
+    x = RgX_addspec(x,yd, nx,a);
     lz = (a>nx)? ny+2: lg(x)+d;
     x += 2; while (xd > x) *--zd = *--xd;
   }
   while (yd > y) gel(--zd,0) = gcopy(gel(--yd,0));
-  *--zd = evalsigne(1);
+  *--zd = x0[1];
   *--zd = evaltyp(t_POL) | evallg(lz); return zd;
 }
 
@@ -1242,8 +1298,8 @@ RgXn_mul_basecase(GEN x, GEN y, long n)
   return normalizepol_lg(z - 2, lz);
 }
 /* Mulders / Karatsuba product f*g mod t^n (Hanrot-Zimmermann variant) */
-GEN
-RgXn_mul(GEN f, GEN g, long n)
+static GEN
+RgXn_mul2(GEN f, GEN g, long n)
 {
   pari_sp av = avma;
   GEN fe,fo, ge,go, l,h,m;
@@ -1265,8 +1321,30 @@ RgXn_mul(GEN f, GEN g, long n)
   /* deg(t^2 h(t^2)) <= 2n0 <= n, truncate to < n */
   if (2*degpol(h)+2 == n) h = normalizepol_lg(h, lg(h)-1);
   h = RgX_inflate(h,2);
-  h = addmulXncopy(addmulXn(h,m,1), l,1);
-  setvarn(h, varn(f)); return gerepileupto(av, h);
+  h = RgX_addmulXn(RgX_addmulXn_shallow(h,m,1), l,1);
+  return gerepileupto(av, h);
+}
+/* (f*g) \/ x^n */
+static GEN
+RgX_mulhigh_i2(GEN f, GEN g, long n)
+{
+  long d = degpol(f)+degpol(g) + 1 - n;
+  GEN h;
+  if (d <= 2) return RgX_shift_shallow(RgX_mul(f,g), -n);
+  h = RgX_recip_shallow(RgXn_mul(RgX_recip_shallow(f),
+                                 RgX_recip_shallow(g), d));
+  return RgX_shift_shallow(h, d-1-degpol(h)); /* possibly (fg)(0) = 0 */
+}
+
+/* (f*g) \/ x^n */
+static GEN
+RgX_sqrhigh_i2(GEN f, long n)
+{
+  long d = 2*degpol(f)+ 1 - n;
+  GEN h;
+  if (d <= 2) return RgX_shift_shallow(RgX_sqr(f), -n);
+  h = RgX_recip_shallow(RgXn_sqr(RgX_recip_shallow(f), d));
+  return RgX_shift_shallow(h, d-1-degpol(h)); /* possibly (fg)(0) = 0 */
 }
 
 /* fast product (Karatsuba) of polynomials a,b. These are not real GENs, a+2,
@@ -1301,19 +1379,19 @@ RgX_mulspec(GEN a, GEN b, long na, long nb)
     c = RgX_mulspec(a,b,n0a,n0b);
     c0 = RgX_mulspec(a0,b0, na,nb);
 
-    c2 = addpol(a0,a, na,n0a);
-    c1 = addpol(b0,b, nb,n0b);
+    c2 = RgX_addspec_shallow(a0,a, na,n0a);
+    c1 = RgX_addspec_shallow(b0,b, nb,n0b);
 
     c1 = RgX_mulspec(c1+2,c2+2, lgpol(c1),lgpol(c2));
     c2 = RgX_sub(c1, RgX_add(c0,c));
-    c0 = addmulXn(c0, c2, n0);
+    c0 = RgX_addmulXn_shallow(c0, c2, n0);
   }
   else
   {
     c = RgX_mulspec(a,b,n0a,nb);
     c0 = RgX_mulspec(a0,b,na,nb);
   }
-  c0 = addmulXncopy(c0,c,n0);
+  c0 = RgX_addmulXn(c0,c,n0);
   return RgX_shift_inplace(gerepileupto(av,c0), v);
 }
 
@@ -1376,13 +1454,13 @@ RgXn_sqr_basecase(GEN x, long n)
   z -= 2; return normalizepol_lg(z, lz);
 }
 /* Mulders / Karatsuba product f^2 mod t^n (Hanrot-Zimmermann variant) */
-GEN
-RgXn_sqr(GEN f, long n)
+static GEN
+RgXn_sqr2(GEN f, long n)
 {
   pari_sp av = avma;
   GEN fe,fo, l,h,m;
   long n0, n1;
-  if (2*degpol(f) < n) return RgX_sqr(f);
+  if (2*degpol(f) < n) return RgX_sqr_i(f);
   if (n < 80) return RgXn_sqr_basecase(f,n);
   n0 = n>>1; n1 = n-n0;
   RgX_even_odd(f, &fe, &fo);
@@ -1398,10 +1476,9 @@ RgXn_sqr(GEN f, long n)
   /* deg(t^2 h(t^2)) <= 2n0 <= n, truncate to < n */
   if (2*degpol(h)+2 == n) h = normalizepol_lg(h, lg(h)-1);
   h = RgX_inflate(h,2);
-  h = addmulXncopy(addmulXn(h,m,1), l,1);
-  setvarn(h, varn(f)); return gerepileupto(av, h);
+  h = RgX_addmulXn(RgX_addmulXn_shallow(h,m,1), l,1);
+  return gerepileupto(av, h);
 }
-
 GEN
 RgX_sqrspec(GEN a, long na)
 {
@@ -1419,8 +1496,8 @@ RgX_sqrspec(GEN a, long na)
   c = RgX_sqrspec(a,n0a);
   c0 = RgX_sqrspec(a0,na);
   c1 = gmul2n(RgX_mulspec(a0,a, na,n0a), 1);
-  c0 = addmulXn(c0,c1, n0);
-  c0 = addmulXncopy(c0,c,n0);
+  c0 = RgX_addmulXn_shallow(c0,c1, n0);
+  c0 = RgX_addmulXn(c0,c,n0);
   return RgX_shift_inplace(gerepileupto(av,c0), v);
 }
 
@@ -1430,23 +1507,23 @@ RgX_mul_normalized(GEN A, long a, GEN B, long b)
 {
   GEN z = RgX_mul(A, B);
   if (a < b)
-    z = addmulXn(addmulXn(A, B, b-a), z, a);
+    z = RgX_addmulXn_shallow(RgX_addmulXn_shallow(A, B, b-a), z, a);
   else if (a > b)
-    z = addmulXn(addmulXn(B, A, a-b), z, b);
+    z = RgX_addmulXn_shallow(RgX_addmulXn_shallow(B, A, a-b), z, b);
   else
-    z = addmulXn(RgX_add(A, B), z, a);
-  setvarn(z,varn(A)); return z;
+    z = RgX_addmulXn_shallow(RgX_add(A, B), z, a);
+  return z;
 }
 
 GEN
-RgX_mul(GEN x, GEN y)
+RgX_mul_i(GEN x, GEN y)
 {
-  GEN z = RgX_mulspec(y+2, x+2, lgpol(y), lgpol(x));
-  setvarn(z,varn(x)); return z;
+  GEN z = RgX_mulspec(x+2, y+2, lgpol(x), lgpol(y));
+  setvarn(z, varn(x)); return z;
 }
 
 GEN
-RgX_sqr(GEN x)
+RgX_sqr_i(GEN x)
 {
   GEN z = RgX_sqrspec(x+2, lgpol(x));
   setvarn(z,varn(x)); return z;
@@ -1477,16 +1554,14 @@ RgX_Rg_div(GEN x, GEN y) {
 GEN
 RgX_normalize(GEN x)
 {
-  GEN d = NULL;
+  GEN z, d = NULL;
   long i, n = lg(x)-1;
-  for (i = n; i > 1; i--)
-  {
-    d = gel(x,i);
-    if (!gequal0(d)) break;
-  }
+  for (i = n; i > 1; i--) { d = gel(x,i); if (!gequal0(d)) break; }
   if (i == 1) return pol_0(varn(x));
   if (i == n && isint1(d)) return x;
-  return normalizepol_lg(RgX_Rg_div(x, d), i+1);
+  n = i; z = cgetg(n+1, t_POL); z[1] = x[1];
+  for (i=2; i<n; i++) gel(z,i) = gdiv(gel(x,i),d);
+  gel(z,n) = Rg_get_1(d); return z;
 }
 GEN
 RgX_divs(GEN x, long y) {
@@ -1512,17 +1587,16 @@ RgX_div_by_X_x(GEN a, GEN x, GEN *r)
   return z;
 }
 /* Polynomial division x / y:
- *   if z = ONLY_REM  return remainder, otherwise return quotient
- *   if z != NULL set *z to remainder
- *   *z is the last object on stack (and thus can be disposed of with cgiv
- *   instead of gerepile) */
+ *   if pr = ONLY_REM return remainder, otherwise return quotient
+ *   if pr = ONLY_DIVIDES return quotient if division is exact, else NULL
+ *   if pr != NULL set *pr to remainder, as the last object on stack */
 /* assume, typ(x) = typ(y) = t_POL, same variable */
-GEN
-RgX_divrem(GEN x, GEN y, GEN *pr)
+static GEN
+RgX_divrem_i(GEN x, GEN y, GEN *pr)
 {
   pari_sp avy, av, av1;
   long dx,dy,dz,i,j,sx,lr;
-  GEN z,p1,p2,rem,y_lead,mod;
+  GEN z,p1,p2,rem,y_lead,mod,p;
   GEN (*f)(GEN,GEN);
 
   if (!signe(y)) pari_err_INV("RgX_divrem",y);
@@ -1558,6 +1632,18 @@ RgX_divrem(GEN x, GEN y, GEN *pr)
 
   /* x,y in R[X], y non constant */
   av = avma;
+  p = NULL;
+  if (RgX_is_FpX(x, &p) && RgX_is_FpX(y, &p) && p)
+  {
+    z = FpX_divrem(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p, pr);
+    if (!z) { avma = av; return NULL; }
+    z = FpX_to_mod(z, p);
+    if (!pr || pr == ONLY_REM || pr == ONLY_DIVIDES)
+      return gerepileupto(av, z);
+    *pr = FpX_to_mod(*pr, p);
+    gerepileall(av, 2, pr, &z);
+    return z;
+  }
   switch(typ(y_lead))
   {
     case t_REAL:
@@ -1692,6 +1778,13 @@ RgX_divrem(GEN x, GEN y, GEN *pr)
   }
 }
 
+GEN
+RgX_divrem(GEN x, GEN y, GEN *pr)
+{
+  if (pr == ONLY_REM) return RgX_rem(x, y);
+  return RgX_divrem_i(x, y, pr);
+}
+
 /* x and y in (R[Y]/T)[X]  (lifted), T in R[Y]. y preferably monic */
 GEN
 RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
@@ -2047,6 +2140,18 @@ RgXQX_sqr(GEN x, GEN T)
   return RgXQX_red(RgX_sqr(x), T);
 }
 
+GEN
+RgXQX_powers(GEN P, long n, GEN T)
+{
+  GEN v = cgetg(n+2, t_VEC);
+  long i;
+  gel(v, 1) = pol_1(varn(T));
+  if (n==0) return v;
+  gel(v, 2) = gcopy(P);
+  for (i = 2; i <= n; i++) gel(v,i+1) = RgXQX_mul(P, gel(v,i), T);
+  return v;
+}
+
 static GEN
 _add(void *data, GEN x, GEN y) { (void)data; return RgX_add(x, y); }
 static GEN
@@ -2168,18 +2273,45 @@ RgXn_eval(GEN Q, GEN x, long n)
   return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&S,&RgXn_algebra,_cmul);
 }
 
-GEN
-RgXn_inv(GEN f, long e)
+/* (f*g mod t^n) \ t^n2, assuming 2*n2 >= n */
+static GEN
+RgXn_mulhigh(GEN f, GEN g, long n2, long n)
 {
-  pari_sp av = avma, av2;
+  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
+  return RgX_add(RgX_mulhigh_i(fl, g, n2), RgXn_mul(fh, g, n - n2));
+}
+
+/* (f^2 mod t^n) \ t^n2, assuming 2*n2 >= n */
+static GEN
+RgXn_sqrhigh(GEN f, long n2, long n)
+{
+  GEN F = RgX_blocks(f, n2, 2), fl = gel(F,1), fh = gel(F,2);
+  return RgX_add(RgX_mulhigh_i(fl, f, n2), RgXn_mul(fh, f, n - n2));
+}
+
+GEN
+RgXn_inv_i(GEN f, long e)
+{
+  pari_sp av;
   ulong mask;
-  GEN W;
-  long v = varn(f), n=1;
-  if (signe(f)==0)
-    pari_err_INV("RgXn_inv",f);
-  W = scalarpol(ginv(gel(f,2)),v);
+  GEN W, a;
+  long v = varn(f), n = 1;
+
+  if (!signe(f)) pari_err_INV("RgXn_inv",f);
+  a = ginv(gel(f,2));
+  if (e == 1) return scalarpol(a, v);
+  else if (e == 2)
+  {
+    GEN b;
+    if (degpol(f) <= 0 || gequal0(b = gel(f,3))) return scalarpol(a, v);
+    av = avma; b = gneg(b);
+    if (!gequal1(a)) b = gmul(b, gsqr(a));
+    W = deg1pol_shallow(b, a, v);
+    return gcopy(W);
+  }
+  W = scalarpol_shallow(a,v);
   mask = quadratic_prec_mask(e);
-  av2 = avma;
+  av = avma;
   for (;mask>1;)
   {
     GEN u, fr;
@@ -2187,15 +2319,66 @@ RgXn_inv(GEN f, long e)
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
     fr = RgXn_red_shallow(f, n);
-    u = RgX_shift_shallow(RgXn_mul(W, fr, n), -n2);
-    W = RgX_sub(W, RgX_shift_shallow(RgXn_mul(u, W, n-n2), n2));
-    if (gc_needed(av2,2))
+    u = RgXn_mul(W, RgXn_mulhigh(fr, W, n2, n), n-n2);
+    W = RgX_sub(W, RgX_shift_shallow(u, n2));
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_inv, e = %ld", n);
-      W = gerepileupto(av2, W);
+      W = gerepileupto(av, W);
     }
   }
-  return gerepileupto(av, W);
+  return W;
+}
+
+static GEN
+RgXn_inv_FpX(GEN x, long e, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flxn_inv(RgX_to_Flx(x, pp), e, pp));
+  }
+  else
+    r = FpXn_inv(RgX_to_FpX(x, p), e, p);
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+RgXn_inv_FpXQX(GEN x, long n, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  GEN r, T = RgX_to_FpX(pol, p);
+  if (signe(T) == 0) pari_err_OP("/", gen_1, x);
+  r = FpXQXn_inv(RgX_to_FpXQX(x, T, p), n, T, p);
+  return gerepileupto(av, FpXQX_to_mod(r, pol, p));
+}
+
+#define code(t1,t2) ((t1 << 6) | t2)
+
+static GEN
+RgXn_inv_fast(GEN x, long e)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type(x,&p,&pol,&pa);
+  switch(t)
+  {
+    case t_INTMOD: return RgXn_inv_FpX(x, e, p);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgXn_inv_FpXQX(x, e, pol, p);
+    default:       return NULL;
+  }
+}
+#undef code
+
+GEN
+RgXn_inv(GEN f, long e)
+{
+  GEN h = RgXn_inv_fast(f, e);
+  if (h) return h;
+  return RgXn_inv_i(f, e);
 }
 
 GEN
@@ -2272,6 +2455,36 @@ RgXn_reverse(GEN f, long e)
   return gerepileupto(av, a);
 }
 
+GEN
+RgXn_sqrt(GEN h, long e)
+{
+  pari_sp av = avma, av2;
+  long v = varn(h), n = 1;
+  GEN f = scalarpol(gen_1, v), df = f;
+  ulong mask = quadratic_prec_mask(e);
+  if (degpol(h)<0 || !gequal1(gel(h,2)))
+    pari_err_SQRTN("RgXn_sqrt",h);
+  av2 = avma;
+  while(1)
+  {
+    long n2 = n, m;
+    GEN g;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    m = n-n2;
+    g = RgX_sub(RgXn_sqrhigh(f, n2, n), RgX_shift_shallow(RgXn_red_shallow(h, n),-n2));
+    f = RgX_sub(f, RgX_shift_shallow(RgXn_mul(gmul2n(df, -1), g, m), n2));
+    if (mask==1) return gerepileupto(av, f);
+    g = RgXn_mul(df, RgXn_mulhigh(df, f, n2, n), m);
+    df = RgX_sub(df, RgX_shift_shallow(g, n2));
+    if (gc_needed(av2,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_sqrt, e = %ld", n);
+      gerepileall(av2, 2, &f, &df);
+    }
+  }
+}
+
 /* x,T in Rg[X], n in N, compute lift(x^n mod T)) */
 GEN
 RgXQ_powu(GEN x, ulong n, GEN T)
@@ -2364,16 +2577,6 @@ RgXQ_matrix_pow(GEN y, long n, long m, GEN P)
   return RgXV_to_RgM(RgXQ_powers(y,m-1,P),n);
 }
 
-GEN
-RgXQ_minpoly_naive(GEN y, GEN P)
-{
-  pari_sp ltop=avma;
-  long n=lgpol(P);
-  GEN M=ker(RgXQ_matrix_pow(y,n,n,P));
-  M=content(RgM_to_RgXV(M,varn(P)));
-  return gerepileupto(ltop,M);
-}
-
 GEN
 RgXQ_norm(GEN x, GEN T)
 {
@@ -2479,3 +2682,289 @@ Kronecker_to_mod(GEN z, GEN T)
   gel(x,i) = mkpolmod(RgX_rem(normalizepol_lg(t,N), T), T);
   return normalizepol_lg(x, i+1);
 }
+
+/*******************************************************************/
+/*                                                                 */
+/*                        Domain detection                         */
+/*                                                                 */
+/*******************************************************************/
+
+static GEN
+zero_FpX_mod(GEN p, long v)
+{
+  GEN r = cgetg(3,t_POL);
+  r[1] = evalvarn(v);
+  gel(r,2) = mkintmod(gen_0, icopy(p));
+  return r;
+}
+
+static GEN
+RgX_mul_FpX(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flx_mul(RgX_to_Flx(x, pp),
+                                  RgX_to_Flx(y, pp), pp));
+  }
+  else
+    r = FpX_mul(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+zero_FpXQX_mod(GEN pol, GEN p, long v)
+{
+  GEN r = cgetg(3,t_POL);
+  r[1] = evalvarn(v);
+  gel(r,2) = mkpolmod(mkintmod(gen_0, icopy(p)), gcopy(pol));
+  return r;
+}
+
+static GEN
+RgX_mul_FpXQX(GEN x, GEN y, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  long dT;
+  GEN kx, ky, r;
+  GEN T = RgX_to_FpX(pol, p);
+  if (signe(T)==0) pari_err_OP("*", x, y);
+  dT = degpol(T);
+  kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
+  ky = ZXX_to_Kronecker(RgX_to_FpXQX(y, T, p), dT);
+  r = FpX_mul(kx, ky, p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
+}
+
+static GEN
+RgX_liftred(GEN x, GEN T)
+{ return RgXQX_red(liftpol_shallow(x), T); }
+
+static GEN
+RgX_mul_QXQX(GEN x, GEN y, GEN T)
+{
+  pari_sp av = avma;
+  long dT = degpol(T);
+  GEN r = QX_mul(ZXX_to_Kronecker(RgX_liftred(x, T), dT),
+                 ZXX_to_Kronecker(RgX_liftred(y, T), dT));
+  return gerepileupto(av, Kronecker_to_mod(r, T));
+}
+
+static GEN
+RgX_sqr_FpX(GEN x, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flx_sqr(RgX_to_Flx(x, pp), pp));
+  }
+  else
+    r = FpX_sqr(RgX_to_FpX(x, p), p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+RgX_sqr_FpXQX(GEN x, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  long dT;
+  GEN kx, r, T = RgX_to_FpX(pol, p);
+  if (signe(T)==0) pari_err_OP("*",x,x);
+  dT = degpol(T);
+  kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
+  r = FpX_sqr(kx, p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
+}
+
+static GEN
+RgX_sqr_QXQX(GEN x, GEN T)
+{
+  pari_sp av = avma;
+  long dT = degpol(T);
+  GEN r = QX_sqr(ZXX_to_Kronecker(RgX_liftred(x, T), dT));
+  return gerepileupto(av, Kronecker_to_mod(r, T));
+}
+
+static GEN
+RgX_rem_FpX(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flx_rem(RgX_to_Flx(x, pp),
+                                  RgX_to_Flx(y, pp), pp));
+  }
+  else
+    r = FpX_rem(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+RgX_rem_QXQX(GEN x, GEN y, GEN T)
+{
+  pari_sp av = avma;
+  GEN r;
+  r = RgXQX_rem(RgX_liftred(x, T), RgX_liftred(y, T), T);
+  return gerepilecopy(av, QXQX_to_mod_shallow(r, T));
+}
+static GEN
+RgX_rem_FpXQX(GEN x, GEN y, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  GEN T = RgX_to_FpX(pol, p);
+  if (signe(T) == 0) pari_err_OP("%", x, y);
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    GEN Tp = ZX_to_Flx(T, pp);
+    r = FlxX_to_ZXX(FlxqX_rem(RgX_to_FlxqX(x, Tp, pp),
+                              RgX_to_FlxqX(y, Tp, pp), Tp, pp));
+  }
+  else
+    r = FpXQX_rem(RgX_to_FpXQX(x, T, p), RgX_to_FpXQX(y, T, p), T, p);
+  if (signe(r)==0)
+  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  return gerepileupto(av, FpXQX_to_mod(r, T, p));
+}
+
+#define code(t1,t2) ((t1 << 6) | t2)
+static GEN
+RgX_mul_fast(GEN x, GEN y)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type2(x,y, &p,&pol,&pa);
+  switch(t)
+  {
+    case t_INT:    return ZX_mul(x,y);
+    case t_FRAC:   return QX_mul(x,y);
+    case t_FFELT:  return FFX_mul(x, y, pol);
+    case t_INTMOD: return RgX_mul_FpX(x, y, p);
+    case code(t_POLMOD, t_INT):
+    case code(t_POLMOD, t_FRAC):
+                   return RgX_mul_QXQX(x, y, pol);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgX_mul_FpXQX(x, y, pol, p);
+    default:       return NULL;
+  }
+}
+static GEN
+RgX_sqr_fast(GEN x)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type(x,&p,&pol,&pa);
+  switch(t)
+  {
+    case t_INT:    return ZX_sqr(x);
+    case t_FRAC:   return QX_sqr(x);
+    case t_FFELT:  return FFX_sqr(x, pol);
+    case t_INTMOD: return RgX_sqr_FpX(x, p);
+    case code(t_POLMOD, t_INT):
+    case code(t_POLMOD, t_FRAC):
+                   return RgX_sqr_QXQX(x, pol);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgX_sqr_FpXQX(x, pol, p);
+    default:       return NULL;
+  }
+}
+
+static GEN
+RgX_rem_fast(GEN x, GEN y)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type2(x,y, &p,&pol,&pa);
+  switch(t)
+  {
+    case t_INT:    return ZX_is_monic(y) ? ZX_rem(x,y): NULL;
+    case t_FRAC:   return RgX_is_ZX(y) && ZX_is_monic(y) ? QX_ZX_rem(x,y): NULL;
+    case t_FFELT:  return FFX_rem(x, y, pol);
+    case t_INTMOD: return RgX_rem_FpX(x, y, p);
+    case code(t_POLMOD, t_INT):
+    case code(t_POLMOD, t_FRAC):
+                   return RgX_rem_QXQX(x, y, pol);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgX_rem_FpXQX(x, y, pol, p);
+    default:       return NULL;
+  }
+}
+
+#undef code
+
+GEN
+RgX_mul(GEN x, GEN y)
+{
+  GEN z = RgX_mul_fast(x,y);
+  if (!z) z = RgX_mul_i(x,y);
+  return z;
+}
+
+GEN
+RgX_sqr(GEN x)
+{
+  GEN z = RgX_sqr_fast(x);
+  if (!z) z = RgX_sqr_i(x);
+  return z;
+}
+
+GEN
+RgX_rem(GEN x, GEN y)
+{
+  GEN z = RgX_rem_fast(x, y);
+  if (!z) z = RgX_divrem_i(x, y, ONLY_REM);
+  return z;
+}
+
+GEN
+RgXn_mul(GEN f, GEN g, long n)
+{
+  pari_sp av = avma;
+  GEN h = RgX_mul_fast(f,g);
+  if (!h) return RgXn_mul2(f,g,n);
+  if (degpol(h) < n) return h;
+  return gerepilecopy(av, RgXn_red_shallow(h, n));
+}
+
+GEN
+RgXn_sqr(GEN f, long n)
+{
+  pari_sp av = avma;
+  GEN g = RgX_sqr_fast(f);
+  if (!g) return RgXn_sqr2(f,n);
+  if (degpol(g) < n) return g;
+  return gerepilecopy(av, RgXn_red_shallow(g, n));
+}
+
+/* (f*g) \/ x^n */
+GEN
+RgX_mulhigh_i(GEN f, GEN g, long n)
+{
+  GEN h = RgX_mul_fast(f,g);
+  return h? RgX_shift_shallow(h, -n): RgX_mulhigh_i2(f,g,n);
+}
+
+/* (f*g) \/ x^n */
+GEN
+RgX_sqrhigh_i(GEN f, long n)
+{
+  GEN h = RgX_sqr_fast(f);
+  return h? RgX_shift_shallow(h, -n): RgX_sqrhigh_i2(f,n);
+}
