@@ -159,10 +159,11 @@ modulereltoabs(GEN rnf, GEN x)
 {
   GEN W=gel(x,1), I=gel(x,2), rnfeq = rnf_get_map(rnf), polabs = gel(rnfeq,1);
   long i, j, k, m, N = lg(W)-1;
-  GEN zknf, czknf, M;
+  GEN zknf, dzknf, M;
 
   if (!N) return cgetg(1, t_VEC);
-  rnf_get_nfzk(rnf, &zknf,&czknf);
+  zknf = rnf_get_nfzk(rnf);
+  dzknf = gel(zknf,1);
   m = rnf_get_nfdegree(rnf);
   M = cgetg(N*m+1, t_VEC);
   for (k=i=1; i<=N; i++)
@@ -172,7 +173,7 @@ modulereltoabs(GEN rnf, GEN x)
     if (lg(id) == 1) continue; /* must be a t_MAT */
     id = Q_primitive_part(id, &cid);
     w = Q_primitive_part(eltreltoabs(rnfeq,w), &c0);
-    c0 = mul_content(c0, mul_content(cid,czknf));
+    c0 = mul_content(c0, mul_content(cid,inv_content(dzknf)));
     if (typ(id) == t_INT)
       for (j=1; j<=m; j++)
       {
@@ -225,16 +226,16 @@ mknfabs(GEN rnf, long prec)
 static GEN
 mkupdown(GEN rnf)
 {
-  GEN NF = obj_check(rnf, rnf_NFABS), M, zknf, czknf;
-  long i, m;
-  rnf_get_nfzk(rnf, &zknf, &czknf);
-  if (isint1(czknf)) czknf = NULL;
-  m = lg(zknf)-1; M = cgetg(m+1, t_MAT);
+  GEN NF = obj_check(rnf, rnf_NFABS), M, zknf, dzknf;
+  long i, l;
+  zknf = rnf_get_nfzk(rnf);
+  dzknf = gel(zknf,1); if (gequal1(dzknf)) dzknf = NULL;
+  l = lg(zknf); M = cgetg(l, t_MAT);
   gel(M,1) = vec_ei(nf_get_degree(NF), 1);
-  for (i = 2; i <= m; i++)
+  for (i = 2; i < l; i++)
   {
     GEN c = poltobasis(NF, gel(zknf,i));
-    if (czknf) c = gmul(c, czknf);
+    if (dzknf) c = gdiv(c, dzknf);
     gel(M,i) = c;
   }
   return Qevproj_init(M);
@@ -251,38 +252,50 @@ void
 rnfcomplete(GEN rnf)
 { (void)rnf_build_nfabs(rnf, nf_get_prec(rnf_get_nf(rnf))); }
 
-void
-nf_nfzk(GEN nf, GEN rnfeq, GEN *zknf, GEN *czknf)
+GEN
+nf_nfzk(GEN nf, GEN rnfeq)
 {
   GEN pol = gel(rnfeq,1), a = gel(rnfeq,2);
-  GEN zk = QXV_QXQ_eval(nf_get_zk(nf), a, pol);
-  *zknf = Q_primitive_part(zk, czknf);
-  if (!*czknf) *czknf = gen_1;
+  return Q_primpart(QXV_QXQ_eval(nf_get_zkprimpart(nf), a, pol));
+}
+
+/* true nf */
+GEN
+check_polrel(GEN nf, GEN P, ulong *lim)
+{
+  if (typ(P) != t_VEC || lg(P) != 3) *lim = 0;
+  else { *lim = gtou(gel(P,2)); P = gel(P,1); }
+  if (typ(P) != t_POL) pari_err_TYPE("rnfinit",P);
+  P = RgX_nffix("rnfinit", nf_get_pol(nf), P, 0);
+  if (!gequal1(leading_coeff(P)))
+    pari_err_IMPL("non-monic relative polynomials");
+  return P;
 }
 
 GEN
-rnfinit0(GEN nf, GEN polrel, long flag)
+rnfinit0(GEN nf, GEN T, long flag)
 {
   pari_sp av = avma;
-  GEN rnf, bas, D,d,f, B, rnfeq, zknf,czknf;
+  GEN bas, D, f, B, T0, rnfeq, rnf = obj_init(11, 2);
+  ulong lim;
   nf = checknf(nf);
-  bas = rnfallbase(nf,&polrel, &D,&d, &f);
+  T0 = check_polrel(nf, T, &lim);
+  T = lift_shallow(T0);
+  gel(rnf,11) = rnfeq = nf_rnfeq(nf,T);
+  gel(rnf,2) = nf_nfzk(nf, rnfeq);
+  bas = rnfallbase(nf, T0, lim, rnf, &D, &f);
   B = matbasistoalg(nf,gel(bas,1));
-  gel(bas,1) = lift_if_rational( RgM_to_RgXV(B,varn(polrel)) );
-  rnfeq = nf_rnfeq(nf,polrel);
-  nf_nfzk(nf, rnfeq, &zknf, &czknf);
-  rnf = obj_init(11, 2);
-  gel(rnf,1) = polrel;
-  gel(rnf,2) = mkvec2(zknf, czknf);
-  gel(rnf,3) = mkvec2(D, d);
+  gel(bas,1) = lift_if_rational( RgM_to_RgXV(B,varn(T)) );
+  gel(rnf,1) = T;
+  gel(rnf,3) = D;
   gel(rnf,4) = f;
   gel(rnf,5) = cgetg(1, t_VEC); /* dummy */
   gel(rnf,6) = cgetg(1, t_VEC); /* dummy */
   gel(rnf,7) = bas;
   gel(rnf,8) = lift_if_rational( RgM_inv_upper(B) );
-  gel(rnf,9) = typ(f) == t_INT? gen_1: RgM_det_triangular(f);
+  gel(rnf,9) = typ(f) == t_INT? powiu(f, nf_get_degree(nf))
+                              : RgM_det_triangular(f);
   gel(rnf,10)= nf;
-  gel(rnf,11)= rnfeq;
   rnf = gerepilecopy(av, rnf);
   if (flag) rnfcomplete(rnf);
   return rnf;
@@ -294,7 +307,7 @@ GEN
 rnfeltup0(GEN rnf, GEN x, long flag)
 {
   pari_sp av = avma;
-  GEN zknf, czknf, nf, NF, POL;
+  GEN zknf, nf, NF, POL;
   long tx = typ(x);
   checkrnf(rnf);
   if (flag) rnfcomplete(rnf);
@@ -325,8 +338,8 @@ rnfeltup0(GEN rnf, GEN x, long flag)
   }
   else
   {
-    rnf_get_nfzk(rnf, &zknf, &czknf);
-    x = nfeltup(nf, x, zknf, czknf);
+    zknf = rnf_get_nfzk(rnf);
+    x = nfeltup(nf, x, zknf);
     if (typ(x) == t_POL) x = mkpolmod(x, POL);
   }
   return gerepilecopy(av, x);
@@ -335,14 +348,15 @@ GEN
 rnfeltup(GEN rnf, GEN x) { return rnfeltup0(rnf,x,0); }
 
 GEN
-nfeltup(GEN nf, GEN x, GEN zknf, GEN czknf)
+nfeltup(GEN nf, GEN x, GEN zknf)
 {
-  GEN c;
+  GEN c, dzknf = gel(zknf,1);
   x = nf_to_scalar_or_basis(nf, x);
   if (typ(x) != t_COL) return x;
   x = Q_primitive_part(x, &c);
   if (!RgV_is_ZV(x)) pari_err_TYPE("rnfeltup", x);
-  c = mul_content(c, czknf);
+  if (gequal1(dzknf)) dzknf = NULL;
+  c = mul_content(c, inv_content(dzknf));
   x = RgV_RgC_mul(zknf, x); if (c) x = RgX_Rg_mul(x, c);
   return x;
 }
@@ -491,7 +505,7 @@ rnfidealhnf(GEN rnf, GEN x)
       return rnfprincipaltohnf(rnf,x);
   }
   pari_err_TYPE("rnfidealhnf",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -560,7 +574,7 @@ rnfidealreltoabs0(GEN rnf, GEN x, long flag)
 GEN
 rnfidealabstorel(GEN rnf, GEN x)
 {
-  long N, j, tx = typ(x);
+  long n, N, j, tx = typ(x);
   pari_sp av = avma;
   GEN A, I, invbas;
 
@@ -573,12 +587,17 @@ rnfidealabstorel(GEN rnf, GEN x)
     if (!N) return rnfideal0();
     pari_err_DIM("rnfidealabstorel");
   }
+  n = rnf_get_degree(rnf);
   A = cgetg(N+1,t_MAT);
   I = cgetg(N+1,t_VEC);
   for (j=1; j<=N; j++)
   {
     GEN t = lift_shallow( rnfeltabstorel(rnf, gel(x,j)) );
-    gel(A,j) = mulmat_pol(invbas, t);
+    if (typ(t) == t_POL)
+      t = RgM_RgX_mul(invbas, t);
+    else
+      t = scalarcol_shallow(t, n);
+    gel(A,j) = t;
     gel(I,j) = gen_1;
   }
   return gerepileupto(av, nfhnf(rnf_get_nf(rnf), mkvec2(A,I)));
@@ -684,7 +703,7 @@ rnfidealtwoelement(GEN rnf, GEN x)
   y = idealtwoelt(NF, y);
   if (cy) y = RgV_Rg_mul(y, cy);
   z = gel(y,2);
-  if (typ(z) == t_COL) z = rnfeltabstorel(rnf, coltoliftalg(NF, z));
+  if (typ(z) == t_COL) z = rnfeltabstorel(rnf, nf_to_scalar_or_alg(NF, z));
   return gerepilecopy(av, mkvec2(gel(y,1), z));
 }
 
@@ -708,7 +727,7 @@ rnfidealmul(GEN rnf,GEN x,GEN y)
 }
 
 static GEN
-rnfidealprimedec_1(GEN rnf, GEN L, GEN SL, GEN prK)
+rnfidealprimedec_1(GEN rnf, GEN SL, GEN prK)
 {
   GEN v, piL = rnfeltup0(rnf, pr_get_gen(prK), 1);
   long i, c, l;
@@ -717,7 +736,7 @@ rnfidealprimedec_1(GEN rnf, GEN L, GEN SL, GEN prK)
   for (i = c = 1; i < l; i++)
   {
     GEN P = gel(SL,i);
-    if (ZC_prdvd(L, piL, P)) gel(v,c++) = P;
+    if (ZC_prdvd(piL, P)) gel(v,c++) = P;
   }
   setlg(v, c); return v;
 }
@@ -741,15 +760,13 @@ rnfidealprimedec(GEN rnf, GEN pr)
     p = pr_get_p(pr);
   }
   SL = idealprimedec(NF, p);
-  if (pr)
-    z = rnfidealprimedec_1(rnf, NF, SL, pr);
+  if (pr) z = rnfidealprimedec_1(rnf, SL, pr);
   else
   {
     GEN vK = idealprimedec(nf, p), vL;
     long l = lg(vK), i;
     vL = cgetg(l, t_VEC);
-    for (i = 1; i < l; i++)
-      gel(vL,i) = rnfidealprimedec_1(rnf, NF, SL, gel(vK,i));
+    for (i = 1; i < l; i++) gel(vL,i) = rnfidealprimedec_1(rnf, SL, gel(vK,i));
     z = mkvec2(vK, vL);
   }
   return gerepilecopy(av, z);
@@ -785,7 +802,7 @@ rnfequationall(GEN A, GEN B, long *pk, GEN *pLPRS)
     pari_err_DOMAIN("rnfequation","issquarefree(B)","=",gen_0,B);
 
   *pk = 0; C = ZX_ZXY_resultant_all(A, B, pk, pLPRS);
-  if (gsigne(leading_coeff(C)) < 0) C = RgX_neg(C);
+  if (signe(leading_coeff(C)) < 0) C = ZX_neg(C);
   *pk = -*pk; return Q_primpart(C);
 }
 
@@ -825,7 +842,9 @@ GEN
 nf_rnfeqsimple(GEN nf, GEN relpol)
 {
   long sa;
-  GEN junk, pol = rnfequationall(nf, relpol, &sa, NULL);
+  GEN junk, pol;
+  relpol = liftpol_shallow(relpol);
+  pol = rnfequationall(nf, relpol, &sa, NULL);
   return mkvec5(pol,gen_0/*dummy*/,stoi(sa),get_nfpol(nf, &junk),relpol);
 }
 
@@ -904,7 +923,7 @@ rnfscal(GEN m, GEN x, GEN y)
   long i, l = lg(m);
   GEN z = cgetg(l, t_COL);
   for (i = 1; i < l; i++)
-    gel(z,i) = gmul(gconj(shallowtrans(gel(x,i))), gmul(gel(m,i), gel(y,i)));
+    gel(z,i) = gmul(conj_i(shallowtrans(gel(x,i))), gmul(gel(m,i), gel(y,i)));
   return z;
 }
 
@@ -937,7 +956,7 @@ findmin(GEN nf, GEN x, GEN muf)
   m = grndtoi(m, &e);
   if (e >= 0) return NULL; /* precision problem */
   m = ZM_ZC_mul(x, m);
-  if (cx) m = RgC_Rg_mul(m, cx);
+  if (cx) m = ZC_Q_mul(m, cx);
   return gerepileupto(av, m);
 }
 
@@ -988,7 +1007,7 @@ do_SWAP(GEN I, GEN MC, GEN MCS, GEN h, GEN mu, GEN B, long kmax, long k,
   swap(gel(I,k-1), gel(I,k));
   for (j=1; j<=k-2; j++) swap(gcoeff(mu,k-1,j),gcoeff(mu,k,j));
   muf = gcoeff(mu,k,k-1);
-  mufc = gconj(muf);
+  mufc = conj_i(muf);
   Bf = gadd(gel(B,k), vecmul(real_i(vecmul(muf,mufc)), gel(B,k-1)));
   if (check_0(Bf)) return 1; /* precision problem */
 
@@ -1036,14 +1055,14 @@ rel_T2(GEN nf, GEN pol, long lx, long prec)
       {
         s = gen_0;
         for (j = 1; j < lx; j++)
-          s = gadd(s, gmul(gconj(gmael(powreorder,k,j)),
-                                 gmael(powreorder,l,j)));
+          s = gadd(s, gmul(conj_i(gmael(powreorder,k,j)),
+                                  gmael(powreorder,l,j)));
         if (l == k)
           gcoeff(m, l, l) = real_i(s);
         else
         {
           gcoeff(m, k, l) = s;
-          gcoeff(m, l, k) = gconj(s);
+          gcoeff(m, l, k) = conj_i(s);
         }
       }
     gel(T2,i) = m;
@@ -1184,11 +1203,7 @@ rnfpolred(GEN nf, GEN pol, long prec)
   {
     GEN newpol, L, a, Ij = gel(I,j);
     a = RgC_Rg_mul(gel(O,j), (typ(Ij) == t_MAT)? gcoeff(Ij,1,1): Ij);
-    for (i=n; i; i--)
-    {
-      GEN c = gel(a,i);
-      if (typ(c) == t_COL) gel(a,i) = coltoliftalg(nf, c);
-    }
+    for (i=n; i; i--) gel(a,i) = nf_to_scalar_or_alg(nf, gel(a,i));
     a = RgV_to_RgX(a, v);
     newpol = RgXQX_red(RgXQ_charpoly(a, pol, v), nfpol);
     newpol = Q_primpart(newpol);
@@ -1274,7 +1289,7 @@ element_mulvecrow(GEN nf, GEN x, GEN m, long i, long lim)
         if (typ(t) == t_COL)
           t = RgM_RgC_mul(x, t);
         else
-          t = RgC_Rg_mul(gel(x,1), t);
+          t = ZC_Q_mul(gel(x,1), t);
         if (dx) t = gdiv(t, dx);
         t = nf_to_scalar_or_basis(nf,t);
       }
@@ -1329,9 +1344,10 @@ zero_nfbezout(GEN nf,GEN bB, GEN b, GEN A,GEN B,GEN *u,GEN *v,GEN *w,GEN *di)
  * di=d^-1, w=A.B.di, u, v such that au+bv=1 and u in A.di, v in B.di.
  * Assume A, B non-zero, but a or b can be zero (not both) */
 static GEN
-nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi)
+nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi,
+         int red)
 {
-  GEN w, u,v,uv, d, di, aA, bB;
+  GEN w, u, v, d, di, aA, bB;
 
   if (isintzero(a)) return zero_nfbezout(nf,NULL,b,A,B,pu,pv,pw,pdi);
   if (isintzero(b)) return zero_nfbezout(nf,NULL,a,B,A,pv,pu,pw,pdi);
@@ -1341,28 +1357,26 @@ nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi)
     a = nf_to_scalar_or_basis(nf,a);
     if (isint1(a)) a = gen_1;
   }
-  aA = (a == gen_1)? A: idealmul(nf,a,A);
+  aA = (a == gen_1)? idealhnf_shallow(nf,A): idealmul(nf,a,A);
   bB = idealmul(nf,b,B);
   d = idealadd(nf,aA,bB);
-  if (gequal(aA, d)) return zero_nfbezout(nf,aA, a,B,A,pv,pu,pw,pdi);
-  if (gequal(bB, d)) return zero_nfbezout(nf,bB, b,A,B,pu,pv,pw,pdi);
+  if (gequal(aA, d)) return zero_nfbezout(nf,d, a,B,A,pv,pu,pw,pdi);
+  if (gequal(bB, d)) return zero_nfbezout(nf,d, b,A,B,pu,pv,pw,pdi);
   /* general case is slow */
   di = idealHNF_inv(nf,d);
-  w = idealmul(nf,aA,di); /* integral */
-  uv = idealaddtoone(nf, w, idealmul(nf,bB,di));
-  w = idealmul(nf,w,B);
-  u = gel(uv,1);
-  v = nfdiv(nf,gel(uv,2),b);
+  aA = idealmul(nf,aA,di); /* integral */
+  bB = idealmul(nf,bB,di); /* integral */
+
+  u = red? idealaddtoone_i(nf, aA, bB): idealaddtoone_raw(nf, aA, bB);
+  w = idealmul(nf,aA,B);
+  v = nfdiv(nf, nfsub(nf, gen_1, u), b);
   if (a != gen_1)
   {
     GEN inva = nfinv(nf, a);
     u =  nfmul(nf,u,inva);
     w = idealmul(nf, inva, w); /* AB/d */
   }
-  *pu = u;
-  *pv = v;
-  *pw = w;
-  *pdi = di; return d;
+  *pu = u; *pv = v; *pw = w; *pdi = di; return d;
 }
 /* v a vector of ideals, simplify in place the ones generated by elts of Q */
 static void
@@ -1420,7 +1434,7 @@ nfhnf0(GEN nf, GEN x, long flag)
       b = gel(T0,i); if (isintzero(b)) continue;
 
       S0 = gel(A,def); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di);
+      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di,1);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       gel(A,def) = S; gel(A,j) = T;
@@ -1531,7 +1545,7 @@ nfsnf0(GEN nf, GEN x, long flag)
       b = gel(T0,i); if (gequal0(b)) continue;
 
       S0 = gel(A,i); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(J,i),gel(J,j), &u,&v,&w,&dinv);
+      d = nfbezout(nf, a,b, gel(J,i),gel(J,j), &u,&v,&w,&dinv,1);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       gel(A,i) = S; gel(A,j) = T;
@@ -1550,7 +1564,7 @@ nfsnf0(GEN nf, GEN x, long flag)
       b = gcoeff(A,j,i); if (gequal0(b)) continue;
 
       a = gcoeff(A,i,i);
-      d = nfbezout(nf, a,b, gel(I,i),gel(I,j), &u,&v,&w,&dinv);
+      d = nfbezout(nf, a,b, gel(I,i),gel(I,j), &u,&v,&w,&dinv,1);
       ri = rowcomb(nf, u,v,       i,j, A, i);
       rj = rowcomb(nf, a,gneg(b), j,i, A, i);
       for (k=1; k<=i; k++) {
@@ -1744,7 +1758,7 @@ nfcleanmod(GEN nf, GEN x, long lim, GEN D)
 }
 
 GEN
-nfhnfmod(GEN nf, GEN x, GEN detmat)
+nfhnfmod(GEN nf, GEN x, GEN D)
 {
   long li, co, i, j, def, ldef;
   pari_sp av0=avma, av;
@@ -1757,16 +1771,16 @@ nfhnfmod(GEN nf, GEN x, GEN detmat)
   co = lg(A); if (co==1) return cgetg(1,t_MAT);
 
   li = lgcols(A);
-  if (typ(detmat)!=t_MAT) detmat = idealhnf_shallow(nf, detmat);
-  detmat = Q_remove_denom(detmat, NULL);
-  RgM_check_ZM(detmat, "nfhnfmod");
+  if (typ(D)!=t_MAT) D = idealhnf_shallow(nf, D);
+  D = Q_remove_denom(D, NULL);
+  RgM_check_ZM(D, "nfhnfmod");
 
   av = avma;
   A = RgM_to_nfM(nf, A);
   A = Q_remove_denom(A, &dA);
   I = Q_remove_denom(leafcopy(I), &dI);
   dA = mul_denom(dA,dI);
-  if (dA) detmat = ZM_Z_mul(detmat, powiu(dA, minss(li,co)));
+  if (dA) D = ZM_Z_mul(D, powiu(dA, minss(li,co)));
 
   def = co; ldef = (li>co)? li-co+1: 1;
   for (i=li-1; i>=ldef; i--)
@@ -1784,29 +1798,29 @@ nfhnfmod(GEN nf, GEN x, GEN detmat)
       b = gel(T0,i); if (isintzero(b)) continue;
 
       S0 = gel(A,def); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di);
+      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di,0);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       if (u != gen_0 && v != gen_0) /* already reduced otherwise */
-        nfcleanmod(nf, S, i, idealmul(nf,detmat,di));
-      nfcleanmod(nf, T, i, idealdiv(nf,detmat,w));
+        nfcleanmod(nf, S, i, idealmul(nf,D,di));
+      nfcleanmod(nf, T, i, idealdiv(nf,D,w));
       gel(A,def) = S; gel(A,j) = T;
       gel(I,def) = d; gel(I,j) = w;
     }
     if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"[1]: nfhnfmod, i = %ld", i);
-      gerepileall(av,dA? 4: 3, &A,&I,&detmat,&dA);
+      gerepileall(av,dA? 4: 3, &A,&I,&D,&dA);
     }
   }
-  def--; d0 = detmat;
+  def--; d0 = D;
   A += def; A[0] = evaltyp(t_MAT)|evallg(li);
   I += def; I[0] = evaltyp(t_VEC)|evallg(li);
   J = cgetg(li,t_VEC);
   for (i=li-1; i>=1; i--)
   {
     GEN b = gcoeff(A,i,i);
-    d = nfbezout(nf, gen_1,b, d0,gel(I,i), &u,&v,&w,&di);
+    d = nfbezout(nf, gen_1,b, d0,gel(I,i), &u,&v,&w,&di,0);
     p1 = nfC_nf_mul(nf,gel(A,i),v);
     if (i > 1)
     {
