@@ -18,14 +18,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /**                  GENERIC ABELIAN CHARACTERS                     **/
 /**                                                                 **/
 /*********************************************************************/
-
-/* check whether G is a bidZ */
-static int
-checkbidZ_i(GEN G)
+/* check whether G is a znstar */
+int
+checkznstar_i(GEN G)
 {
   return (typ(G) == t_VEC && lg(G) == 6
-      && typ(bid_get_fact(G)) == t_VEC
-      && typ(bid_get_mod(G)) == t_VEC && lg(bid_get_mod(G)) == 3);
+      && typ(znstar_get_faN(G)) == t_VEC
+      && typ(gel(G,1)) == t_VEC && lg(gel(G,1)) == 3);
 }
 
 int
@@ -56,7 +55,8 @@ char_simplify(GEN D, GEN C)
     GEN t = gcdii(d, ZV_content(C));
     if (!equali1(t))
     {
-      C = ZC_Z_divexact(C, t);
+      long tc = typ(C);
+      C = ZC_Z_divexact(C, t); settyp(C, tc);
       d = diviiexact(d, t);
     }
   }
@@ -84,7 +84,7 @@ char_normalize(GEN chi, GEN ncyc)
 static GEN
 get_cyc(GEN x, GEN chi, const char *s)
 {
-  if (nftyp(x) == typ_BID && checkbidZ_i(x))
+  if (nftyp(x) == typ_BIDZ)
   {
     if (!zncharcheck(x, chi)) pari_err_TYPE(s, chi);
     return NULL;
@@ -117,26 +117,6 @@ charconj0(GEN x, GEN chi)
   return cyc? charconj(cyc, chi): zncharconj(x, chi);
 }
 
-/* exp(2iPi/d), assume d a t_INT */
-GEN
-rootsof1_cx(GEN d, long prec)
-{
-  if (lgefint(d) == 3) return rootsof1u_cx((ulong)d[2], prec);
-  return expIr(divri(Pi2n(1,prec), d));
-}
-/* exp(2iPi/d) */
-GEN
-rootsof1u_cx(ulong d, long prec)
-{
-  switch(d)
-  {
-    case 1: return gen_1;
-    case 2: return gen_m1;
-    case 4: return gen_I();
-  }
-  return expIr(divru(Pi2n(1,prec), d));
-}
-
 GEN
 charorder(GEN cyc, GEN x)
 {
@@ -144,11 +124,13 @@ charorder(GEN cyc, GEN x)
   long i, l = lg(cyc);
   GEN f = gen_1;
   for (i = 1; i < l; i++)
-  {
-    GEN o = gel(cyc,i), c = gcdii(o, gel(x,i));
-    if (!is_pm1(c)) o = diviiexact(o,c);
-    f = lcmii(f, o);
-  }
+    if (signe(gel(x,i)))
+    {
+      GEN c, o = gel(cyc,i);
+      c = gcdii(o, gel(x,i));
+      if (!is_pm1(c)) o = diviiexact(o,c);
+      f = lcmii(f, o);
+    }
   return gerepileuptoint(av, f);
 }
 GEN
@@ -181,6 +163,14 @@ charker0(GEN x, GEN chi)
   return cyc? charker(cyc, chi): zncharker(x, chi);
 }
 
+GEN
+charpow(GEN cyc, GEN a, GEN N)
+{
+  long i, l;
+  GEN v = cgetg_copy(a, &l);
+  for (i = 1; i < l; i++) gel(v,i) = Fp_mul(gel(a,i), N, gel(cyc,i));
+  return v;
+}
 GEN
 charmul(GEN cyc, GEN a, GEN b)
 {
@@ -198,6 +188,12 @@ chardiv(GEN cyc, GEN a, GEN b)
   return v;
 }
 GEN
+charpow0(GEN x, GEN a, GEN N)
+{
+  GEN cyc = get_cyc(x, a, "charpow");
+  return cyc? charpow(cyc, a, N): zncharpow(x, a, N);
+}
+GEN
 charmul0(GEN x, GEN a, GEN b)
 {
   const char *s = "charmul";
@@ -303,16 +299,236 @@ chareval(GEN G, GEN chi, GEN x, GEN z)
       L = isprincipal(G, x);
       nchi = get_chi(bnf_get_cyc(G), chi);
       break;
-    case typ_BID:
-      if (checkbidZ_i(G)) return gerepileupto(av, znchareval(G, chi, x, z));
+    case typ_BIDZ:
+      if (checkznstar_i(G)) return gerepileupto(av, znchareval(G, chi, x, z));
       /* don't implement chars on general bid: need an nf... */
     default:
       pari_err_TYPE("chareval", G);
-      return NULL;/* not reached */
+      return NULL;/* LCOV_EXCL_LINE */
   }
   return gerepileupto(av, chareval_i(nchi, L, z));
 }
 
+/* nchi = [ord,D] a quasi-normalized character (ord may be a multiple of
+ * the character order); return v such that v[n] = -1 if (n,N) > 1 else
+ * chi(n) = e(v[n]/ord), 1 <= n <= N */
+GEN
+ncharvecexpo(GEN G, GEN nchi)
+{
+  long N = itou(znstar_get_N(G)), ord = itou(gel(nchi,1)), i, j, l;
+  GEN cyc, gen, d, t, t1, t2, t3, e, u, u1, u2, u3;
+  GEN D = gel(nchi,2), v = const_vecsmall(N,-1);
+  pari_sp av = avma;
+  if (typ(D) == t_COL) {
+    cyc = znstar_get_conreycyc(G);
+    gen = znstar_get_conreygen(G);
+  } else {
+    cyc = znstar_get_cyc(G);
+    gen = znstar_get_gen(G);
+  }
+  l = lg(cyc);
+  e = u = cgetg(N+1,t_VECSMALL);
+  d = t = cgetg(N+1,t_VECSMALL);
+  *++d = 1;
+  *++e = 0; v[*d] = *e;
+  for (i = 1; i < l; i++)
+  {
+    ulong g = itou(gel(gen,i)), c = itou(gel(cyc,i)), x = itou(gel(D,i));
+    for (t1=t,u1=u,j=c-1; j; j--,t1=t2,u1=u2)
+      for (t2=d,u2=e, t3=t1,u3=u1; t3<t2; )
+      {
+        *++d = Fl_mul(*++t3, g, N);
+        *++e = Fl_add(*++u3, x, ord); v[*d] = *e;
+      }
+  }
+  avma = av; return v;
+}
+
+/*****************************************************************************/
+
+static ulong
+lcmuu(ulong a, ulong b) { return (a/ugcd(a,b)) * b; }
+static ulong
+zv_charorder(GEN cyc, GEN x)
+{
+  long i, l = lg(cyc);
+  ulong f = 1;
+  for (i = 1; i < l; i++)
+    if (x[i])
+    {
+      ulong o = cyc[i];
+      f = lcmuu(f, o / ugcd(o, x[i]));
+    }
+  return f;
+}
+
+/* N > 0 */
+GEN
+coprimes_zv(ulong N)
+{
+  GEN v = const_vecsmall(N,1);
+  pari_sp av = avma;
+  GEN P = gel(factoru(N),1);
+  long i, l = lg(P);
+  for (i = 1; i < l; i++)
+  {
+    ulong p = P[i], j;
+    for (j = p; j <= N; j += p) v[j] = 0;
+  }
+  avma = av; return v;
+}
+/* cf zv_cyc_minimal: return k such that g*k is minimal (wrt lex) */
+long
+zv_cyc_minimize(GEN cyc, GEN g, GEN coprime)
+{
+  pari_sp av = avma;
+  long d, k, e, i, k0, bestk, l = lg(g), o = lg(coprime)-1;
+  GEN best, gk, gd;
+  ulong t;
+  if (o == 1) return 1;
+  for (i = 1; i < l; i++)
+    if (g[i]) break;
+  if (g[i] == 1) return 1;
+  k0 = Fl_invgen(g[i], cyc[i], &t);
+  d = cyc[i] / (long)t;
+  if (k0 > 1) g = vecmoduu(Flv_Fl_mul(g, k0, cyc[i]), cyc);
+  for (i++; i < l; i++)
+    if (g[i]) break;
+  if (i == l || d >= cyc[i]) return k0;
+  cyc = vecslice(cyc,i,l-1);
+  g   = vecslice(g,  i,l-1);
+  e = cyc[1];
+  gd = Flv_Fl_mul(g, d, e);
+  bestk = 1; best = g;
+  for (gk = g, k = d+1; k < e; k += d)
+  {
+    long ko = k % o;
+    gk = Flv_add(gk, gd, e); if (!ko || !coprime[ko]) continue;
+    gk = vecmoduu(gk, cyc);
+    if (vecsmall_lexcmp(gk, best) < 0) { best = gk; bestk = k; }
+  }
+  avma = av; return bestk == 1? k0: Fl_mul(k0, bestk, o);
+}
+/* g of order o in abelian group G attached to cyc. Is g a minimal generator
+ * [wrt lex order] of the cyclic subgroup it generates;
+ * coprime = coprimes_zv(o) */
+long
+zv_cyc_minimal(GEN cyc, GEN g, GEN coprime)
+{
+  pari_sp av = avma;
+  long i, d, k, e, l = lg(g), o = lg(coprime)-1; /* elt order */
+  GEN gd, gk;
+  if (o == 1) return 1;
+  for (k = 1; k < l; k++)
+    if (g[k]) break;
+  if (g[k] == 1) return 1;
+  if (cyc[k] % g[k]) return 0;
+  d = cyc[k] / g[k]; /* > 1 */
+  for (k++; k < l; k++) /* skip following 0s */
+    if (g[k]) break;
+  if (k == l) return 1;
+  cyc = vecslice(cyc,k,l-1);
+  g   = vecslice(g,  k,l-1);
+  e = cyc[1];
+  /* find k in (Z/e)^* such that g*k mod cyc is lexicographically minimal,
+   * k = 1 mod d to fix the first non-zero entry */
+  gd = Flv_Fl_mul(g, d, e);
+  for (gk = g, k = d+1, i = 1; i < e; i++, k += d)
+  {
+    long ko = k % o;
+    gk = Flv_add(gk, gd, e); if (!coprime[ko]) continue;
+    gk = vecmoduu(gk, cyc);
+    if (vecsmall_lexcmp(gk, g) < 0) { avma = av; return 0; }
+  }
+  avma = av; return 1;
+}
+
+static GEN
+coprime_tables(long N)
+{
+  GEN D = divisorsu(N), v = const_vec(N, NULL);
+  long i, l = lg(D);
+  for (i = 1; i < l; i++) gel(v, D[i]) = coprimes_zv(D[i]);
+  return v;
+}
+/* enumerate all group elements, modulo (Z/cyc[1])^* */
+static GEN
+cyc2elts_normal(GEN cyc, long maxord, GEN ORD)
+{
+  long i, n, o, N, j = 1;
+  GEN z, vcoprime;
+
+  if (typ(cyc) != t_VECSMALL) cyc = gtovecsmall(cyc);
+  n = lg(cyc)-1;
+  if (n == 0) return cgetg(1, t_VEC);
+  N = zv_prod(cyc);
+  z = cgetg(N+1, t_VEC);
+  if (1 <= maxord && (!ORD|| zv_search(ORD,1)))
+    gel(z,j++) = zero_zv(n);
+  vcoprime = coprime_tables(cyc[1]);
+  for (i = n; i > 0; i--)
+  {
+    GEN cyc0 = vecslice(cyc,i+1,n), pre = zero_zv(i);
+    GEN D = divisorsu(cyc[i]), C = cyc2elts(cyc0);
+    long s, t, lD = lg(D), nC = lg(C)-1; /* remove last element */
+    for (s = 1; s < lD-1; s++)
+    {
+      long o0 = D[lD-s]; /* cyc[i] / D[s] */
+      if (o0 > maxord) continue;
+      pre[i] = D[s];
+      if (!ORD || zv_search(ORD,o0))
+      {
+        GEN c = vecsmall_concat(pre, zero_zv(n-i));
+        gel(z,j++) = c;
+      }
+      for (t = 1; t < nC; t++)
+      {
+        GEN chi0 = gel(C,t);
+        o = lcmuu(o0, zv_charorder(cyc0,chi0));
+        if (o <= maxord && (!ORD || zv_search(ORD,o)))
+        {
+          GEN c = vecsmall_concat(pre, chi0);
+          if (zv_cyc_minimal(cyc, c, gel(vcoprime,o))) gel(z,j++) = c;
+        }
+      }
+    }
+  }
+  setlg(z,j); return z;
+}
+
+GEN
+chargalois(GEN G, GEN ORD)
+{
+  pari_sp av = avma;
+  long maxord, i, l;
+  GEN v, cyc = (typ(G) == t_VEC && RgV_is_ZVpos(G))? G: member_cyc(G);
+  if (lg(cyc) == 1) retmkvec(cgetg(1,t_VEC));
+  maxord = itou(gel(cyc,1));
+  if (ORD && gequal0(ORD)) ORD = NULL;
+  if (ORD)
+    switch(typ(ORD))
+    {
+      long l;
+      case t_VEC:
+        ORD = ZV_to_zv(ORD);
+      case t_VECSMALL:
+        ORD = leafcopy(ORD);
+        vecsmall_sort(ORD);
+        l = lg(ORD);
+        if (l == 1) return cgetg(1, t_VECSMALL);
+        maxord = minss(maxord, ORD[l-1]);
+        break;
+      case t_INT:
+        maxord = minss(maxord, itos(ORD));
+        ORD = NULL;
+        break;
+      default: pari_err_TYPE("chargalois", ORD);
+    }
+  v = cyc2elts_normal(cyc, maxord, ORD); l = lg(v);
+  for(i = 1; i < l; i++) gel(v,i) = zv_to_ZV(gel(v,i));
+  return gerepileupto(av, v);
+}
+
 /*********************************************************************/
 /**                                                                 **/
 /**                  (Z/NZ)^* AND DIRICHLET CHARACTERS              **/
@@ -320,13 +536,13 @@ chareval(GEN G, GEN chi, GEN x, GEN z)
 /*********************************************************************/
 
 GEN
-ZNstar(GEN N, long flag)
+znstar0(GEN N, long flag)
 {
   GEN F = NULL, P, E, cyc, gen, mod, G;
   long i, i0, l, nbprimes;
   pari_sp av = avma;
-  const long dogen = flag & nf_GEN;
 
+  if (flag && flag != 1) pari_err_FLAG("znstar");
   if ((F = check_arith_all(N,"znstar")))
   {
     F = clean_Z_factor(F);
@@ -334,20 +550,15 @@ ZNstar(GEN N, long flag)
   }
   if (!signe(N))
   {
+    if (flag) pari_err_IMPL("znstar(0,1)");
     avma = av;
-    if (flag & nf_GEN)
-      retmkvec3(gen_2, mkvec(gen_2), mkvec(gen_m1));
-    else
-      retmkvec2(gen_2, mkvec(gen_2));
+    retmkvec3(gen_2, mkvec(gen_2), mkvec(gen_m1));
   }
-  if (signe(N) < 0) N = absi(N);
+  N = absi_shallow(N);
   if (abscmpiu(N,2) <= 0)
   {
-    if (flag & nf_GEN)
-      G = mkvec3(gen_1, cgetg(1,t_VEC), cgetg(1,t_VEC));
-    else
-      G = mkvec2(gen_1, cgetg(1,t_VEC));
-    if (flag & nf_INIT)
+    G = mkvec3(gen_1, cgetg(1,t_VEC), cgetg(1,t_VEC));
+    if (flag)
     {
       GEN v = const_vec(6,cgetg(1,t_VEC));
       gel(v,3) = cgetg(1,t_MAT);
@@ -415,7 +626,7 @@ ZNstar(GEN N, long flag)
     gel(mod,i) = Q;
   }
   /* gen[i] has order cyc[i] and generates (Z/mod[i]Z)^* */
-  if (dogen && nbprimes > 1) /* lift generators to (Z/NZ)^*, = 1 mod N/mod[i] */
+  if (nbprimes > 1) /* lift generators to (Z/NZ)^*, = 1 mod N/mod[i] */
     for (i=1; i<l; i++)
     {
       GEN Q = gel(mod,i), g = gel(gen,i), qinv = Fp_inv(Q, diviiexact(N,Q));
@@ -424,25 +635,22 @@ ZNstar(GEN N, long flag)
     }
 
   /* cyc[i] > 1 and remain so in the loop, gen[i] = 1 mod (N/mod[i]) */
-  if (!(flag & nf_INIT))
+  if (!flag)
   { /* update generators in place; about twice faster */
     G = gen;
     for (i=l-1; i>=2; i--)
     {
-      GEN ci = gel(cyc,i), gi = dogen? gel(G,i): NULL;
+      GEN ci = gel(cyc,i), gi = gel(G,i);
       long j;
       for (j=i-1; j>=1; j--) /* we want cyc[i] | cyc[j] */
       {
-        GEN cj = gel(cyc,j), qj, v, d;
+        GEN cj = gel(cyc,j), gj, qj, v, d;
 
-        d = dogen? bezout(ci,cj,NULL,&v): gcdii(ci,cj); /* > 1 */
+        d = bezout(ci,cj,NULL,&v); /* > 1 */
         if (absequalii(ci, d)) continue; /* ci | cj */
         if (absequalii(cj, d)) { /* cj | ci */
-          if (dogen)
-          {
-            swap(gel(G,j),gel(G,i));
-            gi = gel(G,i);
-          }
+          swap(gel(G,j),gel(G,i));
+          gi = gel(G,i);
           swap(gel(cyc,j),gel(cyc,i));
           ci = gel(cyc,i); continue;
         }
@@ -453,24 +661,18 @@ ZNstar(GEN N, long flag)
 
         /* [1,v*cj/d; 0,1]*[1,0;-1,1]*diag(cj,ci)*[ci/d,-v; cj/d,u]
          * = diag(lcm,gcd), with u ci + v cj = d */
-        if (dogen)
-        {
-          GEN gj = gel(G,j);
-          /* (gj, gi) *= [1,0; -1,1]^-1 */
-          gj = Fp_mul(gj, gi, N); /* order ci*qj = lcm(ci,cj) */
-          /* (gj,gi) *= [1,v*qj; 0,1]^-1 */
-          togglesign_safe(&v);
-          if (signe(v) < 0) v = modii(v,ci); /* >= 0 to avoid inversions */
-          gel(G,i) = gi = Fp_mul(gi, Fp_pow(gj, mulii(qj, v), N), N);
-          gel(G,j) = gj;
-        }
+        gj = gel(G,j);
+        /* (gj, gi) *= [1,0; -1,1]^-1 */
+        gj = Fp_mul(gj, gi, N); /* order ci*qj = lcm(ci,cj) */
+        /* (gj,gi) *= [1,v*qj; 0,1]^-1 */
+        togglesign_safe(&v);
+        if (signe(v) < 0) v = modii(v,ci); /* >= 0 to avoid inversions */
+        gel(G,i) = gi = Fp_mul(gi, Fp_pow(gj, mulii(qj, v), N), N);
+        gel(G,j) = gj;
         ci = d; if (absequaliu(ci, 2)) break;
       }
     }
-    if (dogen)
-      G = mkvec3(ZV_prod(cyc), cyc, FpV_to_mod(G,N));
-    else
-      G = mkvec2(ZV_prod(cyc), cyc);
+    G = mkvec3(ZV_prod(cyc), cyc, FpV_to_mod(G,N));
   }
   else
   { /* keep matrices between generators, return an 'init' structure */
@@ -492,33 +694,16 @@ ZNstar(GEN N, long flag)
       }
       gel(lo,i) = t;
     }
-    if (dogen)
-    {
-      G = cgetg(l, t_VEC);
-      for (i = 1; i < l; i++) gel(G,i) = FpV_factorback(gen, gel(Ui,i), N);
-      G = mkvec3(ZV_prod(D), D, G);
-    }
-    else
-      G = mkvec2(ZV_prod(D), D);
+    G = cgetg(l, t_VEC);
+    for (i = 1; i < l; i++) gel(G,i) = FpV_factorback(gen, gel(Ui,i), N);
+    G = mkvec3(ZV_prod(D), D, G);
     G = mkvec5(mkvec2(N,mkvec(gen_0)), G, F,
                mkvecn(6,mod,fao,Ui,gen,cyc,lo), U);
   }
   return gerepilecopy(av, G);
 }
 GEN
-znstar(GEN N) { return ZNstar(N, nf_GEN); }
-GEN
-znstar0(GEN N,long flag)
-{
-  switch(flag)
-  {
-    case 0: return ZNstar(N, nf_GEN);
-    case 1: return ZNstar(N, nf_INIT);
-    case 2: return ZNstar(N, nf_INIT|nf_GEN);
-    default: pari_err_FLAG("znstar");
-  }
-  return NULL; /* not reached */
-}
+znstar(GEN N) { return znstar0(N, 0); }
 
 /* g has order 2^(e-2), g,h = 1 (mod 4); return x s.t. g^x = h (mod 2^e) */
 static GEN
@@ -557,17 +742,14 @@ int
 znconrey_check(GEN cyc, GEN chi)
 { return typ(chi) == t_COL && lg(chi) == lg(cyc) && RgV_is_ZV(chi); }
 
-static GEN
-bidZ_get_cycg(GEN G) { return gmael(G,4,5); }
-
 int
 zncharcheck(GEN G, GEN chi)
 {
   switch(typ(chi))
   {
     case t_INT: return 1;
-    case t_COL: return znconrey_check(bidZ_get_cycg(G), chi);
-    case t_VEC: return char_check(bid_get_cyc(G), chi);
+    case t_COL: return znconrey_check(znstar_get_conreycyc(G), chi);
+    case t_VEC: return char_check(znstar_get_cyc(G), chi);
   }
   return 0;
 }
@@ -575,11 +757,11 @@ zncharcheck(GEN G, GEN chi)
 GEN
 znconreyfromchar_normalized(GEN bid, GEN chi)
 {
-  GEN nchi, U = bid_get_U(bid);
+  GEN nchi, U = znstar_get_U(bid);
   long l = lg(chi);
   if (l == 1) retmkvec2(gen_1,cgetg(1,t_VEC));
   if (!RgV_is_ZV(chi) || lgcols(U) != l) pari_err_TYPE("lfunchiZ", chi);
-  nchi = char_normalize(chi, cyc_normalize(bid_get_cyc(bid)));
+  nchi = char_normalize(chi, cyc_normalize(znstar_get_cyc(bid)));
   gel(nchi,2) = ZV_ZM_mul(gel(nchi,2),U); return nchi;
 }
 
@@ -587,7 +769,7 @@ GEN
 znconreyfromchar(GEN bid, GEN chi)
 {
   GEN nchi = znconreyfromchar_normalized(bid, chi);
-  GEN v = char_denormalize(bidZ_get_cycg(bid), gel(nchi,1), gel(nchi,2));
+  GEN v = char_denormalize(znstar_get_conreycyc(bid), gel(nchi,1), gel(nchi,2));
   settyp(v, t_COL); return v;
 }
 
@@ -599,12 +781,11 @@ znconreylog(GEN bid, GEN x)
   pari_sp av = avma;
   GEN N, L, F, P,E, y, pe, fao, gen, lo, cycg;
   long i, l;
-  if (!checkbidZ_i(bid)) pari_err_TYPE("znconreylog", bid);
-  N = bid_get_ideal(bid);
+  if (!checkznstar_i(bid)) pari_err_TYPE("znconreylog", bid);
+  N = znstar_get_N(bid);
   if (typ(N) != t_INT) pari_err_TYPE("znconreylog", N);
   if (abscmpiu(N, 2) <= 0) return cgetg(1, t_COL);
-  L = gel(bid,4);
-  cycg = bidZ_get_cycg(bid);
+  cycg = znstar_get_conreycyc(bid);
   switch(typ(x))
   {
     GEN Ui;
@@ -612,19 +793,20 @@ znconreylog(GEN bid, GEN x)
       if (!signe(x)) pari_err_COPRIME("znconreylog", x, N);
       break;
     case t_COL: /* log_bid(x) */
-      Ui = gel(L,3);
+      Ui = znstar_get_Ui(bid);
       if (!RgV_is_ZV(x) || lg(x) != lg(Ui)) pari_err_TYPE("znconreylog", x);
       return gerepileupto(av, vecmodii(ZM_ZC_mul(Ui,x), cycg));
     case t_VEC:
       return gerepilecopy(av, znconreyfromchar(bid, x));
     default: pari_err_TYPE("znconreylog", x);
   }
-  F = bid_get_fact(bid); /* factor(N) */
+  F = znstar_get_faN(bid); /* factor(N) */
   P = gel(F, 1); /* prime divisors of N */
   E = gel(F, 2); /* exponents */
-  pe = gel(L,1);
+  L = gel(bid,4);
+  pe = znstar_get_pe(bid);
   fao = gel(L,2);
-  gen = gel(L,4); /* local generators of (Z/p^k)^* */
+  gen = znstar_get_conreygen(bid); /* local generators of (Z/p^k)^* */
   lo = gel(L,6); /* 1/log_p((g_i)^(p_i-1)) */
 
   l = lg(gen); i = 1;
@@ -661,7 +843,7 @@ GEN
 Zideallog(GEN bid, GEN x)
 {
   pari_sp av = avma;
-  GEN y = znconreylog(bid, x), U = bid_get_U(bid);
+  GEN y = znconreylog(bid, x), U = znstar_get_U(bid);
   return gerepileupto(av, ZM_ZC_mul(U, y));
 }
 GEN
@@ -671,8 +853,8 @@ znlog0(GEN h, GEN g, GEN o)
   {
     GEN N;
     if (o) pari_err_TYPE("znlog [with znstar]", o);
-    if (!checkbidZ_i(g)) pari_err_TYPE("znlog", h);
-    N = bid_get_ideal(g);
+    if (!checkznstar_i(g)) pari_err_TYPE("znlog", h);
+    N = znstar_get_N(g);
     h = Rg_to_Fp(h,N);
     return Zideallog(g, h);
   }
@@ -684,10 +866,10 @@ znconreyexp(GEN bid, GEN x)
 {
   pari_sp av = avma;
   long i, l;
-  GEN N, L, pe, gen, cycg, v, vmod;
+  GEN N, pe, gen, cycg, v, vmod;
   int e2;
-  if (!checkbidZ_i(bid)) pari_err_TYPE("znconreyexp", bid);
-  cycg = bidZ_get_cycg(bid);
+  if (!checkznstar_i(bid)) pari_err_TYPE("znconreyexp", bid);
+  cycg = znstar_get_conreycyc(bid);
   switch(typ(x))
   {
     case t_VEC:
@@ -697,11 +879,11 @@ znconreyexp(GEN bid, GEN x)
       if (RgV_is_ZV(x) && lg(x) == lg(cycg)) break;
     default: pari_err_TYPE("znconreyexp",x);
   }
-  L = gel(bid,4);
-  pe = gel(L,1);
-  gen = gel(L,4); /* local generators of (Z/p^k)^* */
+  pe = znstar_get_pe(bid);
+  gen = znstar_get_conreygen(bid); /* local generators of (Z/p^k)^* */
+  cycg = znstar_get_conreycyc(bid);
   l = lg(x); v = cgetg(l, t_VEC);
-  N = bid_get_ideal(bid);
+  N = znstar_get_N(bid);
   e2 = !mod8(N); /* 2 generators at p = 2 */
   for (i = 1; i < l; i++)
   {
@@ -729,7 +911,7 @@ GEN
 znconreychar(GEN bid, GEN m)
 {
   pari_sp av = avma;
-  GEN c, d, L, Ui, nchi;
+  GEN c, d, nchi;
 
   switch(typ(m))
   {
@@ -739,13 +921,11 @@ znconreychar(GEN bid, GEN m)
       break;
     default:
       pari_err_TYPE("znconreychar",m);
-      return NULL;/*not reached*/
+      return NULL;/*LCOV_EXCL_LINE*/
   }
-  L = gel(bid,4);
-  Ui = gel(L,3);
   d = gel(nchi,1);
-  c = ZV_ZM_mul(gel(nchi,2), Ui); /* images of bid gens */
-  return gerepilecopy(av, char_denormalize(bid_get_cyc(bid),d,c));
+  c = ZV_ZM_mul(gel(nchi,2), znstar_get_Ui(bid)); /* images of bid gens */
+  return gerepilecopy(av, char_denormalize(znstar_get_cyc(bid),d,c));
 }
 
 /* chi a t_INT or Conrey log describing a character. Return conductor, as an
@@ -761,16 +941,16 @@ znconreyconductor(GEN bid, GEN chi, GEN *pm)
   long i, j, l;
   int e2, primitive = 1;
 
-  if (!checkbidZ_i(bid)) pari_err_TYPE("znconreyconductor", bid);
+  if (!checkznstar_i(bid)) pari_err_TYPE("znconreyconductor", bid);
   if (typ(chi) == t_COL)
   {
-    if (!znconrey_check(bidZ_get_cycg(bid), chi))
+    if (!znconrey_check(znstar_get_conreycyc(bid), chi))
       pari_err_TYPE("znconreyconductor",chi);
   }
   else
     chi = znconreylog(bid, chi);
   l = lg(chi);
-  F = bid_get_fact(bid);
+  F = znstar_get_faN(bid);
   P = gel(F,1);
   E = gel(F,2);
   if (l == 1)
@@ -829,7 +1009,7 @@ znconreyconductor(GEN bid, GEN chi, GEN *pm)
   if (pm) *pm = m; /* attached primitive  character */
   if (primitive)
   {
-    q = bid_get_ideal(bid);
+    q = znstar_get_N(bid);
     if (mod4(q) == 2) primitive = 0;
   }
   if (!primitive)
@@ -839,7 +1019,7 @@ znconreyconductor(GEN bid, GEN chi, GEN *pm)
       P = vecsplice(P,1);
       E = vecsplice(E,1);
     }
-    E = zv_to_ZV(E);
+    E = zc_to_ZC(E);
     q = mkvec2(factorback2(P,E), mkmat2(P,E));
   }
   gerepileall(av, pm? 2: 1, &q, pm);
@@ -854,16 +1034,16 @@ zncharinduce(GEN G, GEN chi, GEN N)
   long i, j, l;
   int e2;
 
-  if (!checkbidZ_i(G)) pari_err_TYPE("zncharinduce", G);
+  if (!checkznstar_i(G)) pari_err_TYPE("zncharinduce", G);
   if (!zncharcheck(G, chi)) pari_err_TYPE("zncharinduce", chi);
-  q = bid_get_ideal(G);
+  q = znstar_get_N(G);
   if (typ(chi) != t_COL) chi = znconreylog(G, chi);
-  if (checkbidZ_i(N))
+  if (checkznstar_i(N))
   {
-    GEN faN = bid_get_fact(N);
+    GEN faN = znstar_get_faN(N);
     P = gel(faN,1); l = lg(P);
     E = gel(faN,2);
-    N = bid_get_ideal(N);
+    N = znstar_get_N(N);
     if (l > 2 && equalii(gel(P,1),gel(P,2)))
     { /* remove duplicate 2 */
       l--;
@@ -902,7 +1082,7 @@ zncharinduce(GEN G, GEN chi, GEN N)
     return equali1(N)? cgetg(1, t_COL): zerocol(l+e2 - 1);
   }
 
-  faq = bid_get_fact(G);
+  faq = znstar_get_faN(G);
   Pq = gel(faq,1);
   Eq = gel(faq,2);
   CHI = cgetg(l+e2, t_COL);
@@ -942,7 +1122,7 @@ zncharinduce(GEN G, GEN chi, GEN N)
 GEN
 znconreylog_normalize(GEN G, GEN m)
 {
-  GEN cycg = bidZ_get_cycg(G);
+  GEN cycg = znstar_get_conreycyc(G);
   long i, l;
   GEN d, M = cgetg_copy(m, &l);
   if (typ(cycg) != t_VEC || lg(cycg) != l)
@@ -970,7 +1150,7 @@ znconrey_normalized(GEN G, GEN chi)
       return znconreyfromchar_normalized(G, chi);
   }
   pari_err_TYPE("znchareval",chi);
-  return NULL;/* not reached */
+  return NULL;/* LCOV_EXCL_LINE */
 }
 
 /* return 1 iff chi(-1) = -1, and 0 otherwise */
@@ -979,10 +1159,10 @@ zncharisodd(GEN G, GEN chi)
 {
   long i, l, s;
   GEN N;
-  if (!checkbidZ_i(G)) pari_err_TYPE("zncharisodd", G);
+  if (!checkznstar_i(G)) pari_err_TYPE("zncharisodd", G);
   if (!zncharcheck(G, chi)) pari_err_TYPE("zncharisodd", chi);
   if (typ(chi) != t_COL) chi = znconreylog(G, chi);
-  N = bid_get_ideal(G);
+  N = znstar_get_N(G);
   l = lg(chi);
   s = 0;
   if (!mod8(N))
@@ -1007,7 +1187,7 @@ znchartokronecker(GEN G, GEN chi, long flag)
   s = zncharisodd(G, chi)? -1: 1;
   if (typ(chi) != t_COL) chi = znconreylog(G, chi);
   o = zncharorder(G, chi);
-  if (cmpiu(o,2) > 0) { avma = av; return gen_0; }
+  if (abscmpiu(o,2) > 0) { avma = av; return gen_0; }
   F = znconreyconductor(G, chi, NULL);
   if (typ(F) == t_INT)
   {
@@ -1018,7 +1198,7 @@ znchartokronecker(GEN G, GEN chi, long flag)
   F = (s < 0)? negi(F): icopy(F);
   if (!flag)
   {
-    GEN MF = bid_get_fact(G), P = gel(MF,1);
+    GEN MF = znstar_get_faN(G), P = gel(MF,1);
     long i, l = lg(P);
     for (i = 1; i < l; i++)
     {
@@ -1029,11 +1209,66 @@ znchartokronecker(GEN G, GEN chi, long flag)
   return gerepileuptoint(av, F);
 }
 
-/* G a bidZ, not stack clean */
+/* (D/.) as a character mod N; assume |D| divides N and D = 0,1 mod 4*/
+GEN
+znchar_quad(GEN G, GEN D)
+{
+  GEN cyc = znstar_get_conreycyc(G);
+  GEN gen = znstar_get_conreygen(G);
+  long i, l = lg(cyc);
+  GEN chi = cgetg(l, t_COL);
+  for (i = 1; i < l; i++)
+  {
+    long k = kronecker(D, gel(gen,i));
+    gel(chi,i) = (k==1)? gen_0: shifti(gel(cyc,i), -1);
+  }
+  return chi;
+}
+
+GEN
+znchar(GEN D)
+{
+  pari_sp av = avma;
+  GEN G, chi;
+  switch(typ(D))
+  {
+    case t_INT:
+      if (!signe(D) || Mod4(D) > 1) pari_err_TYPE("znchar", D);
+      G = znstar0(D,1);
+      chi = mkvec2(G, znchar_quad(G,D));
+      break;
+    case t_INTMOD:
+      G = znstar0(gel(D,1), 1);
+      chi = mkvec2(G, znconreylog(G, gel(D,2)));
+      break;
+    case t_VEC:
+      if (checkMF_i(D)) { chi = vecslice(MF_get_CHI(D),1,2); break; }
+      else if (checkmf_i(D)) { chi = vecslice(mf_get_CHI(D),1,2); break; }
+      if (lg(D) != 3) pari_err_TYPE("znchar", D);
+      G = gel(D,1);
+      if (!checkznstar_i(G)) pari_err_TYPE("znchar", D);
+      chi = gel(D,2);
+      if (typ(chi) == t_VEC && lg(chi) == 3 && is_vec_t(typ(gel(chi,2))))
+      { /* normalized character */
+        GEN n = gel(chi,1), chic = gel(chi,2);
+        GEN cyc = typ(chic)==t_VEC? znstar_get_cyc(G): znstar_get_conreycyc(G);
+        if (!char_check(cyc, chic)) pari_err_TYPE("znchar",D);
+        chi = char_denormalize(cyc, n, chic);
+      }
+      if (!zncharcheck(G, chi)) pari_err_TYPE("znchar", D);
+      chi = mkvec2(G,chi); break;
+    default:
+      pari_err_TYPE("znchar", D);
+      return NULL; /*LCOV_EXCL_LINE*/
+  }
+  return gerepilecopy(av, chi);
+}
+
+/* G a znstar, not stack clean */
 GEN
 znchareval(GEN G, GEN chi, GEN n, GEN z)
 {
-  GEN nchi, N = bid_get_ideal(G);
+  GEN nchi, N = znstar_get_N(G);
   /* avoid division by 0 */
   if (typ(n) == t_FRAC && !equali1(gcdii(gel(n,2), N))) return not_coprime(z);
   n = Rg_to_Fp(n, N);
@@ -1043,74 +1278,145 @@ znchareval(GEN G, GEN chi, GEN n, GEN z)
   return chareval_i(nchi, znconreylog(G,n), z);
 }
 
-/* G is a bidZ, chi a Dirichlet character */
+/* G is a znstar, chi a Dirichlet character */
 GEN
 zncharconj(GEN G, GEN chi)
 {
   switch(typ(chi))
   {
     case t_INT: chi = znconreylog(G, chi); /* fall through */
-    case t_COL: return charconj(bidZ_get_cycg(G), chi);
-    case t_VEC: return charconj(bid_get_cyc(G), chi);
+    case t_COL: return charconj(znstar_get_conreycyc(G), chi);
+    case t_VEC: return charconj(znstar_get_cyc(G), chi);
   }
   pari_err_TYPE("zncharconj",chi);
-  return NULL; /*not reached*/
+  return NULL; /*LCOV_EXCL_LINE*/
 }
 
-/* G is a bidZ, chi a Dirichlet character */
+/* G is a znstar, chi a Dirichlet character */
 GEN
 zncharorder(GEN G,  GEN chi)
 {
   switch(typ(chi))
   {
     case t_INT: chi = znconreylog(G, chi); /*fall through*/
-    case t_COL: return charorder(bidZ_get_cycg(G), chi);
-    case t_VEC: return charorder(bid_get_cyc(G), chi);
+    case t_COL: return charorder(znstar_get_conreycyc(G), chi);
+    case t_VEC: return charorder(znstar_get_cyc(G), chi);
     default: pari_err_TYPE("zncharorder",chi);
-             return NULL; /* not reached */
+             return NULL; /* LCOV_EXCL_LINE */
   }
 }
 
-/* G is a bidZ, chi a Dirichlet character */
+/* G is a znstar, chi a Dirichlet character */
 GEN
 zncharker(GEN G, GEN chi)
 {
   if (typ(chi) != t_VEC) chi = znconreychar(G, chi);
-  return charker(bid_get_cyc(G), chi);
+  return charker(znstar_get_cyc(G), chi);
 }
 
-/* G is a bidZ, 'a' and 'b' are Dirichlet character */
+/* G is a znstar, 'a' is a Dirichlet character */
+GEN
+zncharpow(GEN G, GEN a, GEN n)
+{
+  switch(typ(a))
+  {
+    case t_INT: return Fp_pow(a, n, znstar_get_N(G));
+    case t_VEC: return charpow(znstar_get_cyc(G), a, n);
+    case t_COL: return charpow(znstar_get_conreycyc(G), a, n);
+    default: pari_err_TYPE("znchapow",a);
+             return NULL; /* LCOV_EXCL_LINE */
+  }
+}
+/* G is a znstar, 'a' and 'b' are Dirichlet character */
 GEN
 zncharmul(GEN G, GEN a, GEN b)
 {
   long ta = typ(a), tb = typ(b);
   if (ta == tb) switch(ta)
   {
-    case t_INT: return Fp_mul(a, b, bid_get_ideal(G));
-    case t_VEC: return charmul(bid_get_cyc(G), a, b);
-    case t_COL: return charmul(bidZ_get_cycg(G), a, b);
+    case t_INT: return Fp_mul(a, b, znstar_get_N(G));
+    case t_VEC: return charmul(znstar_get_cyc(G), a, b);
+    case t_COL: return charmul(znstar_get_conreycyc(G), a, b);
     default: pari_err_TYPE("zncharmul",a);
-             return NULL; /* not reached */
+             return NULL; /* LCOV_EXCL_LINE */
   }
   if (ta != t_COL) a = znconreylog(G, a);
   if (tb != t_COL) b = znconreylog(G, b);
-  return charmul(bidZ_get_cycg(G), a, b);
+  return charmul(znstar_get_conreycyc(G), a, b);
 }
 
-/* G is a bidZ, 'a' and 'b' are Dirichlet character */
+/* G is a znstar, 'a' and 'b' are Dirichlet character */
 GEN
 znchardiv(GEN G, GEN a, GEN b)
 {
   long ta = typ(a), tb = typ(b);
   if (ta == tb) switch(ta)
   {
-    case t_INT: return Fp_div(a, b, bid_get_ideal(G));
-    case t_VEC: return chardiv(bid_get_cyc(G), a, b);
-    case t_COL: return chardiv(bidZ_get_cycg(G), a, b);
+    case t_INT: return Fp_div(a, b, znstar_get_N(G));
+    case t_VEC: return chardiv(znstar_get_cyc(G), a, b);
+    case t_COL: return chardiv(znstar_get_conreycyc(G), a, b);
     default: pari_err_TYPE("znchardiv",a);
-             return NULL; /* not reached */
+             return NULL; /* LCOV_EXCL_LINE */
   }
   if (ta != t_COL) a = znconreylog(G, a);
   if (tb != t_COL) b = znconreylog(G, b);
-  return chardiv(bidZ_get_cycg(G), a, b);
+  return chardiv(znstar_get_conreycyc(G), a, b);
+}
+
+/* CHI mod N = \prod_p p^e; let CHI = \prod CHI_p, CHI_p mod p^e
+ * return \prod_{p | (Q,N)} CHI_p. E.g if Q = p, return chi_p */
+GEN
+znchardecompose(GEN G, GEN chi, GEN Q)
+{
+  GEN c, P, E, F;
+  long l, lP, i;
+
+  if (!checkznstar_i(G)) pari_err_TYPE("znchardecompose", G);
+  if (typ(Q) != t_INT) pari_err_TYPE("znchardecompose", Q);
+  if (typ(chi) == t_COL)
+  { if (!zncharcheck(G, chi)) pari_err_TYPE("znchardecompose", chi); }
+  else
+    chi = znconreylog(G, chi);
+  l = lg(chi);
+  F = znstar_get_faN(G);
+  c = zerocol(l-1);
+  P = gel(F,1); /* prime divisors of N */
+  lP = lg(P);
+  E = gel(F,2); /* exponents */
+  for (i = 1; i < lP; i++)
+  {
+    GEN p = gel(P,i);
+    if (i == 1 && equaliu(p,2) && E[1] >= 3)
+    {
+      if (!mpodd(Q))
+      {
+        gel(c,1) = icopy(gel(chi,1));
+        gel(c,2) = icopy(gel(chi,2));
+      }
+      i = 2; /* skip P[2] = P[1] = 2 */
+    }
+    else
+      if (dvdii(Q, p)) gel(c,i) = icopy(gel(chi,i));
+  }
+  return c;
+}
+
+GEN
+zncharconductor(GEN G, GEN chi)
+{
+  pari_sp av = avma;
+  GEN F = znconreyconductor(G, chi, NULL);
+  if (typ(F) == t_INT) return F;
+  return gerepilecopy(av, gel(F,1));
+}
+GEN
+znchartoprimitive(GEN G, GEN chi)
+{
+  pari_sp av = avma;
+  GEN chi0, F = znconreyconductor(G, chi, &chi0);
+  if (typ(F) == t_INT)
+    chi = mkvec2(G,chi);
+  else
+    chi = mkvec2(znstar0(F,1), chi0);
+  return gerepilecopy(av, chi);
 }
