@@ -155,7 +155,7 @@ logboundG0(long e, double aY)
   double cla, loggam;
   cla = 1 + 1/sqrt(aY);
   if (e) cla = ( cla + 1/(2*aY) ) / (2*sqrt(aY));
-  loggam = (e) ? LOG2-aY : -aY + log( log( 1+1/aY) );
+  loggam = (e) ? M_LN2-aY : -aY + log( log( 1+1/aY) );
   return log(cla) + loggam;
 }
 
@@ -166,11 +166,11 @@ param_points(GEN N, double Y, double tmax, long bprec, long *cprec, long *L,
   double D, a, aY, X, logM;
   long d = 2, w = 1;
   tmax *= d;
-  D = bprec * LOG2 + M_PI/4*tmax + 2;
-  *cprec = nbits2prec(ceil(D / LOG2) + 5);
+  D = bprec * M_LN2 + M_PI/4*tmax + 2;
+  *cprec = nbits2prec(ceil(D / M_LN2) + 5);
   a = 2 * M_PI / sqrt(gtodouble(N));
   aY = a * cos(M_PI/2*Y);
-  logM = 2*LOG2 + logboundG0(w+1, aY) + tmax * Y * M_PI/2;
+  logM = 2*M_LN2 + logboundG0(w+1, aY) + tmax * Y * M_PI/2;
   *h = ( 2 * M_PI * M_PI / 2 * Y ) / ( D + logM );
   X = log( D / a);
   *L = ceil( X / *h);
@@ -266,7 +266,7 @@ ellL1_add(void *E, GEN n, GEN a)
   for (j = 1; j < l; j++)
     if (cmpii(n, gel(bb->bnd,j)) <= 0)
     {
-      ulong r, q = udiviu_rem(n, bb->rbnd[j], &r);
+      ulong r, q = uabsdiviu_rem(n, bb->rbnd[j], &r);
       GEN giant = gel(bb->giant, j), baby = gel(bb->baby, j);
       affrr(addrr(gel(giant, q+1), mulri(gel(baby, r+1), a)), gel(giant, q+1));
       avma = av;
@@ -340,12 +340,12 @@ glambda(GEN t, GEN vec, GEN h, long real, long prec)
   /* assume vec is a grid */
   ehs = gexp(gmul(gen_I(),gmul(h, t)), prec);
   elhs = (real == 1) ? gen_1 : mkcomplex(gen_0, gen_m1);
-  r = gmul2n(greal(gmul(greal(gel(vec, 1)), elhs)), -1);
+  r = gmul2n(real_i(gmul(real_i(gel(vec, 1)), elhs)), -1);
   /* FIXME: summing backward may be more stable */
   for (l = 2; l <= L; ++l)
   {
     elhs = gmul(elhs, ehs);
-    r = gadd(r, greal(gmul(gel(vec, l), elhs)));
+    r = gadd(r, real_i(gmul(gel(vec, l), elhs)));
   }
   return gmul(mulsr(4, h), r);
 }
@@ -379,8 +379,8 @@ ellgammafactor(GEN N, GEN s, long prec)
 static GEN
 ellL1_eval(GEN e, GEN vec, struct lcritical *C, GEN t, long prec)
 {
-  GEN gam = ellgammafactor(ellQ_get_N(e), gaddgs(gmul(gen_I(),t), 1), prec);
-  return gdiv(Llambda(vec, C, t, prec), gam);
+  GEN g = ellgammafactor(ellQ_get_N(e), gaddgs(gmul(gen_I(),t), 1), prec);
+  return gdiv(Llambda(vec, C, t, prec), g);
 }
 
 static GEN
@@ -388,7 +388,7 @@ ellL1_der(GEN e, GEN vec, struct lcritical *C, GEN t, long der, long prec)
 {
   GEN r = polcoeff0(ellL1_eval(e, vec, C, t, prec), der, 0);
   r = gmul(r,powIs(C->real == 1 ? -der: 1-der));
-  return gmul(r, mpfact(der));
+  return gmul(real_i(r), mpfact(der));
 }
 
 GEN
@@ -466,7 +466,7 @@ heegner_L1_bg(void*E, GEN n, GEN a)
   for (j = 1; j < l; j++)
     if (cmpii(n, gel(bb->bnd,j)) <= 0)
     {
-      ulong r, q = udiviu_rem(n, bb->rbnd[j], &r);
+      ulong r, q = uabsdiviu_rem(n, bb->rbnd[j], &r);
       GEN giant = gel(bb->giant, j), baby = gel(bb->baby, j);
       gaffect(gadd(gel(giant, q+1), gdiv(gmul(gel(baby, r+1), a), n)), gel(giant, q+1));
     }
@@ -480,7 +480,7 @@ heegner_L1(void*E, GEN n, GEN a)
   for (j = 1; j < l; j++)
     if (cmpii(n, gel(bb->bnd,j)) <= 0)
     {
-      ulong r, q = udiviu_rem(n, bb->rbnd[j], &r);
+      ulong r, q = uabsdiviu_rem(n, bb->rbnd[j], &r);
       GEN giant = gel(bb->giant, j), baby = gel(bb->baby, j);
       GEN ex = mulreal(gel(baby, r+1), gel(giant, q+1));
       affrr(addrr(gel(bb->sum, j), divri(mulri(ex, a), n)), gel(bb->sum, j));
@@ -552,7 +552,7 @@ heegner_psi(GEN E, GEN N, GEN points, long bitprec)
   Q  = cgetg(L+1, t_VEC);
   for (l = 1; l <= L; ++l)
   {
-    gel(bnd,l) = ceil_safe(divrr(B,gimag(gel(points, l))));
+    gel(bnd,l) = ceil_safe(divrr(B,imag_i(gel(points, l))));
     rbnd[l] = itou(sqrtint(gel(bnd,l)))+1;
     gel(Q, l) = expIxy(pi2, gel(points, l), prec);
   }
@@ -661,7 +661,7 @@ heightQ(GEN P, long prec)
   }
   s = signe(P);
   if (!s) return real_0(prec);
-  if (s < 0) P = absi(P);
+  if (s < 0) P = negi(P);
   return glog(P, prec);
 }
 
@@ -672,7 +672,7 @@ logplusQ(GEN t, long prec)
   if (typ(t) == t_INT)
   {
     if (!signe(t)) return real_1(prec);
-    if (signe(t) < 0) t = absi(t);
+    if (signe(t) < 0) t = negi(t);
   }
   else
   {
@@ -690,7 +690,7 @@ hnaive_max(GEN ell, GEN ht)
 {
   const long prec = LOWDEFAULTPREC; /* minimal accuracy */
   GEN b2     = ell_get_b2(ell), j = ell_get_j(ell);
-  GEN logd   = glog(absi(ell_get_disc(ell)), prec);
+  GEN logd   = glog(absi_shallow(ell_get_disc(ell)), prec);
   GEN logj   = logplusQ(j, prec);
   GEN hj     = heightQ(j, prec);
   GEN logb2p = signe(b2)? addrr(logplusQ(gdivgs(b2, 12),prec), mplog2(prec))
@@ -758,16 +758,16 @@ best_point_old(GEN Q, GEN NQ, GEN f, GEN *u, GEN *v)
   {
     for (k = -n; k<=n; k++)
     {
-      *u = addis(c, k); *v = addis(d, n);
+      *u = addis(c, k); *v = addiu(d, n);
       if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
         return;
-      *v= subis(d, n);
+      *v= subiu(d, n);
       if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
         return;
-      *u = addis(c, n); *v = addis(d, k);
+      *u = addiu(c, n); *v = addis(d, k);
       if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
         return;
-      *u = subis(c, n);
+      *u = subiu(c, n);
       if (equali1(ggcd(mulii(*u, NQ), mulii(*v, Q))))
         return;
     }
@@ -795,18 +795,15 @@ qfmin_eval(GEN q0, GEN q1, GEN q2, long i)
 static ulong
 gcduii(ulong a, GEN b, GEN c)
 {
-  ulong d = a;
-  d = ugcd(umodiu(b, d), d );
-  if (d == 1) return 1;
-  d = ugcd(umodiu(c, d), d );
-  return d;
+  a = ugcdiu(b, a);
+  return a == 1? 1: ugcdiu(c, a);
 }
 
 static void
 best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
 {
   GEN a = mulii(NQ, gel(f,3)), b = negi(gel(f,2)), c = diviiexact(gel(f,1), NQ);
-  GEN D = absi( qfb_disc(f) );
+  GEN D = absi_shallow( qfb_disc(f) );
   GEN U, qr = redimagsl2(qfi(a,b,c), &U);
   GEN A = gel(qr,1), B = gel(qr,2), A2 = shifti(A,1), AA4 = sqri(A2);
   GEN V, best;
@@ -837,7 +834,7 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
   if (is_pm1(gcdii(*pu, Q)) && is_pm1(gcdii(*pv, NQ))) return;
 
   /* (X,Y) = (1, \pm1) always works. Try to do better now */
-  best = subii(addii(a, c), absi(b));
+  best = subii(addii(a, c), absi_shallow(b));
   *pu = gen_1;
   *pv = signe(b) < 0? gen_1: gen_m1;
 
@@ -853,7 +850,7 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
     Dy2 = mulii(D, sqru(y));
     if (cmpii(Dy2, best) >= 0) break; /* we won't improve. STOP */
     By = muliu(B,y), x0 = truedvmdii(negi(By), A2, &r);
-    if (cmpii(r, A) >= 0) { x0 = subis(x0,1); r = subii(r, A2); }
+    if (cmpii(r, A) >= 0) { x0 = subiu(x0,1); r = subii(r, A2); }
     /* (2A x + By)^2 + Dy^2, minimal at x = x0. Assume A2 > 0 */
     /* r = 2A x0 + By */
     q0 = addii(sqri(r), Dy2); /* minimal value for this y, at x0 */
@@ -867,7 +864,7 @@ best_point(GEN Q, GEN NQ, GEN f, GEN *pu, GEN *pv)
       GEN x, N = qfmin_eval(q0, q1, AA4, i);
       if (cmpii(N , best) >= 0) break;
       x = addis(x0, i);
-      if (ugcd(umodiu(x, y), y) == 1)
+      if (ugcdiu(x, y) == 1)
       {
         GEN u, v;
         V = ZC_add(V0, ZC_z_mul(gel(U,1), i)); /* [X, Y] */
@@ -948,7 +945,7 @@ static GEN
 heegner_indexmult(GEN om, long t, GEN tam, long prec)
 {
   pari_sp av = avma;
-  GEN Ovr = gabs(gimag(gel(om, 2)), prec); /* O_vol/O_re, t_REAL */
+  GEN Ovr = gabs(imag_i(gel(om, 2)), prec); /* O_vol/O_re, t_REAL */
   return gerepileupto(av, divru(divir(tam, Ovr), 4*t*t));
 }
 
@@ -985,7 +982,7 @@ heegner_try_point(GEN E, GEN lambdas, GEN ht, GEN z, long prec)
 {
   long l = lg(lambdas);
   long i, eps;
-  GEN P = greal(pointell(E, z, prec)), x = gel(P,1);
+  GEN P = real_i(pointell(E, z, prec)), x = gel(P,1);
   GEN rh = subrr(ht, shiftr(ellheightoo(E, P, prec),1));
   for (i = 1; i < l; ++i)
   {
@@ -1039,7 +1036,7 @@ heegner_find_point(GEN e, GEN om, GEN ht, GEN z1, long k, long prec)
     avma = av;
   }
   pari_err_BUG("ellheegner, point not found");
-  return NULL; /* NOT REACHED */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* N > 1, fa = factor(N), return factor(4*N) */
@@ -1050,7 +1047,7 @@ fa_shift2(GEN fa)
   if (absequaliu(gcoeff(fa,1,1), 2))
   {
     E = shallowcopy(E);
-    gel(E,1) = addis(gel(E,1), 2);
+    gel(E,1) = addiu(gel(E,1), 2);
   }
   else
   {
@@ -1095,8 +1092,7 @@ find_div(GEN N, GEN faN)
 }
 
 static long
-testDisc(GEN bad, long d)
-{ return !bad || ugcd(umodiu(bad, -d), -d) == 1; }
+testDisc(GEN bad, long d) { return !bad || ugcdiu(bad, -d) == 1; }
 /* bad = product of bad primes. Return the NDISC largest fundamental
  * discriminants D < d such that (D,bad) = 1 and d is a square mod 4N */
 static GEN
@@ -1283,7 +1279,7 @@ ellheegner(GEN E)
     pari_err_DOMAIN("ellheegner", "(analytic rank)%2","=",gen_0,E);
   torsion = elltors(E);
   wtor = itos( gel(torsion,1) ); /* #E(Q)_tor */
-  etor = wtor > 1? itos(gmael(torsion, 2, 1)): 1; /* exponent of E(Q)_tor */
+  etor = wtor > 1? itou(gmael(torsion, 2, 1)): 1; /* exponent of E(Q)_tor */
   while (1)
   {
     GEN hnaive, l1;
@@ -1314,7 +1310,7 @@ ellheegner(GEN E)
   for (k = 2; k < l; ++k) z = addrr(z, mulsr(coefs[k], gel(s, k)));
   if (DEBUGLEVEL) err_printf("z=%Ps\n", z);
   z = gsub(z, gmul(gel(om,1), ground(gdiv(z, gel(om,1)))));
-  lint = wtor > 1 ? cgcd(ind, etor): 1;
+  lint = wtor > 1 ? ugcd(ind, etor): 1;
   P = heegner_find_point(E, om, ht, gmulsg(2*lint, z), lint*2*ind, prec);
   if (DEBUGLEVEL) timer_printf(&ti,"heegner_find_point");
   if (cb) P = ellchangepointinv(P, cb);
@@ -1323,18 +1319,38 @@ ellheegner(GEN E)
 
 /* Modular degree */
 
-/* Modular degree of elliptic curve e over Q, assuming Manin constant = 1
-   (otherwise multiply by square of Manin constant). */
-GEN
-ellmoddegree(GEN e, long bitprec)
+static GEN
+ellisobound(GEN e)
 {
-  pari_sp ltop = avma;
-  long prec = nbits2prec(bitprec);
-  GEN E = ellminimalmodel(e, NULL);
-  GEN nor = lfunellmfpeters(E, bitprec);
-  GEN deg = gdiv(gmul(nor, sqrr(Pi2n(1, prec))), member_area(E));
-  GEN degr = bestappr(deg, int2n(bitprec>>1));
-  long err = gexpo(gsub(gen_1, gdiv(deg,degr)));
-  obj_free(E);
-  return gerepilecopy(ltop, mkvec2(degr, stoi(err)));
+  GEN v, M = gel(ellisomat(e,0,1),2);
+  long i, j, k, n = lg(M)-1;
+  v = cgetg((n-1)*(n-2) / 2 + 1, t_VEC);
+  for (i = 1, k = 1; i <= n; i++)
+    for (j = i+1; j <= n; j++) gel(v,k++) = gcoeff(M,i,j);
+  return glcm0(v, NULL);
+}
+
+/* Modular degree of elliptic curve e over Q, assuming Manin constant = 1
+ * (otherwise multiply by square of Manin constant). */
+GEN
+ellmoddegree(GEN E)
+{
+  pari_sp av = avma;
+  GEN N, tam, mc2, d;
+  long b;
+  E = ellanal_globalred_all(E, NULL, &N, &tam);
+  mc2 = sqri(ellisobound(E));
+  b = expi(mulii(N, mc2))+16;
+  while(1)
+  {
+    long  prec = nbits2prec(b), err, s;
+    GEN deg, nor = lfunellmfpeters(E, b);
+    deg = gmul(gdiv(gmul(nor, sqrr(Pi2n(1, prec))), ellR_area(E, prec)), mc2);
+    d = grndtoi(deg, &err);
+    if (DEBUGLEVEL)
+      err_printf("ellmoddegree: %Ps : bitprec=%ld: err=%ld\n",deg,b,err);
+    s = expi(d);
+    if (err <= -8 && s <= b-8) return gerepileupto(av, gdiv(d,mc2));
+    b = maxss(s, b+err) + 16;
+  }
 }
