@@ -21,8 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 static int
 smaller_x(GEN p, GEN q)
 {
-  int s = abscmpii(denom(p), denom(q));
-  return (s<0 || (s==0 && abscmpii(numer(p),numer(q)) < 0));
+  int s = abscmpii(denom_i(p), denom_i(q));
+  return (s<0 || (s==0 && abscmpii(numer_i(p),numer_i(q)) < 0));
 }
 
 /* best generator in cycle of length k */
@@ -344,7 +344,7 @@ ellisdivisible(GEN E, GEN P, GEN n, GEN *pQ)
   {
     case t_INT:
       N = n;
-      if (!isprime(absi(n)))
+      if (!isprime(absi_shallow(n)))
       {
         GEN f = absZ_factor(n), LP = gel(f,1), LE = gel(f,2);
         l = lg(LP);
@@ -399,7 +399,7 @@ ellisdivisible(GEN E, GEN P, GEN n, GEN *pQ)
   avma = av; return 0;
 }
 
-/* 2-torsion point of abcissa x */
+/* 2-torsion point of abscissa x */
 static GEN
 tor2(GEN E, GEN x) { return mkvec2(x, gmul2n(gneg(ec_h_evalx(E,x)), -1)); }
 
@@ -481,7 +481,7 @@ ellnftorsprimary(GEN E, long p, long N1, long N2, long v)
     P2 = Q2;
   }
 
-  /* compute E[p^\infty] = < P1, P2 > */
+  /* compute E[p^oo] = < P1, P2 > */
   n1 = n2;
   if (n2 == N2)
   {
@@ -526,6 +526,17 @@ ellnftorsprimary(GEN E, long p, long N1, long N2, long v)
   return ptor2(p, n1,n2, P1,P2);
 }
 
+/* P affine point */
+static GEN
+nfpt(GEN e, GEN P)
+{
+  GEN T = nf_get_pol(ellnf_get_nf(e));
+  GEN x = gel(P,1), y = gel(P,2);
+  long tx = typ(x), ty = typ(y);
+  if (tx == ty) return P;
+  if (tx != t_POLMOD) x = mkpolmod(x,T); else y = mkpolmod(y,T);
+  return mkvec2(x,y);
+}
 /* Computes the torsion subgroup of E(K), as [order, cyc, gen] */
 static GEN
 ellnftors(GEN e)
@@ -549,8 +560,8 @@ ellnftors(GEN e)
   }
   (void)delete_var();
   if (is_pm1(d1)) return mkvec3(gen_1,cgetg(1,t_VEC),cgetg(1,t_VEC));
-  if (is_pm1(d2)) return mkvec3(d1, mkvec(d1), mkvec(P1));
-  return mkvec3(mulii(d1,d2), mkvec2(d1,d2), mkvec2(P1,P2));
+  if (is_pm1(d2)) return mkvec3(d1, mkvec(d1), mkvec(nfpt(e,P1)));
+  return mkvec3(mulii(d1,d2), mkvec2(d1,d2), mkvec2(nfpt(e,P1),nfpt(e,P2)));
 }
 
 GEN
@@ -588,6 +599,7 @@ ellorder_Q(GEN E, GEN P)
   ulong a4, p;
   long k;
   if (ell_is_inf(P)) return 1;
+  if (gequal(P, ellneg(E,P))) return 2;
 
   dx = Q_denom(gel(P,1));
   dy = Q_denom(gel(P,2));
@@ -631,6 +643,7 @@ ellorder_nf(GEN E, GEN P)
   forprime_t S;
   ulong a4, p;
   if (ell_is_inf(P)) return gen_1;
+  if (gequal(P, ellneg(E,P))) return gen_2;
 
   B = gel(nftorsbound(E), 1);
   dx = Q_denom(gel(P,1));
@@ -680,10 +693,12 @@ ellorder(GEN E, GEN P, GEN o)
   if (ell_is_inf(P)) return gen_1;
   if (ell_get_type(E)==t_ELL_Q)
   {
+    long tx = typ(gel(P,1)), ty = typ(gel(P,2));
     GEN p = NULL;
-    if (is_rational_t(typ(gel(P,1))) && is_rational_t(typ(gel(P,2))))
-      return utoi( ellorder_Q(E, P) );
-    if (RgV_is_FpV(P,&p) && p)
+    if (is_rational_t(tx) && is_rational_t(ty)) return utoi(ellorder_Q(E, P));
+    if (tx == t_INTMOD || tx == t_FFELT) p = gel(P,1);
+    if (!p && (ty == t_INTMOD || ty == t_FFELT)) p = gel(P,2);
+    if (p)
     {
       E = ellinit(E,p,0);
       if (lg(E)==1) pari_err_IMPL("ellorder for curve with singular reduction");
