@@ -21,11 +21,11 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "paripriv.h"
 
 /* assume z[1] was created last */
-#define fix_frac_if_int(z) if (is_pm1(gel(z,2)))\
+#define fix_frac_if_int(z) if (equali1(gel(z,2)))\
   z = gerepileupto((pari_sp)(z+3), gel(z,1));
 
 /* assume z[1] was created last */
-#define fix_frac_if_int_GC(z,tetpil) { if (is_pm1(gel(z,2)))\
+#define fix_frac_if_int_GC(z,tetpil) { if (equali1(gel(z,2)))\
   z = gerepileupto((pari_sp)(z+3), gel(z,1));\
 else\
   gerepilecoeffssp((pari_sp)z, tetpil, z+1, 2); }
@@ -288,13 +288,13 @@ gred_rfrac_simple(GEN n, GEN d)
   {
     z = c;
     do { z = content(z); } while (typ(z) == t_POL);
-    cd = denom(z);
+    cd = denom_i(z);
     cn = gmul(c, cd);
   }
   else
   {
-    cn = numer(c);
-    cd = denom(c);
+    cn = numer_i(c);
+    cd = denom_i(c);
   }
   z = cgetg(3,t_RFRAC);
   gel(z,1) = gmul(n, cn);
@@ -371,7 +371,7 @@ Qdivii(GEN x, GEN y)
   GEN r, q;
 
   if (is_pm1(y)) return (signe(y) < 0)? negi(x): icopy(x);
-  if (is_pm1(x)) {
+  if (equali1(x)) {
     long s = signe(y);
     if (!s) pari_err_INV("gdiv",y);
     if (signe(x) < 0) s = -s;
@@ -413,85 +413,71 @@ static GEN
 quad_polmod_conj(GEN x, GEN y)
 {
   GEN z, u, v, a, b;
-  pari_sp av;
   if (typ(x) != t_POL) return gcopy(x);
   if (varn(x) != varn(y) || degpol(x) <= 0) return RgX_copy(x);
   a = gel(y,4); u = gel(x,3); /*Mod(ux + v, ax^2 + bx + c)*/
   b = gel(y,3); v = gel(x,2);
-  z = cgetg(4, t_POL); z[1] = x[1]; av = avma;
-  gel(z,2) = gerepileupto(av, gsub(v, gdiv(gmul(u,b), a)));
+  z = cgetg(4, t_POL); z[1] = x[1];
+  gel(z,2) = gsub(v, gdiv(gmul(u,b), a));
   gel(z,3) = gneg(u); return z;
 }
 static GEN
 quad_polmod_norm(GEN x, GEN y)
 {
   GEN z, u, v, a, b, c;
-  pari_sp av;
-  if (typ(x) != t_POL || varn(x) != varn(y) || degpol(x) <= 0)
-    return gsqr(x);
+  if (typ(x) != t_POL || varn(x) != varn(y) || degpol(x) <= 0) return gsqr(x);
   a = gel(y,4); u = gel(x,3); /*Mod(ux + v, ax^2 + bx + c)*/
   b = gel(y,3); v = gel(x,2);
-  c = gel(y,2); av = avma;
+  c = gel(y,2);
   z = gmul(u, gsub(gmul(c,u), gmul(b,v)));
   if (!gequal1(a)) z = gdiv(z, a);
-  return gerepileupto(av, gadd(z, gsqr(v)));
+  return gadd(z, gsqr(v));
 }
 
 GEN
-gconj(GEN x)
+conj_i(GEN x)
 {
   long lx, i;
   GEN y;
 
   switch(typ(x))
   {
-    case t_INT: case t_REAL:
-      return mpcopy(x);
+    case t_INT: case t_REAL: case t_INTMOD: case t_FRAC: case t_PADIC:
+      return x;
 
-    case t_INTMOD: case t_FRAC: case t_PADIC:
-      return gcopy(x);
-
-    case t_COMPLEX:
-      y = cgetg(3,t_COMPLEX);
-      gel(y,1) = gcopy(gel(x,1));
-      gel(y,2) = gneg(gel(x,2));
-      break;
+    case t_COMPLEX: return mkcomplex(gel(x,1), gneg(gel(x,2)));
 
     case t_QUAD:
       y = cgetg(4,t_QUAD);
-      gel(y,1) = ZX_copy(gel(x,1));
-      gel(y,2) = gequal0(gmael(x,1,3))? gcopy(gel(x,2))
-                                    : gadd(gel(x,2), gel(x,3));
-      gel(y,3) = gneg(gel(x,3));
-      break;
+      gel(y,1) = gel(x,1);
+      gel(y,2) = gequal0(gmael(x,1,3))? gel(x,2)
+                                      : gadd(gel(x,2), gel(x,3));
+      gel(y,3) = gneg(gel(x,3)); return y;
 
     case t_POL: case t_SER:
       y = cgetg_copy(x, &lx); y[1] = x[1];
-      for (i=2; i<lx; i++) gel(y,i) = gconj(gel(x,i));
-      break;
+      for (i=2; i<lx; i++) gel(y,i) = conj_i(gel(x,i));
+      return y;
 
     case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
       y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gconj(gel(x,i));
-      break;
+      for (i=1; i<lx; i++) gel(y,i) = conj_i(gel(x,i));
+      return y;
 
     case t_POLMOD:
     {
-      GEN y, X = gel(x,1);
+      GEN X = gel(x,1);
       long d = degpol(X);
-      if (d < 2) return RgX_copy(x);
-      if (d == 2) {
-        y = cgetg(3, t_POLMOD);
-        gel(y,1) = RgX_copy(X);
-        gel(y,2) = quad_polmod_conj(gel(x,2), X); return y;
-      }
+      if (d < 2) return x;
+      if (d == 2) return mkpolmod(quad_polmod_conj(gel(x,2), X), X);
     }
-    default:
-      pari_err_TYPE("gconj",x);
-      return NULL; /* not reached */
   }
-  return y;
+  pari_err_TYPE("gconj",x);
+  return NULL; /* LCOV_EXCL_LINE */
 }
+GEN
+gconj(GEN x)
+{ pari_sp av = avma; return gerepilecopy(av, conj_i(x)); }
 
 GEN
 conjvec(GEN x,long prec)
@@ -566,7 +552,7 @@ conjvec(GEN x,long prec)
 
     default:
       pari_err_TYPE("conjvec",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return z;
 }
@@ -703,11 +689,7 @@ addsub_polmod(GEN X, GEN Y, GEN x, GEN y, GEN(*op)(GEN,GEN))
 /* Mod(y, Y) +/- x,  x scalar or polynomial in same var and reduced degree */
 static GEN
 addsub_polmod_scal(GEN Y, GEN y, GEN x, GEN(*op)(GEN,GEN))
-{
-  GEN z = cgetg(3,t_POLMOD);
-  gel(z,1) = RgX_copy(Y);
-  gel(z,2) = op(y, x); return z;
-}
+{ retmkpolmod(degpol(Y)? op(y, x): gen_0, RgX_copy(Y)); }
 
 /* typ(y) == t_SER, x "scalar" [e.g object in lower variable] */
 static GEN
@@ -762,7 +744,7 @@ ser_add(GEN x, GEN y)
   /* valp(x) <= valp(y) */
   lx = _serprec(x);
   if (lx == 2) /* don't lose type information */
-    return scalarser(gadd(RgX_get_0(x), RgX_get_0(y)), varn(x), valp(x));
+    return scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), valp(x));
   ly = _serprec(y) + n; if (lx < ly) ly = lx;
   if (n)
   {
@@ -803,7 +785,7 @@ add_scal(GEN y, GEN x, long ty)
       break;
   }
   pari_err_TYPE2("+",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -822,7 +804,7 @@ addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
     q = dvmdii(n, d, &r);
     if (r == gen_0) { avma = av; return icopy(q); }
     r = gcdii(d, r);
-    if (!is_pm1(r)) { n = diviiexact(n, r); d = diviiexact(d, r); }
+    if (!equali1(r)) { n = diviiexact(n, r); d = diviiexact(d, r); }
     gel(z,1) = icopy_avma(n, (pari_sp)z);
     gel(z,2) = icopy_avma(d, (pari_sp)gel(z,1));
     avma = (pari_sp)gel(z,2); return z;
@@ -841,7 +823,7 @@ addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
         gel(z,2) = Q; return z;
       }
       r = gcdii(x2, r);
-      if (!is_pm1(r)) { n = diviiexact(n, r); x2 = diviiexact(x2, r); }
+      if (!equali1(r)) { n = diviiexact(n, r); x2 = diviiexact(x2, r); }
       d = mulii(x2,Q);
       gel(z,1) = icopy_avma(n, (pari_sp)z);
       gel(z,2) = icopy_avma(d, (pari_sp)gel(z,1));
@@ -863,7 +845,7 @@ addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
         gel(z,2) = Q; return z;
       }
       r = gcdii(y2, r);
-      if (!is_pm1(r)) { n = diviiexact(n, r); y2 = diviiexact(y2, r); }
+      if (!equali1(r)) { n = diviiexact(n, r); y2 = diviiexact(y2, r); }
       d = mulii(y2,Q);
       gel(z,1) = icopy_avma(n, (pari_sp)z);
       gel(z,2) = icopy_avma(d, (pari_sp)gel(z,1));
@@ -872,7 +854,7 @@ addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
     delta = gcdii(y2,r);
   }
   /* delta = gcd(x2,y2) */
-  if (is_pm1(delta))
+  if (equali1(delta))
   { /* numerator is non-zero */
     gel(z,1) = gerepileuptoint((pari_sp)z, op(mulii(x1,y2), mulii(y1,x2)));
     gel(z,2) = mulii(x2,y2); return z;
@@ -885,13 +867,13 @@ addsub_frac(GEN x, GEN y, GEN (*op)(GEN,GEN))
   q = dvmdii(n, delta, &r);
   if (r == gen_0)
   {
-    if (is_pm1(d)) { avma = av; return icopy(q); }
+    if (equali1(d)) { avma = av; return icopy(q); }
     avma = (pari_sp)z;
     gel(z,2) = icopy(d);
     gel(z,1) = icopy(q); return z;
   }
   r = gcdii(delta, r);
-  if (!is_pm1(r))
+  if (!equali1(r))
   {
     n     = diviiexact(n, r);
     delta = diviiexact(delta, r);
@@ -1158,9 +1140,9 @@ gadd(GEN x, GEN y)
       {
         case t_SER:
           if (lg(x) == 2) return gcopy(y);
-          i = lg(y) + valp(y) - RgX_val(x);
+          i = RgX_val(x); if (i == LONG_MAX) i = 0; /* e.g. x = Mod(0,3)*x^0 */
+          i = lg(y) + valp(y) - i;
           if (i < 3) return gcopy(y);
-
           p1 = RgX_to_ser(x,i); y = ser_add(p1,y);
           settyp(p1, t_VECSMALL); /* p1 left on stack */
           return y;
@@ -1194,7 +1176,7 @@ gadd(GEN x, GEN y)
       break;
   }
   pari_err_TYPE2("+",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -1334,7 +1316,7 @@ static GEN
 mul_ser_scal(GEN y, GEN x) {
   long ly, i;
   GEN z;
-  if (isexactzero(x)) return gmul(RgX_get_0(y), x);
+  if (isexactzero(x)) return gmul(Rg_get_0(y), x);
   if (ser_isexactzero(y))
   {
     if (lg(y) == 2) return gcopy(y);
@@ -1393,7 +1375,7 @@ mul_scal(GEN y, GEN x, long ty)
       if (typ(x) == t_INT && gequal1(x)) return gcopy(y); /* fall through */
   }
   pari_err_TYPE2("*",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -1449,11 +1431,7 @@ div_rfrac_pol(GEN x1, GEN x2, GEN y2)
 /* Mod(y, Y) * x,  assuming x scalar */
 static GEN
 mul_polmod_scal(GEN Y, GEN y, GEN x)
-{
-  GEN z = cgetg(3,t_POLMOD);
-  gel(z,1) = RgX_copy(Y);
-  gel(z,2) = gmul(x,y); return z;
-}
+{ retmkpolmod(gmul(x,y), RgX_copy(Y)); }
 
 /* cf mulqq */
 static GEN
@@ -1511,26 +1489,7 @@ mul_polmod_same(GEN T, GEN x, GEN y)
     if (lg(T) == 5 && isint1(gel(T,4))) /* quadratic fields */
       a = quad_polmod_mul(T, x, y);
     else
-    {
-      pari_sp av = avma;
-      GEN p = NULL;
-      if (RgX_is_FpX(T, &p) && RgX_is_FpX(x, &p) && RgX_is_FpX(y, &p) && p)
-      {
-        T = RgX_to_FpX(T, p); x = RgX_to_FpX(x, p); y = RgX_to_FpX(y, p);
-        if (lgefint(p) == 3)
-        {
-          ulong pp = p[2];
-          a = Flxq_mul(ZX_to_Flx(x, pp), ZX_to_Flx(y, pp), ZX_to_Flx(T, pp), pp);
-          a = Flx_to_ZX(a);
-        }
-        else
-          a = FpXQ_mul(x, y, T, p);
-        a = FpX_to_mod(a, p);
-      }
-      else
-        a = RgXQ_mul(x, y, gel(z,1));
-      a = gerepileupto(av, a);
-    }
+      a = RgXQ_mul(x, y, gel(z,1));
   }
   gel(z,2) = a; return z;
 }
@@ -1544,22 +1503,7 @@ sqr_polmod(GEN T, GEN x)
   else
   {
     pari_sp av = avma;
-    GEN p = NULL;
-    if (RgX_is_FpX(T, &p) && RgX_is_FpX(x, &p) && p)
-    {
-      T = RgX_to_FpX(T, p); x = RgX_to_FpX(x, p);
-      if (lgefint(p) == 3)
-      {
-        ulong pp = p[2];
-        a = Flxq_sqr(ZX_to_Flx(x, pp), ZX_to_Flx(T, pp), pp);
-        a = Flx_to_ZX(a);
-      }
-      else
-        a = FpXQ_sqr(x, T, p);
-      a = FpX_to_mod(a, p);
-    }
-    else
-      a = RgXQ_sqr(x, gel(z,1));
+    a = RgXQ_sqr(x, gel(z,1));
     a = gerepileupto(av, a);
   }
   gel(z,2) = a; return z;
@@ -1814,6 +1758,18 @@ mulcxmI(GEN x)
       return gmul(mkcomplex(gen_0, gen_m1), x);
   }
 }
+/* x * I^k */
+GEN
+mulcxpowIs(GEN x, long k)
+{
+  switch (k & 3)
+  {
+    case 1: return mulcxI(x);
+    case 2: return gneg(x);
+    case 3: return mulcxmI(x);
+  }
+  return x;
+}
 
 /* fill in coefficients of t_SER z from coeffs of t_POL y */
 static GEN
@@ -1857,9 +1813,9 @@ gmul(GEN x, GEN y)
       GEN y1 = gel(y,1), y2 = gel(y,2);
       z=cgetg(3,t_FRAC);
       p1 = gcdii(x1, y2);
-      if (!is_pm1(p1)) { x1 = diviiexact(x1,p1); y2 = diviiexact(y2,p1); }
+      if (!equali1(p1)) { x1 = diviiexact(x1,p1); y2 = diviiexact(y2,p1); }
       p1 = gcdii(x2, y1);
-      if (!is_pm1(p1)) { x2 = diviiexact(x2,p1); y1 = diviiexact(y1,p1); }
+      if (!equali1(p1)) { x2 = diviiexact(x2,p1); y1 = diviiexact(y1,p1); }
       tetpil = avma;
       gel(z,2) = mulii(x2,y2);
       gel(z,1) = mulii(x1,y1);
@@ -1880,11 +1836,9 @@ gmul(GEN x, GEN y)
         if (varncmp(vx, vy) < 0) return RgX_Rg_mul(x, y);
         else                     return RgX_Rg_mul(y, x);
       }
-      if (RgX_is_ZX(x) && RgX_is_ZX(y)) return ZX_mul(x,y);
       return RgX_mul(x, y);
 
     case t_SER: {
-      GEN p = NULL;
       vx = varn(x);
       vy = varn(y);
       if (vx != vy) {
@@ -1898,19 +1852,7 @@ gmul(GEN x, GEN y)
       z[1] = evalvalp(valp(x)+valp(y)) | evalvarn(vx) | evalsigne(1);
       x = ser2pol_i(x, lx);
       y = ser2pol_i(y, lx);
-      if (RgX_is_FpX(x,&p) && RgX_is_FpX(y,&p))
-      {
-        if (!p) y = ZX_mul(x,y);
-        else
-        {
-          x = RgX_to_FpX(x, p);
-          y = RgX_to_FpX(y, p);
-          y = FpX_to_mod(ZX_mul(x,y), p);
-        }
-      }
-      else
-        y = RgXn_mul(x, y, lx-2);
-      z = fill_ser(z, y);
+      y = RgXn_mul(x, y, lx-2);
       return gerepilecopy(av, fill_ser(z,y));
     }
     case t_QFI: return qficomp(x,y);
@@ -1948,7 +1890,7 @@ gmul(GEN x, GEN y)
           if (!signe(x)) return gen_0;
           z=cgetg(3,t_FRAC);
           p1 = gcdii(x,gel(y,2));
-          if (is_pm1(p1))
+          if (equali1(p1))
           {
             avma = (pari_sp)z;
             gel(z,2) = icopy(gel(y,2));
@@ -2127,21 +2069,7 @@ gmul(GEN x, GEN y)
       break;
   }
   pari_err_TYPE2("*",x,y);
-  return NULL; /* not reached */
-}
-
-int
-ff_poltype(GEN *x, GEN *p, GEN *pol)
-{
-  GEN Q, P = *x;
-  if (!signe(P)) return 0;
-  if (!RgX_is_FpXQX(P,pol,p) || !*p || !*pol || typ(*pol)!=t_POL)
-    return 0;
-  Q = RgX_to_FpX(*pol, *p);
-  P = RgX_to_FpXQX(P, Q, *p);
-  *x = ZXX_to_Kronecker(P, degpol(Q));
-  *pol = Q;
-  return 1;
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* return a non-normalized result */
@@ -2245,21 +2173,7 @@ gsqr(GEN x)
 
     case t_FFELT: return FF_sqr(x);
 
-    case t_POL:
-    {
-      GEN a = x, p = NULL, pol = NULL;
-      av = avma;
-      if (RgX_is_ZX(x)) return ZX_sqr(x);
-      if (ff_poltype(&x,&p,&pol))
-      {
-        z = ZX_sqr(x);
-        if (p) z = FpX_to_mod(z,p);
-        if (pol) z = Kronecker_to_mod(z,pol);
-        z = gerepileupto(av, z);
-      }
-      else { avma = av; z = RgX_sqr(a); }
-      return z;
-    }
+    case t_POL: return RgX_sqr(x);
 
     case t_SER:
       lx = lg(x);
@@ -2273,22 +2187,11 @@ gsqr(GEN x)
       else
       {
         pari_sp av = avma;
-        GEN z = cgetg(lx,t_SER), p = NULL;
+        GEN z = cgetg(lx,t_SER);
         z[1] = evalvalp(2*valp(x)) | evalvarn(varn(x)) | evalsigne(1);
         x = ser2pol_i(x,lx);
-        if (RgX_is_FpX(x,&p))
-        {
-          if (!p) x = ZX_sqr(x);
-          else
-          {
-            x = RgX_to_FpX(x, p);
-            x = FpX_to_mod(ZX_sqr(x), p);
-          }
-        }
-        else
-          x = RgXn_sqr(x, lx-2);
-        z = fill_ser(z, x);
-        return gerepilecopy(av, z);
+        x = RgXn_sqr(x, lx-2);
+        return gerepilecopy(av, fill_ser(z,x));
       }
 
     case t_RFRAC: z = cgetg(3,t_RFRAC);
@@ -2309,7 +2212,7 @@ gsqr(GEN x)
       return z;
   }
   pari_err_TYPE2("*",x,x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -2374,7 +2277,7 @@ div_T_scal(GEN x, GEN y, long tx) {
     case t_RFRAC: return div_rfrac_scal(x,y);
   }
   pari_err_TYPE2("/",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -2406,7 +2309,7 @@ div_scal_T(GEN x, GEN y, long ty) {
     case t_RFRAC: return div_scal_rfrac(x, y);
   }
   pari_err_TYPE2("/",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* assume tx = ty = t_SER, same variable vx */
@@ -2420,18 +2323,19 @@ div_ser(GEN x, GEN y, long vx)
   if (ser_isexactzero(x))
   {
     if (lx == 2) return zeroser(vx, l);
-    return scalarser(gmul(gel(x,2),RgX_get_0(y)), varn(x), l);
+    return scalarser(gmul(gel(x,2),Rg_get_0(y)), varn(x), l);
   }
   y_lead = gel(y,2);
   if (gequal0(y_lead)) /* normalize denominator if leading term is 0 */
   {
+    GEN y0 = y;
     pari_warn(warner,"normalizing a series with 0 leading term");
     for (l--, ly--,y++; ly > 2; l--, ly--, y++)
     {
       y_lead = gel(y,2);
       if (!gequal0(y_lead)) break;
     }
-    if (ly <= 2) pari_err_INV("div_ser", y);
+    if (ly <= 2) pari_err_INV("div_ser", y0);
   }
   if (ly < lx) lx = ly;
   p2 = cgetg(lx, t_VECSMALL); /* left on stack for efficiency */
@@ -2495,24 +2399,7 @@ div_polmod_same(GEN T, GEN x, GEN y)
   else
   {
     pari_sp av = avma;
-    GEN p = NULL;
-    if (RgX_is_FpX(T, &p) && RgX_is_FpX(x, &p) && RgX_is_FpX(y, &p) && p)
-    {
-      T = RgX_to_FpX(T, p); x = RgX_to_FpX(x, p); y = RgX_to_FpX(y, p);
-      if (lgefint(p) == 3)
-      {
-        ulong pp = p[2];
-        x = ZX_to_Flx(x, pp);
-        y = ZX_to_Flx(y, pp);
-        T = ZX_to_Flx(T, pp);
-        a = Flxq_mul(x, Flxq_inv(y,T,pp),T,pp);
-        a = Flx_to_ZX(a);
-      }
-      else
-        a = FpXQ_div(x, y, T, p);
-      a = FpX_to_mod(a, p);
-    } else
-      a = RgXQ_mul(x, ginvmod(y, gel(z,1)), gel(z,1));
+    a = RgXQ_mul(x, ginvmod(y, gel(z,1)), gel(z,1));
     a = gerepileupto(av, a);
   }
   gel(z,2) = a; return z;
@@ -2544,9 +2431,9 @@ gdiv(GEN x, GEN y)
       GEN y1 = gel(y,1), y2 = gel(y,2);
       z = cgetg(3, t_FRAC);
       p1 = gcdii(x1, y1);
-      if (!is_pm1(p1)) { x1 = diviiexact(x1,p1); y1 = diviiexact(y1,p1); }
+      if (!equali1(p1)) { x1 = diviiexact(x1,p1); y1 = diviiexact(y1,p1); }
       p1 = gcdii(x2, y2);
-      if (!is_pm1(p1)) { x2 = diviiexact(x2,p1); y2 = diviiexact(y2,p1); }
+      if (!equali1(p1)) { x2 = diviiexact(x2,p1); y2 = diviiexact(y2,p1); }
       tetpil = avma;
       gel(z,2) = mulii(x2,y1);
       gel(z,1) = mulii(x1,y2);
@@ -2565,7 +2452,7 @@ gdiv(GEN x, GEN y)
         gel(z,2) = gerepileupto(av, gneg(gdiv(gel(x,1), y)));
         return z;
       }
-      av = avma; p1 = cxnorm(y); p2 = mulcc(x, gconj(y)); tetpil = avma;
+      av = avma; p1 = cxnorm(y); p2 = mulcc(x, conj_i(y)); tetpil = avma;
       return gerepile(av, tetpil, gdiv(p2,p1));
 
     case t_PADIC:
@@ -2574,7 +2461,7 @@ gdiv(GEN x, GEN y)
 
     case t_QUAD:
       if (!ZX_equal(gel(x,1),gel(y,1))) pari_err_OP("/",x,y);
-      av = avma; p1 = quadnorm(y); p2 = mulqq(x, gconj(y)); tetpil = avma;
+      av = avma; p1 = quadnorm(y); p2 = mulqq(x, conj_i(y)); tetpil = avma;
       return gerepile(av, tetpil, gdiv(p2,p1));
 
     case t_FFELT: return FF_div(x,y);
@@ -2655,7 +2542,7 @@ gdiv(GEN x, GEN y)
         return div_intmod_same(z, gel(y,1), modii(x, gel(y,1)), gel(y,2));
       case t_FRAC:
         z = cgetg(3,t_FRAC); p1 = gcdii(x,gel(y,1));
-        if (is_pm1(p1))
+        if (equali1(p1))
         {
           avma = (pari_sp)z;
           gel(z,2) = icopy(gel(y,1));
@@ -2677,7 +2564,7 @@ gdiv(GEN x, GEN y)
       case t_COMPLEX: return divRc(x,y);
       case t_PADIC: return divTp(x, y);
       case t_QUAD:
-        av = avma; p1 = quadnorm(y); p2 = mulRq(x, gconj(y)); tetpil = avma;
+        av = avma; p1 = quadnorm(y); p2 = mulRq(x, conj_i(y)); tetpil = avma;
         return gerepile(av, tetpil, gdiv(p2,p1));
     }
   }
@@ -2718,10 +2605,10 @@ gdiv(GEN x, GEN y)
 
         case t_COMPLEX:
           av = avma;
-          return gerepileupto(av, mulRc_direct(gdiv(x,cxnorm(y)), gconj(y)));
+          return gerepileupto(av, mulRc_direct(gdiv(x,cxnorm(y)), conj_i(y)));
 
         case t_QUAD:
-          av = avma; p1 = quadnorm(y); p2 = gmul(x,gconj(y)); tetpil = avma;
+          av = avma; p1 = quadnorm(y); p2 = gmul(x,conj_i(y)); tetpil = avma;
           return gerepile(av,tetpil, gdiv(p2,p1));
 
         case t_PADIC: { GEN X = gel(x,1);
@@ -2736,7 +2623,7 @@ gdiv(GEN x, GEN y)
       {
         case t_INT: z = cgetg(3, t_FRAC);
         p1 = gcdii(y,gel(x,1));
-        if (is_pm1(p1))
+        if (equali1(p1))
         {
           avma = (pari_sp)z; tetpil = 0;
           gel(z,1) = icopy(gel(x,1));
@@ -2770,7 +2657,7 @@ gdiv(GEN x, GEN y)
           return divTp(x, y);
 
         case t_QUAD:
-          av=avma; p1=quadnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
+          av=avma; p1=quadnorm(y); p2=gmul(x,conj_i(y)); tetpil=avma;
           return gerepile(av,tetpil,gdiv(p2,p1));
       }
 
@@ -2812,7 +2699,7 @@ gdiv(GEN x, GEN y)
           return div_intmod_same(z, Y, padic_to_Fp(x, Y), gel(y,2));
         }
         case t_COMPLEX: case t_QUAD:
-          av=avma; p1=gmul(x,gconj(y)); p2=gnorm(y); tetpil=avma;
+          av=avma; p1=gmul(x,conj_i(y)); p2=gnorm(y); tetpil=avma;
           return gerepile(av,tetpil,gdiv(p1,p2));
 
         case t_REAL: pari_err_TYPE2("/",x,y);
@@ -2858,9 +2745,7 @@ gdiv(GEN x, GEN y)
     vx = varn(X);
     if (vx != vy) {
       if (varncmp(vx, vy) > 0) return div_scal_T(x, y, ty);
-      z = cgetg(3,t_POLMOD);
-      gel(z,1) = RgX_copy(X);
-      gel(z,2) = gdiv(gel(x,2), y); return z;
+      retmkpolmod(gdiv(gel(x,2), y), RgX_copy(X));
     }
     /* y is POL, SER or RFRAC */
     av = avma;
@@ -2930,7 +2815,7 @@ gdiv(GEN x, GEN y)
       break;
   }
   pari_err_TYPE2("/",x,y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -2958,7 +2843,7 @@ gmulsg(long s, GEN y)
     case t_FRAC:
       if (!s) return gen_0;
       z = cgetg(3,t_FRAC);
-      i = cgcd(s, smodis(gel(y,2), s));
+      i = labs(s); i = ugcd(i, umodiu(gel(y,2), i));
       if (i == 1)
       {
         gel(z,2) = icopy(gel(y,2));
@@ -2985,20 +2870,19 @@ gmulsg(long s, GEN y)
       gel(z,2) = gmulsg(s,gel(y,2));
       gel(z,3) = gmulsg(s,gel(y,3)); return z;
 
-    case t_POLMOD: z = cgetg(3, t_POLMOD);
-      gel(z,1) = RgX_copy(gel(y,1));
-      gel(z,2) = gmulsg(s,gel(y,2)); return z;
+    case t_POLMOD:
+      retmkpolmod(gmulsg(s,gel(y,2)), RgX_copy(gel(y,1)));
 
     case t_POL:
       if (!signe(y)) return RgX_copy(y);
-      if (!s) return scalarpol(RgX_get_0(y), varn(y));
+      if (!s) return scalarpol(Rg_get_0(y), varn(y));
       z = cgetg_copy(y, &ly); z[1]=y[1];
       for (i=2; i<ly; i++) gel(z,i) = gmulsg(s,gel(y,i));
       return normalizepol_lg(z, ly);
 
     case t_SER:
       if (ser_isexactzero(y)) return gcopy(y);
-      if (!s) return RgX_get_0(y);
+      if (!s) return Rg_get_0(y);
       z = cgetg_copy(y, &ly); z[1]=y[1];
       for (i=2; i<ly; i++) gel(z,i) = gmulsg(s,gel(y,i));
       return normalize(z);
@@ -3015,7 +2899,7 @@ gmulsg(long s, GEN y)
       return z;
   }
   pari_err_TYPE("gmulsg",y);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -3058,7 +2942,7 @@ gdivgs(GEN x, long s)
     case t_FFELT: return FF_Z_Z_muldiv(x,gen_1,stoi(s));
 
     case t_FRAC: z = cgetg(3, t_FRAC);
-      i = cgcd(s, smodis(gel(x,1), s));
+      i = labs(s); i = ugcd(i, umodiu(gel(x,1), i));
       if (i == 1)
       {
         gel(z,2) = mulsi(s, gel(x,2));
@@ -3089,9 +2973,8 @@ gdivgs(GEN x, long s)
       gel(z,2) = gdivgs(gel(x,2),s);
       gel(z,3) = gdivgs(gel(x,3),s); return z;
 
-    case t_POLMOD: z = cgetg(3, t_POLMOD);
-      gel(z,1) = RgX_copy(gel(x,1));
-      gel(z,2) = gdivgs(gel(x,2),s); return z;
+    case t_POLMOD:
+      retmkpolmod(gdivgs(gel(x,2),s), RgX_copy(gel(x,1)));
 
     case t_RFRAC:
       if (s == 1) return gcopy(x);
@@ -3109,7 +2992,7 @@ gdivgs(GEN x, long s)
 
   }
   pari_err_TYPE2("/",x, stoi(s));
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* True shift (exact multiplication by 2^n) */
@@ -3166,9 +3049,8 @@ gmul2n(GEN x, long n)
       gel(z,2) = gmul2n(gel(x,2),n);
       gel(z,3) = gmul2n(gel(x,3),n); return z;
 
-    case t_POLMOD: z = cgetg(3,t_POLMOD);
-      gel(z,1) = RgX_copy(gel(x,1));
-      gel(z,2) = gmul2n(gel(x,2),n); return z;
+    case t_POLMOD:
+      retmkpolmod(gmul2n(gel(x,2),n), RgX_copy(gel(x,1)));
 
     case t_POL:
       z = cgetg_copy(x, &lx); z[1] = x[1];
@@ -3191,7 +3073,7 @@ gmul2n(GEN x, long n)
       return gmul(gmul2n(gen_1,n),x);
   }
   pari_err_TYPE("gmul2n",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /*******************************************************************/
@@ -3208,36 +3090,10 @@ inv_polmod(GEN T, GEN x)
     a = ginv(x);
   else
   {
-    pari_sp av = avma;
     if (lg(T) == 5) /* quadratic fields */
       a = RgX_Rg_div(quad_polmod_conj(x,T), quad_polmod_norm(x,T));
     else
-    {
-      GEN p = NULL;
-      if (RgX_is_FpX(T, &p) && RgX_is_FpX(x, &p) && p)
-      {
-        T = RgX_to_FpX(T, p); x = RgX_to_FpX(x, p);
-        if (lgefint(p) == 3)
-        {
-          ulong pp = p[2];
-          a = Flxq_inv(ZX_to_Flx(x, pp), ZX_to_Flx(T, pp), pp);
-          a = Flx_to_ZX(a);
-        }
-        else
-          a = FpXQ_inv(x, T, p);
-        a = FpX_to_mod(a, p);
-        a = gerepileupto(av, a);
-      }
-      else if (RgX_is_ZX(T) && RgX_is_QX(x))
-      {
-        avma = av;
-        a = QXQ_inv(x, T);
-      }
-      else {
-        avma = av;
-        a = RgXQ_inv(x, T);
-      }
-    }
+      a = RgXQ_inv(x, T);
   }
   gel(z,2) = a; return z;
 }
@@ -3281,7 +3137,7 @@ ginv(GEN x)
       return gerepile(av,tetpil,divcR(p2,p1));
 
     case t_QUAD:
-      av=avma; p1=gnorm(x); p2=gconj(x); tetpil=avma;
+      av=avma; p1=gnorm(x); p2=conj_i(x); tetpil=avma;
       return gerepile(av,tetpil,gdiv(p2,p1));
 
     case t_PADIC: z = cgetg(5,t_PADIC);
@@ -3350,5 +3206,5 @@ ginv(GEN x)
     }
   }
   pari_err_TYPE("inverse",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
