@@ -30,7 +30,6 @@ typedef struct {
   long k;    /* input known mod P^k */
   GEN p, pk; /* p^k = denom(prk^-1) [ assume pr unramified ]*/
   GEN prk;   /* |.|^2 LLL-reduced basis (b_i) of P^k  (NOT T2-reduced) */
-  GEN prkHNF;/* HNF basis of P^k */
   GEN iprk;  /* den * prk^-1 */
   GEN GSmin; /* min |b_i^*|^2 */
 
@@ -51,7 +50,6 @@ typedef struct
   GEN nf;
   GEN pol, polbase; /* leading coeff is a t_INT */
   GEN fact;
-  GEN pr;
   GEN Br, bound, ZC, BS_2;
 } nfcmbf_t;
 
@@ -60,10 +58,24 @@ typedef struct
 /*******************************************************************/
 /* NOT stack clean. a, b stay on the stack */
 static GEN
-lift_to_frac(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom)
+lift_to_frac_tdenom(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom, GEN tdenom)
 {
   GEN a, b;
   if (signe(t) < 0) t = addii(t, mod); /* in case t is a centerlift */
+  if (tdenom)
+  {
+    pari_sp av = avma;
+    a = Fp_center_i(Fp_mul(t, tdenom, mod), mod, shifti(mod,-1));
+    if (abscmpii(a, amax) < 0)
+    {
+      GEN d = gcdii(a, tdenom);
+      a = diviiexact(a, d);
+      b = diviiexact(tdenom, d);
+      if (is_pm1(b)) { return gerepileuptoint(av, a); }
+      return gerepilecopy(av, mkfrac(a, b));
+    }
+    avma = av;
+  }
   if (!Fp_ratlift(t, mod, amax,bmax, &a,&b)
      || (denom && !dvdii(denom,b))
      || !is_pm1(gcdii(a,b))) return NULL;
@@ -71,6 +83,12 @@ lift_to_frac(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom)
   return mkfrac(a, b);
 }
 
+static GEN
+lift_to_frac(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom)
+{
+  return lift_to_frac_tdenom(t, mod, amax, bmax, denom, NULL);
+}
+
 /* Compute rational lifting for all the components of M modulo mod.
  * Assume all Fp_ratlift preconditions are met; we allow centerlifts but in
  * that case are no longer stack clean. If one component fails, return NULL.
@@ -79,39 +97,39 @@ lift_to_frac(GEN t, GEN mod, GEN amax, GEN bmax, GEN denom)
  * We suppose gcd(mod, denom) = 1, then a and b are coprime; so we can use
  * mkfrac rather than gdiv */
 GEN
-FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
-{
-  pari_sp av = avma;
-  long i, j, h, l = lg(M);
-  GEN a, N = cgetg_copy(M, &l);
-  if (l == 1) return N;
-  h = lgcols(M);
-  for (j = 1; j < l; ++j)
-  {
-    gel(N,j) = cgetg(h, t_COL);
-    for (i = 1; i < h; ++i)
-    {
-      a = lift_to_frac(gcoeff(M,i,j), mod, amax,bmax,denom);
-      if (!a) { avma = av; return NULL; }
-      gcoeff(N,i,j) = a;
-    }
-  }
-  return N;
-}
-GEN
 FpC_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
 {
   pari_sp ltop = avma;
   long j, l;
-  GEN a, Q = cgetg_copy(P, &l);
+  GEN a, d, tdenom = NULL, Q = cgetg_copy(P, &l);
+  if (l==1) return Q;
   for (j = 1; j < l; ++j)
   {
-    a = lift_to_frac(gel(P,j), mod, amax,bmax,denom);
+    a = lift_to_frac_tdenom(gel(P,j), mod, amax, bmax, denom, tdenom);
     if (!a) { avma = ltop; return NULL; }
+    d = Q_denom(a);
+    tdenom = tdenom ? cmpii(tdenom, d)<0? d: tdenom : d;
     gel(Q,j) = a;
   }
   return Q;
 }
+
+GEN
+FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
+{
+  pari_sp av = avma;
+  long j, l = lg(M);
+  GEN N = cgetg_copy(M, &l);
+  if (l == 1) return N;
+  for (j = 1; j < l; ++j)
+  {
+    GEN a = FpC_ratlift(gel(M, j), mod, amax, bmax, denom);
+    if (!a) { avma = av; return NULL; }
+    gel(N,j) = a;
+  }
+  return N;
+}
+
 GEN
 FpX_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
 {
@@ -153,32 +171,36 @@ GEN
 nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
 {
   pari_sp btop, ltop = avma;
-  GEN lP, lQ, M, dsol, R, bo, sol, mod = NULL;
+  GEN lP, lQ, M, dsol, R, bo, sol, mod = NULL, lden = NULL;
   long vP = varn(P), vT = varn(T), dT = degpol(T), dM = 0, dR;
   forprime_t S;
 
   if (!signe(P)) { if (Pnew) *Pnew = pol_0(vT); return gcopy(Q); }
   if (!signe(Q)) { if (Pnew) *Pnew = pol_1(vT);   return gcopy(P); }
   /*Compute denominators*/
-  if (!den) den = ZX_disc(T);
   lP = leading_coeff(P);
   lQ = leading_coeff(Q);
   if ( !((typ(lP)==t_INT && is_pm1(lP)) || (typ(lQ)==t_INT && is_pm1(lQ))) )
-    den = mulii(den, gcdii(ZX_resultant(lP, T), ZX_resultant(lQ, T)));
+  {
+    lden = gcdii(ZX_resultant(lP, T), ZX_resultant(lQ, T));
+    if (den) den = mulii(den, lden);
+  }
 
   init_modular_small(&S);
   btop = avma;
   for(;;)
   {
     ulong p = u_forprime_next(&S);
+    GEN Tp;
     if (!p) pari_err_OVERFLOW("nfgcd [ran out of primes]");
     /*Discard primes dividing disc(T) or lc(PQ) */
-    if (!umodiu(den, p)) continue;
-    if (DEBUGLEVEL>5) err_printf("nfgcd: p=%lu\n",p);
+    if (lden && !umodiu(lden, p)) continue;
+    Tp = ZX_to_Flx(T,p);
+    if (!Flx_is_squarefree(Tp, p)) continue;
     /*Discard primes when modular gcd does not exist*/
     if ((R = FlxqX_safegcd(ZXX_to_FlxX(P,p,vT),
                            ZXX_to_FlxX(Q,p,vT),
-                           ZX_to_Flx(T,p), p)) == NULL) continue;
+                           Tp, p)) == NULL) continue;
     dR = degpol(R);
     if (dR == 0) { avma = ltop; if (Pnew) *Pnew = P; return pol_1(vP); }
     if (mod && dR > dM) continue; /* p divides Res(P/gcd, Q/gcd). Discard. */
@@ -441,7 +463,7 @@ nf_bestlift(GEN elt, GEN bound, nflift_t *L)
   long i,l = lg(L->prk), t = typ(elt);
   if (t != t_INT)
   {
-    if (t == t_POL) elt = mulmat_pol(L->tozk, elt);
+    if (t == t_POL) elt = ZM_ZX_mul(L->tozk, elt);
     u = ZM_ZC_mul(L->iprk,elt);
     for (i=1; i<l; i++) gel(u,i) = diviiround(gel(u,i), L->pk);
   }
@@ -643,7 +665,7 @@ nf_Mignotte_bound(GEN nf, GEN polbase)
   long prec = nf_get_prec(nf), n = nf_get_degree(nf), r1 = nf_get_r1(nf);
   long i, j, d = degpol(polbase);
 
-  binlS = bin = vecbinome(d-1);
+  binlS = bin = vecbinomial(d-1);
   if (!isint1(lS)) binlS = ZC_Z_mul(bin,lS);
 
   N2 = cgetg(n+1, t_VEC);
@@ -698,7 +720,7 @@ nf_Beauzamy_bound(GEN nf, GEN polbase)
 {
   GEN lt, C, s, POL, bin;
   long d = degpol(polbase), n = nf_get_degree(nf), prec = nf_get_prec(nf);
-  bin = vecbinome(d);
+  bin = vecbinomial(d);
   POL = polbase + 2;
   /* compute [POL]_2 */
   for (;;)
@@ -749,8 +771,7 @@ nf_root_bounds(GEN nf, GEN P)
   long lR, i, j, l, prec, r1;
   GEN Ps, R, V;
 
-  if (RgX_is_rational(P)) return logmax_modulus_bound(P);
-
+  if (RgX_is_rational(P)) return polrootsbound(P, NULL);
   r1 = nf_get_r1(nf);
   P = Q_primpart(P);
   prec = ZXX_max_lg(P) + 1;
@@ -767,22 +788,20 @@ nf_root_bounds(GEN nf, GEN P)
   {
     GEN r = gel(R,j);
     for (i=2; i<l; i++) gel(Ps,i) = poleval(gel(P,i), r);
-    gel(V,j) = logmax_modulus_bound(Ps);
+    gel(V,j) = polrootsbound(Ps, NULL);
   }
   return mkvec2(vecslice(V,1,r1), vecslice(V,r1+1,lg(V)-1));
 }
 
-/* return B such that if x in O_K, K = Z[X]/(T), then the L2-norm of the
- * coordinates of the numerator of x [on the power, resp. integral, basis if T
- * is a polynomial, resp. an nf] is  <= B T_2(x)
- * den = multiplicative bound for denom(x) */
+/* return B such that, if x = sum x_i K.zk[i] in O_K, then ||x||_2^2 <= B T_2(x)
+ * den = multiplicative bound for denom(x) [usually NULL, for 1, but when we
+ * use nf_PARTIALFACT K.zk may not generate O_K] */
 static GEN
 L2_bound(GEN nf, GEN den)
 {
   GEN M, L, prep, T = nf_get_pol(nf), tozk = nf_get_invzk(nf);
-  long bit = bit_accuracy(ZX_max_lg(T)) + bit_accuracy(ZM_max_lg(tozk));
-  long prec = nbits2prec(bit + degpol(T));
-  (void)initgaloisborne(nf, den, prec, &L, &prep, NULL);
+  long prec = ZM_max_lg(tozk) + ZX_max_lg(T) + nbits2prec(degpol(T));
+  (void)initgaloisborne(nf, den? den: gen_1, prec, &L, &prep, NULL);
   M = vandermondeinverse(L, RgX_gtofp(T,prec), den, prep);
   return RgM_fpnorml2(RgM_mul(tozk,M), DEFAULTPREC);
 }
@@ -883,10 +902,9 @@ init_trace(trace_data *T, GEN S, nflift_t *L, GEN q)
   l = lg(S);
   h = lgcols(T->dPinvS);
   T->PinvSdbl = (double**)cgetg(l, t_MAT);
-  init_dalloc();
   for (j = 1; j < l; j++)
   {
-    double *t = dalloc(h * sizeof(double));
+    double *t = (double *) stack_malloc_align(h * sizeof(double), sizeof(double));
     GEN c = gel(T->dPinvS,j);
     pari_sp av = avma;
     T->PinvSdbl[j] = t;
@@ -920,9 +938,9 @@ FqX_centermod(GEN z, GEN T, GEN pk, GEN pks2)
   {
     GEN c = gel(y,i);
     if (typ(c) == t_INT)
-      c = centermodii(c, pk, pks2);
+      c = Fp_center_i(c, pk, pks2);
     else
-      c = FpX_center(c, pk, pks2);
+      c = FpX_center_i(c, pk, pks2);
     gel(y,i) = c;
   }
   return y;
@@ -935,8 +953,8 @@ typedef struct {
 static void
 init_div_data(div_data *D, GEN pol, nflift_t *L)
 {
-  GEN C = mul_content(L->topowden, L->dn);
-  GEN C2lt, Clt, lc = leading_coeff(pol), lt = is_pm1(lc)? NULL: absi(lc);
+  GEN C2lt, Clt, C = mul_content(L->topowden, L->dn);
+  GEN lc = leading_coeff(pol), lt = is_pm1(lc)? NULL: absi_shallow(lc);
   if (C)
   {
     GEN C2 = sqri(C);
@@ -1203,7 +1221,7 @@ nf_to_Zq(GEN x, GEN T, GEN pk, GEN pks2, GEN proj)
   }
   y = ZM_ZC_mul(proj, x);
   y = RgV_to_RgX(y, varn(T));
-  return FpX_center(FpX_rem(y, T, pk), pk, pks2);
+  return FpX_center_i(FpX_rem(y, T, pk), pk, pks2);
 }
 
 /* Assume P in nfX form, lc(P) != 0 mod p. Reduce P to Zp[X]/(T) mod p^a */
@@ -1229,14 +1247,19 @@ ZqX_normalize(GEN P, GEN lt, nflift_t *L)
 /* return log [  2sqrt(C/d) * ( (3/2)sqrt(gamma) )^(d-1) ] ^d / log N(pr)
  * cf. Belabas relative van Hoeij algorithm, lemma 3.12 */
 static double
-bestlift_bound(GEN C, long d, double alpha, GEN Npr)
+bestlift_bound(GEN C, long d, double alpha, GEN p, long f)
 {
-  const double y = 1 / (alpha - 0.25); /* = 2 if alpha = 3/4 */
-  double t;
-  C = gtofp(C,DEFAULTPREC);
+  const double g = 1 / (alpha - 0.25); /* = 2 if alpha = 3/4 */
+  GEN C4 = shiftr(gtofp(C,DEFAULTPREC), 2);
+  double t, logp = log(gtodouble(p));
+  if (f == d)
+  { /* p inert, no LLL fudge factor: p^(2k) / 4 > C */
+    t = 0.5 * rtodbl(mplog(C4));
+    return ceil(t / logp);
+  }
   /* (1/2)log (4C/d) + (d-1)(log 3/2 sqrt(gamma)) */
-  t = rtodbl(mplog(gmul2n(divru(C,d), 2))) * 0.5 + (d-1) * log(1.5 * sqrt(y));
-  return ceil((t * d) / log(gtodouble(Npr)));
+  t = 0.5 * rtodbl(mplog(divru(C4,d))) + (d-1) * log(1.5 * sqrt(g));
+  return ceil((t * d) / (logp * f));
 }
 
 static GEN
@@ -1257,15 +1280,15 @@ get_R(GEN M)
 }
 
 static void
-init_proj(nflift_t *L, GEN nfT)
+init_proj(nflift_t *L, GEN prkHNF, GEN nfT)
 {
-  if (L->Tp)
+  if (degpol(L->Tp)>1)
   {
     GEN coTp = FpX_div(FpX_red(nfT, L->p), L->Tp,  L->p); /* Tp's cofactor */
     GEN z, proj;
     z = ZpX_liftfact(nfT, mkvec2(L->Tp, coTp), L->pk, L->p, L->k);
     L->Tpk = gel(z,1);
-    proj = get_proj_modT(L->topow, L->Tpk, L->pk);
+    proj = QXQV_to_FpM(L->topow, L->Tpk, L->pk);
     if (L->topowden)
       proj = FpM_red(ZM_Z_mul(proj, Fp_inv(L->topowden, L->pk)), L->pk);
     L->ZqProj = proj;
@@ -1273,13 +1296,14 @@ init_proj(nflift_t *L, GEN nfT)
   else
   {
     L->Tpk = NULL;
-    L->ZqProj = dim1proj(L->prkHNF);
+    L->ZqProj = dim1proj(prkHNF);
   }
 }
 
 /* Square of the radius of largest ball inscript in PRK's fundamental domain,
  *   whose orthogonalized vector's norms are the Bi
- * Rmax ^2 =  min 1/4T_i where T_i = sum ( s_ij^2 / B_j) */
+ * Rmax ^2 =  min 1/4T_i where T_i = sum_j ( s_ij^2 / B_j)
+ * For p inert, S = Id, T_i = 1 / p^{2k} and Rmax = p^k / 2 */
 static GEN
 max_radius(GEN PRK, GEN B)
 {
@@ -1299,38 +1323,50 @@ max_radius(GEN PRK, GEN B)
 }
 
 static void
-bestlift_init(long a, GEN nf, GEN pr, GEN C, nflift_t *L)
+bestlift_init(long a, GEN nf, GEN C, nflift_t *L)
 {
   const double alpha = 0.99; /* LLL parameter */
   const long d = nf_get_degree(nf);
   pari_sp av = avma, av2;
-  GEN prk, PRK, B, GSmin, pk;
+  GEN prk, PRK, iPRK, GSmin, T = L->Tp, p = L->p;
+  long f = degpol(T);
   pari_timer ti;
 
+  if (f == d)
+  { /* inert p, much simpler */
+    long a0 = bestlift_bound(C, d, alpha, p, f);
+    GEN q;
+    if (a < a0) a = a0; /* guarantees GSmin >= C */
+    if (DEBUGLEVEL>2) err_printf("exponent %ld\n",a);
+    q = powiu(p,a);
+    PRK = prk = scalarmat_shallow(q, d);
+    GSmin = shiftr(itor(q, DEFAULTPREC), -1);
+    iPRK = matid(d); goto END;
+  }
   timer_start(&ti);
-  if (!a) a = (long)bestlift_bound(C, d, alpha, pr_norm(pr));
-
+  if (!a) a = (long)bestlift_bound(C, d, alpha, p, f);
   for (;; avma = av, a += (a==1)? 1: (a>>1)) /* roughly a *= 1.5 */
   {
+    GEN B, q = powiu(p,a), Tq = FpXQ_powu(T, a, FpX_red(nf_get_pol(nf), q), q);
     if (DEBUGLEVEL>2) err_printf("exponent %ld\n",a);
-    prk = idealpows(nf, pr, a);
+    prk = idealhnf_two(nf, mkvec2(q, Tq));
     av2 = avma;
-    pk = gcoeff(prk,1,1);
     PRK = ZM_lll_norms(prk, alpha, LLL_INPLACE, &B);
     GSmin = max_radius(PRK, B);
     if (gcmp(GSmin, C) >= 0) break;
   }
   gerepileall(av2, 2, &PRK, &GSmin);
+  iPRK = ZM_inv(PRK, NULL);
   if (DEBUGLEVEL>2)
     err_printf("for this exponent, GSmin = %Ps\nTime reduction: %ld\n",
-      GSmin, timer_delay(&ti));
+               GSmin, timer_delay(&ti));
+END:
   L->k = a;
-  L->pk = pk;
+  L->pk = gcoeff(prk,1,1);
   L->prk = PRK;
-  L->iprk = ZM_inv(PRK, pk);
+  L->iprk = iPRK;
   L->GSmin= GSmin;
-  L->prkHNF = prk;
-  init_proj(L, nf_get_pol(nf));
+  init_proj(L, prk, nf_get_pol(nf));
 }
 
 /* Let X = Tra * M_L, Y = bestlift(X) return V s.t Y = X - PRK V
@@ -1370,7 +1406,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
   long ti_LLL = 0, ti_CF = 0;
   pari_timer ti2, TI;
 
-  lP = absi(leading_coeff(P));
+  lP = absi_shallow(leading_coeff(P));
   if (is_pm1(lP)) lP = NULL;
 
   n0 = lg(famod) - 1;
@@ -1407,7 +1443,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
     {
       GEN polred;
 
-      bestlift_init((L->k)<<1, T->nf, T->pr, Btra, L);
+      bestlift_init((L->k)<<1, T->nf, Btra, L);
       polred = ZqX_normalize(T->polbase, lP, L);
       famod = ZqX_liftfact(polred, famod, L->Tpk, L->pk, L->p, L->k);
       for (i=1; i<=n0; i++) TT[i] = 0;
@@ -1495,14 +1531,14 @@ AGAIN:
       if (DEBUGLEVEL>2) ti_CF += timer_delay(&ti);
       if (list) break;
     }
-    CM_L = gerepilecopy(av2, CM_L);
     if (gc_needed(av,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"nf_LLL_cmbf");
-      gerepileall(av, L->Tpk? 8: 7,
-                      &CM_L,&TT,&Tra,&famod,&L->GSmin,&L->prk,&L->iprk,&L->Tpk);
-      L->pk = gcoeff(L->prk,1,1);
+      gerepileall(av, L->Tpk? 9: 8,
+                      &CM_L,&TT,&Tra,&famod,&L->GSmin,&L->pk,&L->prk,&L->iprk,
+                      &L->Tpk);
     }
+    else CM_L = gerepilecopy(av2, CM_L);
   }
   if (DEBUGLEVEL>2)
     err_printf("* Time LLL: %ld\n* Time Check Factor: %ld\n",ti_LLL,ti_CF);
@@ -1543,57 +1579,79 @@ static GEN
 nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, long fl, nflift_t *L)
 {
   GEN z, Cltx_r, ltdn;
-  long i, m;
+  long i, m, lz;
   div_data D;
 
   init_div_data(&D, pol, L);
   ltdn = mul_content(D.lt, L->dn);
   z = ZqX_roots(polred, L->Tpk, L->p, L->k);
   Cltx_r = deg1pol_shallow(D.Clt? D.Clt: gen_1, NULL, varn(pol));
-  for (m=1,i=1; i<lg(z); i++)
+  lz = lg(z);
+  if (DEBUGLEVEL > 3) err_printf("Checking %ld roots:",lz-1);
+  for (m=1,i=1; i<lz; i++)
   {
-    GEN q, r = gel(z,i);
+    GEN r = gel(z,i);
+    int dvd;
     pari_sp av;
+    if (DEBUGLEVEL > 3) err_printf(" %ld",i);
     /* lt*dn*topowden * r = Clt * r */
     r = nf_bestlift_to_pol(ltdn? gmul(ltdn,r): r, NULL, L);
     av = avma;
     gel(Cltx_r,2) = gneg(r); /* check P(r) == 0 */
-    q = RgXQX_divrem(D.C2ltpol, Cltx_r, nfpol, ONLY_DIVIDES); /* integral */
+    dvd = ZXQX_dvd(D.C2ltpol, Cltx_r, nfpol); /* integral */
     avma = av;
     /* don't go on with q, usually much larger that C2ltpol */
-    if (q) {
+    if (dvd) {
       if (D.Clt) r = gdiv(r, D.Clt);
       gel(z,m++) = r;
     }
     else if (fl == ROOTS_SPLIT) return cgetg(1, t_VEC);
   }
+  if (DEBUGLEVEL > 3) err_printf(" done\n");
   z[0] = evaltyp(t_VEC) | evallg(m);
   return z;
 }
 
-/* returns a factor of T in Fp of degree <= maxf, NULL if none exist */
+/* returns a few factors of T in Fp of degree <= maxf, NULL if none exist */
 static GEN
 get_good_factor(GEN T, ulong p, long maxf)
 {
   pari_sp av = avma;
-  GEN r, list = gel(Flx_factor(ZX_to_Flx(T,p),p), 1);
+  GEN r, R = gel(Flx_factor(ZX_to_Flx(T,p),p), 1);
   if (maxf == 1)
-  { /* deg.1 factors are best */
-    r = gel(list,1);
-    if (degpol(r) == 1) return r;
+  { /* degree 1 factors are best */
+    r = gel(R,1);
+    if (degpol(r) == 1) return mkvec(r);
   }
   else
   { /* otherwise, pick factor of largish degree */
-    long i, dr;
-    for (i = lg(list)-1; i > 0; i--)
+    long i, j, dr, d = 0, l = lg(R);
+    GEN v;
+    if (l == 2) return mkvec(gel(R,1)); /* inert is fine */
+    v = cgetg(l, t_VEC);
+    for (i = j = 1; i < l; i++)
     {
-      r = gel(list,i); dr = degpol(r);
-      if (dr <= maxf) return r;
+      r = gel(R,i); dr = degpol(r);
+      if (dr > maxf) break;
+      if (dr != d) { gel(v,j++) = r; d = dr; }
     }
+    setlg(v,j); if (j > 1) return v;
   }
   avma = av; return NULL; /* failure */
 }
 
+/* n = number of modular factors, f = residue degree; nold/fold current best
+ * return 1 if new values are "better" than old ones */
+static int
+record(long nold, long n, long fold, long f)
+{
+  if (!nold) return 1; /* uninitialized */
+  if (fold == f) return n < nold;
+  /* if f increases, allow increasing n a little */
+  if (fold < f) return n <= 20 || n < 1.1*nold;
+  /* f decreases, only allow if decreasing n a lot */
+  return n < 0.7*nold;
+}
 /* Optimization problem: factorization of polynomials over large Fq is slow,
  * BUT bestlift correspondingly faster.
  * Return maximal residue degree to be considered when picking a prime ideal */
@@ -1606,90 +1664,89 @@ get_maxf(long nfdeg)
   else if (nfdeg >= 15) maxf = 8;
   return maxf;
 }
-
-/* Select a prime ideal pr over which to factor polbase.
- * Return the number of factors (or roots, according to flag fl) mod pr,
- * Input:
- *   ct: number of attempts to find best
+/* number of maximal ideals to test before settling on best prime and number
+ * of factors; B = [K:Q]*deg(P) */
+static long
+get_nbprimes(long B)
+{
+  if (B <= 128) return 5;
+  if (B <= 1024) return 20;
+  if (B <= 2048) return 65;
+  return 100;
+}
+/* Select a prime ideal pr over which to factor pol.
+ * Return the number of factors (or roots, according to flag fl) mod pr.
  * Set:
  *   lt: leading term of polbase (t_INT or NULL [ for 1 ])
  *   pr: a suitable maximal ideal
  *   Fa: factors found mod pr
  *   Tp: polynomial defining Fq/Fp */
 static long
-nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
-              GEN *lt, GEN *pr, GEN *Tp)
+nf_pick_prime(GEN nf, GEN pol, long fl, GEN *lt, GEN *Tp, ulong *pp)
 {
   GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
-  long maxf, nfdeg = degpol(nfpol), dpol = degpol(polbase), nbf = 0;
-  ulong pp;
+  long nfdeg = degpol(nfpol), dpol = degpol(pol), nold = 0, fold = 1;
+  long maxf = get_maxf(nfdeg), ct = get_nbprimes(nfdeg * dpol);
+  ulong p;
   forprime_t S;
   pari_timer ti_pr;
 
   if (DEBUGLEVEL>3) timer_start(&ti_pr);
-  *lt  = leading_coeff(polbase); /* t_INT */
+  *lt  = leading_coeff(pol); /* t_INT */
   if (gequal1(*lt)) *lt = NULL;
-  *pr = NULL;
+  *pp = 0;
   *Tp = NULL;
-
-  maxf = get_maxf(nfdeg);
   (void)u_forprime_init(&S, 2, ULONG_MAX);
   /* select pr such that pol has the smallest number of factors, ct attempts */
-  while ((pp = u_forprime_next(&S)))
+  while ((p = u_forprime_next(&S)))
   {
-    GEN aT, apr, ap, amodpr, red, r;
-    long anbf;
+    GEN vT;
+    long n, i, l, ok = 0;
     ulong ltp = 0;
-    pari_sp av2 = avma;
 
-    /* first step : select prime of high inertia degree */
-    if (! umodiu(bad,pp)) continue;
-    if (*lt) { ltp = umodiu(*lt, pp); if (!ltp) continue; }
-    r = get_good_factor(nfpol, pp, maxf);
-    if (!r) continue;
-
-    ap = utoipos(pp);
-    apr = idealprimedec_kummer(nf, Flx_to_ZX(r), 1, ap);
-    amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
-
-    /* second step : evaluate factorisation mod apr */
-    red = nfX_to_FqX(polbase, nf, amodpr);
-    if (!aT)
-    { /* degree 1 */
-      red = ZX_to_Flx(red, pp);
-      if (ltp) red = Flx_normalize(red, pp);
-      if (!Flx_is_squarefree(red, pp)) { avma = av2; continue; }
-      anbf = fl == FACTORS? Flx_nbfact(red, pp): Flx_nbroots(red, pp);
-    }
-    else
+    if (! umodiu(bad,p)) continue;
+    if (*lt) { ltp = umodiu(*lt, p); if (!ltp) continue; }
+    vT = get_good_factor(nfpol, p, maxf);
+    if (!vT) continue;
+    l = lg(vT);
+    for (i = 1; i < l; i++)
     {
-      if (ltp) red = FqX_normalize(red, aT,ap);
-      if (!FqX_is_squarefree(red,aT,ap)) { avma = av2; continue; }
-      anbf = fl == FACTORS? FqX_nbfact(red, aT, ap)
-                          : FqX_nbroots(red, aT, ap);
-    }
-    if (fl == ROOTS_SPLIT && anbf < dpol) return anbf;
-    if (anbf <= 1)
-    {
-      if (fl == FACTORS) return anbf; /* irreducible */
-      if (!anbf) return 0; /* no root */
-    }
-    if (DEBUGLEVEL>3)
-      err_printf("%3ld %s at prime\n  %Ps\nTime: %ld\n",
-                 anbf, fl == FACTORS?"factors": "roots", apr, timer_delay(&ti_pr));
+      pari_sp av2 = avma;
+      GEN T = gel(vT,i), red = RgX_to_FlxqX(pol, T, p);
+      long f = degpol(T);
+      if (f == 1)
+      { /* degree 1 */
+        red = FlxX_to_Flx(red);
+        if (ltp) red = Flx_normalize(red, p);
+        if (!Flx_is_squarefree(red, p)) { avma = av2; continue; }
+        ok = 1;
+        n = (fl == FACTORS)? Flx_nbfact(red,p): Flx_nbroots(red,p);
+      }
+      else
+      {
+        if (ltp) red = FlxqX_normalize(red, T, p);
+        if (!FlxqX_is_squarefree(red, T, p)) { avma = av2; continue; }
+        ok = 1;
+        n = (fl == FACTORS)? FlxqX_nbfact(red,T,p): FlxqX_nbroots(red,T,p);
+      }
+      if (fl == ROOTS_SPLIT && n < dpol) return n; /* not split */
+      if (n <= 1)
+      {
+        if (fl == FACTORS) return n; /* irreducible */
+        if (!n) return 0; /* no root */
+      }
+      if (DEBUGLEVEL>3)
+        err_printf("%3ld %s at prime (%ld,x^%ld+...)\n Time: %ld\n",
+            n, (fl == FACTORS)? "factors": "roots", p,f, timer_delay(&ti_pr));
 
-    if (!nbf || anbf < nbf
-             || (anbf == nbf && pr_get_f(apr) > pr_get_f(*pr)))
-    {
-      nbf = anbf;
-      *pr = apr;
-      *Tp = aT;
+      if (fl == ROOTS && f==nfdeg) { *Tp = T; *pp = p; return n; }
+      if (record(nold, n, fold, f)) { nold = n; fold = f; *Tp = T; *pp = p; }
+      else avma = av2;
     }
-    else avma = av2;
-    if (--ct <= 0) break;
+    if (ok && --ct <= 0) break;
   }
-  if (!nbf) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
-  return nbf;
+  if (!nold) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
+  return nold;
 }
 
 /* Assume lt(T) is a t_INT and T square free. Return t_VEC of irred. factors */
@@ -1772,8 +1829,9 @@ static GEN
 nfsqff(GEN nf, GEN pol, long fl, GEN den)
 {
   long n, nbf, dpol = degpol(pol);
-  GEN pr, C0, polbase;
+  GEN C0, polbase;
   GEN N2, res, polred, lt, nfpol = typ(nf)==t_POL?nf:nf_get_pol(nf);
+  ulong pp;
   nfcmbf_t T;
   nflift_t L;
   pari_timer ti, ti_tot;
@@ -1807,7 +1865,13 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   }
 
   polbase = RgX_to_nfX(nf, pol);
-  nbf = nf_pick_prime(5, nf, polbase, fl, &lt, &pr, &L.Tp);
+  nbf = nf_pick_prime(nf, pol, fl, &lt, &L.Tp, &pp);
+  if (L.Tp)
+  {
+    L.Tp = Flx_to_ZX(L.Tp);
+    L.p = utoi(pp);
+  }
+
   if (fl == ROOTS_SPLIT && nbf < dpol) return cgetg(1,t_VEC);
   if (nbf <= 1)
   {
@@ -1817,10 +1881,11 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
 
   if (DEBUGLEVEL>2) {
     timer_printf(&ti, "choice of a prime ideal");
-    err_printf("Prime ideal chosen: %Ps\n", pr);
+    err_printf("Prime ideal chosen: (%lu,x^%ld+...)\n", pp, degpol(L.Tp));
   }
   L.tozk = nf_get_invzk(nf);
-  L.topow= Q_remove_denom(nf_get_zk(nf), &L.topowden);
+  L.topow= nf_get_zkprimpart(nf);
+  L.topowden = nf_get_zkden(nf);
   if (is_pm1(den)) den = NULL;
   L.dn = den;
   T.ZC = L2_bound(nf, den);
@@ -1842,9 +1907,7 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
     err_printf("  3) Final bound: %Ps\n", T.bound);
   }
 
-  L.p = pr_get_p(pr);
-  if (L.Tp && degpol(L.Tp) == 1) L.Tp = NULL;
-  bestlift_init(0, nf, pr, T.bound, &L);
+  bestlift_init(0, nf, T.bound, &L);
   if (DEBUGLEVEL>2) timer_start(&ti);
   polred = ZqX_normalize(polbase, lt, &L); /* monic */
 
@@ -1855,8 +1918,8 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   }
 
   T.fact = gel(FqX_factor(polred, L.Tp, L.p), 1);
-  if (DEBUGLEVEL>2) timer_printf(&ti, "splitting mod %Ps", pr);
-  T.pr = pr;
+  if (DEBUGLEVEL>2)
+    timer_printf(&ti, "splitting mod %Ps^%ld", L.p, degpol(L.Tp));
   T.L  = &L;
   T.polbase = polbase;
   T.pol   = pol;
@@ -1899,19 +1962,12 @@ nfroots_if_split(GEN *pnf, GEN pol)
  *           Hensel lift to pr^k and find the representative in the ball
  *           If there is it is a primitive root */
 
-typedef struct {
-  GEN q;
-  GEN modpr;
-  GEN pr;
-  nflift_t *L;
-} prklift_t;
-
 /* Choose prime ideal unramified with "large" inertia degree */
 static void
-nf_pick_prime_for_units(GEN nf, prklift_t *P)
+nf_pick_prime_for_units(GEN nf, nflift_t *L)
 {
   GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
-  GEN aT, amodpr, apr, ap = NULL, r = NULL;
+  GEN ap = NULL, r = NULL;
   long nfdeg = degpol(nfpol), maxf = get_maxf(nfdeg);
   ulong pp;
   forprime_t S;
@@ -1924,16 +1980,13 @@ nf_pick_prime_for_units(GEN nf, prklift_t *P)
     if (r) break;
   }
   if (!r) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
+  r = gel(r,lg(r)-1); /* largest inertia degree */
   ap = utoipos(pp);
-  apr = idealprimedec_kummer(nf, Flx_to_ZX(r), 1, ap);
-  amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
-  P->pr = apr;
-  P->q = pr_norm(apr);
-  P->modpr = amodpr;
-  P->L->p = ap;
-  P->L->Tp = aT;
-  P->L->tozk = nf_get_invzk(nf);
-  P->L->topow = Q_remove_denom(nf_get_zk(nf), &(P->L->topowden));
+  L->p = ap;
+  L->Tp = Flx_to_ZX(r);
+  L->tozk = nf_get_invzk(nf);
+  L->topow = nf_get_zkprimpart(nf);
+  L->topowden = nf_get_zkden(nf);
 }
 
 /* *Heuristic* exponent k such that the fundamental domain of pr^k
@@ -1942,13 +1995,6 @@ static double
 mybestlift_bound(GEN C)
 {
   C = gtofp(C,DEFAULTPREC);
-#if 0 /* d = nf degree, Npr = Norm(pr) */
-  const double alpha = 0.99; /* LLL parameter */
-  const double y = 1 / (alpha - 0.25); /* = 2 if alpha = 3/4 */
-  double t;
-  t = rtodbl(mplog(gmul2n(divru(C,d), 4))) * 0.5 + (d-1) * log(1.5 * sqrt(y));
-  return ceil((t * d) / log(gtodouble(Npr))); /* proved upper bound */
-#endif
   return ceil(log(gtodouble(C)) / 0.2) + 3;
 }
 
@@ -2002,10 +2048,10 @@ guess_roots(GEN nf)
     /* the gcd of the p^f - 1 is p^(gcd of the f's) - 1 */
     for (i = 1; i <= nfdegree; i++)
       if (F[i]) {
-        gcdf = gcdf? cgcd(gcdf, i): i;
+        gcdf = gcdf? ugcd(gcdf, i): i;
         if (gcdf == 1) break;
       }
-    pf_1 = subis(powuu(p, gcdf), 1);
+    pf_1 = subiu(powuu(p, gcdf), 1);
     old = nbroots;
     nbroots = nbroots? gcdii(pf_1, nbroots): pf_1;
     if (DEBUGLEVEL>5)
@@ -2057,7 +2103,7 @@ ZXirred_is_cyclo_translate(GEN T, long n_squarefree)
   T1 = ZX_graeffe(T);
   if (ZX_equal(T, T1)) /* T = Phi_n, n odd */
     return deg1pol_shallow(gen_m1, negi(c), varn(T));
-  else if (ZX_equal(T1, ZX_unscale(T, gen_m1))) /* T = Phi_{2n}, nodd */
+  else if (ZX_equal(T1, ZX_z_unscale(T, -1))) /* T = Phi_{2n}, nodd */
     return deg1pol_shallow(gen_1, c, varn(T));
   return NULL;
 }
@@ -2068,9 +2114,8 @@ trivroots(void) { return mkvec2(gen_2, gen_m1); }
 GEN
 rootsof1(GEN nf)
 {
-  prklift_t P;
   nflift_t L;
-  GEN fa, LP, LE, C0, z, prim_root, disc, nfpol;
+  GEN T, q, fa, LP, LE, C0, z, disc;
   pari_timer ti;
   long i, l, nbguessed, nbroots, nfdegree;
   pari_sp av;
@@ -2127,60 +2172,59 @@ rootsof1(GEN nf)
   av = avma;
 
   /* Step 1.5 : test if nf.pol == subst(polcyclo(nbguessed), x, \pm x+c) */
+  T = nf_get_pol(nf);
   if (eulerphiu_fact(fa) == (ulong)nfdegree)
   {
-    GEN elt = ZXirred_is_cyclo_translate(nf_get_pol(nf),
-                                         uissquarefree_fact(fa));
-    if (elt)
+    z = ZXirred_is_cyclo_translate(T, uissquarefree_fact(fa));
+    if (DEBUGLEVEL>2) timer_printf(&ti, "checking for cyclotomic polynomial");
+    if (z)
     {
-      if (DEBUGLEVEL>2)
-        timer_printf(&ti, "checking for cyclotomic polynomial [yes]");
-      return gerepilecopy(av, mkvec2(utoipos(nbguessed), elt));
+      z = nf_to_scalar_or_basis(nf,z);
+      return gerepilecopy(av, mkvec2(utoipos(nbguessed), z));
     }
     avma = av;
   }
-  if (DEBUGLEVEL>2)
-    timer_printf(&ti, "checking for cyclotomic polynomial [no]");
 
   /* Step 2 : choose a prime ideal for local lifting */
-  P.L = &L; nf_pick_prime_for_units(nf, &P);
+  nf_pick_prime_for_units(nf, &L);
   if (DEBUGLEVEL>2)
     timer_printf(&ti, "choosing prime %Ps, degree %ld",
-             P.L->p, P.L->Tp? degpol(P.L->Tp): 1);
+             L.p, L.Tp? degpol(L.Tp): 1);
 
   /* Step 3 : compute a reduced pr^k allowing lifting of local solutions */
   /* evaluate maximum L2 norm of a root of unity in nf */
   C0 = gmulsg(nfdegree, L2_bound(nf, gen_1));
   /* lift and reduce pr^k */
   if (DEBUGLEVEL>2) err_printf("Lift pr^k; GSmin wanted: %Ps\n",C0);
-  bestlift_init((long)mybestlift_bound(C0), nf, P.pr, C0, P.L);
-  P.L->dn = NULL;
+  bestlift_init((long)mybestlift_bound(C0), nf, C0, &L);
+  L.dn = NULL;
   if (DEBUGLEVEL>2) timer_start(&ti);
 
   /* Step 4 : actual computation of roots */
-  nbroots = 2; prim_root = gen_m1;
-  nfpol = nf_get_pol(nf);
+  nbroots = 2; z = gen_m1;
+  q = powiu(L.p,degpol(L.Tp));
   for (i = 1; i < l; i++)
   { /* for all prime power factors of nbguessed, find a p^k-th root of unity */
     long k, p = LP[i];
-    for (k = minss(LE[i], Z_lval(subiu(P.q,1UL),p)); k > 0; k--)
+    for (k = minss(LE[i], Z_lval(subiu(q,1UL),p)); k > 0; k--)
     { /* find p^k-th roots */
       pari_sp av = avma;
       long pk = upowuu(p,k);
+      GEN r;
       if (pk==2) continue; /* no need to test second roots ! */
-      z = nfcyclo_root(pk, nfpol, P.L);
+      r = nfcyclo_root(pk, T, &L);
       if (DEBUGLEVEL>2) timer_printf(&ti, "for factoring Phi_%ld^%ld", p,k);
-      if (z) {
-        if (DEBUGLEVEL>2) err_printf("  %ld-th root of unity found.\n", pk);
-        if (p==2) { nbroots = pk; prim_root = z; }
-        else     { nbroots *= pk; prim_root = nfmul(nf, prim_root,z); }
+      if (r) {
+        if (DEBUGLEVEL>2) err_printf("  %s root of unity found\n",uordinal(pk));
+        if (p==2) { nbroots = pk; z = r; }
+        else     { nbroots *= pk; z = nfmul(nf, z,r); }
         break;
       }
       avma = av;
       if (DEBUGLEVEL) pari_warn(warner,"rootsof1: wrong guess");
     }
   }
-  return gerepilecopy(av, mkvec2(utoi(nbroots), prim_root));
+  return gerepilecopy(av, mkvec2(utoi(nbroots), z));
 }
 
 static long
@@ -2242,5 +2286,5 @@ rootsof1_kannan(GEN nf)
     if (z) return gerepilecopy(av, mkvec2(w, z));
   }
   pari_err_BUG("rootsof1");
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
