@@ -22,6 +22,25 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 static const double pariINFINITY = 1./0.;
 
+static long
+isvalidcoeff(GEN x)
+{
+  switch (typ(x))
+  {
+    case t_INT: case t_REAL: case t_FRAC: return 1;
+    case t_COMPLEX: return isvalidcoeff(gel(x,1)) && isvalidcoeff(gel(x,2));
+  }
+  return 0;
+}
+
+static void
+checkvalidpol(GEN p, const char *f)
+{
+  long i,n = lg(p);
+  for (i=2; i<n; i++)
+    if (!isvalidcoeff(gel(p,i))) pari_err_TYPE(f, gel(p,i));
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                   FAST ARITHMETIC over Z[i]                    **/
@@ -154,29 +173,6 @@ CX_square_spec(GEN P, long lP)
   }
   return normalizepol_lg(s, nn+3);
 }
-/* not stack clean */
-static GEN
-RgX_addspec(GEN x, long nx, GEN y, long ny)
-{
-  GEN z, t;
-  long i;
-  if (nx == ny) {
-    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
-    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
-    return normalizepol_lg(z, nx+2);
-  }
-  if (ny < nx) {
-    z = cgetg(nx+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
-    for (i=0; i < ny; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
-    for (   ; i < nx; i++) gel(t,i) = gel(x,i);
-    return normalizepol_lg(z, nx+2);
-  } else {
-    z = cgetg(ny+2,t_POL); z[1] = evalsigne(1)|evalvarn(0); t = z+2;
-    for (i=0; i < nx; i++) gel(t,i) = gadd(gel(x,i),gel(y,i));
-    for (   ; i < ny; i++) gel(t,i) = gel(y,i);
-    return normalizepol_lg(z, ny+2);
-  }
-}
 /* nx = lgpol(x) */
 static GEN
 RgX_s_mulspec(GEN x, long nx, long s)
@@ -213,7 +209,7 @@ karasquare(GEN P, long nP)
   n0 = (n>>1) + 1; n1 = nP - n0;
   s0 = karasquare(P, n0); Q = P + n0;
   s2 = karasquare(Q, n1);
-  s1 = RgX_addspec(P, n0, Q, n1);
+  s1 = RgX_addspec_shallow(P, Q, n0, n1);
   s1 = RgX_sub(karasquare(s1+2, lgpol(s1)), RgX_add(s0,s2));
   N = (n<<1) + 1;
   a = cgetg(N + 2, t_POL); a[1] = evalsigne(1)|evalvarn(0);
@@ -246,16 +242,16 @@ cook_square(GEN P, long nP)
 
   q = cgetg(8,t_VEC) + 4;
   Q = cook_square(p0, n0);
-  r = RgX_addspec(p0,n0, p2,n0);
-  t = RgX_addspec(p1,n0, p3,n3);
+  r = RgX_addspec_shallow(p0,p2, n0,n0);
+  t = RgX_addspec_shallow(p1,p3, n0,n3);
   gel(q,-1) = RgX_sub(r,t);
   gel(q,1)  = RgX_add(r,t);
-  r = RgX_addspec(p0,n0, RgX_shiftspec(p2,n0, 2)+2,n0);
-  t = gmul2n(RgX_addspec(p1,n0, RgX_shiftspec(p3,n3, 2)+2,n3), 1);
+  r = RgX_addspec_shallow(p0,RgX_shiftspec(p2,n0, 2)+2, n0,n0);
+  t = gmul2n(RgX_addspec_shallow(p1,RgX_shiftspec(p3,n3, 2)+2, n0,n3), 1);
   gel(q,-2) = RgX_sub(r,t);
   gel(q,2)  = RgX_add(r,t);
-  r = RgX_addspec(p0,n0, RgX_s_mulspec(p2,n0, 9)+2,n0);
-  t = gmulsg(3, RgX_addspec(p1,n0, RgX_s_mulspec(p3,n3, 9)+2,n3));
+  r = RgX_addspec_shallow(p0,RgX_s_mulspec(p2,n0, 9)+2, n0,n0);
+  t = gmulsg(3, RgX_addspec_shallow(p1,RgX_s_mulspec(p3,n3, 9)+2, n0,n3));
   gel(q,-3) = RgX_sub(r,t);
   gel(q,3)  = RgX_add(r,t);
 
@@ -389,7 +385,7 @@ mydbllog2i(GEN x)
 static double
 mydbllogr(GEN x) {
   if (!signe(x)) return -pariINFINITY;
-  return LOG2*dbllog2r(x);
+  return M_LN2*dbllog2r(x);
 }
 
 /* return log2(|x|) or -pariINFINITY */
@@ -453,8 +449,7 @@ newton_polygon(GEN p, long k)
   double *logcoef, slope;
   long n = degpol(p), i, j, h, l, *vertex;
 
-  init_dalloc();
-  logcoef = (double*)stack_malloc((n+1)*sizeof(double));
+  logcoef = (double*)stack_malloc_align((n+1)*sizeof(double), sizeof(double));
   vertex = (long*)new_chunk(n+1);
 
   /* vertex[i] = 1 if i a vertex of convex hull, 0 otherwise */
@@ -565,7 +560,7 @@ eval_rel_pol(GEN p, long bit)
 {
   long i;
   for (i = 2; i < lg(p); i++)
-    if (gequal0(gel(p,i))) gel(p,i) = gen_0; /* bad behaviour of gexpo */
+    if (gequal0(gel(p,i))) gel(p,i) = gen_0; /* bad behavior of gexpo */
   return pol_to_gaussint(p, bit-gexpo(p)+1);
 }
 
@@ -648,7 +643,7 @@ lower_bound(GEN p, long *k, double eps)
   avma = ltop; return R;
 }
 
-/* log of modulus of the largest root of p with relative error tau. Assume
+/* return R such that exp(R - tau) <= rho_n(P) <= exp(R + tau)
  * P(0) != 0 and P non constant */
 static double
 logmax_modulus(GEN p, double tau)
@@ -691,14 +686,46 @@ logmax_modulus(GEN p, double tau)
   }
   if (!signe(r)) { avma = ltop; return 0.; }
   r = itor(r, DEFAULTPREC); shiftr_inplace(r, -M);
-  avma = ltop; return -rtodbl(r) * LOG2; /* -log(2) sum e_i 2^-i */
+  avma = ltop; return -rtodbl(r) * M_LN2; /* -log(2) sum e_i 2^-i */
 }
-/* assume P non constant */
-GEN
-logmax_modulus_bound(GEN P)
+
+static GEN
+RgX_normalize1(GEN x)
 {
+  long i, n = lg(x)-1;
+  GEN y;
+  for (i = n; i > 1; i--)
+    if (!gequal0( gel(x,i) )) break;
+  if (i == n) return x;
+  pari_warn(warner,"normalizing a polynomial with 0 leading term");
+  if (i == 1) pari_err_ROOTS0("roots");
+  y = cgetg(i+1, t_POL); y[1] = x[1];
+  for (; i > 1; i--) gel(y,i) = gel(x,i);
+  return y;
+}
+
+static GEN
+polrootsbound_i(GEN P, double TAU)
+{
+  pari_sp av = avma;
+  double d;
   (void)RgX_valrem_inexact(P,&P);
-  return lg(P)==3? gen_0: dblexp(logmax_modulus(P, 0.01) + 0.01);
+  P = RgX_normalize1(P);
+  switch(degpol(P))
+  {
+    case -1: pari_err_ROOTS0("roots");
+    case 0:  avma = av; return gen_0;
+  }
+  d = logmax_modulus(P, TAU) + TAU;
+  /* not dblexp: result differs on ARM emulator */
+  return gerepileuptoleaf(av, mpexp(dbltor(d)));
+}
+GEN
+polrootsbound(GEN P, GEN tau)
+{
+  if (typ(P) != t_POL) pari_err_TYPE("polrootsbound",P);
+  checkvalidpol(P, "polrootsbound");
+  return polrootsbound_i(P, tau? gtodouble(tau): 0.01);
 }
 
 /* log of modulus of the smallest root of p, with relative error tau */
@@ -746,7 +773,7 @@ logmodulus(GEN p, long k, double tau)
     tau2 *= 1.5; if (tau2 > 1.) tau2 = 1.;
     bit = 1 + (long)(nn*(2. + log2(3.*nn/tau2)));
   }
-  avma = ltop; return -r * LOG2;
+  avma = ltop; return -r * M_LN2;
 }
 
 /* return the log of the k-th modulus r_k of p, rel. error tau, knowing that
@@ -766,7 +793,7 @@ logpre_modulus(GEN p, long k, double tau, double lrmin, double lrmax)
 
   lrho  = (lrmin + lrmax) / 2;
   av = avma;
-  bit = (long)(n*(2. + aux / LOG2 - log2(tau2)));
+  bit = (long)(n*(2. + aux / M_LN2 - log2(tau2)));
   q = homothetie(p, lrho, bit);
   imax2 = (long)(log2(3./tau * log(4.*n))) + 1;
   if (imax > imax2) imax = imax2;
@@ -778,7 +805,7 @@ logpre_modulus(GEN p, long k, double tau, double lrmin, double lrmax)
     q = gerepileupto(av, graeffe(q));
     aux = 2*aux + 2*tau2;
     tau2 *= 1.5;
-    bit = (long)(n*(2. + aux / LOG2 - log2(1-exp(-tau2))));
+    bit = (long)(n*(2. + aux / M_LN2 - log2(1-exp(-tau2))));
     q = RgX_gtofp_bit(q, bit);
   }
   aux = exp2((double)imax);
@@ -1203,7 +1230,7 @@ split_fromU(GEN p, long k, double delta, long bit,
   NN *= Lmax; ltop = avma;
   for(;;)
   {
-    bit2 = (long)(((double)NN*delta-mu)/LOG2) + gexpo(pp) + 8;
+    bit2 = (long)(((double)NN*delta-mu)/M_LN2) + gexpo(pp) + 8;
     dft(pp, k, NN, Lmax, bit2, FF, H, polreal);
     if (refine_F(pp,&FF,&GG,H,bit,gamma)) break;
     NN <<= 1; avma = ltop;
@@ -1262,7 +1289,7 @@ scalepol(GEN p, GEN R, long bit)
 static GEN
 conformal_pol(GEN p, GEN a)
 {
-  GEN z, r, ma = gneg(a), ca = gconj(a);
+  GEN z, r, ma = gneg(a), ca = conj_i(a);
   long n = degpol(p), i;
   pari_sp av = avma;
 
@@ -1270,10 +1297,10 @@ conformal_pol(GEN p, GEN a)
   r = scalarpol(gel(p,2+n), 0);
   for (i=n-1; ; i--)
   {
-    r = addmulXn(r, gmul(ma,r), 1); /* r *= (X - a) */
+    r = RgX_addmulXn_shallow(r, gmul(ma,r), 1); /* r *= (X - a) */
     r = gadd(r, gmul(z, gel(p,2+i)));
     if (i == 0) return gerepileupto(av, r);
-    z = addmulXn(gmul(z,ca), gneg(z), 1); /* z *= conj(a)X - 1 */
+    z = RgX_addmulXn_shallow(gmul(z,ca), gneg(z), 1); /* z *= conj(a)X - 1 */
     if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"conformal_pol");
@@ -1372,7 +1399,7 @@ conformal_mapping(double *radii, GEN ctr, GEN p, long k, long bit,
   lrho = logradius(radii, q,k,aux/10., &delta);
   update_radius(n, radii, lrho, &param, &param2);
 
-  bit2 += (long)(n * fabs(lrho)/LOG2 + 1.);
+  bit2 += (long)(n * fabs(lrho)/M_LN2 + 1.);
   R = mygprec(dblexp(-lrho), bit2);
   q = scalepol(q,R,bit2);
   gerepileall(av,2, &q,&R);
@@ -1404,8 +1431,7 @@ split_2(GEN p, long bit, GEN ctr, double thickness, GEN *F, GEN *G)
   long n = degpol(p), i, j, k, bit2;
   double lrmin, lrmax, lrho, *radii;
 
-  init_dalloc();
-  radii = (double*) stack_malloc((n+1) * sizeof(double));
+  radii = (double*) stack_malloc_align((n+1) * sizeof(double), sizeof(double));
 
   for (i=2; i<n; i++) radii[i] = UNDEF;
   aux = thickness/(double)(4 * n);
@@ -1446,7 +1472,7 @@ split_2(GEN p, long bit, GEN ctr, double thickness, GEN *F, GEN *G)
     for (i=1; i<=n; i++)
       if (radii[i] != UNDEF) radii[i] -= lrho;
 
-    bit2 = bit + (long)(n * fabs(lrho)/LOG2 + 1.);
+    bit2 = bit + (long)(n * fabs(lrho)/M_LN2 + 1.);
     R = mygprec(dblexp(-lrho), bit2);
     q = scalepol(p,R,bit2);
     conformal_mapping(radii, ctr, q, k, bit2, aux, &FF, &GG);
@@ -1456,7 +1482,7 @@ split_2(GEN p, long bit, GEN ctr, double thickness, GEN *F, GEN *G)
     lrho = logradius(radii, p, k, aux/10., &delta);
     update_radius(n, radii, lrho, &param, &param2);
 
-    bit2 = bit + (long)(n * fabs(lrho)/LOG2 + 1.);
+    bit2 = bit + (long)(n * fabs(lrho)/M_LN2 + 1.);
     R = mygprec(dblexp(-lrho), bit2);
     q = scalepol(p,R,bit2);
     optimize_split(q, k, delta, bit2, &FF, &GG, param, param2);
@@ -1501,10 +1527,10 @@ split_1(GEN p, long bit, GEN *F, GEN *G)
       double lquo = logmax_modulus(qq,0.05) - lrmin;
       if (lquo > lthick) { lthick = lquo; newq = qq; ctr = gel(v,i); }
     }
-    if (lthick > LOG2) break;
-    if (polreal && i==2 && lthick > LOG3 - LOG2) break;
+    if (lthick > M_LN2) break;
+    if (polreal && i==2 && lthick > LOG3 - M_LN2) break;
   }
-  bit2 = bit + gexpo(newq) - ep + (long)(n*LOG3/LOG2 + 1);
+  bit2 = bit + gexpo(newq) - ep + (long)(n*LOG3/M_LN2 + 1);
   split_2(newq, bit2, ctr, lthick, &FF, &GG);
   r = gneg(mygprec(ctr,bit2));
   FF = RgX_translate(FF,r);
@@ -1775,7 +1801,7 @@ quicktofp(GEN x)
       b = cxcompotor(b, prec); return sqrtr(addrr(sqrr(a), sqrr(b)));
     }
     default: pari_err_TYPE("quicktofp",x);
-      return NULL;/*not reached*/
+      return NULL;/*LCOV_EXCL_LINE*/
   }
 
 }
@@ -1918,21 +1944,6 @@ fix_roots(GEN r, GEN *m, long h, long bit)
   return allr;
 }
 
-static GEN
-RgX_normalize1(GEN x)
-{
-  long i, n = lg(x)-1;
-  GEN y;
-  for (i = n; i > 1; i--)
-    if (!gequal0( gel(x,i) )) break;
-  if (i == n) return x;
-  pari_warn(warner,"normalizing a polynomial with 0 leading term");
-  if (i == 1) pari_err_ROOTS0("roots");
-  y = cgetg(i+1, t_POL); y[1] = x[1];
-  for (; i > 1; i--) gel(y,i) = gel(x,i);
-  return y;
-}
-
 static GEN
 all_roots(GEN p, long bit)
 {
@@ -1982,25 +1993,6 @@ isexactpol(GEN p)
   return 1;
 }
 
-static long
-isvalidcoeff(GEN x)
-{
-  switch (typ(x))
-  {
-    case t_INT: case t_REAL: case t_FRAC: return 1;
-    case t_COMPLEX: return isvalidcoeff(gel(x,1)) && isvalidcoeff(gel(x,2));
-  }
-  return 0;
-}
-
-static void
-checkvalidpol(GEN p)
-{
-  long i,n = lg(p);
-  for (i=2; i<n; i++)
-    if (!isvalidcoeff(gel(p,i))) pari_err_TYPE("roots", gel(p,i));
-}
-
 static GEN
 solve_exact_pol(GEN p, long bit)
 {
@@ -2152,7 +2144,7 @@ roots_aux(GEN p, long l, long clean)
     return cgetg(1,t_COL); /* constant polynomial */
   }
   if (!signe(p)) pari_err_ROOTS0("roots");
-  checkvalidpol(p);
+  checkvalidpol(p,"roots");
   if (lg(p) == 3) return cgetg(1,t_COL); /* constant polynomial */
   if (l < LOWDEFAULTPREC) l = LOWDEFAULTPREC;
   bit = prec2nbits(l);
@@ -2174,7 +2166,7 @@ polmod_to_embed(GEN x, long prec)
   long i, l;
   if (typ(A) != t_POL || varn(A) != varn(T))
   {
-    checkvalidpol(T);
+    checkvalidpol(T,"polmod_to_embed");
     return const_col(degpol(T), A);
   }
   v = cleanroots(T,prec); l = lg(v);
@@ -2364,14 +2356,11 @@ splitcauchy(GEN Pp, GEN Pm, long prec)
 static GEN
 polsolve(GEN P, long bitprec)
 {
-  pari_sp av = avma;
-  GEN Pp, Pm, Pprimep, Pprimem, pows;
-  GEN Pprime, Pprime2;
-  GEN ra, rb, rc, rcold = NULL, Pc, Ppc;
-  long deg = degpol(P), D, rt;
-  long s0, cprec = DEFAULTPREC, prec = nbits2prec(bitprec);
-  long bitaddprec, addprec;
-  long expoold = LONG_MAX, iter;
+  pari_sp av = avma, av2;
+  GEN Pp, Pm, Pprimep, Pprimem, Pprime, Pprime2, ra, rb, rc, Pc;
+  long deg = degpol(P);
+  long expoold = LONG_MAX, cprec = DEFAULTPREC, prec = nbits2prec(bitprec);
+  long iter, D, rt, s0, bitaddprec, addprec;
   if (deg == 1)
     return gerepileuptoleaf(av, rdivii(negi(gel(P,2)), gel(P,3), prec));
   Pprime = ZX_deriv(P);
@@ -2383,7 +2372,7 @@ polsolve(GEN P, long bitprec)
   rb = splitcauchy(Pp, Pm, DEFAULTPREC);
   for(;;)
   {
-    pows = gen_powers(rb, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+    GEN pows = gen_powers(rb, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
     Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec);
     if (!Pc) { cprec++; rb = rtor(rb, cprec); continue; }
     if (signe(Pc) != s0) break;
@@ -2401,7 +2390,7 @@ polsolve(GEN P, long bitprec)
       rc = shiftr(rb, -1);
     do
     {
-      pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+      GEN pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
       Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec+2);
       if (Pc) break;
       cprec++; rc = rtor(rc, cprec);
@@ -2424,25 +2413,22 @@ polsolve(GEN P, long bitprec)
       break;
   }
   rc = rb;
-  iter = 0;
-  for(;;)
+  av2 = avma;
+  for(;; rc = gerepileuptoleaf(av2, rc))
   {
     long exponew;
-    GEN dist;
-    iter++;
-    rcold = rc;
-    pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
+    GEN Ppc, dist, rcold = rc;
+    GEN pows = gen_powers(rc, rt, 1, NULL, _mp_sqr, _mp_mul, _gen_one);
     Ppc = splitpoleval(Pprimep, Pprimem, pows, D-1, bitaddprec+4);
     if (Ppc)
       Pc = splitpoleval(Pp, Pm, pows, D, bitaddprec+4);
     if (!Ppc || !Pc)
     {
       if (cprec >= prec+addprec)
-        cprec++;
+        cprec += EXTRAPRECWORD;
       else
         cprec = minss(2*cprec, prec+addprec);
-      rc = rtor(rc, cprec); /* Backtrack one step */
-      continue;
+      rc = rtor(rc, cprec); continue; /* backtrack one step */
     }
     dist = typ(Ppc) == t_REAL? divrr(Pc, Ppc): divri(Pc, Ppc);
     rc = subrr(rc, dist);
@@ -2450,8 +2436,7 @@ polsolve(GEN P, long bitprec)
     {
       if (cprec >= prec+addprec) break;
       cprec = minss(2*cprec, prec+addprec);
-      rc = rtor(rcold, cprec); /* Backtrack one step */
-      continue;
+      rc = rtor(rcold, cprec); continue; /* backtrack one step */
     }
     if (expoold == LONG_MAX) { expoold = expo(dist); continue; }
     exponew = expo(dist);
@@ -2464,9 +2449,9 @@ polsolve(GEN P, long bitprec)
     if (exponew > expoold - 2)
     {
       if (cprec >= prec+addprec) break;
+      expoold = LONG_MAX;
       cprec = minss(2*cprec, prec+addprec);
-      rc = rtor(rc, cprec);
-      expoold = LONG_MAX; continue;
+      rc = rtor(rc, cprec); continue;
     }
     expoold = exponew;
   }
@@ -2481,9 +2466,9 @@ usp(GEN Q0, long deg, long flag, long bitprec)
   long listsize = 64, nbr = 0, nb_todo, ind, deg0, indf, i, k, nb;
 
 
-  sol = const_col(deg, gen_0);
+  sol = zerocol(deg);
   deg0 = deg;
-  Lc = const_vec(listsize, gen_0);
+  Lc = zerovec(listsize);
   Lk = cgetg(listsize+1, t_VECSMALL);
   c = gen_0;
   k = Lk[1] = 0;
@@ -2598,7 +2583,7 @@ ZX_Uspensky_cst_pol(long nbz, long flag, long bitprec)
   switch(flag)
   {
     case 0:  return zerocol(nbz);
-    case 1:  retconst_col(nbz, real_0_bit(bitprec));
+    case 1:  retconst_col(nbz, real_0_bit(-bitprec));
     default: return utoi(nbz);
   }
 }
@@ -2671,7 +2656,7 @@ ZX_Uspensky(GEN P, GEN ab, long flag, long bitprec)
       sol = zerocol(nbz);
       break;
     case 1:
-      sol = const_col(nbz, real_0_bit(bitprec));
+      sol = const_col(nbz, real_0_bit(-bitprec));
       break;
     /* case 2: nothing */
   }
@@ -2765,7 +2750,7 @@ ZX_Uspensky(GEN P, GEN ab, long flag, long bitprec)
     GEN Pcurm, unscaledres;
     long i, bm = (long)ceil(fb);
     if (bm < 0) bm = 0;
-    Pcurm = ZX_unscale2n(ZX_unscale(Pcur, gen_m1), bm);
+    Pcurm = ZX_unscale2n(ZX_z_unscale(Pcur, -1), bm);
     unscaledres = usp(Pcurm, deg, flag, bitprec);
     if (flag <= 1)
     {
@@ -2792,7 +2777,7 @@ ZX_Uspensky(GEN P, GEN ab, long flag, long bitprec)
 static GEN
 rootsdeg0(GEN x)
 {
-  if (!is_rational_t(typ(x))) pari_err_TYPE("realroots",x);
+  if (!is_real_t(typ(x))) pari_err_TYPE("realroots",x);
   if (gequal0(x)) pari_err_ROOTS0("realroots");
   return cgetg(1,t_COL); /* constant polynomial */
 }
@@ -2832,8 +2817,8 @@ realroots(GEN P, GEN ab, long prec)
     case -1: return rootsdeg0(gen_0);
     case 0: return rootsdeg0(gel(P,2));
   }
+  if (!RgX_is_ZX(P)) P = RgX_rescale_to_int(P);
   P = Q_primpart(P);
-  RgX_check_ZX(P,"realroots");
   v = ZX_valrem(P,&P);
   if (v && (!ab || (gsigne(gel(ab,1)) <= 0 && gsigne(gel(ab,2)) >= 0)))
     sol = const_col(v, real_0(prec));
