@@ -274,7 +274,7 @@ consteuler(long prec)
 
   incrprec(prec);
 
-  l = prec+EXTRAPRECWORD; x = (long) (1 + prec2nbits_mul(l, LOG2/4));
+  l = prec+EXTRAPRECWORD; x = (long) (1 + prec2nbits_mul(l, M_LN2/4));
   a = utor(x,l); u=logr_abs(a); setsigne(u,-1); affrr(u,a);
   b = real_1(l);
   v = real_1(l);
@@ -418,13 +418,13 @@ static GEN
 powr0(GEN x)
 { return signe(x)? real_1(realprec(x)): mpexp0(x); }
 
-/* x t_POL or t_SER, return scalarpol(RgX_get_1(x)) */
+/* x t_POL or t_SER, return scalarpol(Rg_get_1(x)) */
 static GEN
 scalarpol_get_1(GEN x)
 {
   GEN y = cgetg(3,t_POL);
   y[1] = evalvarn(varn(x)) | evalsigne(1);
-  gel(y,2) = RgX_get_1(x); return y;
+  gel(y,2) = Rg_get_1(x); return y;
 }
 /* to be called by the generic function gpowgs(x,s) when s = 0 */
 static GEN
@@ -456,9 +456,7 @@ gpowg0(GEN x)
     case t_FFELT: return FF_1(x);
 
     case t_POLMOD:
-      y = cgetg(3,t_POLMOD);
-      gel(y,1) = gcopy(gel(x,1));
-      gel(y,2) = scalarpol_get_1(gel(x,1)); return y;
+      retmkpolmod(scalarpol_get_1(gel(x,1)), gcopy(gel(x,1)));
 
     case t_RFRAC:
       return scalarpol_get_1(gel(x,2));
@@ -476,7 +474,7 @@ gpowg0(GEN x)
     case t_VECSMALL: return identity_perm(lg(x) - 1);
   }
   pari_err_TYPE("gpow",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -805,7 +803,7 @@ pow_monome(GEN x, long n)
   b = gpowgs(gel(x,dx+2), n); /* not memory clean if (n < 0) */
   if (!y) y = A;
   else {
-    GEN c = denom(b);
+    GEN c = denom_i(b);
     gel(y,1) = c; if (c != gen_1) b = gmul(b,c);
     gel(y,2) = A;
   }
@@ -1102,7 +1100,7 @@ gpow0(GEN x, GEN n, long prec)
 GEN
 gpow(GEN x, GEN n, long prec)
 {
-  long i, lx, tx, tn = typ(n);
+  long prec0, i, lx, tx, tn = typ(n);
   pari_sp av;
   GEN y;
 
@@ -1130,6 +1128,7 @@ gpow(GEN x, GEN n, long prec)
   if (tn == t_FRAC)
   {
     GEN z, d = gel(n,2), a = gel(n,1);
+    long D;
     switch (tx)
     {
     case t_INTMOD:
@@ -1145,21 +1144,42 @@ gpow(GEN x, GEN n, long prec)
       }
 
     case t_PADIC:
-      z = Qp_sqrtn(x, d, NULL);
-      if (!z) pari_err_SQRTN("gpow",x);
+      z = Qp_sqrtn(x, d, NULL); if (!z) pari_err_SQRTN("gpow",x);
       return gerepileupto(av, powgi(z, a));
 
     case t_FFELT:
       return gerepileupto(av,FF_pow(FF_sqrtn(x,d,NULL),a));
     default:
-      if (gequal(d, gen_2))
-        return gerepileupto(av, gmul(powgi(x, shifti(subis(a, 1), -1)),
-                                     gsqrt(x, prec)));
+      D = itos_or_0(d);
+      if (!D) break;
+      if (D == 2)
+      {
+        GEN y = gsqrt(x,prec), t = shifti(subiu(a,1), -1);
+        if (signe(t)) y = gmul(y, powgi(x,t));
+        return gerepileupto(av, y);
+      }
+      if (is_real_t(tx) && gsigne(x) > 0)
+      {
+        prec += nbits2extraprec(expi(a));
+        if (tx != t_REAL) x = gtofp(x, prec);
+        z = sqrtnr(x, D);
+        if (!equali1(a)) z = powgi(z, a);
+        return gerepileuptoleaf(av, z);
+      }
     }
   }
-  i = (long) precision(n); if (i) prec=i;
+  i = precision(n);
+  if (i) prec = i;
+  prec0 = prec;
+  if (!gprecision(x))
+  {
+    long e = gexpo_safe(n); /* avoided if n = 0 or gexpo not defined */
+    if (e > 2) prec += nbits2extraprec(e);
+  }
   y = gmul(n, glog(x,prec));
-  return gerepileupto(av, gexp(y,prec));
+  y = gexp(y,prec);
+  if (prec0 == prec) return gerepileupto(av, y);
+  return gerepilecopy(av, gprec_wtrunc(y,prec0));
 }
 
 GEN
@@ -1223,12 +1243,18 @@ grootsof1(long N, long prec)
   long i, k;
 
   if ((N & 3) == 0) return grootsof1_4(N, prec);
-  if (N == 1) return mkvec(gen_1);
+  if (N <= 2) return N == 1? mkvec(gen_1): mkvec2(gen_1, gen_m1);
   k = (N+3)>>1;
   RU = cgetg(N+1,t_VEC);
   v  = ((GEN*)RU) + 1;
   v[0] = gen_1; v[1] = z = rootsof1u_cx(N, prec);
-  for (i=2; i<k; i++) v[i] = gmul(z, v[i-1]);
+  if (odd(N))
+    for (i=2; i<k; i++) v[i] = gmul(z, v[i-1]);
+  else
+  {
+    for (i=2; i<k-1; i++) v[i] = gmul(z, v[i-1]);
+    v[i++] = gen_m1; /*avoid loss of accuracy*/
+  }
   for (   ; i<N; i++) v[i] = gconj(v[N-i]);
   return RU;
 }
@@ -1353,7 +1379,7 @@ sqrt_ser(GEN b, long prec)
   a = leafcopy(b);
   x = cgetg_copy(b, &lx);
   if (e & 1)
-    pari_err_DOMAIN("sqrtn", "valuation", "!=", mkintmod(gen_0, gen_2), x);
+    pari_err_DOMAIN("sqrtn", "valuation", "!=", mkintmod(gen_0, gen_2), b);
   a[1] = x[1] = evalsigne(1) | evalvarn(0) | _evalvalp(0);
   lta = gel(a,2);
   if (gequal1(lta)) ltx = lta;
@@ -1535,7 +1561,7 @@ Qp_log(GEN x)
     y = gmul2n(palogaux(y), 1);
   else
   { /* compute log(x^(p-1)) / (p-1) */
-    GEN mod = gel(y,3), p1 = subis(p,1);
+    GEN mod = gel(y,3), p1 = subiu(p,1);
     gel(y,4) = Fp_pow(a, p1, mod);
     p1 = diviiexact(subsi(1,mod), p1); /* 1/(p-1) */
     y = gmul(palogaux(y), shifti(p1,1));
@@ -1577,7 +1603,7 @@ Qp_sqrtn_ram(GEN x, long e)
     if (!a) return NULL;
     /* n=p^e and a^n=z*x where z is a (p-1)th-root of 1.
      * Since z^n=z, we have (a/z)^n = x. */
-    a = gdiv(x, powgi(a,addis(n,-1))); /* = a/z = x/a^(n-1)*/
+    a = gdiv(x, powgi(a,subiu(n,1))); /* = a/z = x/a^(n-1)*/
     if (v) setvalp(a,v);
   }
   return gerepileupto(ltop,a);
@@ -1682,42 +1708,31 @@ sqrtnint(GEN a, long n)
     flag = cmpii(a, powuu(3, n)); avma = ltop;
     return (flag < 0) ? gen_2: stoi(3);
   }
-  if (e < n*(BITS_IN_LONG - 1))
+  if (e < n*BITS_IN_LONG - 1)
   {
-    ulong s, xs, qs;
-    s = 1 + e/n; xs = 1UL << s;
-    qs = itou(shifti(a, -nm1*s));
-    while (qs < xs) {
-      xs -= (xs - qs + nm1)/n;
+    ulong xs, qs;
+    b = itor(a, (2*e < n*BITS_IN_LONG)? DEFAULTPREC: MEDDEFAULTPREC);
+    x = mpexp(divru(logr_abs(b), n));
+    xs = itou(floorr(x)) + 1; /* >= a^(1/n) */
+    for(;;) {
       q = divii(a, powuu(xs, nm1));
       if (lgefint(q) > 3) break;
-      qs = itou(q);
+      qs = itou(q); if (qs >= xs) break;
+      xs -= (xs - qs + nm1)/n;
     }
     return utoi(xs);
   }
-  b = addsi(1, shifti(a, -n*k));
-  x = shifti(addsi(1, sqrtnint(b, n)), k);
+  b = addui(1, shifti(a, -n*k));
+  x = shifti(addui(1, sqrtnint(b, n)), k);
   q = divii(a, powiu(x, nm1));
   while (cmpii(q, x) < 0) /* a priori one iteration, no GC necessary */
   {
-    x = subii(x, divis(addsi(nm1, subii(x, q)), n));
+    x = subii(x, divis(addui(nm1, subii(x, q)), n));
     q = divii(a, powiu(x, nm1));
   }
   return gerepileuptoleaf(ltop, x);
 }
 
-GEN
-cbrtr_abs(GEN x)
-{
-  long prec = realprec(x), n = bit_accuracy(prec), e = expo(x), er = e / 3;
-  /* x 2^3(n-er) = b t_INT */
-  GEN b = mantissa2nr(x, (e-3*er+1) + n*2);
-  b = sqrtnint(b, 3);
-  b = itor(b, prec);
-  setexpo(b, expo(b)+er-n);
-  return b;
-}
-
 ulong
 usqrtn(ulong a, ulong n)
 {
@@ -1736,6 +1751,198 @@ usqrtn(ulong a, ulong n)
   return x;
 }
 
+static ulong
+cubic_prec_mask(long n)
+{
+  long a = n, i;
+  ulong mask = 0;
+  for(i = 1;; i++, mask *= 3)
+  {
+    long c = a%3;
+    if (c) mask += 3 - c;
+    a = (a+2)/3;
+    if (a==1) return mask + upowuu(3, i);
+  }
+}
+
+/* cubic Newton iteration, |a|^(1/n), assuming a != 0 */
+GEN
+sqrtnr_abs(GEN a, long n)
+{
+  pari_sp av;
+  GEN x, b;
+  long eextra, eold, n1, n2, prec, B, v;
+  ulong mask;
+
+  if (n == 1) return mpabs(a);
+  if (n == 2) return sqrtr_abs(a);
+
+  prec = realprec(a);
+  B = prec2nbits(prec);
+  eextra = expu(n)-1;
+  n1 = n+1;
+  n2 = 2*n; av = avma;
+  v = expo(a) / n;
+  if (v) a = shiftr(a, -n*v);
+
+  b = rtor(a, DEFAULTPREC);
+  x = mpexp(divru(logr_abs(b), n));
+  if (prec == DEFAULTPREC)
+  {
+    if (v) shiftr_inplace(x, v);
+    return gerepileuptoleaf(av, x);
+  }
+  mask = cubic_prec_mask(B + 63);
+  eold = 1;
+  for(;;)
+  { /* reach 64 */
+    long enew = eold * 3;
+    enew -= mask % 3;
+    if (enew > 64) break; /* back up one step */
+    mask /= 3;
+    eold = enew;
+  }
+  for(;;)
+  {
+    long pr, enew = eold * 3;
+    GEN y, z;
+    enew -= mask % 3;
+    mask /= 3;
+    pr = nbits2prec(enew + eextra);
+    b = rtor(a, pr); setsigne(b,1);
+    x = rtor(x, pr);
+    y = subrr(powru(x, n), b);
+    z = divrr(y, addrr(mulur(n1, y), mulur(n2, b)));
+    shiftr_inplace(z,1);
+    x = mulrr(x, subsr(1,z));
+    if (mask == 1)
+    {
+      if (v) shiftr_inplace(x, v);
+      return gerepileuptoleaf(av, gprec_wtrunc(x,prec));
+    }
+    eold = enew;
+  }
+}
+
+static void
+shiftc_inplace(GEN z, long d)
+{
+  shiftr_inplace(gel(z,1), d);
+  shiftr_inplace(gel(z,2), d);
+}
+
+/* exp(2*Pi*I/n), same iteration as sqrtnr_abs, different initial point */
+static GEN
+sqrtnof1(ulong n, long prec)
+{
+  pari_sp av;
+  GEN x;
+  long eold, n1, n2, B;
+  ulong mask;
+
+  B = prec2nbits(prec);
+  n1 = n+1;
+  n2 = 2*n; av = avma;
+
+  x = expIr(divru(Pi2n(1, LOWDEFAULTPREC), n));
+  if (prec == LOWDEFAULTPREC) return gerepileupto(av, x);
+  mask = cubic_prec_mask(B + BITS_IN_LONG-1);
+  eold = 1;
+  for(;;)
+  { /* reach BITS_IN_LONG */
+    long enew = eold * 3;
+    enew -= mask % 3;
+    if (enew > BITS_IN_LONG) break; /* back up one step */
+    mask /= 3;
+    eold = enew;
+  }
+  for(;;)
+  {
+    long pr, enew = eold * 3;
+    GEN y, z;
+    enew -= mask % 3;
+    mask /= 3;
+    pr = nbits2prec(enew);
+    x = cxtofp(x, pr);
+    y = gsub(gpowgs(x, n), gen_1);
+    z = gdiv(y, gaddgs(gmulsg(n1, y), n2));
+    shiftc_inplace(z,1);
+    x = gmul(x, gsubsg(1, z));
+    if (mask == 1) return gerepileupto(av, gprec_w(x,prec));
+    eold = enew;
+  }
+}
+
+/* exp(2iPi/d) */
+GEN
+rootsof1u_cx(ulong n, long prec)
+{
+  switch(n)
+  {
+    case 1: return gen_1;
+    case 2: return gen_m1;
+    case 4: return gen_I();
+    case 3: case 6: case 12:
+    {
+      pari_sp av = avma;
+      GEN a = (n == 3)? mkfrac(gen_m1,gen_2): ghalf;
+      GEN sq3 = sqrtr_abs(utor(3, prec));
+      shiftr_inplace(sq3, -1);
+      a = (n == 12)? mkcomplex(sq3, a): mkcomplex(a, sq3);
+      return gerepilecopy(av, a);
+    }
+    case 8:
+    {
+      pari_sp av = avma;
+      GEN sq2 = sqrtr_abs(utor(2, prec));
+      shiftr_inplace(sq2,-1);
+      return gerepilecopy(av, mkcomplex(sq2, sq2));
+    }
+  }
+  return sqrtnof1(n, prec);
+}
+/* e(a/b) */
+GEN
+rootsof1q_cx(long a, long b, long prec)
+{
+  long g = cgcd(a,b);
+  GEN z;
+  if (g != 1) { a /= g; b /= g; }
+  if (b < 0) { b = -b; a = -a; }
+  z = rootsof1u_cx(b, prec);
+  if (a < 0) { z = conj_i(z); a = -a; }
+  return gpowgs(z, a);
+}
+
+/* initializes powers of e(a/b) */
+GEN
+rootsof1powinit(long a, long b, long prec)
+{
+  long g = cgcd(a,b);
+  if (g != 1) { a /= g; b /= g; }
+  if (b < 0) { b = -b; a = -a; }
+  a %= b; if (a < 0) a += b;
+  return mkvec2(grootsof1(b,prec), mkvecsmall2(a,b));
+}
+/* T = rootsof1powinit(a,b); return  e(a/b)^c */
+GEN
+rootsof1pow(GEN T, long c)
+{
+  GEN vz = gel(T,1), ab = gel(T,2);
+  long a = ab[1], b = ab[2]; /* a >= 0, b > 0 */
+  c %= b; if (c < 0) c += b;
+  a = Fl_mul(a, c, b);
+  return gel(vz, a + 1);
+}
+
+/* exp(2iPi/d), assume d a t_INT */
+GEN
+rootsof1_cx(GEN d, long prec)
+{
+  if (lgefint(d) == 3) return rootsof1u_cx((ulong)d[2], prec);
+  return expIr(divri(Pi2n(1,prec), d));
+}
+
 GEN
 gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
 {
@@ -1817,7 +2024,7 @@ gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
         y = gexp(gdiv(glog(x,prec), n), prec);
       y = gerepileupto(av, y);
     }
-    if (zetan) *zetan = rootsof1_cx(n,prec);
+    if (zetan) *zetan = rootsof1_cx(n, prec);
     return y;
 
   case t_QUAD:
@@ -1828,7 +2035,7 @@ gsqrtn(GEN x, GEN n, GEN *zetan, long prec)
     return gerepileupto(av, ser_powfrac(y, ginv(n), prec));
   }
   pari_err_TYPE("sqrtn",x);
-  return NULL;/* not reached */
+  return NULL;/* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -1871,7 +2078,7 @@ exp1r_abs(GEN x)
   * error bounded by 1/6(n+1) <= 1/12. Finally, we want
   * n (-1/log(2) -log_2 |Y| + log_2(n+1)) >= b  */
   b += m;
-  d = m-dbllog2(x)-1/LOG2; /* ~ -log_2 Y - 1/log(2) */
+  d = m-dbllog2(x)-1/M_LN2; /* ~ -log_2 Y - 1/log(2) */
   n = (long)(b / d);
   if (n > 1)
     n = (long)(b / (d + log2((double)n+1))); /* log~constant in small ranges */
@@ -1968,9 +2175,9 @@ gexpm1(GEN x, long prec)
 static GEN
 modlog2(GEN x, long *sh)
 {
-  double d = rtodbl(x);
-  long q = (long) ((fabs(d) + (LOG2/2))/LOG2);
-  if (d > LOG2 * LONG_MAX)
+  double d = rtodbl(x), da = fabs(d);
+  long q = (long) ((da + (M_LN2/2))/M_LN2);
+  if (da > M_LN2 * LONG_MAX)
     pari_err_OVERFLOW("expo()"); /* avoid overflow in  q */
   if (d < 0) q = -q;
   *sh = q;
@@ -2060,8 +2267,8 @@ Qp_exp_prec(GEN x)
   }
   else
   { /* e > 0, n > 0 */
-    GEN r, t = subis(p, 1);
-    k = itos(dvmdii(subis(muliu(t,n), 1), subis(muliu(t,e), 1), &r));
+    GEN r, t = subiu(p, 1);
+    k = itos(dvmdii(subiu(muliu(t,n), 1), subiu(muliu(t,e), 1), &r));
     if (!signe(r)) k--;
   }
   return k;
@@ -2159,6 +2366,30 @@ serchop0(GEN s)
   return normalize(y);
 }
 
+GEN
+serchop_i(GEN s, long n)
+{
+  long i, m, l = lg(s);
+  GEN y;
+  if (l == 2 || (l == 3 && isexactzero(gel(s,2))))
+  {
+    if (valp(s) < n) { s = shallowcopy(s); setvalp(s,n); }
+    return s;
+  }
+  m = n - valp(s); if (m < 0) return s;
+  if (l-m <= 2) return zeroser(varn(s), n);
+  y = cgetg(l-m, t_SER); y[1] = s[1]; setvalp(y, valp(y)+m);
+  for (i=m+2; i < l; i++) gel(y,i-m) = gel(s,i);
+  return normalize(y);
+}
+GEN
+serchop(GEN s, long n)
+{
+  pari_sp av = avma;
+  if (typ(s) != t_SER) pari_err_TYPE("serchop",s);
+  return gerepilecopy(av, serchop_i(s,n));
+}
+
 static GEN
 serexp(GEN x, long prec)
 {
@@ -2340,7 +2571,7 @@ ser_cmp_expo(GEN A, GEN B)
     GEN a = gel(A,i), b;
     long ei;
     if (isexactzero(a)) continue;
-    b = polcoeff_i(B, i-2 + d, v);
+    b = polcoef_i(B, i-2 + d, v);
     ei = gexpo(a);
     if (!isexactzero(b)) ei -= gexpo(b);
     e = maxss(e, ei);
@@ -2463,7 +2694,8 @@ atanhQ_split(ulong u, ulong v, long prec)
   abpq_sum(&R, 0, nmax, &A);
   return rdivii(R.T, mulii(R.B,R.Q),prec);
 }
-/* log(2) = 10*atanh(1/17)+4*atanh(13/499) */
+/* log(2) = 10*atanh(1/17)+4*atanh(13/499); faster than logagmr_abs()
+ * and Pi/2M(1,4/2^n) ~ n log(2) */
 static GEN
 log2_split(long prec)
 {
@@ -2472,16 +2704,6 @@ log2_split(long prec)
   shiftr_inplace(v, 2);
   return addrr(mulur(10, u), v);
 }
-#if 0 /* slower ! */
-/* cf logagmr_abs(). Compute Pi/2agm(1, 4/2^n) ~ log(2^n) = n log(2) */
-static GEN
-log2_agm(long prec)
-{
-  long n = prec2nbits(prec) >> 1;
-  GEN y = divrr(Pi2n(-1, prec), agm1r_abs( real2n(2 - n, prec) ));
-  return divru(y, n);
-}
-#endif
 GEN
 constlog2(long prec)
 {
@@ -2499,15 +2721,15 @@ constlog2(long prec)
 GEN
 mplog2(long prec) { return rtor(constlog2(prec), prec); }
 
+/* dont check that q != 2^expo(q), done in logr_abs */
 static GEN
 logagmr_abs(GEN q)
 {
-  long prec = realprec(q), lim, e = expo(q);
-  GEN z, y, Q, _4ovQ;
-  pari_sp av;
+  long prec = realprec(q), e = expo(q), lim;
+  GEN z = cgetr(prec), y, Q, _4ovQ;
+  pari_sp av = avma;
 
-  if (absrnz_equal2n(q)) return e? mulsr(e, mplog2(prec)): real_0(prec);
-  z = cgetr(prec); av = avma; incrprec(prec);
+  incrprec(prec);
   lim = prec2nbits(prec) >> 1;
   Q = rtor(q,prec);
   shiftr_inplace(Q,lim-e); setsigne(Q,1);
@@ -2518,18 +2740,51 @@ logagmr_abs(GEN q)
   y = addrr(y, mulsr(e - lim, mplog2(prec)));
   affrr_fixlg(y, z); avma = av; return z;
 }
+
+/* sum_{k >= 0} y^(2k+1) / (2k+1), y close to 0 */
+static GEN
+logr_aux(GEN y)
+{
+  long k, L = realprec(y); /* should be ~ l+1 - (k-2) */
+  /* log(x) = log(1+y) - log(1-y) = 2 sum_{k odd} y^k / k
+   * Truncate the sum at k = 2n+1, the remainder is
+   *   2 sum_{k >= 2n+3} y^k / k < 2y^(2n+3) / (2n+3)(1-y) < y^(2n+3)
+   * We want y^(2n+3) < y 2^(-prec2nbits(L)), hence
+   *   n+1 > -prec2nbits(L) /-log_2(y^2) */
+  double d = -2*dbllog2r(y); /* ~ -log_2(y^2) */
+  k = (long)(2*(prec2nbits(L) / d));
+  k |= 1;
+  if (k >= 3)
+  {
+    GEN T, S = cgetr(L), y2 = sqrr(y), unr = real_1(L);
+    pari_sp av = avma;
+    long s = 0, incs = (long)d, l1 = nbits2prec((long)d);
+    setprec(S,  l1);
+    setprec(unr,l1); affrr(divru(unr,k), S);
+    for (k -= 2;; k -= 2) /* k = 2n+1, ..., 1 */
+    { /* S = y^(2n+1-k)/(2n+1) + ... + 1 / k */
+      setprec(y2, l1); T = mulrr(S,y2);
+      if (k == 1) break;
+
+      l1 += dvmdsBIL(s + incs, &s); if (l1>L) l1=L;
+      setprec(S, l1);
+      setprec(unr,l1);
+      affrr(addrr(divru(unr, k), T), S); avma = av;
+    }
+    /* k = 1 special-cased for eficiency */
+    y = mulrr(y, addsr(1,T)); /* = log(X)/2 */
+  }
+  return y;
+}
 /*return log(|x|), assuming x != 0 */
 GEN
 logr_abs(GEN X)
 {
-  pari_sp ltop;
   long EX, L, m, k, a, b, l = realprec(X);
   GEN z, x, y;
   ulong u;
   double d;
 
-  if (l > LOGAGM_LIMIT) return logagmr_abs(X);
-
  /* Assuming 1 < x < 2, we want delta = x-1, 1-x/2, 1-1/x, or 2/x-1 small.
   * We have 2/x-1 > 1-x/2, 1-1/x < x-1. So one should be choosing between
   * 1-1/x and 1-x/2 ( crossover sqrt(2), worse ~ 0.29 ). To avoid an inverse,
@@ -2545,9 +2800,14 @@ logr_abs(GEN X)
     while (!u && ++k < l) u = uel(X,k);
   }
   if (k == l) return EX? mulsr(EX, mplog2(l)): real_0(l);
-  z = cgetr(EX? l: l - (k-2)); ltop = avma;
-
   a = prec2nbits(k) + bfffo(u); /* ~ -log2 |1-x| */
+  L = l+1;
+  b = prec2nbits(L - (k-2)); /* take loss of accuracy into account */
+  if (b > 24*a*log2(L) &&
+      prec2nbits(l) > prec2nbits(LOGAGM_LIMIT)) return logagmr_abs(X);
+
+  z = cgetr(EX? l: l - (k-2));
+
  /* Multiplication is quadratic in this range (l is small, otherwise we
   * use AGM). Set Y = x^(1/2^m), y = (Y - 1) / (Y + 1) and compute truncated
   * series sum y^(2k+1)/(2k+1): the costs is less than
@@ -2559,11 +2819,9 @@ logr_abs(GEN X)
   * or b/6e + BITS_IN_LONG/2e + BITS_IN_LONG/2be ~ m
   *    B := (b / 6 + BITS_IN_LONG/2 + BITS_IN_LONG^2 / 2b) ~ m(m+a)
   *     m = min( -a/2 + sqrt(a^2/4 + B),  b - a )
-  * NB: e ~ (b/6)^(1/2) as b -> oo */
-  L = l+1;
-  b = prec2nbits(L - (k-2)); /* take loss of accuracy into account */
-  /* instead of the above pessimistic estimate for the cost of the sum, use
-   * optimistic estimate (BITS_IN_LONG -> 0) */
+  * NB: e ~ (b/6)^(1/2) as b -> oo
+  * Instead of the above pessimistic estimate for the cost of the sum, use
+  * optimistic estimate (BITS_IN_LONG -> 0) */
   d = -a/2.; m = (long)(d + sqrt(d*d + b/6)); /* >= 0 */
 
   if (m > b-a) m = b-a;
@@ -2574,39 +2832,10 @@ logr_abs(GEN X)
   for (k=1; k<=m; k++) x = sqrtr_abs(x);
 
   y = divrr(subrs(x,1), addrs(x,1)); /* = (x-1) / (x+1), close to 0 */
-  L = realprec(y); /* should be ~ l+1 - (k-2) */
-  /* log(x) = log(1+y) - log(1-y) = 2 sum_{k odd} y^k / k
-   * Truncate the sum at k = 2n+1, the remainder is
-   *   2 sum_{k >= 2n+3} y^k / k < 2y^(2n+3) / (2n+3)(1-y) < y^(2n+3)
-   * We want y^(2n+3) < y 2^(-prec2nbits(L)), hence
-   *   n+1 > -prec2nbits(L) /-log_2(y^2) */
-  d = -2*dbllog2r(y); /* ~ -log_2(y^2) */
-  k = (long)(2*(prec2nbits(L) / d));
-  k |= 1;
-  if (k >= 3)
-  {
-    GEN S, T, y2 = sqrr(y), unr = real_1(L);
-    pari_sp av = avma;
-    long s = 0, incs = (long)d, l1 = nbits2prec((long)d);
-    S = x;
-    setprec(S,  l1);
-    setprec(unr,l1); affrr(divru(unr,k), S); /* destroy x, not needed anymore */
-    for (k -= 2;; k -= 2) /* k = 2n+1, ..., 1 */
-    { /* S = y^(2n+1-k)/(2n+1) + ... + 1 / k */
-      setprec(y2, l1); T = mulrr(S,y2);
-      if (k == 1) break;
-
-      l1 += dvmdsBIL(s + incs, &s); if (l1>L) l1=L;
-      setprec(S, l1);
-      setprec(unr,l1);
-      affrr(addrr(divru(unr, k), T), S); avma = av;
-    }
-    /* k = 1 special-cased for eficiency */
-    y = mulrr(y, addsr(1,T)); /* = log(X)/2 */
-  }
+  y = logr_aux(y); /* log(1+y) - log(1-y) = log(x) */
   shiftr_inplace(y, m + 1);
   if (EX) y = addrr(y, mulsr(EX, mplog2(l+1)));
-  affrr_fixlg(y, z); avma = ltop; return z;
+  affrr_fixlg(y, z); avma = (pari_sp)z; return z;
 }
 
 /* assume Im(q) != 0 and precision(q) >= prec. Compute log(q) with accuracy
@@ -2693,9 +2922,9 @@ Zp_teichmuller(GEN x, GEN p, long e, GEN pe)
     }
     else
     {
-      w = diviiexact(addsi(-1,qold),p1); /* -1/(p-1) + O(qold) */
-      t = Fp_mul(w, subis(Fp_pow(z,p1,q), 1), q);
-      z = Fp_mul(z, addsi(1,t), q);
+      w = diviiexact(subiu(qold,1),p1); /* -1/(p-1) + O(qold) */
+      t = Fp_mul(w, subiu(Fp_pow(z,p1,q), 1), q);
+      z = Fp_mul(z, addui(1,t), q);
     }
   }
   return gerepileuptoint(av, z);
@@ -2832,6 +3061,78 @@ glog(GEN x, long prec)
   }
   return trans_eval("log",glog,x,prec);
 }
+
+static GEN
+mplog1p(GEN x)
+{
+  long ex, a, b, l, L;
+  if (!signe(x)) return rcopy(x);
+  ex = expo(x); if (ex >= 0) return glog(addrs(x,1), 0);
+  a = -ex;
+  b = realprec(x); L = b+1;
+  if (b > a*log2(L) && prec2nbits(b) > prec2nbits(LOGAGM_LIMIT))
+  {
+    x = addrs(x,1); l = b + nbits2extraprec(a);
+    if (realprec(x) < l) x = rtor(x,l);
+    return logagmr_abs(x);
+  }
+  x = rtor(x, L);
+  x = logr_aux(divrr(x, addrs(x,2)));
+  if (realprec(x) > b) fixlg(x, b);
+  shiftr_inplace(x,1); return x;
+}
+
+static GEN log1p_i(GEN x, long prec);
+static GEN
+cxlog1p(GEN x, long prec)
+{
+  pari_sp av;
+  GEN z, a, b = gel(x,2);
+  long l;
+  if (ismpzero(b)) return log1p_i(gel(x,1), prec);
+  l = precision(x); if (l > prec) prec = l;
+  if (prec >= LOGAGMCX_LIMIT) return logagmcx(gaddgs(x,1), prec);
+  a = gel(x,1);
+  z = cgetg(3,t_COMPLEX); av = avma;
+  a = gadd(gadd(gmul2n(a,1), gsqr(a)), gsqr(b));
+  a = log1p_i(a, prec); shiftr_inplace(a,-1);
+  gel(z,1) = gerepileupto(av, a);
+  gel(z,2) = garg(gaddgs(x,1),prec); return z;
+}
+static GEN
+log1p_i(GEN x, long prec)
+{
+  switch(typ(x))
+  {
+    case t_REAL: return mplog1p(x);
+    case t_COMPLEX: return cxlog1p(x, prec);
+    case t_PADIC: return Qp_log(gaddgs(x,1));
+    default:
+    {
+      long ey;
+      GEN y;
+      if (!(y = toser_i(x))) break;
+      ey = valp(y);
+      if (ey < 0) pari_err_DOMAIN("log1p","valuation", "<", gen_0, x);
+      if (gequal0(y)) return gcopy(y);
+      if (ey)
+        return glog(gaddgs(y,1),prec);
+      else
+      {
+        GEN a = gel(y,2), a1 = gaddgs(a,1);
+        y = gdiv(y, a1); gel(y,2) = gen_1;
+        return gadd(glog1p(a,prec), glog(y, prec));
+      }
+    }
+  }
+  return trans_eval("log1p",glog1p,x,prec);
+}
+GEN
+glog1p(GEN x, long prec)
+{
+  pari_sp av = avma;
+  return gerepileupto(av, log1p_i(x, prec));
+}
 /********************************************************************/
 /**                                                                **/
 /**                        SINE, COSINE                            **/
@@ -2888,7 +3189,7 @@ mpcosm1(GEN x, long *ptmod8)
   L = l + nbits2extraprec(m);
 
   b += m;
-  d = 2.0 * (m-dbllog2r(x)-1/LOG2); /* ~ 2( - log_2 Y - 1/log(2) ) */
+  d = 2.0 * (m-dbllog2r(x)-1/M_LN2); /* ~ 2( - log_2 Y - 1/log(2) ) */
   n = (long)(b / d);
   if (n > 1)
     n = (long)(b / (d + log2((double)n+1))); /* log~constant in small ranges */
@@ -3352,15 +3653,33 @@ gsinc(GEN x, long prec)
       if (gequal0(x)) return cvtop(gen_1, gel(x,2), valp(x));
       av = avma; y = sin_p(x);
       if (!y) pari_err_DOMAIN("gsinc(t_PADIC)","argument","",gen_0,x);
-      return gerepileuptoleaf(av,gdiv(y,x));
+      return gerepileupto(av,gdiv(y,x));
 
     default:
+    {
+      long ex;
       av = avma; if (!(y = toser_i(x))) break;
       if (gequal0(y)) return gerepileupto(av, gaddsg(1,y));
-      if (valp(y) < 0)
-        pari_err_DOMAIN("sinc","valuation", "<", gen_0, x);
-      gsincos(y,&u,&v,prec);
-      return gerepilecopy(av,gdiv(u,y));
+      ex = valp(y);
+      if (ex < 0) pari_err_DOMAIN("sinc","valuation", "<", gen_0, x);
+      if (ex)
+      {
+        gsincos(y,&u,&v,prec);
+        y = gerepileupto(av, gdiv(u,y));
+        if (lg(y) > 2) gel(y,2) = gen_1;
+        return y;
+      }
+      else
+      {
+        GEN z0, y0 = gel(y,2), y1 = serchop0(y), y10 = y1;
+        if (!gequal1(y0)) y10 = gdiv(y10, y0);
+        gsincos(y1,&u,&v,prec);
+        z0 = gdiv(gcos(y0,prec), y0);
+        y = gaddsg(1, y10);
+        u = gadd(gmul(gsinc(y0, prec),v), gmul(z0, u));
+        return gerepileupto(av,gdiv(u,y));
+      }
+    }
   }
   return trans_eval("sinc",gsinc,x,prec);
 }
