@@ -23,10 +23,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 GEN
 trans_fix_arg(long *prec, GEN *s0, GEN *sig, GEN *tau, pari_sp *av, GEN *res)
 {
-  GEN s, p1;
+  GEN p1, s = *s0 = cxtoreal(*s0);
   long l;
-  if (typ(*s0)==t_COMPLEX && gequal0(gel(*s0,2))) *s0 = gel(*s0,1);
-  s = *s0;
   l = precision(s); if (!l) l = *prec;
   if (l < LOWDEFAULTPREC) l = LOWDEFAULTPREC;
   *res = cgetc(l); *av = avma;
@@ -51,6 +49,10 @@ trans_fix_arg(long *prec, GEN *s0, GEN *sig, GEN *tau, pari_sp *av, GEN *res)
 /**                          ARCTANGENT                            **/
 /**                                                                **/
 /********************************************************************/
+/* atan(b/a), real a and b, suitable for gerepileupto */
+static GEN
+atan2_agm(GEN a, GEN b, long prec)
+{ return gel(logagmcx(mkcomplex(a, b), prec), 2); }
 static GEN
 mpatan(GEN x)
 {
@@ -67,10 +69,7 @@ mpatan(GEN x)
     return y;
   }
   if (l > AGM_ATAN_LIMIT)
-  {
-    av = avma; y = logagmcx(mkcomplex(gen_1, x), l);
-    return gerepileuptoleaf(av, gel(y,2));
-  }
+  { av = avma; return gerepileuptoleaf(av, atan2_agm(gen_1, x, l)); }
 
   e = expo(x); inv = (e >= 0); /* = (|x| > 1 ) */
   if (e > 0) lp += nbits2extraprec(e);
@@ -164,10 +163,7 @@ mpasin(GEN x) {
   pari_sp av = avma;
   GEN z, a = sqrtr(subsr(1, sqrr(x)));
   if (realprec(x) > AGM_ATAN_LIMIT)
-  {
-    z = logagmcx(mkcomplex(a,x), realprec(x));
-    z = gel(z,2);
-  }
+    z = atan2_agm(a, x, realprec(x));
   else
     z = mpatan(divrr(x, a));
   return gerepileuptoleaf(av, z);
@@ -234,11 +230,9 @@ mpacos(GEN x)
   pari_sp av = avma;
   GEN z, a = sqrtr(subsr(1, sqrr(x)));
   if (realprec(x) > AGM_ATAN_LIMIT)
+    z = atan2_agm(x, a, realprec(x));
+  else
   {
-    z = logagmcx(mkcomplex(x,a), realprec(x));
-    z = gel(z,2);
-  }
-  else {
     z = mpatan(divrr(a, x));
     if (signe(x) < 0) z = addrr(mppi(realprec(z)), z);
   }
@@ -838,7 +832,7 @@ mpbern(long nb, long prec)
     /* Not cached, must compute */
     /* huge accuracy ? May as well compute exactly */
     if (n_is_small && (prec == LONG_MAX ||
-                       2*n * log((double)2*n) < prec2nbits_mul(prec, LOG2)))
+                       2*n * log((double)2*n) < prec2nbits_mul(prec, M_LN2)))
       S = bernfrac_using_zeta(2*n);
     else
     {
@@ -947,7 +941,7 @@ bernpol_i(long k, long v)
   if (v < 0) v = 0;
   if (k < 0) pari_err_DOMAIN("bernpol", "index", "<", gen_0, stoi(k));
   mpbern(k >> 1, 0); /* cache B_2, ..., B_2[k/2] */
-  C = vecbinome(k);
+  C = vecbinomial(k);
   B = cgetg(k + 3, t_POL);
   for (i = 0; i <= k; ++i) gel(B, k-i+2) = gmul(gel(C,i+1), bernfrac(i));
   B[1] = evalsigne(1) | evalvarn(v);
@@ -1113,7 +1107,7 @@ cxgamma(GEN s0, int dolog, long prec)
     S = gprec_w(s,LOWDEFAULTPREC);
     /* l2 ~ |lngamma(s))|^2 */
     l2 = gnorm(gmul(S, glog(S, LOWDEFAULTPREC)));
-    l = (prec2nbits_mul(prec, LOG2) - rtodbl(glog(l2,LOWDEFAULTPREC))/2) / 2.;
+    l = (prec2nbits_mul(prec, M_LN2) - rtodbl(glog(l2,LOWDEFAULTPREC))/2) / 2.;
     if (l < 0) l = 0.;
 
     iS = imag_i(S);
@@ -1175,7 +1169,7 @@ cxgamma(GEN s0, int dolog, long prec)
     v = v - st;
     l2 = u*u + v*v;
     if (l2 < 0.000001) l2 = 0.000001;
-    l = (prec2nbits_mul(prec, LOG2) - log(l2)/2) / 2.;
+    l = (prec2nbits_mul(prec, M_LN2) - log(l2)/2) / 2.;
     if (l < 0) l = 0.;
 
     la = 3.; /* FIXME: heuristic... */
@@ -1213,7 +1207,7 @@ cxgamma(GEN s0, int dolog, long prec)
     if (is_bigint(s0)) {
       for (i=1; i < nn; i++)
       {
-        y = mulri(y, addis(s0, i));
+        y = mulri(y, addiu(s0, i));
         if (gc_needed(av2,3))
         {
           if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
@@ -1234,29 +1228,32 @@ cxgamma(GEN s0, int dolog, long prec)
     }
     if (dolog) y = logr_abs(y);
   }
-  else if (!dolog || typ(s) == t_REAL)
+  else
   { /* Compute lngamma mod 2 I Pi */
-    for (i=1; i < nn; i++)
+    GEN sq = gsqr(s);
+    pari_sp av3 = avma;
+    for (i = 1; i < nn - 1; i += 2)
     {
-      y = gmul(y, gaddgs(s,i));
+      y = gmul(y, gaddsg(i*(i + 1), gadd(gmulsg(2*i + 1, s), sq)));
       if (gc_needed(av2,3))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
-        y = gerepileupto(av2, y);
+        y = gerepileupto(av3, y);
       }
     }
-    if (dolog) y = logr_abs(y);
-  }
-  else
-  { /* dolog && complex s: be careful with imaginary part */
-    y = glog(y, prec);
-    for (i=1; i < nn; i++)
+    if (!odd(nn)) y = gmul(y, gaddsg(nn - 1, s));
+    if (dolog)
     {
-      y = gadd(y, glog(gaddgs(s,i), prec));
-      if (gc_needed(av2,3))
-      {
-        if(DEBUGMEM>1) pari_warn(warnmem,"gamma");
-        y = gerepileupto(av2, y);
+      if (typ(s) == t_REAL) y = logr_abs(y);
+      else
+      { /* fix imaginary part */
+        long prec0 = LOWDEFAULTPREC;
+        GEN s0 = gprec_w(s, prec0), y0 = s0, k;
+        y0 = garg(y0, prec0); /* Im log(s) at low accuracy */
+        for (i=1; i < nn; i++) y0 = gadd(y0, garg(gaddgs(s0,i), prec0));
+        y = glog(y, prec);
+        k = ground( gdiv(gsub(y0, imag_i(y)), Pi2n(1,prec0)) );
+        if (signe(k)) y = gadd(y, mulcxI(mulir(k, Pi2n(1, prec))));
       }
     }
   }
@@ -1582,7 +1579,7 @@ ggamma(GEN x, long prec)
         if (is_bigint(a) || labs(m = itos(a)) > 962354)
         {
           pari_err_OVERFLOW("gamma");
-          return NULL; /* not reached */
+          return NULL; /* LCOV_EXCL_LINE */
         }
         return gammahs(m-1, prec);
       }
@@ -1699,7 +1696,7 @@ static GEN
 cxpsi(GEN s0, long prec)
 {
   pari_sp av, av2;
-  GEN sum,z,a,res,tes,in2,sig,tau,s,unr;
+  GEN sum,z,a,res,tes,in2,sig,tau,s,unr,s2,sq;
   long lim,nn,k;
   const long la = 3;
   int funeq = 0;
@@ -1719,7 +1716,7 @@ cxpsi(GEN s0, long prec)
 
     l = rtodbl( gnorm(glog(S, 3)) );
     l = log(l) / 2.;
-    lim = 2 + (long)ceil((prec2nbits_mul(prec, LOG2) - l) / (2*(1+log((double)la))));
+    lim = 2 + (long)ceil((prec2nbits_mul(prec, M_LN2) - l) / (2*(1+log((double)la))));
     if (lim < 2) lim = 2;
 
     l = (2*lim-1)*la / (2.*M_PI);
@@ -1742,7 +1739,7 @@ cxpsi(GEN s0, long prec)
     }
     if (l < 0.000001) l = 0.000001;
     l = log(l) / 2.;
-    lim = 2 + (long)ceil((prec2nbits_mul(prec, LOG2) - l) / (2*(1+log((double)la))));
+    lim = 2 + (long)ceil((prec2nbits_mul(prec, M_LN2) - l) / (2*(1+log((double)la))));
     if (lim < 2) lim = 2;
 
     l = (2*lim-1)*la / (2.*M_PI);
@@ -1753,16 +1750,18 @@ cxpsi(GEN s0, long prec)
     if (DEBUGLEVEL>2) err_printf("lim, nn: [%ld, %ld]\n",lim,nn);
   }
   incrprec(prec); unr = real_1(prec); /* one extra word of precision */
-
+  s2 = gmul2n(s, 1); sq = gsqr(s);
   a = gdiv(unr, gaddgs(s, nn)); /* 1 / (s+n) */
-  av2 = avma; sum = gmul2n(a,-1);
-  for (k = 0; k < nn; k++)
+  av2 = avma; sum = gmul2n(a, -1);
+  for (k = 0; k < nn - 1; k += 2)
   {
-    sum = gadd(sum, gdiv(unr, gaddgs(s, k)));
-    if ((k & 127) == 0) sum = gerepileupto(av2, sum);
+    GEN tmp = gaddsg(k*(k + 1), gadd(gmulsg(2*k + 1, s), sq));
+    sum = gadd(sum, gdiv(gaddsg(2*k + 1, s2), tmp));
+    if ((k & 1023) == 0) sum = gerepileupto(av2, sum);
   }
+  if (odd(nn)) sum = gadd(sum, gdiv(unr, gaddsg(nn - 1, s)));
   z = gsub(glog(gaddgs(s, nn), prec), sum);
-  if (DEBUGLEVEL>2) timer_printf(&T,"sum from 0 to N-1");
+  if (DEBUGLEVEL>2) timer_printf(&T,"sum from 0 to N - 1");
 
   in2 = gsqr(a);
   mpbern(lim,prec);
