@@ -18,7 +18,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*******************************************************************/
 
 #include "pari.h"
-#include "paripriv.h"
 #include "rect.h"
 
 #ifdef HPPA
@@ -37,7 +36,6 @@ BEGINEXTERN
 ENDEXTERN
 
 static Colormap PARI_Colormap;
-static XColor  *PARI_Colors;
 
 struct data_x
 {
@@ -56,10 +54,35 @@ exiterr(const char *str)
   term_color(c_NONE); exit(1);
 }
 
+static long
+rgb_to_pixel(Display *display, int r, int g, int b)
+{
+  XColor X;
+  X.red   = r*65535/255;
+  X.green = g*65535/255;
+  X.blue  = b*65535/255;
+  X.flags = DoRed | DoGreen | DoBlue;
+  if (!XAllocColor(display,PARI_Colormap,&X)) exiterr("cannot allocate color");
+  return X.pixel;
+}
+static long
+colormapindex_to_pixel(Display *display, long i)
+{
+  GEN c = gel(GP_DATA->colormap, i+1);
+  int r,g,b; color_to_rgb(c, &r,&g,&b);
+  return rgb_to_pixel(display, r, g, b);
+}
+static long
+rgb_color(Display *display, long c)
+{
+  int r,g,b; long_to_rgb(c, &r, &g, &b);
+  return rgb_to_pixel(display, r, g, b);
+}
+
 static void SetForeground(void *data, long col)
 {
   struct data_x *dx = (struct data_x *) data;
-  XSetForeground(dx->display,dx->gc, PARI_Colors[col].pixel);
+  XSetForeground(dx->display,dx->gc, rgb_color(dx->display,col));
 }
 
 static void DrawPoint(void *data, long x, long y)
@@ -80,6 +103,12 @@ static void DrawRectangle(void *data, long x, long y, long w, long h)
   XDrawRectangle(dx->display,dx->win,dx->gc, x,y, w,h);
 }
 
+static void FillRectangle(void *data, long x, long y, long w, long h)
+{
+  struct data_x *dx = (struct data_x *) data;
+  XFillRectangle(dx->display,dx->win,dx->gc, x,y, w,h);
+}
+
 static void DrawPoints(void *data, long nb, struct plot_points *p)
 {
   struct data_x *dx = (struct data_x *) data;
@@ -131,31 +160,7 @@ IOerror(Display *d) {
 }
 
 static void
-PARI_ColorSetUp(Display *display, GEN colors)
-{
-  static int init_done = 0;
-  long i, n = lg(colors)-1;
-
-  if (init_done) return;
-  init_done=1;
-
-  PARI_Colormap = DefaultColormap(display, 0);
-  PARI_Colors = (XColor *) pari_malloc((n+1) * sizeof(XColor));
-  for (i=0; i<n; i++)
-  {
-    int r, g, b;
-    color_to_rgb(gel(colors,i+1), &r, &g, &b);
-    PARI_Colors[i].red   = r*65535/255;
-    PARI_Colors[i].green = g*65535/255;
-    PARI_Colors[i].blue  = b*65535/255;
-    PARI_Colors[i].flags = DoRed | DoGreen | DoBlue;
-    if (!XAllocColor(display, PARI_Colormap, &PARI_Colors[i]))
-      exiterr("cannot allocate color");
-  }
-}
-
-void
-rectdraw0(long *w, long *x, long *y, long lw)
+draw(PARI_plot *T, GEN w, GEN x, GEN y)
 {
   long oldwidth,oldheight;
   struct plot_eng plotX;
@@ -172,30 +177,30 @@ rectdraw0(long *w, long *x, long *y, long lw)
   Atom wm_delete_window, wm_protocols;
 
   if (pari_daemon()) return;  /* parent process returns */
-
-  PARI_get_plot();
   pari_close();
 
   display = XOpenDisplay(NULL);
-  font_info = XLoadQueryFont(display, "9x15");
-  if (!font_info) exiterr("cannot open 9x15 font");
+  if (!display) exiterr("cannot open Display");
+  font_info = XLoadQueryFont(display, "7x13");
+  if (!font_info) exiterr("cannot open 7x13 font");
   XSetErrorHandler(Xerror);
   XSetIOErrorHandler(IOerror);
-  PARI_ColorSetUp(display,GP_DATA->colormap);
+  PARI_Colormap = DefaultColormap(display, 0);
 
   screen = DefaultScreen(display);
   win = XCreateSimpleWindow
     (display, RootWindow(display, screen), 0, 0,
-     pari_plot.width, pari_plot.height, 4,
-     PARI_Colors[1].pixel, PARI_Colors[0].pixel);
+     T->width, T->height, 4,
+     colormapindex_to_pixel(display, 1),
+     colormapindex_to_pixel(display, 0));
 
   size_hints.flags = PPosition | PSize;
   size_hints.x = 0;
   size_hints.y = 0;
-  size_hints.width  = pari_plot.width;
-  size_hints.height = pari_plot.height;
+  size_hints.width  = T->width;
+  size_hints.height = T->height;
   XSetStandardProperties
-    (display, win, "rectplot", NULL, None, NULL, 0, &size_hints);
+    (display, win, "PARI plot", NULL, None, NULL, 0, &size_hints);
 
   wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
   wm_protocols = XInternAtom(display, "WM_PROTOCOLS", False);
@@ -214,8 +219,8 @@ rectdraw0(long *w, long *x, long *y, long lw)
 
   XClearWindow(display, win);
   XMapWindow(display, win);
-  oldwidth  = pari_plot.width;
-  oldheight = pari_plot.height;
+  oldwidth  = T->width;
+  oldheight = T->height;
   dx.display= display;
   dx.win = win;
   dx.numcolors = lg(GP_DATA->colormap)-1;
@@ -224,10 +229,11 @@ rectdraw0(long *w, long *x, long *y, long lw)
   plotX.pt = &DrawPoint;
   plotX.ln = &DrawLine;
   plotX.bx = &DrawRectangle;
+  plotX.fb = &FillRectangle;
   plotX.mp = &DrawPoints;
   plotX.ml = &DrawLines;
   plotX.st = &DrawString;
-  plotX.pl = &pari_plot;
+  plotX.pl = T;
   plotX.data = (void*)&dx;
 
   for(;;)
@@ -269,30 +275,41 @@ EXIT:
         oldheight = height;
 
         /* recompute scale */
-        xs = ((double)width)/pari_plot.width;
-        ys = ((double)height)/pari_plot.height;
+        xs = ((double)width)/T->width;
+        ys = ((double)height)/T->height;
       }
       case Expose:
-        gen_rectdraw0(&plotX, w, x, y,lw,xs,ys);
+        gen_draw(&plotX, w, x, y, xs, ys);
     }
   }
 }
 
-void
-PARI_get_plot(void)
+INLINE void
+gp_get_display_sizes(long *dwidth, long *dheight, long *fwidth, long *fheight)
 {
   Display *display;
-  int screen;
 
-  if (pari_plot.init) return;
-  if (!(display = XOpenDisplay(NULL))) pari_err(e_MISC, "no X server");
-  screen = DefaultScreen(display);
-  pari_plot.width  = DisplayWidth(display, screen) - 40;
-  pari_plot.height = DisplayHeight(display, screen) - 60;
-  pari_plot.fheight = 15;
-  pari_plot.fwidth  = 9;
-  pari_plot.hunit   = 5;
-  pari_plot.vunit   = 5;
-  pari_plot.init = 1;
-  XCloseDisplay(display);
+  display = XOpenDisplay(NULL);
+  if (display)
+  {
+    int screen = DefaultScreen(display);
+    *dwidth  = DisplayWidth(display, screen);
+    *dheight = DisplayHeight(display, screen);
+    XCloseDisplay(display);
+  }
+  else
+  {
+    /* Situation looks grim */
+    *dwidth  = 0;
+    *dheight = 0;
+  }
+  *fwidth  = 7;
+  *fheight = 13;
+}
+
+void
+gp_get_plot(PARI_plot *T)
+{
+  gp_get_plot_generic(T,gp_get_display_sizes);
+  T->draw = &draw;
 }
