@@ -129,6 +129,8 @@ pari_close_compiler(void)
   pari_stack_delete(&s_operand);
   pari_stack_delete(&s_data);
   pari_stack_delete(&s_lvar);
+  pari_stack_delete(&s_dbginfo);
+  pari_stack_delete(&s_frame);
 }
 
 struct codepos
@@ -1896,7 +1898,7 @@ call0(GEN fun, GEN args)
       return closure_callgenvec(fun, args);
     default:
       pari_err_TYPE("call", fun);
-      return NULL; /* NOT REACHED */
+      return NULL; /* LCOV_EXCL_LINE */
   }
 }
 
