@@ -13,7 +13,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
-#include "anal.h"
 
 /**********************************************************************/
 /***                                                                ***/
@@ -216,8 +215,8 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
 
      Thus the problem is minimizing (1 + slow2_size/arena)*overhead**0.29.
      This boils down to F=((X+A)/(X+B))X^alpha, X=overhead,
-     B = (1 - fixed_to_cache/cache_arena), A = B +
-     slow2_size/cache_arena, alpha = 0.38, and X>=0.018, X>-B.
+     B = (1 - fixed_to_cache/cache_arena), A = B + slow2_size/cache_arena,
+     alpha = 0.38, and X>=0.018, X>-B.
 
      We need to find the rightmost root of (X+A)*(X+B) - alpha(A-B)X to the
      right of 0.018 (if such exists and is below Xmax).  Then we manually
@@ -227,18 +226,13 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
      function, as a sanity check always prefer fitting into the
      cache (or "almost fitting") if F-law predicts that the larger
      value of the arena provides less than 10% speedup.
-
    */
 
   /* The simplest case: we fit into cache */
-  if (total + fixed_to_cache <= cache_arena)
-      return total;
+  asize = cache_arena - fixed_to_cache;
+  if (total <= asize) return total;
   /* The simple case: fitting into cache doesn't slow us down more than 10% */
-  if (cache_arena - fixed_to_cache > 10 * slow2_size) {
-      asize = cache_arena - fixed_to_cache;
-      if (asize > total) asize = total; /* Automatically false... */
-      return asize;
-  }
+  if (asize > 10 * slow2_size) return asize;
   /* Slowdown of not fitting into cache is significant.  Try to optimize.
      Do not be afraid to spend some time on optimization - in trivial
      cases we do not reach this point; any gain we get should
@@ -250,38 +244,36 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
   C1 = (A + B - 1/alpha*(A - B))/2;
   D = C1*C1 - C2;
   if (D > 0)
-      V = cut_off*cut_off + 2*C1*cut_off + C2; /* Value at CUT_OFF */
+    V = cut_off*cut_off + 2*C1*cut_off + C2; /* Value at CUT_OFF */
   else
-      V = 0;                            /* Peacify the warning */
+    V = 0; /* Peacify the warning */
   Xmin = cut_off;
   Xmax = ((double)total - fixed_to_cache)/cache_arena; /* Two candidates */
 
   if ( D <= 0 || (V >= 0 && C1 + cut_off >= 0) ) /* slowdown increasing */
-      Xmax = cut_off;                   /* Only one candidate */
-  else if (V >= 0 &&                    /* slowdown concave down */
+    Xmax = cut_off; /* Only one candidate */
+  else if (V >= 0 && /* slowdown concave down */
            ((Xmax + C1) <= 0 || (Xmax*Xmax + 2*C1*Xmax + C2) <= 0))
-      /* DO NOTHING */;                 /* Keep both candidates */
-  else if (V <= 0 && (Xmax*Xmax + 2*C1*Xmax + C2) <= 0) /* slowdown decreasing */
-      Xmin = cut_off;                   /* Only one candidate */
+      /* DO NOTHING */;  /* Keep both candidates */
+  else if (V <= 0 && (Xmax*Xmax + 2*C1*Xmax + C2) <= 0) /*slowdown decreasing*/
+      Xmin = cut_off; /* Only one candidate */
   else /* Now we know: 2 roots, the largest is in CUT_OFF..Xmax */
       Xmax = sqrt(D) - C1;
-  if (Xmax != Xmin) {   /* Xmin == CUT_OFF; Check which one is better */
-      double v1 = (cut_off + A)/(cut_off + B);
-      double v2 = 2.33 * (Xmax + A)/(Xmax + B) * pow(Xmax, alpha);
+  if (Xmax != Xmin) { /* Xmin == CUT_OFF; Check which one is better */
+    double v1 = (cut_off + A)/(cut_off + B);
+    double v2 = 2.33 * (Xmax + A)/(Xmax + B) * pow(Xmax, alpha);
 
-      if (1.1 * v2 >= v1) /* Prefer fitting into the cache if slowdown < 10% */
-          V = v1;
-      else {
-          Xmin = Xmax;
-          V = v2;
-      }
-  } else if (B > 0)                     /* We need V */
-      V = 2.33 * (Xmin + A)/(Xmin + B) * pow(Xmin, alpha);
+    if (1.1 * v2 >= v1) /* Prefer fitting into the cache if slowdown < 10% */
+      V = v1;
+    else
+    { Xmin = Xmax; V = v2; }
+  } else if (B > 0) /* We need V */
+    V = 2.33 * (Xmin + A)/(Xmin + B) * pow(Xmin, alpha);
   if (B > 0 && 1.1 * V > A/B)  /* Now Xmin is the minumum.  Compare with 0 */
-      Xmin = 0;
+    Xmin = 0;
 
   asize = (ulong)((1 + Xmin)*cache_arena - fixed_to_cache);
-  if (asize > total) asize = total;     /* May happen due to approximations */
+  if (asize > total) asize = total; /* May happen due to approximations */
   return asize;
 }
 
@@ -513,18 +505,15 @@ init_primepointer_gt(ulong a, byteptr *pd)
 static ulong
 optimize_chunk(ulong a, ulong b)
 {
-  /* TODO: Optimize size (surely < 512k to stay in L1 cache, but not so large */
-  /* as to force recalculating too often). */
-  /* Guesstimate: greater of sqrt(n) * lg(n) or 1M */
-  ulong chunk = maxuu(0x100000, usqrt(b) * expu(b));
+  /* TODO: Optimize size (surely < 512k to stay in L2 cache, but not so large
+   * as to force recalculating too often). */
+  ulong chunk = 0x80000UL;
   ulong tmp = (b - a) / chunk + 1;
 
   if (tmp == 1)
     chunk = b - a + 16;
   else
     chunk = (b - a) / tmp + 15;
-  /* Don't take up more than 2/3 of the stack */
-  chunk = minuu(chunk, avma - stack_lim(avma, 2));
   /* ensure 16 | chunk + 2 */
   return (((chunk + 2)>>4)<<4) - 2;
 }
@@ -570,8 +559,7 @@ arith_set(forprime_t *T)
   T->p = itou_or_0(d); avma = av; /* d = 0 is impossible */
 }
 
-/* run through primes in arithmetic progression = c (mod q).
- * Assume (c,q)=1, 0 <= c < q */
+/* run through primes in arithmetic progression = c (mod q) */
 static int
 u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
                             ulong a, ulong b, ulong c, ulong q)
@@ -587,10 +575,15 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
     return 0;
   }
   maxp = maxprime();
-  if (q != 1 && c != 2 && odd(q)) {
-    /* only allow *odd* primes. If c = 2, then p = 2 must be included :-( */
-    if (!odd(c)) c += q;
-    q <<= 1;
+  if (q != 1)
+  {
+    c %= q;
+    if (ugcd(c,q) != 1) { a = maxuu(a,c); b = minuu(b,c); }
+    if (odd(q) && (a > 2 || c != 2))
+    { /* only *odd* primes. If a <= c = 2, then p = 2 must be included :-( */
+      if (!odd(c)) c += q;
+      q <<= 1;
+    }
   }
   T->q = q;
   T->c = c;
@@ -653,7 +646,7 @@ u_forprime_restrict(forprime_t *T, ulong c) { T->b = c; }
 
 /* b = NULL: loop forever */
 int
-forprime_init(forprime_t *T, GEN a, GEN b)
+forprimestep_init(forprime_t *T, GEN a, GEN b, GEN q)
 {
   long lb;
   a = gceil(a); if (typ(a) != t_INT) pari_err_TYPE("forprime_init",a);
@@ -665,9 +658,7 @@ forprime_init(forprime_t *T, GEN a, GEN b)
     if (signe(b) < 0 || cmpii(a,b) > 0)
     {
       T->strategy = PRST_nextprime; /* paranoia */
-      T->bb = gen_0;
-      T->pp = gen_0;
-      return 0;
+      T->bb = T->pp = gen_0; return 0;
     }
     lb = lgefint(b);
     T->bb = b;
@@ -680,18 +671,39 @@ forprime_init(forprime_t *T, GEN a, GEN b)
   else /* b == -oo */
   {
     T->strategy = PRST_nextprime; /* paranoia */
-    T->bb = gen_0;
-    T->pp = gen_0;
-    return 0;
+    T->bb = T->pp = gen_0; return 0;
   }
   T->pp = cgeti(lb);
+  T->c = 0;
+  T->q = 1;
   /* a, b are positive integers, a <= b */
+  if (q)
+  {
+    switch(typ(q))
+    {
+      case t_INT: break;
+      case t_INTMOD: a = addii(a, modii(subii(gel(q,2),a), gel(q,1)));
+                     q = gel(q,1); break;
+      default: pari_err_TYPE("forprimestep_init",q);
+    }
+    if (signe(q) <= 0) pari_err_TYPE("forprimestep_init (q <= 0)",q);
+    if (equali1(q)) q = NULL;
+    else
+    {
+      T->q = itou(q);
+      T->c = umodiu(a, T->q);
+    }
+  }
   if (lgefint(a) == 3) /* lb == 3 implies b != NULL */
-    return u_forprime_init(T, uel(a,2), lb == 3? uel(b,2): ULONG_MAX);
+    return u_forprime_arith_init(T, uel(a,2), lb == 3? uel(b,2): ULONG_MAX,
+                                    T->c, T->q);
   T->strategy = PRST_nextprime;
-  affii(subiu(a,1), T->pp);
+  affii(subiu(a,T->q), T->pp);
   return 1;
 }
+int
+forprime_init(forprime_t *T, GEN a, GEN b)
+{ return forprimestep_init(T,a,b,NULL); }
 
 /* assume a <= b <= maxprime()^2, a,b odd, sieve[n] corresponds to
  *   a+16*n, a+16*n+2, ..., a+16*n+14 (bits 0 to 7)
@@ -781,12 +793,6 @@ init_modular_big(forprime_t *S)
 #endif
 }
 
-void
-init_modular(forprime_t *S)
-{
-  u_forprime_init(S, 27449, ULONG_MAX);
-}
-
 /* T->cache is a 0-terminated list of primes, return the first one and
  * remove it from list. Most of the time the list contains a single prime */
 static ulong
@@ -909,20 +915,13 @@ NEXT_CHUNK:
 #endif
       T->p = unextprime(T->p + 1);
     }
-    else
-    {
-      do {
-        T->p += T->q;
-        if (T->p < T->q) return 0; /* overflow */
-      } while (!uisprime(T->p));
-    }
-    if (!T->p) /* overflow ulong, switch to GEN */
-      T->strategy = PRST_nextprime;
-    else
-    {
-      if (T->p > T->b) return 0;
-      return T->p;
-    }
+    else do {
+      T->p += T->q;
+      if (T->p < T->q) { T->p = 0; break; } /* overflow */
+    } while (!uisprime(T->p));
+    if (T->p && T->p <= T->b) return T->p;
+    /* overflow ulong, switch to GEN */
+    T->strategy = PRST_nextprime;
   }
   return 0; /* overflow */
 }
@@ -934,26 +933,28 @@ forprime_next(forprime_t *T)
   GEN p;
   if (T->strategy != PRST_nextprime)
   {
-    ulong q = u_forprime_next(T);
-    if (q) { affui(q, T->pp); return T->pp; }
+    ulong u = u_forprime_next(T);
+    if (u) { affui(u, T->pp); return T->pp; }
     /* failure */
     if (T->strategy != PRST_nextprime) return NULL; /* we're done */
     /* overflow ulong, switch to GEN */
-    affui(ULONG_MAX, T->pp); /* initialize */
+    u = ULONG_MAX;
+    if (T->q > 1) u -= (ULONG_MAX-T->c) % T->q;
+    affui(u, T->pp);
   }
   av = avma;
-  p = nextprime(addiu(T->pp, 1));
+  p = nextprime(addiu(T->pp, T->q));
   if (T->bb && abscmpii(p, T->bb) > 0) { avma = av; return NULL; }
   affii(p, T->pp); avma = av; return T->pp;
 }
 
 void
-forprime(GEN a, GEN b, GEN code)
+forprimestep(GEN a, GEN b, GEN q, GEN code)
 {
   pari_sp av = avma;
   forprime_t T;
 
-  if (!forprime_init(&T, a,b)) { avma = av; return; }
+  if (!forprimestep_init(&T, a,b,q)) { avma = av; return; }
 
   push_lex(T.pp,code);
   while(forprime_next(&T))
@@ -965,6 +966,8 @@ forprime(GEN a, GEN b, GEN code)
   }
   pop_lex(1); avma = av;
 }
+void
+forprime(GEN a, GEN b, GEN code) { return forprimestep(a,b,NULL,code); }
 
 int
 forcomposite_init(forcomposite_t *C, GEN a, GEN b)
