@@ -22,12 +22,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #  include "../systems/mingw/mingw.h"
 #  include <process.h>
 #endif
-#ifdef __EMSCRIPTEN__
-#include "../systems/emscripten/emscripten.h"
-#endif
 
 #include "pari.h"
 #include "paripriv.h"
+#ifdef __EMSCRIPTEN__
+#include "../systems/emscripten/emscripten.h"
+#endif
 
 /********************************************************************/
 /**                                                                **/
@@ -168,7 +168,7 @@ print_fun_list(char **list, long nbli)
   if (i) pari_putc('\n');
 }
 
-static const long MAX_SECTION = 14;
+static const long MAX_SECTION = 16;
 static void
 commands(long n)
 {
@@ -270,21 +270,22 @@ menu_commands(void)
   ulong i;
   const char *s[] = {
   "user-defined functions (aliases, installed and user functions)",
+  "PROGRAMMING under GP",
   "Standard monadic or dyadic OPERATORS",
   "CONVERSIONS and similar elementary functions",
-  "TRANSCENDENTAL functions",
+  "functions related to COMBINATORICS",
   "NUMBER THEORETICAL functions",
-  "ELLIPTIC CURVES",
-  "L-FUNCTIONS",
-  /* "MODULAR FORMS", */
-  "MODULAR SYMBOLS",
-  "General NUMBER FIELDS",
-  "Associative and central simple ALGEBRAS",
   "POLYNOMIALS and power series",
   "Vectors, matrices, LINEAR ALGEBRA and sets",
+  "TRANSCENDENTAL functions",
   "SUMS, products, integrals and similar functions",
+  "General NUMBER FIELDS",
+  "Associative and central simple ALGEBRAS",
+  "ELLIPTIC CURVES",
+  "L-FUNCTIONS",
+  "MODULAR FORMS",
+  "MODULAR SYMBOLS",
   "GRAPHIC functions",
-  "PROGRAMMING under GP",
   "The PARI community"
   };
   pari_puts("Help topics: for a list of relevant subtopics, type ?n for n in\n");
@@ -297,6 +298,7 @@ menu_commands(void)
 Extended help (if available):\n\
   ??             (opens the full user's manual in a dvi previewer)\n\
   ??  tutorial / refcard / libpari (tutorial/reference card/libpari manual)\n\
+  ??  refcard-ell (or -lfun/-mf/-nf: specialized reference card)\n\
   ??  keyword    (long help text about \"keyword\" from the user's manual)\n\
   ??? keyword    (a propos: list of related functions).");
 }
@@ -352,7 +354,7 @@ disc : discriminant                                ell,nf,bnf,bnr,rnf\n\
 e, f : inertia/residue  degree           prid\n\
 fu   : fundamental units                                  bnf,bnr\n\
 gen  : generators                    bid,prid,clgp,ell,   bnf,bnr,    gal\n\
-group: group                                       ell,          ,rnf,gal\n\
+group: group                                       ell,               gal\n\
 index: index                                           nf,bnf,bnr\n\
 j    : j-invariant                                 ell\n");
 /* split: some compilers can't handle long constant strings */
@@ -471,7 +473,13 @@ gphelp_keyword_list(void)
   "real",
   "readline",
   "refcard",
+  "refcard-nf",
+  "refcard-ell",
+  "refcard-mf",
+  "refcard-lfun",
   "tutorial",
+  "tutorial-mf",
+  "mf",
   "nf",
   "bnf",
   "bnr",
@@ -528,6 +536,12 @@ digit_help(char *s, long flag)
   return;
 }
 
+long
+pari_community(void)
+{
+  return MAX_SECTION+1;
+}
+
 static void
 simple_help(const char *s1, const char *s2) { pari_printf("%s: %s\n", s1, s2); }
 
@@ -554,10 +568,10 @@ help(const char *s0, int flag)
   if (*s == '\\') { char *t = s+1; pari_skip_alpha(&t); *t = '\0'; }
   if (isalpha((int)*s))
   {
+    char *t = s;
     if (!strncmp(s, "default", 7))
     { /* special-case ?default(dft_name), e.g. default(log) */
-      char *t = s+7;
-      pari_skip_space(&t);
+      t += 7; pari_skip_space(&t);
       if (*t == '(')
       {
         t++; pari_skip_space(&t);
@@ -565,7 +579,9 @@ help(const char *s0, int flag)
         if (pari_is_default(t)) { default_help(t,flag); return; }
       }
     }
-    cut_trailing_garbage(s);
+    if (!strncmp(s, "refcard-", 8)) t += 8;
+    else if (!strncmp(s, "tutorial-", 9)) t += 9;
+    cut_trailing_garbage(t);
   }
 
   if (long_help && (n = ok_external_help(&s))) { external_help(s,n); return; }
@@ -1041,7 +1057,7 @@ color_prompt(const char *prompt)
   *s = 0;
   /* escape sequences bug readline, so use special bracing (if available) */
   brace_color(s, c_PROMPT, 0);
-  s += strlen(s); strncpy(s, prompt, n);
+  s += strlen(s); memcpy(s, prompt, n);
   s += n; *s = 0;
   brace_color(s, c_INPUT, 1);
   return t;
@@ -1069,7 +1085,6 @@ static int
 is_interactive(void)
 { return cb_pari_is_interactive? cb_pari_is_interactive(): 0; }
 
-static const char esc = (0x1f & '['); /* C-[ = escape */
 static char *
 strip_prompt(const char *s)
 {
@@ -1080,7 +1095,7 @@ strip_prompt(const char *s)
   {
     /* RL_PROMPT_START_IGNORE / RL_PROMPT_END_IGNORE */
     if (*s == 1 || *s == 2) continue;
-    if (*s == esc) /* skip ANSI color escape sequence */
+    if (*s == '\x1b') /* skip ANSI color escape sequence */
     {
       while (*++s != 'm')
         if (!*s) goto end;
@@ -1155,7 +1170,7 @@ get_line_from_file(const char *prompt, filtre_t *F, FILE *file)
   }
   s = F->buf->buf;
   /* don't log if from gprc or empty input */
-  if (*s && prompt) gp_echo_and_log(prompt, s);
+  if (*s && prompt && GP_DATA->echo != 2) gp_echo_and_log(prompt, s);
   return 1;
 }
 
@@ -1184,6 +1199,7 @@ gp_read_line(filtre_t *F, const char *PROMPT)
   if (interactive)
   {
     BLOCK_EH_START
+    if (!pari_last_was_newline()) pari_putc('\n');
     if (cb_pari_get_line_interactive)
       res = cb_pari_get_line_interactive(p, GP_DATA->prompt_cont, F);
     else {
@@ -1429,8 +1445,7 @@ static long
 atocolor(const char *s)
 {
   long l = atol(s);
-  if (l <   0) l =   0;
-  if (l > 255) l = 255;
+  if (l & ~0xff) pari_err(e_MISC, "invalid 8bit RGB code: %ld", l);
   return l;
 }
 
@@ -1510,60 +1525,25 @@ sd_graphcolormap(const char *v, long flag)
   }
   if (flag == d_RETURN || flag == d_ACKNOWLEDGE)
   {
-    GEN cols = cgetg(lg(GP_DATA->colormap), t_VEC);
-    long i;
-
-    for (i = 1; i < lg(cols); i++)
+    GEN C = cgetg(lg(GP_DATA->colormap), t_VEC);
+    long i, l = lg(C);
+    for (i = 1; i < l; i++)
     {
       GEN c = gel(GP_DATA->colormap, i);
-      if (typ(c) == t_STR)
-        gel(cols, i) = gcopy(c);
-      else
-        gel(cols, i) = vecsmall_to_vec(c);
+      gel(C, i) = (typ(c) == t_STR)? gcopy(c): zv_to_ZV(c);
     }
-    if (flag == d_RETURN) return cols;
-    pari_printf("   graphcolormap = %Ps\n", cols);
+    if (flag == d_RETURN) return C;
+    pari_printf("   graphcolormap = %Ps\n", C);
   }
   return gnil;
 }
 
 GEN
 sd_graphcolors(const char *v, long flag)
-{
-  long i, l;
-  char *p;
-
-  if (v) {
-    char *t = gp_filter(v);
-    for (p = t+1, l=2; *p != ']'; p++)
-      if (*p == ',') l++;
-      else if (*p < '0' || *p > '9')
-        pari_err(e_SYNTAX, "incorrect value for graphcolors", p, t);
-    if (*++p) pari_err(e_SYNTAX, "incorrect value for graphcolors", p, t);
-    if (GP_DATA->graphcolors) pari_free(GP_DATA->graphcolors);
-    GP_DATA->graphcolors = cgetalloc(t_VECSMALL, l);
-    for (p = t+1, i=0; *p; p++)
-    {
-      long n = 0;
-      while (*p >= '0' && *p <= '9')
-      {
-        n *= 10;
-        n += *p-'0';
-        p++;
-      }
-      GP_DATA->graphcolors[++i] = n;
-    }
-    pari_free(t);
-  }
-  switch(flag)
-  {
-  case d_RETURN:
-    return vecsmall_to_vec(GP_DATA->graphcolors);
-  case d_ACKNOWLEDGE:
-    pari_printf("   graphcolors = %Ps\n", vecsmall_to_vec(GP_DATA->graphcolors));
-  }
-  return gnil;
-}
+{ return sd_intarray(v, flag, &(GP_DATA->graphcolors), "graphcolors"); }
+GEN
+sd_plothsizes(const char *v, long flag)
+{ return sd_intarray(v, flag, &(GP_DATA->plothsizes), "plothsizes"); }
 
 GEN
 sd_help(const char *v, long flag)
@@ -1611,7 +1591,7 @@ sd_breakloop(const char *v, long flag)
 { return sd_toggle(v,flag,"breakloop", &(GP_DATA->breakloop)); }
 GEN
 sd_echo(const char *v, long flag)
-{ return sd_toggle(v,flag,"echo", &(GP_DATA->echo)); }
+{ return sd_ulong(v,flag,"echo", &(GP_DATA->echo), 0,2,NULL); }
 GEN
 sd_timer(const char *v, long flag)
 { return sd_toggle(v,flag,"timer", &(GP_DATA->chrono)); }
@@ -1741,6 +1721,18 @@ print_all_user_fun(int member)
   avma = av;
 }
 
+static char *
+get_name(const char *s)
+{
+  char *t = get_sep(s);
+  if (*t == '"')
+  {
+    long n = strlen(t)-1;
+    if (t[n] == '"') { t[n] = 0; t++; }
+  }
+  return t;
+}
+
 static void
 escape(const char *tch, int ismain)
 {
@@ -1768,7 +1760,7 @@ escape(const char *tch, int ismain)
         case 'a': brute(x, GP_DATA->fmt->format, -1); break;
         case 'x': dbgGEN(x, get_int(s, -1)); break;
         case 'w':
-          s = get_sep(s); if (!*s) s = current_logfile;
+          s = get_name(s); if (!*s) s = current_logfile;
           write0(s, mkvec(x)); return;
       }
       pari_putc('\n'); return;
@@ -1790,7 +1782,7 @@ escape(const char *tch, int ismain)
       break;
     case 'h': print_functions_hash(s); break;
     case 'l':
-      s = get_sep(s);
+      s = get_name(s);
       if (*s)
       {
         (void)sd_logfile(s,d_ACKNOWLEDGE);
@@ -1812,7 +1804,7 @@ escape(const char *tch, int ismain)
       break;
     case 'q': cb_pari_quit(0); break;
     case 'r':
-      s = get_sep(s);
+      s = get_name(s);
       if (!ismain) { (void)gp_read_file(s); break; }
       switchin(s);
       if (file_is_binary(pari_infile))
