@@ -55,7 +55,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "paripriv.h"
 #include "anal.h"
 
-const double LOG2    = 0.6931471805599453; /* log(2) */
 const double LOG10_2 = 0.3010299956639812; /* log_10(2) */
 const double LOG2_10 = 3.321928094887362;  /* log_2(10) */
 
@@ -86,8 +85,7 @@ static const ulong readonly_err_STACK[] = {
   evaltyp(t_ERROR) | _evallg(2),
   e_STACK
 };
-THREAD GEN    bernzone;
-GEN     primetab; /* private primetable */
+THREAD GEN bernzone, primetab;
 byteptr diffptr;
 FILE    *pari_outfile, *pari_errfile, *pari_logfile, *pari_infile;
 char    *current_logfile, *current_psfile, *pari_datadir;
@@ -100,8 +98,6 @@ long    precreal;
 ulong   precdl, logstyle;
 gp_data *GP_DATA;
 
-GEN colormap, pari_graphcolors;
-
 entree  **varentries;
 THREAD long *varpriority;
 
@@ -174,6 +170,7 @@ newblock(size_t n)
 {
   long *x = (long *) pari_malloc((n + BL_HEAD)*sizeof(long)) + BL_HEAD;
 
+  bl_size(x) = n;
   bl_refc(x) = 1;
   bl_next(x) = NULL;
   bl_prev(x) = cur_block;
@@ -182,12 +179,8 @@ newblock(size_t n)
 #ifdef DEBUG
   err_printf("+ %ld\n", ++NUM);
 #endif
-  if (DEBUGMEM)
-  {
-    if (!n) pari_warn(warner,"mallocing NULL object in newblock");
-    if (DEBUGMEM > 2)
-      err_printf("new block, size %6lu (no %ld): %08lx\n", n, next_block-1, x);
-  }
+  if (DEBUGMEM > 2)
+    err_printf("new block, size %6lu (no %ld): %08lx\n", n, next_block-1, x);
   return cur_block = x;
 }
 
@@ -306,19 +299,19 @@ pari_stackcheck_init(void *pari_stack_base)
 /*                         HEAP TRAVERSAL                          */
 /*******************************************************************/
 struct getheap_t { long n, l; };
+/* x is a block, not necessarily a clone [x[0] may not be set] */
 static void
 f_getheap(GEN x, void *D)
 {
   struct getheap_t *T = (struct getheap_t*)D;
   T->n++;
-  T->l += gsizeword(x);
+  T->l += bl_size(x) + BL_HEAD;
 }
 GEN
 getheap(void)
 {
   struct getheap_t T = { 0, 0 };
-  traverseheap(&f_getheap, &T);
-  return mkvec2s(T.n, T.l + BL_HEAD * T.n);
+  traverseheap(&f_getheap, &T); return mkvec2s(T.n, T.l);
 }
 
 void
@@ -342,7 +335,7 @@ pari_daemon(void)
       case -1: return 1; /* father, fork failed */
       case 0:
         (void)setsid(); /* son becomes process group leader */
-        if (fork()) exit(0); /* now son exits, also when fork fails */
+        if (fork()) _exit(0); /* now son exits, also when fork fails */
         break; /* grandson: its father is the son, which exited,
                 * hence father becomes 'init', that'll take care of it */
       default: /* father, fork succeeded */
@@ -395,6 +388,25 @@ pari_handle_SIGINT(void)
 #endif
 }
 
+typedef void (*pari_sighandler_t)(int);
+
+pari_sighandler_t
+os_signal(int sig, pari_sighandler_t f)
+{
+#ifdef HAS_SIGACTION
+  struct sigaction sa, oldsa;
+
+  sa.sa_handler = f;
+  sigemptyset(&sa.sa_mask);
+  sa.sa_flags = SA_NODEFER;
+
+  if (sigaction(sig, &sa, &oldsa)) return NULL;
+  return oldsa.sa_handler;
+#else
+  return signal(sig,f);
+#endif
+}
+
 void
 pari_sighandler(int sig)
 {
@@ -450,7 +462,7 @@ pari_sighandler(int sig)
         pari_outfile = stdout; pari_fclose(f);
         pari_err(e_MISC, "Broken Pipe, resetting file stack...");
       }
-      return; /* not reached */
+      return; /* LCOV_EXCL_LINE */
     }
 #endif
 
@@ -521,7 +533,8 @@ pari_init_defaults(void)
 #endif
 
   precdl = 16;
-  DEBUGFILES = DEBUGLEVEL = DEBUGMEM = 0;
+  DEBUGFILES = DEBUGLEVEL = 0;
+  DEBUGMEM = 1;
   disable_color = 1;
   logstyle = logstyle_none;
 
@@ -541,55 +554,11 @@ pari_init_defaults(void)
   }
   else pari_datadir= pari_strdup(pari_datadir);
   for (i=0; i<c_LAST; i++) gp_colors[i] = c_NONE;
-  colormap = NULL; pari_graphcolors = NULL;
 }
 
 /*********************************************************************/
 /*                   FUNCTION HASHTABLES, MODULES                    */
 /*********************************************************************/
-
-/* Initialize hashtable */
-static void
-init_hashtable(entree **table, long tblsz)
-{
-  long i;
-  for (i = 0; i < tblsz; i++)
-  {
-    entree *last = NULL, *ep = table[i];
-    table[i] = NULL;
-    while (ep)
-    {
-      entree *EP = ep->next;
-      switch(EpVALENCE(ep))
-      {
-        case EpVAR: case EpINSTALL:
-        /* keep: attach it to last entree seen */
-          if (last)
-            last->next = ep;
-          else
-            table[i] = ep;
-          ep->next = NULL; last = ep;
-          break;
-        default: freeep(ep);
-      }
-      ep = EP;
-    }
-  }
-}
-/* Load in hashtable hash the modules contained in A */
-static int
-gp_init_entrees(pari_stack *p_A, entree **hash)
-{
-  long i;
-  entree **v = (entree **)*pari_stack_base(p_A);
-  init_hashtable(hash, functions_tblsz);
-  for (i = 0; i < p_A->n; i++) pari_fill_hashtable(hash, v[i]);
-  return (hash == functions_hash);
-}
-int
-gp_init_functions(void)
-{ return gp_init_entrees(&s_MODULES, functions_hash); }
-
 extern entree functions_basic[], functions_default[];
 static void
 pari_init_functions(void)
@@ -628,9 +597,20 @@ pari_add_defaults_module(entree *ep)
 static void *
 pari_mainstack_malloc(size_t size)
 {
+  /* Check that the system allows reserving "size" bytes. This is just
+   * a check, we immediately free the memory. */
   void *b = mmap(NULL, size, PROT_READ|PROT_WRITE,
-                             MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE,-1,0);
-  return (b == MAP_FAILED) ? NULL: b;
+                             MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+  if (b == MAP_FAILED) return NULL;
+  munmap(b, size);
+
+  /* Map again, this time with MAP_NORESERVE. On some operating systems
+   * like Cygwin, this is needed because remapping with PROT_NONE and
+   * MAP_NORESERVE does not work as expected. */
+  b = mmap(NULL, size, PROT_READ|PROT_WRITE,
+                       MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, -1, 0);
+  if (b == MAP_FAILED) return NULL;
+  return b;
 }
 
 static void
@@ -639,10 +619,64 @@ pari_mainstack_mfree(void *s, size_t size)
   munmap(s, size);
 }
 
+/* Completely discard the memory mapped between the addresses "from"
+ * and "to" (which must be page-aligned).
+ *
+ * We use mmap() with PROT_NONE, which means that the underlying memory
+ * is freed and that the kernel should not commit memory for it. We
+ * still keep the mapping such that we can change the flags to
+ * PROT_READ|PROT_WRITE later.
+ *
+ * NOTE: remapping with MAP_FIXED and PROT_NONE is not the same as
+ * calling mprotect(..., PROT_NONE) because the latter will keep the
+ * memory committed (this is in particular relevant on Linux with
+ * vm.overcommit = 2). This remains true even when calling
+ * madvise(..., MADV_DONTNEED). */
 static void
-pari_mainstack_mreset(void *s, size_t size)
+pari_mainstack_mreset(pari_sp from, pari_sp to)
 {
-  madvise(s, size, MADV_DONTNEED);
+  size_t s = to - from;
+  void *addr, *res;
+  if (!s) return;
+
+  addr = (void*)from;
+  res = mmap(addr, s, PROT_NONE,
+             MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, -1, 0);
+  if (res != addr) pari_err(e_MEM);
+}
+
+/* Commit (make available) the virtual memory mapped between the
+ * addresses "from" and "to" (which must be page-aligned).
+ * Return 0 if successful, -1 if failed. */
+static int
+pari_mainstack_mextend(pari_sp from, pari_sp to)
+{
+  size_t s = to - from;
+  return mprotect((void*)from, s, PROT_READ|PROT_WRITE);
+}
+
+/* Set actual stack size to the given size. This sets st->size and
+ * st->bot. If not enough system memory is available, this can fail.
+ * Return 1 if successful, 0 if failed (in that case, st->size is not
+ * changed) */
+static int
+pari_mainstack_setsize(struct pari_mainstack *st, size_t size)
+{
+  pari_sp newbot = st->top - size;
+  /* Align newbot to pagesize */
+  pari_sp alignbot = newbot & ~(pari_sp)(PARI_STACK_ALIGN - 1);
+  if (pari_mainstack_mextend(alignbot, st->top))
+  {
+    /* Making the memory available did not work: limit vsize to the
+     * current actual stack size. */
+    st->vsize = st->size;
+    pari_warn(warnstack, st->vsize);
+    return 0;
+  }
+  pari_mainstack_mreset(st->vbot, alignbot);
+  st->bot = newbot;
+  st->size = size;
+  return 1;
 }
 
 #else
@@ -656,8 +690,13 @@ pari_mainstack_malloc(size_t s)
 static void
 pari_mainstack_mfree(void *s, size_t size) { (void) size; free(s); }
 
-static void
-pari_mainstack_mreset(void *s, size_t size) { (void) s; (void) size; }
+static int
+pari_mainstack_setsize(struct pari_mainstack *st, size_t size)
+{
+  st->bot = st->top - size;
+  st->size = size;
+  return 1;
+}
 
 #endif
 
@@ -673,7 +712,7 @@ fix_size(size_t a)
 }
 
 static void
-pari_mainstack_alloc(struct pari_mainstack *st, size_t rsize, size_t vsize)
+pari_mainstack_alloc(int numerr, struct pari_mainstack *st, size_t rsize, size_t vsize)
 {
   size_t sizemax = vsize ? vsize: rsize, s = fix_size(sizemax);
   for (;;)
@@ -682,13 +721,16 @@ pari_mainstack_alloc(struct pari_mainstack *st, size_t rsize, size_t vsize)
     if (st->vbot) break;
     if (s == MIN_STACK) pari_err(e_MEM); /* no way out. Die */
     s = fix_size(s >> 1);
-    pari_warn(warnstack, s);
+    pari_warn(numerr, s);
   }
   st->vsize = vsize ? s: 0;
   st->rsize = minuu(rsize, s);
-  st->size = st->rsize;
   st->top = st->vbot+s;
-  st->bot = st->top - st->size;
+  if (!pari_mainstack_setsize(st, st->rsize))
+  {
+    /* This should never happen since we only decrease the allocated space */
+    pari_err(e_MEM);
+  }
   st->memused = 0;
 }
 
@@ -697,7 +739,7 @@ pari_mainstack_free(struct pari_mainstack *st)
 {
   pari_mainstack_mfree((void*)st->vbot, st->vsize ? st->vsize : fix_size(st->rsize));
   st->top = st->bot = st->vbot = 0;
-  st->size = st->vsize =0;
+  st->size = st->vsize = 0;
 }
 
 static void
@@ -705,7 +747,7 @@ pari_mainstack_resize(struct pari_mainstack *st, size_t rsize, size_t vsize)
 {
   BLOCK_SIGINT_START;
   pari_mainstack_free(st);
-  pari_mainstack_alloc(st, rsize, vsize);
+  pari_mainstack_alloc(warnstack, st, rsize, vsize);
   BLOCK_SIGINT_END;
 }
 
@@ -719,7 +761,7 @@ pari_mainstack_use(struct pari_mainstack *st)
 static void
 paristack_alloc(size_t rsize, size_t vsize)
 {
-  pari_mainstack_alloc(pari_mainstack, rsize, vsize);
+  pari_mainstack_alloc(warnstack, pari_mainstack, rsize, vsize);
   pari_mainstack_use(pari_mainstack);
 }
 
@@ -740,7 +782,9 @@ parivstack_resize(ulong newsize)
   evalstate_reset();
   paristack_setsize(pari_mainstack->rsize, newsize);
   s = pari_mainstack->vsize ? pari_mainstack->vsize : pari_mainstack->rsize;
-  pari_warn(warner,"new maximum stack size = %lu (%.3f Mbytes)", s, s/1048576.);
+  if (DEBUGMEM)
+    pari_warn(warner,"new maximum stack size = %lu (%.3f Mbytes)",
+              s, s/1048576.);
   pari_init_errcatch();
   cb_pari_err_recover(-1);
 }
@@ -754,7 +798,8 @@ paristack_newrsize(ulong newsize)
     pari_mainstack_resize(pari_mainstack, newsize, vsize);
   evalstate_reset();
   s = pari_mainstack->rsize;
-  pari_warn(warner,"new stack size = %lu (%.3f Mbytes)", s, s/1048576.);
+  if (DEBUGMEM)
+    pari_warn(warner,"new stack size = %lu (%.3f Mbytes)", s, s/1048576.);
   pari_init_errcatch();
   cb_pari_err_recover(-1);
 }
@@ -762,24 +807,33 @@ paristack_newrsize(ulong newsize)
 void
 paristack_resize(ulong newsize)
 {
-  size_t vsize = pari_mainstack->vsize;
+  long size = pari_mainstack->size;
   if (!newsize)
-    newsize = pari_mainstack->size << 1;
-  newsize = maxuu(minuu(newsize, vsize), pari_mainstack->size);
-  pari_mainstack->size = newsize;
-  pari_mainstack->bot = pari_mainstack->top - pari_mainstack->size;
-  pari_warn(warner,"increasing stack size to %lu",newsize);
+    newsize = 2 * size;
+  newsize = minuu(newsize, pari_mainstack->vsize);
+  if (newsize <= pari_mainstack->size) return;
+  if (pari_mainstack_setsize(pari_mainstack, newsize))
+  {
+    if (DEBUGMEM)
+      pari_warn(warner, "increasing stack size to %lu", pari_mainstack->size);
+  }
+  else
+  {
+    pari_mainstack_setsize(pari_mainstack, size);
+    pari_err(e_STACK);
+  }
 }
 
 void
 parivstack_reset(void)
 {
-  pari_mainstack->size = pari_mainstack->rsize;
-  pari_mainstack->bot = pari_mainstack->top - pari_mainstack->size;
-  pari_mainstack_mreset((void *)pari_mainstack->vbot,
-                        pari_mainstack->bot-pari_mainstack->vbot);
+  pari_mainstack_setsize(pari_mainstack, pari_mainstack->rsize);
+  if (avma < pari_mainstack->bot)
+    pari_err_BUG("parivstack_reset [avma < bot]");
 }
 
+/* Enlarge the stack if needed such that the unused portion of the stack
+ * (between bot and avma) is large enough to contain x longs. */
 void
 new_chunk_resize(size_t x)
 {
@@ -799,7 +853,7 @@ new_chunk_resize(size_t x)
 void
 pari_thread_valloc(struct pari_thread *t, size_t s, size_t v, GEN arg)
 {
-  pari_mainstack_alloc(&t->st,s,v);
+  pari_mainstack_alloc(warnstackthread, &t->st,s,v);
   t->data = arg;
 }
 
@@ -809,7 +863,7 @@ pari_thread_valloc(struct pari_thread *t, size_t s, size_t v, GEN arg)
 void
 pari_thread_alloc(struct pari_thread *t, size_t s, GEN arg)
 {
-  pari_mainstack_alloc(&t->st,s,0);
+  pari_mainstack_alloc(warnstackthread, &t->st,s,0);
   t->data = arg;
 }
 
@@ -830,14 +884,16 @@ pari_thread_init(void)
   pari_init_compiler();
   pari_init_evaluator();
   pari_init_files();
+  pari_thread_init_primetab();
   pari_thread_init_seadata();
 }
 
 void
 pari_thread_sync(void)
 {
-  pari_pthread_init_varstate();
+  pari_pthread_init_primetab();
   pari_pthread_init_seadata();
+  pari_pthread_init_varstate();
 }
 
 void
@@ -905,8 +961,7 @@ pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
   if (init_opts&INIT_DFTm) {
     pari_init_defaults();
     GP_DATA = default_gp_data();
-    gp_expand_path(GP_DATA->path);
-    gp_expand_path(GP_DATA->sopath);
+    pari_init_paths();
   }
 
   pari_mainstack = (struct pari_mainstack *) malloc(sizeof(*pari_mainstack));
@@ -915,8 +970,8 @@ pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
   diffptr = NULL;
   if (!(init_opts&INIT_noPRIMEm))  pari_init_primes(maxprime);
   if (!(init_opts&INIT_noINTGMPm)) pari_kernel_init();
-
-  primetab = cgetalloc(t_VEC, 1);
+  pari_init_graphics();
+  pari_init_primetab();
   pari_init_seadata();
   pari_thread_init();
   pari_init_functions();
@@ -952,7 +1007,7 @@ pari_close_opts(ulong init_opts)
     }
   }
   pari_var_close();
-  free((void*)primetab);
+  pari_close_mf();
   pari_thread_close();
   pari_close_files();
   pari_close_homedir();
@@ -969,13 +1024,13 @@ pari_close_opts(ulong init_opts)
   if (pari_datadir) free(pari_datadir);
   if (init_opts&INIT_DFTm)
   { /* delete GP_DATA */
+    pari_close_paths();
     if (GP_DATA->hist->v) free((void*)GP_DATA->hist->v);
     if (GP_DATA->pp->cmd) free((void*)GP_DATA->pp->cmd);
-    delete_dirs(GP_DATA->path);
-    free((void*)GP_DATA->path->PATH);
-    delete_dirs(GP_DATA->sopath);
-    free((void*)GP_DATA->sopath->PATH);
     if (GP_DATA->help) free((void*)GP_DATA->help);
+    if (GP_DATA->plothsizes) free((void*)GP_DATA->plothsizes);
+    if (GP_DATA->colormap) pari_free(GP_DATA->colormap);
+    if (GP_DATA->graphcolors) pari_free(GP_DATA->graphcolors);
     free((void*)GP_DATA->prompt);
     free((void*)GP_DATA->prompt_cont);
     free((void*)GP_DATA->histfile);
@@ -995,9 +1050,6 @@ pari_close(void)
 void
 gp_context_save(struct gp_context* rec)
 {
-  rec->file = pari_last_tmp_file();
-  if (DEBUGFILES>1)
-    err_printf("gp_context_save: %s\n", rec->file ? rec->file->name: "NULL");
   rec->prettyp = GP_DATA->fmt->prettyp;
   rec->listloc = next_block;
   rec->iferr_env = iferr_env;
@@ -1005,6 +1057,7 @@ gp_context_save(struct gp_context* rec)
   varstate_save(&rec->var);
   evalstate_save(&rec->eval);
   parsestate_save(&rec->parse);
+  filestate_save(&rec->file);
 }
 
 void
@@ -1019,7 +1072,7 @@ gp_context_restore(struct gp_context* rec)
   if (DEBUGMEM>2) err_printf("entering recover(), loc = %ld\n", rec->listloc);
   evalstate_restore(&rec->eval);
   parsestate_restore(&rec->parse);
-  filestate_restore(rec->file);
+  filestate_restore(&rec->file);
   global_err_data = rec->err_data;
   iferr_env = rec->iferr_env;
   GP_DATA->fmt->prettyp = rec->prettyp;
@@ -1117,14 +1170,16 @@ pari_warn(int numerr, ...)
       break;
 
     case warnstack:
+    case warnstackthread:
     {
       ulong  s = va_arg(ap, ulong);
       char buf[128];
-      sprintf(buf,"Warning: not enough memory, new stack %lu", (ulong)s);
+      const char * stk = numerr == warnstackthread
+                         || mt_is_thread() ? "thread": "PARI";
+      sprintf(buf,"Warning: not enough memory, new %s stack %lu", stk, s);
       out_puts(pariErr,buf);
       break;
     }
-
   }
   va_end(ap);
   out_term_color(pariErr, c_NONE);
@@ -1207,6 +1262,11 @@ pari_err2GEN(long numerr, va_list ap)
       const char *f = va_arg(ap, const char*);
       retmkerr3(numerr, strtoGENstr(f), strtoGENstr(va_arg(ap, char*)));
     }
+  case e_FILEDESC:
+    {
+      const char *f = va_arg(ap, const char*);
+      retmkerr3(numerr, strtoGENstr(f), stoi(va_arg(ap, long)));
+    }
   case e_OVERFLOW:
   case e_IMPL:
   case e_DIM:
@@ -1336,6 +1396,8 @@ pari_err2str(GEN e)
     return pari_sprintf("inconsistent dimensions in %Ps.", gel(e,2));
   case e_FILE:
     return pari_sprintf("error opening %Ps: `%Ps'.", gel(e,2), gel(e,3));
+  case e_FILEDESC:
+    return pari_sprintf("invalid file descriptor in %Ps [%Ps]", gel(e,2), gel(e,3));
   case e_FLAG:
     return pari_sprintf("invalid flag in %Ps.", gel(e,2));
   case e_IMPL:
@@ -1465,7 +1527,7 @@ pari_err2str(GEN e)
                           gel(e,2), pol_x(varn(x)), pol_x(varn(y)));
     }
   }
-  return NULL; /*NOT REACHED*/
+  return NULL; /*LCOV_EXCL_LINE*/
 }
 
 static int
@@ -1545,6 +1607,7 @@ numerr_name(long numerr)
   case e_DIM:      return "e_DIM";
   case e_DOMAIN:   return "e_DOMAIN";
   case e_FILE:     return "e_FILE";
+  case e_FILEDESC: return "e_FILEDESC";
   case e_FLAG:     return "e_FLAG";
   case e_IMPL:     return "e_IMPL";
   case e_INV:      return "e_INV";
@@ -1586,6 +1649,7 @@ name_numerr(const char *s)
   if (!strcmp(s,"e_DIM"))      return e_DIM;
   if (!strcmp(s,"e_DOMAIN"))   return e_DOMAIN;
   if (!strcmp(s,"e_FILE"))     return e_FILE;
+  if (!strcmp(s,"e_FILEDESC")) return e_FILEDESC;
   if (!strcmp(s,"e_FLAG"))     return e_FLAG;
   if (!strcmp(s,"e_IMPL"))     return e_IMPL;
   if (!strcmp(s,"e_INV"))      return e_INV;
@@ -1611,7 +1675,7 @@ name_numerr(const char *s)
   if (!strcmp(s,"e_USER"))     return e_USER;
   if (!strcmp(s,"e_VAR"))      return e_VAR;
   pari_err(e_MISC,"unknown error name");
-  return -1; /* NOT REACHED */
+  return -1; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -1669,6 +1733,15 @@ listassign(GEN x, GEN y)
   list_data(y) = list_internal_copy(L, nmax);
 }
 
+/* transform a non-malloced list (e.g. from gtolist or gtomap) to a malloced
+ * list suitable for listput */
+GEN
+listinit(GEN x)
+{
+  GEN y = cgetg(3, t_LIST);
+  listassign(x, y); return y;
+}
+
 /* copy list on the PARI stack */
 GEN
 listcopy(GEN x)
@@ -1749,6 +1822,7 @@ gcopy_avma(GEN x, pari_sp *AVMA)
   switch(typ(x))
   { /* non recursive types */
     case t_INT:
+      if (lgefint(x) == 2) return gen_0;
       *AVMA = (pari_sp)icopy_avma(x, *AVMA);
       return (GEN)*AVMA;
     case t_REAL: case t_STR: case t_VECSMALL:
@@ -1895,15 +1969,13 @@ taille0(GEN x)
   return n;
 }
 
-/* How many words do we need to allocate to copy x ? t_LIST is a special case
- * since list_data() is malloc'ed later, in list_internal_copy() */
 static long
-words_to_allocate(GEN x)
+gsizeclone_i(GEN x)
 {
   long i,n,lx, tx = typ(x);
   switch(tx)
   { /* non recursive types */
-    case t_INT: return lgefint(x);
+    case t_INT: lx = lgefint(x); return lx == 2? 0: lx;;
     case t_REAL:
     case t_STR:
     case t_VECSMALL: return lg(x);
@@ -1911,19 +1983,24 @@ words_to_allocate(GEN x)
     case t_LIST: return 3;
     default:
       n = lx = lg(x);
-      for (i=lontyp[tx]; i<lx; i++) n += words_to_allocate(gel(x,i));
+      for (i=lontyp[tx]; i<lx; i++) n += gsizeclone_i(gel(x,i));
       return n;
   }
 }
 
+/* #words needed to clone x; t_LIST is a special case since list_data() is
+ * malloc'ed later, in list_internal_copy() */
+static long
+gsizeclone(GEN x) { return (typ(x) == t_INT)? lgefint(x): gsizeclone_i(x); }
+
 long
 gsizeword(GEN x)
 {
   GEN L;
-  if (typ(x) != t_LIST) return words_to_allocate(x);
-  /* For t_LIST, return the actual list size, words_to_allocate() is always 3 */
+  if (typ(x) != t_LIST) return gsizeclone(x);
+  /* For t_LIST, return the actual list size, gsizeclone() is always 3 */
   L = list_data(x);
-  return L? 3 + words_to_allocate(L): 3;
+  return L? 3 + gsizeclone(L): 3;
 }
 long
 gsizebyte(GEN x) { return gsizeword(x) * sizeof(long); }
@@ -1957,7 +2034,7 @@ copy_bin_canon(GEN x)
 GEN
 gclone(GEN x)
 {
-  long i,lx,tx = typ(x), t = words_to_allocate(x);
+  long i,lx,tx = typ(x), t = gsizeclone(x);
   GEN y = newblock(t);
   switch(tx)
   { /* non recursive types */
@@ -2040,6 +2117,7 @@ shiftaddress_canon(GEN x, long dec)
         shiftaddress_canon(L, dec);
         list_data(x) = list_internal_copy(L, lg(L)); avma = av;
       }
+      break;
     }
     default:
       lx = lg(x);
@@ -2137,8 +2215,8 @@ obj_free(GEN S)
   for (i = 1; i < lg(v); i++)
   {
     GEN o = gel(v,i);
-    if (isclone(o)) gunclone(o);
     gel(v,i) = gen_0;
+    gunclone_deep(o);
   }
 }
 
@@ -2547,7 +2625,8 @@ pari_version(void)
  *   long menu    : which help section do we belong to
  *                   1: Standard monadic or dyadic OPERATORS
  *                   2: CONVERSIONS and similar elementary functions
- *                   3: TRANSCENDENTAL functions, etc.
+ *                   3: functions related to COMBINATORICS
+ *                   4: TRANSCENDENTAL functions, etc.
  *   char *code   : GP prototype, aka Parser Code (see libpari's manual)
  *                  if NULL, use valence instead.
  *   char *help   : short help text (init to NULL).
