@@ -13,7 +13,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
-#include "anal.h"
 
 GEN
 iferrpari(GEN a, GEN b, GEN c)
@@ -106,6 +105,181 @@ forpari(GEN a, GEN b, GEN code)
   pop_lex(1); avma = ltop;
 }
 
+/* 0 < a <= b. Using small consecutive chunks to 1) limit memory use, 2) allow
+ * cheap early abort */
+static int
+forfactoredpos(ulong a, ulong b, GEN code)
+{
+  const ulong step = 1024;
+  pari_sp av = avma;
+  ulong x1;
+  for(x1 = a;; x1 += step, avma = av)
+  { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
+    ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
+    GEN v = vecfactoru(x1, x2);
+    lv = lg(v);
+    for (j = 1; j < lv; j++)
+    {
+      ulong n = x1-1 + j;
+      set_lex(-1, mkvec2(utoipos(n), Flm_to_ZM(gel(v,j))));
+      closure_evalvoid(code);
+      if (loop_break()) return 1;
+    }
+    if (x2 == b) break;
+    set_lex(-1, gen_0);
+  }
+  return 0;
+}
+
+/* vector of primes to squarefree factorization */
+static GEN
+zv_to_ZM(GEN v)
+{ return mkmat2(zc_to_ZC(v), const_col(lg(v)-1,gen_1)); }
+/* vector of primes to negative squarefree factorization */
+static GEN
+zv_to_mZM(GEN v)
+{
+  long i, l = lg(v);
+  GEN w = cgetg(l+1, t_COL);
+  gel(w,1) = gen_m1; for (i = 1; i < l; i++) gel(w,i+1) = utoipos(v[i]);
+  return mkmat2(w, const_col(l,gen_1));
+}
+/* 0 <= a <= b. Using small consecutive chunks to 1) limit memory use, 2) allow
+ * cheap early abort */
+static void
+forsquarefreepos(ulong a, ulong b, GEN code)
+{
+  const ulong step = 1024;
+  pari_sp av = avma;
+  ulong x1;
+  for(x1 = a;; x1 += step, avma = av)
+  { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
+    ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
+    GEN v = vecfactorsquarefreeu(x1, x2);
+    lv = lg(v);
+    for (j = 1; j < lv; j++) if (gel(v,j))
+    {
+      ulong n = x1-1 + j;
+      set_lex(-1, mkvec2(utoipos(n), zv_to_ZM(gel(v,j))));
+      closure_evalvoid(code); if (loop_break()) return;
+    }
+    if (x2 == b) break;
+    set_lex(-1, gen_0);
+  }
+}
+/* 0 <= a <= b. Loop from -b, ... -a through squarefree integers */
+static void
+forsquarefreeneg(ulong a, ulong b, GEN code)
+{
+  const ulong step = 1024;
+  pari_sp av = avma;
+  ulong x2;
+  for(x2 = b;; x2 -= step, avma = av)
+  { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
+    ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
+    GEN v = vecfactorsquarefreeu(x1, x2);
+    for (j = lg(v)-1; j > 0; j--) if (gel(v,j))
+    {
+      ulong n = x1-1 + j;
+      set_lex(-1, mkvec2(utoineg(n), zv_to_mZM(gel(v,j))));
+      closure_evalvoid(code); if (loop_break()) return;
+    }
+    if (x1 == a) break;
+    set_lex(-1, gen_0);
+  }
+}
+void
+forsquarefree(GEN a, GEN b, GEN code)
+{
+  pari_sp av = avma;
+  long s;
+  if (typ(a) != t_INT) pari_err_TYPE("forsquarefree", a);
+  if (typ(b) != t_INT) pari_err_TYPE("forsquarefree", b);
+  if (cmpii(a,b) > 0) return;
+  s = signe(a);
+  if (s * signe(b) < 0) pari_err_TYPE("forsquarefree [!= signs]", mkvec2(a,b));
+  push_lex(NULL,code);
+  if (s < 0) forsquarefreeneg(itou(b), itou(a), code);
+  else       forsquarefreepos(itou(a), itou(b), code);
+  pop_lex(1); avma = av;
+}
+
+/* convert factoru(n) to factor(-n); M pre-allocated factorization matrix
+ * with (-1)^1 already set */
+static void
+Flm2negfact(GEN v, GEN M)
+{
+  GEN p = gel(v,1), e = gel(v,2), P = gel(M,1), E = gel(M,2);
+  long i, l = lg(p);
+  for (i = 1; i < l; i++)
+  {
+    gel(P,i+1) = utoipos(p[i]);
+    gel(E,i+1) = utoipos(e[i]);
+  }
+  setlg(P,l+1);
+  setlg(E,l+1);
+}
+/* 0 < a <= b, from -b to -a */
+static int
+forfactoredneg(ulong a, ulong b, GEN code)
+{
+  const ulong step = 1024;
+  GEN P, E, M;
+  pari_sp av;
+  ulong x2;
+
+  P = cgetg(18, t_COL); gel(P,1) = gen_m1;
+  E = cgetg(18, t_COL); gel(E,1) = gen_1;
+  M = mkmat2(P,E);
+  av = avma;
+  for(x2 = b;; x2 -= step, avma = av)
+  { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
+    ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
+    GEN v = vecfactoru(x1, x2);
+    for (j = lg(v)-1; j; j--)
+    { /* run backward: from factor(x1..x2) to factor(-x2..-x1) */
+      ulong n = x1-1 + j;
+      Flm2negfact(gel(v,j), M);
+      set_lex(-1, mkvec2(utoineg(n), M));
+      closure_evalvoid(code); if (loop_break()) return 1;
+    }
+    if (x1 == a) break;
+    set_lex(-1, gen_0);
+  }
+  return 0;
+}
+static int
+eval0(GEN code)
+{
+  pari_sp av = avma;
+  set_lex(-1, mkvec2(gen_0, mkmat2(mkcol(gen_0),mkcol(gen_1))));
+  closure_evalvoid(code); avma = av;
+  return loop_break();
+}
+void
+forfactored(GEN a, GEN b, GEN code)
+{
+  pari_sp av = avma;
+  long sa, sb, stop = 0;
+  if (typ(a) != t_INT) pari_err_TYPE("forfactored", a);
+  if (typ(b) != t_INT) pari_err_TYPE("forfactored", b);
+  if (cmpii(a,b) > 0) return;
+  push_lex(NULL,code);
+  sa = signe(a);
+  sb = signe(b);
+  if (sa < 0)
+  {
+    stop = forfactoredneg((sb < 0)? b[2]: 1UL, itou(a), code);
+    if (!stop && sb >= 0) stop = eval0(code);
+    if (!stop && sb > 0) forfactoredpos(1UL, b[2], code);
+  }
+  else
+  {
+    if (!sa) stop = eval0(code);
+    if (!stop && sb) forfactoredpos(sa? a[2]: 1UL, itou(b), code);
+  }
+  pop_lex(1); avma = av;
+}
 void
 whilepari(GEN a, GEN b)
 {
@@ -145,17 +319,19 @@ forstep(GEN a, GEN b, GEN s, GEN code)
   GEN v = NULL;
   int (*cmp)(GEN,GEN);
 
-  b = gcopy(b); s = gcopy(s); av=avma;
-  push_lex(a,code);
-  if (is_vec_t(typ(s)))
+  b = gcopy(b);
+  s = gcopy(s); av = avma;
+  switch(typ(s))
   {
-    v = s; s = gen_0;
-    for (i=lg(v)-1; i; i--) s = gadd(s,gel(v,i));
+    case t_VEC: case t_COL: ss = gsigne(vecsum(s)); v = s; break;
+    case t_INTMOD: a = gadd(a, gmod(gsub(gel(s,2),a), gel(s,1)));
+                   s = gel(s,1);
+    default: ss = gsigne(s);
   }
-  ss = gsigne(s);
   if (!ss) pari_err_DOMAIN("forstep","step","=",gen_0,s);
   cmp = (ss > 0)? &gcmp: &negcmp;
   i = 0;
+  push_lex(a,code);
   while (cmp(a,b) <= 0)
   {
     closure_evalvoid(code); if (loop_break()) break;
@@ -176,13 +352,12 @@ forstep(GEN a, GEN b, GEN s, GEN code)
   pop_lex(1); avma = av0;
 }
 
-void
-fordiv(GEN a, GEN code)
+static void
+_fordiv(GEN a, GEN code, GEN (*D)(GEN))
 {
   long i, l;
   pari_sp av2, av = avma;
-  GEN t = divisors(a);
-
+  GEN t = D(a);
   push_lex(gen_0,code); l=lg(t); av2 = avma;
   for (i=1; i<l; i++)
   {
@@ -192,6 +367,10 @@ fordiv(GEN a, GEN code)
   }
   pop_lex(1); avma=av;
 }
+void
+fordiv(GEN a, GEN code) { return _fordiv(a, code, &divisors); }
+void
+fordivfactored(GEN a, GEN code) { return _fordiv(a, code, &divisors_factored); }
 
 /* Embedded for loops:
  *   fl = 0: execute ch (a), where a = (ai) runs through all n-uplets in
@@ -299,7 +478,7 @@ _next_lt_i(forvec_t *d)
         if (cmpii(d->a[i-1], d->a[i]) < 0) continue;
         av = avma;
         /* M[i] > M[i-1] >= a[i-1] */
-        t = addis(d->a[i-1],1); if (cmpii(t, d->m[i]) < 0) t = d->m[i];
+        t = addiu(d->a[i-1],1); if (cmpii(t, d->m[i]) < 0) t = d->m[i];
         d->a[i] = resetloop(d->a[i], t);/*a[i]:=max(a[i-1]+1,m[i]) <= M[i]*/
         avma = av;
       }
@@ -325,7 +504,7 @@ _next_lt(forvec_t *d)
         i++;
         if (gcmp(d->a[i-1], d->a[i]) < 0) continue;
         /* M[i] > M[i-1] >= a[i-1] >= a[i] */
-        c = addis(gfloor(gsub(d->a[i-1], d->a[i])), 1); /* > a[i-1] - a[i] */
+        c = addiu(gfloor(gsub(d->a[i-1], d->a[i])), 1); /* > a[i-1] - a[i] */
         d->a[i] = gadd(d->a[i], c);
         /* a[i-1] < a[i] <= M[i-1] + 1 => a[i] < M[i]+1 => a[i] <= M[i] */
       }
@@ -375,7 +554,7 @@ forvec_init(forvec_t *d, GEN x, long flag)
       case 2: /* a > m[i-1] - m */
         a = gfloor(gsub(d->m[i-1], m));
         if (typ(a) != t_INT) pari_err_TYPE("forvec",a);
-        a = addis(a, 1);
+        a = addiu(a, 1);
         if (signe(a) > 0) m = gadd(m, a); else m = gcopy(m);
         break;
       default: m = gcopy(m);
@@ -472,31 +651,49 @@ somme(GEN a, GEN b, GEN code, GEN x)
   pop_lex(1); return gerepileupto(av0,x);
 }
 
+static GEN
+sum_init(GEN x0, GEN t)
+{
+  long tp = typ(t);
+  GEN x;
+  if (is_vec_t(tp))
+  {
+    x = const_vec(lg(t)-1, x0);
+    settyp(x, tp);
+  }
+  else
+    x = x0;
+  return x;
+}
+
 GEN
 suminf(void *E, GEN (*eval)(void *, GEN), GEN a, long prec)
 {
-  long fl, G;
+  long fl = 0, G = prec2nbits(prec) + 5;
   pari_sp av0 = avma, av;
-  GEN p1,x = real_1(prec);
+  GEN x = NULL, _1;
 
   if (typ(a) != t_INT) pari_err_TYPE("suminf",a);
   a = setloop(a);
   av = avma;
-  fl=0; G = prec2nbits(prec) + 5;
   for(;;)
   {
-    p1 = eval(E, a); x = gadd(x,p1); a = incloop(a);
-    if (gequal0(p1) || gexpo(p1) <= gexpo(x)-G)
-      { if (++fl==3) break; }
-    else
-      fl=0;
+    GEN t = eval(E, a);
+    if (!x) _1 = x = sum_init(real_1(prec), t);
+
+    x = gadd(x,t);
+    if (!gequal0(t) && gexpo(t) > gexpo(x)-G)
+      fl = 0;
+    else if (++fl == 3)
+      break;
+    a = incloop(a);
     if (gc_needed(av,1))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"suminf");
-      x = gerepileupto(av,x);
+      gerepileall(av,2, &x, &_1);
     }
   }
-  return gerepileupto(av0, gaddgs(x,-1));
+  return gerepileupto(av0, gsub(x, _1));
 }
 GEN
 suminf0(GEN a, GEN code, long prec)
@@ -524,14 +721,16 @@ sumdivmultexpr(GEN num, GEN code)
   GEN y = gen_1, P,E;
   int isint = divisors_init(num, &P,&E);
   long i, l = lg(P);
+  GEN (*mul)(GEN,GEN);
 
   if (l == 1) { avma = av; return gen_1; }
   push_lex(gen_0, code);
+  mul = isint? mulii: gmul;
   for (i=1; i<l; i++)
   {
     GEN p = gel(P,i), q = p, z = gen_1;
     long j, e = E[i];
-    for (j = 1; j <= e; j++, q = isint?mulii(q, p): gmul(q,p))
+    for (j = 1; j <= e; j++, q = mul(q, p))
     {
       set_lex(-1, q);
       z = gadd(z, closure_evalnobrk(code));
@@ -636,7 +835,7 @@ prodinf0(GEN a, GEN code, long flag, long prec)
     case 1: EXPR_WRAP(code, prodinf1(EXPR_ARG, a, prec));
   }
   pari_err_FLAG("prodinf");
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -769,8 +968,8 @@ matrice(GEN nlig, GEN ncol, GEN code)
   GEN c1, c2, y;
   long i, m, n;
 
-  m = gtos(ncol);
   n = gtos(nlig);
+  m = ncol? gtos(ncol): n;
   if (m < 0)  pari_err_DOMAIN("matrix", "nbcols", "<", gen_0, stoi(m));
   if (n < 0)  pari_err_DOMAIN("matrix", "nbrows", "<", gen_0, stoi(n));
   if (!m) return cgetg(1,t_MAT);
@@ -870,7 +1069,7 @@ GEN
 polzag(long n, long m)
 {
   pari_sp av = avma;
-  GEN g = ZX_unscale(polzag1(n,m), gen_m1);
+  GEN g = ZX_z_unscale(polzag1(n,m), -1);
   return gerepileupto(av, RgX_Rg_div(g,gel(g,2)));
 }
 
@@ -942,7 +1141,7 @@ sumalt0(GEN a, GEN code, long flag, long prec)
     case 1: EXPR_WRAP(code, sumalt2(EXPR_ARG,a,prec));
     default: pari_err_FLAG("sumalt");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* For k > 0, set S[k*2^i] <- g(k*2^i), k*2^i <= N = #S.
@@ -979,7 +1178,7 @@ binsum(GEN S, ulong k, void *E, GEN (*f)(void *, GEN), GEN a,
     GEN u;
     av = avma; u = gtofp(f(E, addiu(a, k << i)), prec);
     if (typ(u) != t_REAL) pari_err_TYPE("sumpos",u);
-    t = addrr(gtofp(u,prec), shiftr(t,1)); /* ~ g(k*2^i) */
+    t = addrr(gtofp(u,prec), mpshift(t,1)); /* ~ g(k*2^i) */
     gel(S, k << i) = t = gerepileuptoleaf(av, t);
   }
 }
@@ -1055,7 +1254,7 @@ sumpos0(GEN a, GEN code, long flag, long prec)
     case 1: EXPR_WRAP(code, sumpos2(EXPR_ARG,a,prec));
     default: pari_err_FLAG("sumpos");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -1118,7 +1317,7 @@ zbrent(void *E, GEN (*eval)(void *, GEN), GEN a, GEN b, long prec)
       if (gsigne(p) > 0) q = gneg_i(q); else p = gneg_i(p);
       min1 = gsub(gmulsg(3, gmul(xm,q)), gabs(gmul(q, tol1), 0));
       min2 = gabs(gmul(e, q), 0);
-      if (gcmp(gmul2n(p, 1), gmin(min1, min2)) < 0)
+      if (gcmp(gmul2n(p, 1), gmin_shallow(min1, min2)) < 0)
         { e = d; d = gdiv(p, q); } /* interpolation OK */
       else
         { d = xm; e = d; } /* failed, use bisection */
@@ -1258,33 +1457,257 @@ derivnum(void *E, GEN (*eval)(void *, GEN, long), GEN x, long prec)
   a = eval(E, gsub(x, eps), newprec);
   b = eval(E, gadd(x, eps), newprec);
   y = gmul2n(gsub(b,a), e-1);
-  return gerepileupto(av, gprec_w(y, nbits2prec(fpr)));
+  return gerepilecopy(av, gprec_w(y, nbits2prec(fpr)));
+}
+
+/* Fornberg interpolation algorithm for finite differences coefficients
+* using N+1 equidistant grid points around 0 [ assume N even >= M ].
+* Compute \delta[m]_{N,nu} for all derivation orders m = 0..M such that
+*   h^m * f^{(m)}(0) = \sum_{nu = 0}^n delta[m]_{N,nu}  f(a_nu) + O(h^{N-m+1}),
+* for step size h.
+* Return a = [0,-1,1...,-N2,N2] and vector of vectors d: d[m+1][nu+1]
+* = (M!/m!) * delta[m]_{N,nu}, nu = 0..N */
+static void
+FD(long M, long N, GEN *pd, GEN *pa)
+{
+  GEN d, a, b, W, Wp, t, F, Mfact;
+  long N2, m, nu, i;
+
+  if (odd(N)) N++; /* make it even */
+  N2 = N>>1;
+  F = cgetg(N+2, t_VEC);
+  a = cgetg(N+2, t_VEC);
+  b = cgetg(N2+1, t_VEC);
+  gel(a,1) = gen_0;
+  for (i = 1; i <= N2; i++)
+  {
+    gel(a,2*i)   = utoineg(i);
+    gel(a,2*i+1) = utoipos(i);
+    gel(b,i) = sqru(i);
+  }
+  /* w = \prod (X - a[i]) = x W(x^2) */
+  Mfact = mpfact(M);
+  W = roots_to_pol(b, 0);
+  Wp = ZX_deriv(W);
+  t = gel(W,2); /* w'(0) */
+  t = diviiexact(t, Mfact);
+  gel(F,1) = RgX_Rg_div(RgX_inflate(W,2), t);
+  for (i = 1; i <= N2; i++)
+  { /* t = w'(a_{2i}) = w'(a_{2i+1}) */
+    GEN r, t = mulii(shifti(gel(b,i),1), poleval(Wp, gel(b,i)));
+    GEN U, S, T;
+    U = RgX_inflate(RgX_div_by_X_x(W, gel(b,i), &r), 2);
+    U = RgX_shift_shallow(U, 1);
+    U = RgXn_red_shallow(U, M+1); /* higher terms not needed */
+    t = diviiexact(t, Mfact);
+    U = RgX_Rg_div(U, t);
+    S = RgX_shift_shallow(U,1);
+    T = RgX_Rg_mul(U, gel(a,2*i+1));
+    gel(F,2*i)   = RgX_sub(S, T);
+    gel(F,2*i+1) = RgX_add(S, T);
+  }
+  /* F[i] = M! w(X) / ((X-a[i])w'(a[i])) + O(X^(M+1)) */
+  d = cgetg(M+2, t_VEC);
+  for (m = 0; m <= M; m++)
+  {
+    GEN v = cgetg(N+2, t_VEC); /* coeff(F[nu],X^m) */
+    for (nu = 0; nu <= N; nu++) gel(v, nu+1) = gmael(F, nu+1, m+2);
+    gel(d,m+1) = v;
+  }
+  *pd = d;
+  *pa = a;
+}
+
+static void
+chk_ord(long m)
+{
+  if (m < 0)
+    pari_err_DOMAIN("derivnumk", "derivation order", "<", gen_0, stoi(m));
+}
+
+GEN
+derivnumk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
+{
+  GEN A, D, X, F, ind;
+  long M, fpr, p, i, pr, l, lA, e, ex, eD, newprec;
+  pari_sp av = avma;
+  int allodd = 1;
+
+  ind = gtovecsmall(ind0);
+  l = lg(ind);
+  F = cgetg(l, t_VEC);
+  M = vecsmall_max(ind);
+  chk_ord(M);
+  if (!M) /* silly degenerate case */
+  {
+    X = eval(E, x, prec);
+    for (i = 1; i < l; i++) { chk_ord(ind[i]); gel(F,i) = X; }
+    if (typ(ind0) == t_INT) F = gel(F,1);
+    return gerepilecopy(av, F);
+  }
+  FD(M, 3*M-1, &D,&A); /* optimal if 'eval' uses quadratic time */
+
+  p = precision(x);
+  fpr = p ? prec2nbits(p): prec2nbits(prec);
+  eD = gexpo(gel(D,M));
+  e = (fpr + 3*M*log2((double)M)) / (2*M);
+  ex = gexpo(x);
+  if (ex < 0) ex = 0; /* near 0 */
+  pr = (long)ceil(fpr + e * M); /* ~ 3fpr/2 */
+  newprec = nbits2prec(pr + eD + ex + BITS_IN_LONG);
+  switch(typ(x))
+  {
+    case t_REAL:
+    case t_COMPLEX:
+      x = gprec_w(x, newprec);
+  }
+  lA = lg(A); X = cgetg(lA, t_VEC);
+  for (i = 1; i < l; i++)
+    if (!odd(ind[i])) { allodd = 0; break; }
+  /* if only odd derivation orders, the value at 0 (A[1]) is not needed */
+  gel(X, 1) = gen_0;
+  for (i = allodd? 2: 1; i < lA; i++)
+  {
+    GEN t = eval(E, gadd(x, gmul2n(gel(A,i), -e)), newprec);
+    if (!gprecision(t)) t = gtofp(t, newprec);
+    gel(X, i) = t;
+  }
+
+  for (i = 1; i < l; i++)
+  {
+    GEN t;
+    long m = ind[i]; chk_ord(m);
+    t = gmul2n(RgV_dotproduct(gel(D,m+1), X), e*m);
+    if (m < M) t = gdiv(t, mulu_interval(m+1,M));
+    gel(F,i) = t;
+  }
+  if (typ(ind0) == t_INT) F = gel(F,1);
+  return gerepileupto(av, gprec_w(F, nbits2prec(fpr)));
+}
+/* v(t') */
+static long
+rfrac_val_deriv(GEN t)
+{
+  long v = varn(gel(t,2));
+  return gvaluation(deriv(t, v), pol_x(v));
+}
+
+GEN
+derivfunk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
+{
+  pari_sp av;
+  GEN ind, xp, ixp, F, G;
+  long i, l, vx, M;
+  if (!ind0) return derivfun(E, eval, x, prec);
+  switch(typ(x))
+  {
+  case t_REAL: case t_INT: case t_FRAC: case t_COMPLEX:
+    return derivnumk(E,eval, x, ind0, prec);
+  case t_POL:
+    ind = gtovecsmall(ind0);
+    M = vecsmall_max(ind);
+    xp = RgX_deriv(x);
+    x = RgX_to_ser(x, precdl+2 + M * (1+RgX_val(xp)));
+    break;
+  case t_RFRAC:
+    ind = gtovecsmall(ind0);
+    M = vecsmall_max(ind);
+    x = rfrac_to_ser(x, precdl+2 + M * (1+rfrac_val_deriv(x)));
+    xp = derivser(x);
+    break;
+  case t_SER:
+    ind = gtovecsmall(ind0);
+    M = vecsmall_max(ind);
+    xp = derivser(x);
+    break;
+  default: pari_err_TYPE("numerical derivation",x);
+    return NULL; /*LCOV_EXCL_LINE*/
+  }
+  av = avma; chk_ord(M);
+  vx = varn(x);
+  ixp = M? ginv(xp): NULL;
+  F = cgetg(M+2, t_VEC);
+  gel(F,1) = eval(E, x, prec);
+  for (i = 1; i <= M; i++) gel(F,i+1) = gmul(deriv(gel(F,i),vx), ixp);
+  l = lg(ind); G = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    long m = ind[i]; chk_ord(m);
+    gel(G,i) = gel(F,m+1);
+  }
+  if (typ(ind0) == t_INT) G = gel(G,1);
+  return gerepilecopy(av, G);
 }
 
 GEN
 derivfun(void *E, GEN (*eval)(void *, GEN, long), GEN x, long prec)
 {
   pari_sp av = avma;
+  GEN xp;
   long vx;
   switch(typ(x))
   {
   case t_REAL: case t_INT: case t_FRAC: case t_COMPLEX:
     return derivnum(E,eval, x, prec);
   case t_POL:
-    x = RgX_to_ser(x, precdl+2+1); /* +1 because deriv reduce the precision by 1 */
-  case t_SER: /* FALL THROUGH */
-    vx = varn(x);
-    return gerepileupto(av, gdiv(deriv(eval(E, x, prec),vx), deriv(x,vx)));
+    xp = RgX_deriv(x);
+    x = RgX_to_ser(x, precdl+2+ (1 + RgX_val(xp)));
+    break;
+  case t_RFRAC:
+    x = rfrac_to_ser(x, precdl+2+ (1 + rfrac_val_deriv(x)));
+    /* fall through */
+  case t_SER:
+    xp = derivser(x);
+    break;
   default: pari_err_TYPE("formal derivation",x);
-    return NULL; /*NOT REACHED*/
+    return NULL; /*LCOV_EXCL_LINE*/
   }
+  vx = varn(x);
+  return gerepileupto(av, gdiv(deriv(eval(E, x, prec),vx), xp));
 }
 
 GEN
-derivnum0(GEN a, GEN code, long prec)
+laurentseries(void *E, GEN (*f)(void*,GEN x, long), long M, long v, long prec)
 {
-  EXPR_WRAP(code, derivfun (EXPR_ARGPREC,a,prec));
+  pari_sp av = avma;
+  long d;
+
+  if (v < 0) v = 0;
+  d = maxss(M+1,1);
+  for (;;)
+  {
+    long i, dr, vr;
+    GEN s;
+    s = cgetg(d+2, t_SER); s[1] = evalsigne(1) | evalvalp(1) | evalvarn(v);
+    gel(s, 2) = gen_1; for (i = 3; i <= d+1; i++) gel(s, i) = gen_0;
+    s = f(E, s, prec);
+    if (typ(s) != t_SER || varn(s) != v) pari_err_TYPE("laurentseries", s);
+    vr = valp(s);
+    if (M < vr) { avma = av; return zeroser(v, M); }
+    dr = lg(s) + vr - 3 - M;
+    if (dr >= 0) return gerepileupto(av, s);
+    avma = av; d -= dr;
+  }
 }
+static GEN
+_evalclosprec(void *E, GEN x, long prec)
+{
+  GEN s;
+  push_localprec(prec); s = closure_callgen1((GEN)E, x);
+  pop_localprec(); return s;
+}
+#define CLOS_ARGPREC __E, &_evalclosprec
+GEN
+laurentseries0(GEN f, long M, long v, long prec)
+{
+  if (typ(f) != t_CLOSURE || closure_arity(f) != 1 || closure_is_variadic(f))
+    pari_err_TYPE("laurentseries",f);
+  EXPR_WRAP(f, laurentseries(CLOS_ARGPREC,M,v,prec));
+}
+
+GEN
+derivnum0(GEN a, GEN code, GEN ind, long prec)
+{ EXPR_WRAP(code, derivfunk(EXPR_ARGPREC,a,ind,prec)); }
 
 GEN
 derivfun0(GEN code, GEN args, long prec)
@@ -1297,13 +1720,31 @@ derivfun0(GEN code, GEN args, long prec)
 /********************************************************************/
 /**                   Numerical extrapolation                      **/
 /********************************************************************/
-
+/* for t_CLOSURE */
 static double
-extgetmf(long muli)
+fun_getmf(long mul)
 {
-  double mulfact[] = {0.5,0.5,0.48,0.43,0.41,0.39,0.38,0.37,0.36,0.36,0.35};
-  if (muli > 100) return 0.35*LOG10_2;
-  return mulfact[muli/10]*LOG10_2;
+  const double A[] = {0,
+    0.331,0.260,0.228,0.210,0.198,
+    0.188,0.181,0.175,0.170,0.166 };
+  const double B[] = {0,
+    0.166,0.142,0.132,0.125,0.120,
+    0.116,0.113,0.111,0.109,0.107 };
+  if (mul <=  10) return A[mul];
+  if (mul <= 109) return B[mul/10]; else return 0.105;
+}
+/* for t_VEC */
+static double
+vec_getmf(long mul)
+{
+  const double A[] = {0,
+    0.2062, 0.1760, 0.1613, 0.1519, 0.1459,
+    0.1401, 0.1360, 0.1327, 0.1299, 0.1280 };
+  const double B[] = {0,
+    0.1280, 0.1133, 0.1064, 0.1019, 0.0987,
+    0.0962, 0.0942, 0.0925, 0.0911, 0.0899 };
+  if (mul <=  10) return A[mul];
+  if (mul <= 109) return B[mul/10]; else return 0.0899;
 }
 
 /* [u(n*muli), u <= N], muli = 1 unless f!=NULL */
@@ -1311,18 +1752,17 @@ static GEN
 get_u(void *E, GEN (*f)(void *, GEN, long), long N, long muli, long prec)
 {
   long n;
-  GEN u;
+  GEN u = cgetg(N+1, t_VEC);
   if (f)
   {
-    u = cgetg(N+1, t_VEC);
     for (n = 1; n <= N; n++) gel(u,n) = f(E, stoi(muli*n), prec);
   }
   else
   {
-    u = (GEN)E;
-    n = lg(u)-1;
-    if (n < N) pari_err_COMPONENT("limitnum","<",stoi(N), stoi(n));
-    u = vecslice(u, 1, N);
+    GEN v = (GEN)E;
+    long t = lg(v)-1;
+    if (t < N*muli) pari_err_COMPONENT("limitnum","<",stoi(N), stoi(t));
+    for (n = 1; n <= N; n++) gel(u,n) = gel(v, muli*n);
   }
   for (n = 1; n <= N; n++)
   {
@@ -1347,48 +1787,40 @@ static void
 limit_init(struct limit *L, void *E, GEN (*f)(void*,GEN,long),
            long muli, GEN alpha, long prec)
 {
-  long bitprec = prec2nbits(prec), N;
+  long bitprec = prec2nbits(prec), n, N;
   GEN na;
-  long n;
 
   if (muli <= 0) muli = 20;
-  if (!f) muli = 1;
-  L->N = N = (long)ceil(extgetmf(muli)*bitprec);
-  L->prec = nbits2prec((long)ceil(1.25*bitprec) + 32);
+  L->N = N = (long)ceil((f? fun_getmf(muli): vec_getmf(muli)) * bitprec);
+  L->prec = nbits2prec(bitprec + (long)ceil(1.844*N));
   L->prec0 = prec;
   L->u = get_u(E, f, N, muli, L->prec);
-  if (alpha && gequal1(alpha)) alpha = NULL;
-  L->na = na  = cgetg(N+1, t_VEC);
-  for (n = 1; n <= N; n++)
+  if (alpha && !gequal1(alpha))
   {
-    GEN c = utoipos(n*muli);
-    if (alpha) c = gpow(c, alpha, L->prec);
-    gel(na,n) = c;
-  }
-  if (alpha)
-  {
-    GEN nma, malpha = gneg(alpha);
+    long prec2 = gprecision(alpha);
+    GEN nma;
+    if (!prec2) prec2 = L->prec;
+    na = vecpowug(N, alpha, prec2);
     L->coef = NULL;
-    L->nma= nma = cgetg(N+1, t_VEC);
-    for (n = 1; n <= N; n++)
-    {
-      GEN c = gpow(utoipos(n),malpha,L->prec);
-      if (typ(c) != t_REAL) c = gtofp(c, L->prec);
-      gel(nma, n) = c;
-    }
+    L->nma = nma = cgetg(N+1, t_VEC);
+    for (n = 1; n <= N; n++) gel(nma, n) = ginv(gel(na, n));
+    if (muli != 1) na = gmul(na, gpow(utor(muli,prec2), alpha, prec2));
   }
   else
   {
-    GEN coef, C = vecbinome(N);
+    GEN coef, C = vecbinomial(N), T = vecpowuu(N, N);
+    na = cgetg(N+1, t_VEC);
     L->coef = coef = cgetg(N+1, t_VEC);
     L->nma = NULL;
     for (n = 1; n <= N; n++)
     {
-      GEN c = mulii(gel(C,n+1), powuu(n, N));
+      GEN c = mulii(gel(C,n+1), gel(T,n));
       if (odd(N-n)) togglesign_safe(&c);
       gel(coef, n) = c;
+      gel(na, n) = utoipos(n*muli);
     }
   }
+  L->na = na;
 }
 
 /* Zagier/Lagrange extrapolation */
@@ -1431,17 +1863,20 @@ asympnum(void *E, GEN (*f)(void *, GEN, long), long muli, GEN alpha, long prec)
   const long MAX = 100;
   pari_sp av = avma;
   GEN u, vres = vectrunc_init(MAX);
-  long i;
+  long i, B = prec2nbits(prec);
+  double LB = 0.9*expu(B); /* 0.9 and 0.95 below are heuristic */
   struct limit L;
   limit_init(&L, E,f, muli, alpha, prec);
+  if (alpha) LB *= gtodouble(alpha);
   u = L.u;
   for(i = 1; i <= MAX; i++)
   {
     GEN a, s, v, p, q;
     long n;
     s = limitnum_i(&L);
-    /* NOT bestappr: lindep will "properly" ignore the lower bits */
-    v = lindep(mkvec2(gen_1, s));
+    /* NOT bestappr: lindep properly ignores the lower bits */
+    v = lindep_bit(mkvec2(gen_1, s), maxss((long)(0.95*floor(B - i*LB)), 32));
+    if (lg(v) == 1) break;
     p = negi(gel(v,1));
     q = gel(v,2);
     if (!signe(q)) break;
